package org.maachang.userManager.client ;

import java.io.IOException;
import java.net.URLDecoder;
import java.util.HashMap;
import java.util.Map;

import org.maachang.seabass.service.client.HttpResult;
import org.maachang.seabass.service.client.HttpService;
import org.maachang.util.BaseRandom;
import org.maachang.util.StringUtil;

/**
 * ユーザ管理情報にアクセスする処理群(Seabass32暗号).
 *
 * @version 2010/01/11
 * @author masahito suzuki
 * @since UserManager 1.0.0
 */
public abstract class UserManagerService32 {
    private UserManagerService32(){}
    
    /**
     * エラーメッセージ.
     */
    private static final String ERROR_CODE = "@error," ;
    
    /**
     * ログイン処理.
     * @param rand ランダムオブジェクトを設定します.
     * @param sessionId セッションIDを設定します.
     * @param host 対象のホスト名を設定します.
     * @param user 対象のユーザ名を設定します.
     * @param passwd 対象のパスワードを設定します.
     * @return String[] ログイン処理に成功した場合、セッションIDが返されます.
     * @exception Exception 例外.
     */
    public static final String[] login( BaseRandom rand,String sessionId,String host,String user,String passwd )
        throws Exception {
        if( rand == null || sessionId == null || sessionId.length() <= 0 ||
            host == null || ( host = host.trim() ).length() <= 0 ||
            user == null || ( user = user.trim() ).length() <= 0 ||
            passwd == null || ( passwd = passwd.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        String url = UserManagerService.getHost( host ) ;
        // 新しいセッションIDを取得.
        HttpResult res ;
        res = HttpService.seabass32( rand,sessionId,url+"/um/remote/session/createAjax",null,null,null ) ;
        if( res == null || res.getState() != 200 ) {
            throw new IOException( "ログイン処理に失敗しました" ) ;
        }
        String session = StringUtil.trim( new String( res.getBinary(),"UTF8" ) ) ;
        if( session == null ) {
            throw new IOException( "ログイン処理に失敗しました" ) ;
        }
        // 新しいセッション条件でログイン処理.
        Map<String,Object> pms = new HashMap<String,Object>() ;
        pms.put( "user",user ) ;
        pms.put( "passwd",passwd ) ;
        pms.put( "session",session ) ;
        res = HttpService.seabass32( rand,sessionId,url+"/um/remote/session/loginAjax",pms,null,null ) ;
        if( res == null || res.getState() != 200 ) {
            throw new IOException( "ログイン処理に失敗しました" ) ;
        }
        String ret = StringUtil.trim( new String( res.getBinary(),"UTF8" ) ) ;
        if( ret == null ) {
            throw new IOException( "ログイン処理に失敗しました" ) ;
        }
        ret = URLDecoder.decode( ret,"UTF8" ) ;
        if( ret.startsWith( ERROR_CODE ) ) {
            throw new IOException( ret.substring( ERROR_CODE.length() ) ) ;
        }
        return new String[]{session,ret} ;
    }
    
    /**
     * ログアウト処理.
     * @param rand ランダムオブジェクトを設定します.
     * @param sessionId セッションIDを設定します.
     * @param host 対象のホスト名を設定します.
     * @param session 対象のセッションIDを設定します.
     * @param user 対象のユーザ名を設定します.
     * @return boolean [true]の場合、削除に成功しました.
     * @exception Exception 例外.
     */
    public static final boolean logout( BaseRandom rand,String sessionId,String host,String session,String user )
        throws Exception {
        if( rand == null || sessionId == null || sessionId.length() <= 0 ||
            host == null || ( host = host.trim() ).length() <= 0 ||
            session == null || ( session = session.trim() ).length() <= 0 ||
            user == null || ( user = user.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        String url = UserManagerService.getHost( host ) ;
        HttpResult res ;
        Map<String,Object> pms = new HashMap<String,Object>() ;
        pms.put( "session",session ) ;
        pms.put( "user",user ) ;
        res = HttpService.seabass32( rand,sessionId,url+"/um/remote/session/removeAjax",pms,null,null ) ;
        if( res == null || res.getState() != 200 ) {
            throw new IOException( "ログアウト処理に失敗しました" ) ;
        }
        String ret = StringUtil.trim( new String( res.getBinary(),"UTF8" ) ) ;
        if( ret == null ) {
            throw new IOException( "ログアウト処理に失敗しました" ) ;
        }
        ret = URLDecoder.decode( ret,"UTF8" ) ;
        if( ret.startsWith( ERROR_CODE ) ) {
            throw new IOException( ret.substring( ERROR_CODE.length() ) ) ;
        }
        return "true".equals( ret ) ;
    }
    
    /**
     * ユーザ情報を取得.
     * @param rand ランダムオブジェクトを設定します.
     * @param sessionId セッションIDを設定します.
     * @param host 対象のホスト名を設定します.
     * @param session 対象のセッションIDを設定します.
     * @param user 対象のユーザ名を設定します.
     * @return String ユーザ情報が返されます.
     * @exception Exception 例外.
     */
    public static final String getUserInfo( BaseRandom rand,String sessionId,String host,String session,String user )
        throws Exception {
        if( rand == null || sessionId == null || sessionId.length() <= 0 ||
            host == null || ( host = host.trim() ).length() <= 0 ||
            session == null || ( session = session.trim() ).length() <= 0 ||
            user == null || ( user = user.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        String url = UserManagerService.getHost( host ) ;
        HttpResult res ;
        Map<String,Object> pms = new HashMap<String,Object>() ;
        pms.put( "session",session ) ;
        pms.put( "user",user ) ;
        res = HttpService.seabass32( rand,sessionId,url+"/um/remote/session/userInfoAjax",pms,null,null ) ;
        if( res == null || res.getState() != 200 ) {
            throw new IOException( "ユーザ情報取得処理に失敗しました" ) ;
        }
        String ret = StringUtil.trim( new String( res.getBinary(),"UTF8" ) ) ;
        if( ret == null ) {
            throw new IOException( "ユーザ情報取得処理に失敗しました" ) ;
        }
        ret = URLDecoder.decode( ret,"UTF8" ) ;
        if( ret.startsWith( ERROR_CODE ) ) {
            throw new IOException( ret.substring( ERROR_CODE.length() ) ) ;
        }
        return ret ;
    }
    
    /**
     * ログイン継続処理.
     * @param rand ランダムオブジェクトを設定します.
     * @param sessionId セッションIDを設定します.
     * @param host 対象のホスト名を設定します.
     * @param session 対象のセッションIDを設定します.
     * @param user 対象のユーザ名を設定します.
     * @return boolean [true]の場合、ログイン条件は継続します.
     * @exception Exception 例外.
     */
    public static final boolean update( BaseRandom rand,String sessionId,String host,String session,String user )
        throws Exception {
        if( rand == null || sessionId == null || sessionId.length() <= 0 ||
            host == null || ( host = host.trim() ).length() <= 0 ||
            session == null || ( session = session.trim() ).length() <= 0 ||
            user == null || ( user = user.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        String url = UserManagerService.getHost( host ) ;
        HttpResult res ;
        Map<String,Object> pms = new HashMap<String,Object>() ;
        pms.put( "session",session ) ;
        pms.put( "user",user ) ;
        res = HttpService.seabass32( rand,sessionId,url+"/um/remote/session/updateAjax",pms,null,null ) ;
        if( res == null || res.getState() != 200 ) {
            throw new IOException( "ログイン継続処理に失敗しました" ) ;
        }
        String ret = StringUtil.trim( new String( res.getBinary(),"UTF8" ) ) ;
        if( ret == null ) {
            throw new IOException( "ログイン継続処理に失敗しました" ) ;
        }
        ret = URLDecoder.decode( ret,"UTF8" ) ;
        if( ret.startsWith( ERROR_CODE ) ) {
            throw new IOException( ret.substring( ERROR_CODE.length() ) ) ;
        }
        return "true".equals( ret ) ;
    }
    
    /**
     * セッションが存在するかチェック.
     * @param rand ランダムオブジェクトを設定します.
     * @param sessionId セッションIDを設定します.
     * @param host 対象のホスト名を設定します.
     * @param session 対象のセッションIDを設定します.
     * @param reservation セッションが存在しない場合にそのセッションIDを予約する場合は、[true]を設定します.
     * @return boolean [true]の場合、既にセッション情報が存在します.
     * @exception Exception 例外.
     */
    public static final boolean useSession( BaseRandom rand,String sessionId,String host,String session,boolean reservation )
        throws Exception {
        if( rand == null || sessionId == null || sessionId.length() <= 0 ||
            host == null || ( host = host.trim() ).length() <= 0 ||
            session == null || ( session = session.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        String url = UserManagerService.getHost( host ) ;
        HttpResult res ;
        Map<String,Object> pms = new HashMap<String,Object>() ;
        pms.put( "session",session ) ;
        pms.put( "reservation",String.valueOf( reservation ) ) ;
        res = HttpService.seabass32( rand,sessionId,url+"/um/remote/session/useAjax",pms,null,null ) ;
        if( res == null || res.getState() != 200 ) {
            throw new IOException( "セッション存在確認処理に失敗しました" ) ;
        }
        String ret = StringUtil.trim( new String( res.getBinary(),"UTF8" ) ) ;
        if( ret == null ) {
            throw new IOException( "セッション存在確認処理に失敗しました" ) ;
        }
        ret = URLDecoder.decode( ret,"UTF8" ) ;
        if( ret.startsWith( ERROR_CODE ) ) {
            throw new IOException( ret.substring( ERROR_CODE.length() ) ) ;
        }
        return "true".equals( ret ) ;
    }
    
    /** 指定マスター情報を取得. **/
    private static final String getMaster( BaseRandom rand,String sessionId,String host,String urlPath,String msg )
        throws Exception {
        if( rand == null || sessionId == null || sessionId.length() <= 0 ||
            host == null || ( host = host.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        String url = UserManagerService.getHost( host ) ;
        HttpResult res ;
        res = HttpService.seabass32( rand,sessionId,url,null,null,null ) ;
        if( res == null || res.getState() != 200 ) {
            throw new IOException( msg+"取得に失敗しました" ) ;
        }
        String ret = StringUtil.trim( new String( res.getBinary(),"UTF8" ) ) ;
        if( ret == null ) {
            throw new IOException( msg+"取得に失敗しました" ) ;
        }
        ret = URLDecoder.decode( ret,"UTF8" ) ;
        if( ret.startsWith( ERROR_CODE ) ) {
            throw new IOException( ret.substring( ERROR_CODE.length() ) ) ;
        }
        return ret ;
    }
    
    /**
     * 会社マスター情報を取得.
     * @param rand ランダムオブジェクトを設定します.
     * @param sessionId セッションIDを設定します.
     * @param host 対象のホスト名を設定します.
     * @return String 権限マスター情報が返されます.
     * @exception Exception 例外.
     */
    public static final String getCompanyMaster( BaseRandom rand,String sessionId,String host )
        throws Exception {
        return getMaster( rand,sessionId,host,"/um/remote/master/companyMasterAjax","会社マスター" ) ;
    }
    
    /**
     * 部署マスター情報を取得.
     * @param rand ランダムオブジェクトを設定します.
     * @param sessionId セッションIDを設定します.
     * @param host 対象のホスト名を設定します.
     * @return String 権限マスター情報が返されます.
     * @exception Exception 例外.
     */
    public static final String getPartMaster( BaseRandom rand,String sessionId,String host )
        throws Exception {
        return getMaster( rand,sessionId,host,"/um/remote/master/partMasterAjax","部署マスター" ) ;
    }
    
    /**
     * 社員マスター情報を取得.
     * @param rand ランダムオブジェクトを設定します.
     * @param sessionId セッションIDを設定します.
     * @param host 対象のホスト名を設定します.
     * @return String 社員マスター情報が返されます.
     * @exception Exception 例外.
     */
    public static final String getPersonMaster( BaseRandom rand,String sessionId,String host )
        throws Exception {
        return getMaster( rand,sessionId,host,"/um/remote/master/personMasterAjax","社員マスター" ) ;
    }
}
