package org.maachang.userManager.server ;

import java.io.IOException;
import java.util.Iterator;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.bdb.Bdb;
import org.maachang.bdb.BdbKeyValue;
import org.maachang.bdb.BdbManager;
import org.maachang.seabass.dbm.BdbService;
import org.maachang.util.ConvertParam;
import org.maachang.util.atomic.AtomicOBJECT;
import org.maachang.util.thread.LoopThread;

/**
 * ユーザを管理するオブジェクト.
 *
 * @version 2010/01/11
 * @author masahito suzuki
 * @since UserManager 1.0.0
 */
public class UserManager {
    
    /**
     * 予約ユーザ名.
     */
    public static final String RESERVATION_USER_NAME = "_$reservation$_$user$_$name$_" ;
    
    /**
     * ログ.
     */
    private static final Log LOG = LogFactory.getLog( UserManager.class ) ;
    
    /**
     * 監視処理Wait.
     */
    private static final long WAIT = 250L ;
    
    /**
     * 最小セッションタイムアウト値.
     * 30分.
     */
    private static final long MIN_SESSION_TIMEOUT = 1800000L ;
    
    /**
     * 最大セッションタイムアウト値.
     * 1週間.
     */
    private static final long MAX_SESSION_TIMEOUT = 604800000L ;
    
    /**
     * ユーザ管理用Bdb.
     */
    private final AtomicOBJECT<Bdb> bdb = new AtomicOBJECT<Bdb>() ;
    
    /**
     * セッションタイムアウト値.
     */
    private long timeout = -1L ;
    
    /** セッション監視. **/
    private SessionMonThread mon = null ;
    
    /**
     * コンストラクタ.
     */
    private UserManager() {
        
    }
    
    private static final UserManager SNGL = new UserManager() ;
    
    /**
     * オブジェクトを取得.
     * @return UserManager オブジェクト情報が返されます.
     */
    public static final UserManager getInstance() {
        return SNGL ;
    }
    
    /**
     * 初期化処理.
     * @param timeout セッションタイムアウト値を設定します.
     * @param userManagerName Bdbでオープンする名前を設定します.
     * @exception Exception 例外.
     */
    public void init( long timeout,String userManagerName ) throws Exception {
        if( userManagerName == null || ( userManagerName = userManagerName.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        if( timeout <= MIN_SESSION_TIMEOUT ) {
            timeout = MIN_SESSION_TIMEOUT ;
        }
        else if( timeout >= MAX_SESSION_TIMEOUT ) {
            timeout = MAX_SESSION_TIMEOUT ;
        }
        BdbManager manager = BdbService.getInstance().get() ;
        if( manager != null || !manager.isDestroy() ) {
            this.bdb.set( manager.get( userManagerName ) ) ;
            this.timeout = timeout ;
            this.mon = new SessionMonThread( this.bdb,this.timeout ) ;
        }
        else {
            throw new IOException( "Bdbサービスはオープンされていません" ) ;
        }
    }
    
    /**
     * オブジェクト破棄.
     */
    public void destroy() {
        if( mon != null ) {
            mon.stopThread() ;
        }
    }
    
    /**
     * ユーザ情報を追加／更新
     * @param key 対象のキーを設定します.
     * @param userName 対象のユーザ名を設定します.
     * @exception Exception 例外.
     */
    public void put( BdbKeyValue kv,String key,String userName )
        throws Exception {
        Bdb b = bdb.get() ;
        if( b == null || b.isClose() ) {
            throw new IOException( "オブジェクトは既に閉じています" ) ;
        }
        if( key == null || key.length() <= 0 || userName == null ||
            ( userName = userName.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        UserBean bean = null ;
        kv.setKey( key.getBytes( "UTF8" ) ) ;
        // セッションKeyのユーザ情報が既に存在する場合.
        if( b.get( kv ) ) {
            bean = new UserBean( kv.getValue() ) ;
            if( !userName.equals( bean.getUserName() ) ) {
                bean = new UserBean( userName ) ;
                LOG.info( "@@@@ セッション[" + key + "]-user:" + userName +" を追加" ) ;
            }
            else {
                bean.update() ;
            }
        }
        // セッションKeyのユーザ情報が存在しない場合.
        else {
            bean = new UserBean( userName ) ;
            LOG.info( "@@@@ セッション[" + key + "]-user:" + userName +" を追加" ) ;
        }
        b.put( kv.setValue( bean.toBinary() ) ) ;
    }
    
    /**
     * セッション条件を更新.
     * @param key 対象のキーを設定します.
     * @return boolean [true]の場合、更新成功です.
     * @exception Exception 例外.
     */
    public boolean update( BdbKeyValue kv,String key,String userName )
        throws Exception {
        Bdb b = bdb.get() ;
        if( b == null || b.isClose() ) {
            throw new IOException( "オブジェクトは既に閉じています" ) ;
        }
        if( key == null || key.length() <= 0 || userName == null ||
            ( userName = userName.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        UserBean bean = null ;
        kv.setKey( key.getBytes( "UTF8" ) ) ;
        if( b.get( kv ) ) {
            bean = new UserBean( kv.getValue() ) ;
            if( userName.equals( bean.getUserName() ) ) {
                bean.update() ;
                b.put( kv.create( key.getBytes( "UTF8" ),bean.toBinary() ) ) ;
                return true ;
            }
        }
        return false ;
    }
    
    /**
     * ユーザ情報を削除.
     * @param key 対象のキーを設定します.
     * @exception Exception 例外.
     */
    public void remove( BdbKeyValue kv,String key )
        throws Exception {
        Bdb b = bdb.get() ;
        if( b == null || b.isClose() ) {
            throw new IOException( "オブジェクトは既に閉じています" ) ;
        }
        if( key == null || key.length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        kv.setKey( key.getBytes( "UTF8" ) ) ;
        if( b.remove( kv ) ) {
            LOG.info( "**** セッション[" + key + "]を破棄" ) ;
        }
    }
    
    /**
     * ユーザ情報を取得.
     * @param key 対象のキーを設定します.
     * @return UserBean ユーザ情報が返されます.
     * @exception Exception 例外.
     */
    public UserBean get( BdbKeyValue kv,String key )
        throws Exception {
        Bdb b = bdb.get() ;
        if( b == null || b.isClose() ) {
            throw new IOException( "オブジェクトは既に閉じています" ) ;
        }
        if( key == null || key.length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        kv.setKey( key.getBytes( "UTF8" ) ) ;
        if( b.get( kv ) ) {
            return new UserBean( kv.getValue() ) ;
        }
        return null ;
    }
    
    /**
     * 指定キーが存在するかチェック.
     * @param key 対象のキーを設定します.
     * @return boolean [true]の場合、存在します.
     * @exception Exception 例外.
     */
    public boolean containsKey( BdbKeyValue kv,String key )
        throws Exception {
        Bdb b = bdb.get() ;
        if( b == null || b.isClose() ) {
            throw new IOException( "オブジェクトは既に閉じています" ) ;
        }
        if( key == null || key.length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        kv.setKey( key.getBytes( "UTF8" ) ) ;
        return b.containsKey( kv ) ;
    }
    
    /**
     * Iteratorを取得.
     * @return Iterator<BdbKeyValue> KeyValue情報用Iterator.
     * @exception Exception 例外.
     */
    public Iterator<BdbKeyValue> iterator( BdbKeyValue kv )
        throws Exception {
        Bdb b = bdb.get() ;
        if( b == null || b.isClose() ) {
            throw new IOException( "オブジェクトは既に閉じています" ) ;
        }
        return b.keyValues( kv ) ;
    }
    
    /**
     * 現在のセッション数を取得.
     * @return long 現在のセッション数が返されます.
     * @exception Exception 例外.
     */
    public long size() throws Exception {
        Bdb b = bdb.get() ;
        if( b == null || b.isClose() ) {
            throw new IOException( "オブジェクトは既に閉じています" ) ;
        }
        return b.size() ;
    }
    
    /**
     * 設定されているセッションタイムアウト値を取得.
     * @return long セッションタイムアウト値が返されます.
     */
    public long getTimeout() {
        return timeout ;
    }
    
    /**
     * オブジェクトが有効かチェック.
     * @return boolean [true]の場合、有効です.
     */
    public boolean isUse() {
        Bdb b = bdb.get() ;
        return b != null && !b.isClose() ;
    }
    
    /** セッション監視用スレッド. **/
    private static class SessionMonThread extends LoopThread {
        private static final long RESERVATION_USER_TIMEOUT = 30000L ;
        private final BdbKeyValue kv = new BdbKeyValue() ;
        private AtomicOBJECT<Bdb> bdb = null ;
        private long timeout = -1L ;
        private UserBean bean = new UserBean() ;
        SessionMonThread( AtomicOBJECT<Bdb> bdb,long timeout ) throws Exception {
            this.bdb = bdb ;
            this.timeout = timeout ;
            startThread() ;
        }
        protected void clear() {
            bdb.set( null ) ;
        }
        protected void toException( Exception e ) {
            
        }
        protected boolean execution() throws Exception {
            Thread.sleep( WAIT ) ;
            BdbKeyValue keyValue ;
            Bdb b = bdb.get() ;
            if( b == null || b.isClose() ) {
                stopThread() ;
                return false ;
            }
            if( b.size() <= 0 ) {
                return true ;
            }
            byte[] value ;
            long tm ;
            Iterator<BdbKeyValue> it = b.keyValues( kv ) ;
            while( it.hasNext() ) {
                if( isStop() ) {
                    return true ;
                }
                tm = System.currentTimeMillis() ;
                keyValue = it.next() ;
                value = keyValue.getValue() ;
                if( ConvertParam.convertLong( 0,value ) + timeout <= tm ) {
                    it.remove() ;
                    LOG.info( "**** セッション[" + new String( keyValue.getKey(),"UTF8" ) + "]を破棄" ) ;
                }
                // 予約名の可能性がある場合はチェックする.
                else if( value[ 17 ] == (byte)'_' && value[18] == (byte)'$' ) {
                    bean.create( value ) ;
                    if( RESERVATION_USER_NAME.equals( bean.getUserName() ) && bean.getLastAccess() + RESERVATION_USER_TIMEOUT <= tm ) {
                        bean.clear() ;
                        // 予約セッションが３０秒アクセスがないものは、削除.
                        it.remove() ;
                    }
                    bean.clear() ;
                }
                Thread.sleep( WAIT ) ;
                value = null ;
            }
            return true ;
        }
    }
}

