package org.maachang.conf;

import java.io.IOException;
import java.net.InetAddress;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.maachang.seabass.SeaBassContext;
import org.maachang.seabass.SeaBassScriptEngine;
import org.maachang.seabass.SeaBassStackContext;
import org.maachang.util.FileUtil;
import org.maachang.util.atomic.AtomicARRAY;
import org.maachang.util.atomic.ReadAtomicList;

/**
 * Iniパラメータ管理.
 * 
 * @version 2007/10/18
 * @author masahito suzuki
 * @since MaachangBase 1.00
 */
public class Config {
    
    /**
     * パラメータ管理.
     */
    private Map<String, Map<String, ReadAtomicList>> params = new ConcurrentHashMap<String, Map<String, ReadAtomicList>>();

    /**
     * パラメータキー位置格納.
     */
    private Map<String, ReadAtomicList> keyByNum = new ConcurrentHashMap<String, ReadAtomicList>() ;
    
    /**
     * Pnuts実行キャッシュ用パラメータ管理.
     */
    private Map<String, Map<String, AtomicARRAY<Object>>> pnutsCache = new ConcurrentHashMap<String, Map<String, AtomicARRAY<Object>>>();

    /** Pnuts実行用Engine. **/
    private static final SeaBassScriptEngine ENGINE = new SeaBassScriptEngine() ;

    /** Pnuts実行名. **/
    private static final String EXEC_NAME = "config" ;

    
    /**
     * コンフィグファイルを読み込み.
     * @param name 対象のファイル名を呼び出します.
     * @return Config コンフィグ情報が返されます.
     * @exception Exception 例外.
     */
    public static final Config read( String name ) throws Exception {
        if( name == null || ( name = name.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "コンフィグファイル名が設定されていません" ) ;
        }
        if( !FileUtil.isFileExists( name ) ) {
            throw new IOException( "ファイル名[" + name + "]のコンフィグ情報は存在しません" ) ;
        }
        Config ret = new Config() ;
        ReadConfigUtil.readConfig( ret,name ) ;
        return ret ;
    }

    /** コンフィグキャッシュ用子要素 **/
    private static class CacheConfigChild {
        protected long time = -1 ;
        protected Config conf = null ;
    }

    /** コンフィグキャッシュ用Map. **/
    private static final Map<String,CacheConfigChild> cache = new ConcurrentHashMap<String,CacheConfigChild>() ;

    /**
     * コンフィグファイル読み込み(キャッシュ対応).
     * @param name 対象のファイル名を呼び出します.
     * @return Config コンフィグ情報が返されます.
     * @exception Exception 例外.
     */
    public static final Config readByCache( String name ) throws Exception {
        if( name == null || ( name = name.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "コンフィグファイル名が設定されていません" ) ;
        }
        if( !FileUtil.isFileExists( name ) ) {
            throw new IOException( "ファイル名[" + name + "]のコンフィグ情報は存在しません" ) ;
        }
        name = FileUtil.getFullPath( name ) ;
        CacheConfigChild ch = cache.get( name ) ;
        long time = FileUtil.getLastTime( name ) ;
        if( ch != null ) {
            if( time != ch.time ) {
                ch = null ;
            }
            else {
                return ch.conf ;
            }
        }
        Config ret = new Config() ;
        ReadConfigUtil.readConfig( ret,name ) ;
        ch = new CacheConfigChild() ;
        ch.time = time ;
        ch.conf = ret ;
        cache.put( name,ch ) ;
        return ret ;
    }

    /**
     * コンストラクタ.
     */
    public Config() {

    }

    /**
     * 情報クリア. <BR>
     * <BR>
     * 情報をクリアします.
     */
    public void clear() {
        params.clear();
        keyByNum.clear();
    }

    /**
     * 新しいパラメータを追加. <BR>
     * <BR>
     * 新しいパラメータを追加します. <BR>
     * 
     * @param section
     *            対象のセクション名を設定します.
     * @param key
     *            対象のキー名を設定します.
     * @param value
     *            対象の要素を追加します.
     */
    public void put(String section, String key, String value) {
        if (section == null || key == null) {
            return;
        }
        if (value == null) {
            value = "";
        }
        section = section.trim();
        key = key.trim();
        ReadAtomicList keys = keyByNum.get(section);
        Map<String, ReadAtomicList> sectionValue = params.get(section);
        if (sectionValue == null) {
            sectionValue = new ConcurrentHashMap<String, ReadAtomicList>();
            params.put(section, sectionValue);
            keys = new ReadAtomicList() ;
            keyByNum.put(section, keys);
        }
        ReadAtomicList keyValue = sectionValue.get(key);
        if (keyValue == null) {
            keyValue = new ReadAtomicList() ;
            sectionValue.put(key, keyValue);
            keys.add(key);
        }
        keyValue.add(value);
    }

    /**
     * 新しいパラメータを設定. <BR>
     * <BR>
     * 新しいパラメータを設定します. <BR>
     * この条件の場合、既に同一条件が存在する場合、上書きします.
     * 
     * @param section
     *            対象のセクション名を設定します.
     * @param key
     *            対象のキー名を設定します.
     * @param value
     *            対象の要素を追加します.
     */
    public void set(String section, String key, String value) {
        if (section == null || key == null) {
            return;
        }
        if (value == null) {
            value = "";
        }
        section = section.trim();
        key = key.trim();
        Map<String, ReadAtomicList> sectionValue = params.get(section);
        if (sectionValue == null) {
            sectionValue = new ConcurrentHashMap<String, ReadAtomicList>();
            params.put(section, sectionValue);
        }
        ReadAtomicList keyValue = sectionValue.get(key);
        if (keyValue == null) {
            keyValue = new ReadAtomicList() ;
            sectionValue.put(key, keyValue);
        }
        else {
            keyValue.clear() ;
        }
        keyValue.add(value);
    }

    /**
     * 指定要素の内容を取得. <BR>
     * <BR>
     * 指定要素の内容を取得します. <BR>
     * 
     * @param section
     *            対象のセクション名を設定します.
     * @param key
     *            対象のキー名を設定します.
     * @param no
     *            対象の項番を設定します.
     * @return String 対象の要素が返されます.
     */
    public String get(String section, String key, int no) {
        if (section == null || key == null || no < 0) {
            return null;
        }
        section = section.trim();
        key = key.trim();
        Map<String, ReadAtomicList> sectionValue = params.get(section);
        if (sectionValue != null) {
            ReadAtomicList value = sectionValue.get(key);
            if (value != null) {
                if (value.size() > no) {
                    return ( String )value.get(no);
                }
            }
        }
        return null;
    }

    /**
     * 指定要素の内容を取得. <BR>
     * <BR>
     * 指定要素の内容を取得します. <BR>
     * 
     * @param section
     *            対象のセクション名を設定します.
     * @param key
     *            対象のキー名を設定します.
     * @param no
     *            対象の項番を設定します.
     * @return boolean 対象の要素が返されます.
     */
    public boolean getBoolean(String section, String key, int no) {
        return ConvIniParam.getBoolean( get( section,key,no ) ) ;
    }

    /**
     * 指定要素の内容を取得. <BR>
     * <BR>
     * 指定要素の内容を取得します. <BR>
     * 
     * @param section
     *            対象のセクション名を設定します.
     * @param key
     *            対象のキー名を設定します.
     * @param no
     *            対象の項番を設定します.
     * @return Boolean 対象の要素が返されます.
     */
    public Boolean getBooleanObject(String section, String key, int no) {
        Boolean ret = null ;
        String s = get( section,key,no ) ;
        if( s != null ) {
            try {
                ret = new Boolean( s ) ;
            } catch( Exception e ) {
                ret = null ;
            }
        }
        return ret ;
    }

    /**
     * 指定要素の内容を取得. <BR>
     * <BR>
     * 指定要素の内容を取得します. <BR>
     * 
     * @param section
     *            対象のセクション名を設定します.
     * @param key
     *            対象のキー名を設定します.
     * @param no
     *            対象の項番を設定します.
     * @return int 対象の要素が返されます.
     */
    public int getInt(String section, String key, int no) {
        return ConvIniParam.getInt( get( section,key,no ) ) ;
    }

    /**
     * 指定要素の内容を取得. <BR>
     * <BR>
     * 指定要素の内容を取得します. <BR>
     * 
     * @param section
     *            対象のセクション名を設定します.
     * @param key
     *            対象のキー名を設定します.
     * @param no
     *            対象の項番を設定します.
     * @return Integer 対象の要素が返されます.
     */
    public Integer getIntObject(String section, String key, int no) {
        Integer ret = null ;
        String s = get( section,key,no ) ;
        if( s != null ) {
            try {
                ret = new Integer( s ) ;
            } catch( Exception e ) {
                ret = null ;
            }
        }
        return ret ;
    }

    /**
     * 指定要素の内容を取得. <BR>
     * <BR>
     * 指定要素の内容を取得します. <BR>
     * 
     * @param section
     *            対象のセクション名を設定します.
     * @param key
     *            対象のキー名を設定します.
     * @param no
     *            対象の項番を設定します.
     * @return long 対象の要素が返されます.
     */
    public long getLong(String section, String key, int no) {
        return ConvIniParam.getLong( get( section,key,no ) ) ;
    }

    /**
     * 指定要素の内容を取得. <BR>
     * <BR>
     * 指定要素の内容を取得します. <BR>
     * 
     * @param section
     *            対象のセクション名を設定します.
     * @param key
     *            対象のキー名を設定します.
     * @param no
     *            対象の項番を設定します.
     * @return Long 対象の要素が返されます.
     */
    public Long getLongObject(String section, String key, int no) {
        Long ret = null ;
        String s = get( section,key,no ) ;
        if( s != null ) {
            try {
                ret = new Long( s ) ;
            } catch( Exception e ) {
                ret = null ;
            }
        }
        return ret ;
    }

    /**
     * 指定要素の内容を取得. <BR>
     * <BR>
     * 指定要素の内容を取得します. <BR>
     * 
     * @param section
     *            対象のセクション名を設定します.
     * @param key
     *            対象のキー名を設定します.
     * @param no
     *            対象の項番を設定します.
     * @return double 対象の要素が返されます.
     */
    public double getDouble(String section, String key, int no) {
        return ConvIniParam.getDouble( get( section,key,no ) ) ;
    }

    /**
     * 指定要素の内容を取得. <BR>
     * <BR>
     * 指定要素の内容を取得します. <BR>
     * 
     * @param section
     *            対象のセクション名を設定します.
     * @param key
     *            対象のキー名を設定します.
     * @param no
     *            対象の項番を設定します.
     * @return Double 対象の要素が返されます.
     */
    public Double getDoubleObject(String section, String key, int no) {
        Double ret = null ;
        String s = get( section,key,no ) ;
        if( s != null ) {
            try {
                ret = new Double( s ) ;
            } catch( Exception e ) {
                ret = null ;
            }
        }
        return ret ;
    }

    /**
     * 指定要素の内容を取得. <BR>
     * <BR>
     * 指定要素の内容を取得します. <BR>
     * 
     * @param section
     *            対象のセクション名を設定します.
     * @param key
     *            対象のキー名を設定します.
     * @param no
     *            対象の項番を設定します.
     * @return InetAddress 対象の要素が返されます.
     */
    public InetAddress getInetAddress(String section, String key, int no) {
        return ConvIniParam.getInetAddress( get( section,key,no ) ) ;
    }

    /**
     * Pnuts実行オブジェクトを取得.
     * @param section 対象のセクション名を設定します.
     * @param key 対象のキー名を設定します.
     * @param no 対象の項番を設定します.
     * @return Object 対象の要素が返されます.
     * @exception Exception 例外.
     */
    public Object getPnuts(String section, String key, int no)
        throws Exception {
        return getPnuts( null,section,key,no ) ;
    }
    
    /**
     * Pnuts実行オブジェクトを取得.
     * @param section 対象のセクション名を設定します.
     * @param key 対象のキー名を設定します.
     * @param no 対象の項番を設定します.
     * @return Object 対象の要素が返されます.
     * @exception Exception 例外.
     */
    public Object getPnuts(SeaBassContext context,String section, String key, int no)
        throws Exception {
        if (section == null || key == null || no < 0) {
            return null;
        }
        int valLen ;
        if( ( valLen = size( section,key ) ) > no ) {
            section = section.trim();
            key = key.trim();
            Map<String, AtomicARRAY<Object>> sval = pnutsCache.get(section);
            if (sval == null) {
                sval = new ConcurrentHashMap<String,AtomicARRAY<Object>>() ;
            }
            AtomicARRAY<Object> value = sval.get(key);
            if (value == null ) {
                value = new AtomicARRAY<Object>( valLen ) ;
            }
            Object o = value.get( no );
            if( o == null ) {
                String s = get( section,key,no ) ;
                if( s == null ) {
                    return s ;
                }
                if( context == null ) {
                    context = SeaBassStackContext.getInstance().getContext( null ) ;
                }
                context.setName( EXEC_NAME ) ;
                //o = ENGINE.eval( PACKAGE_LIST + s,context ) ;
                o = ENGINE.eval( s,context ) ;
                value.set( o,no ) ;
            }
            return o ;
        }
        return null;
    }

    /**
     * 指定要素の内容を取得. <BR>
     * <BR>
     * 指定要素の内容を取得します. <BR>
     * 
     * @param section
     *            対象のセクション名を設定します.
     * @param key
     *            対象のキー名を設定します.
     * @param no
     *            対象の項番を設定します.
     * @return String[] 対象の要素群が返されます.
     */
    public String[] getAll(String section, String key) {
        if (section == null || key == null) {
            return null;
        }
        section = section.trim();
        key = key.trim();
        Map<String, ReadAtomicList> sectionValue = params.get(section);
        if (sectionValue != null) {
            ReadAtomicList value = sectionValue.get(key);
            if (value != null && value.size() > 0) {
                int len = value.size();
                String[] ret = new String[len];
                for (int i = 0; i < len; i++) {
                    ret[i] = ( String )value.get(i);
                }
                return ret;
            }
        }
        return null;
    }

    /**
     * 指定セクションを削除. <BR>
     * <BR>
     * 指定セクションを削除します. <BR>
     * 
     * @param section
     *            対象のセクション名を設定します.
     */
    public void removeSection(String section) {
        if (section == null) {
            return;
        }
        params.remove(section);
        keyByNum.remove(section) ;
    }

    /**
     * 指定キー情報を削除. <BR>
     * <BR>
     * 指定キー情報を削除します. <BR>
     * 
     * @param section
     *            対象のセクション名を設定します.
     * @param key
     *            対象のキー名を設定します.
     */
    public void removeKey(String section,String key) {
        if (section == null || key == null) {
            return;
        }
        section = section.trim();
        key = key.trim();
        Map<String, ReadAtomicList> sectionValue = params.get(section);
        if (sectionValue != null) {
            sectionValue.remove(key) ;
            ReadAtomicList keys = keyByNum.get(section);
            if( keys != null && keys.size() > 0 ) {
                int len = keys.size() ;
                for( int i = 0 ; i < len ; i ++ ) {
                    if( key.equals( keys.get( i ) ) ) {
                        keys.remove( i ) ;
                        break ;
                    }
                }
            }
        }
    }

    /**
     * 指定要素数を取得. <BR>
     * <BR>
     * 指定要素数を取得します. <BR>
     * 
     * @param section
     *            対象のセクション名を設定します.
     * @param key
     *            対象のキー名を設定します.
     * @return int 指定要素数が返されます.
     */
    public int size(String section, String key) {
        if (section == null || key == null) {
            return -1;
        }
        section = section.trim();
        key = key.trim();
        Map<String, ReadAtomicList> sectionValue = params.get(section);
        if (sectionValue != null) {
            ReadAtomicList value = sectionValue.get(key);
            if (value != null) {
                return value.size();
            }
        }
        return -1;
    }

    /**
     * 指定キー数を取得. <BR>
     * <BR>
     * 指定キー数を取得します. <BR>
     * 
     * @param section
     *            対象のセクション名を設定します.
     * @return int 指定キー数が返されます.
     */
    public int size(String section) {
        if (section == null) {
            return -1;
        }
        section = section.trim();
        Map<String, ReadAtomicList> sectionValue = params.get(section);
        if (sectionValue != null) {
            return sectionValue.size() ;
        }
        return -1;
    }

    /**
     * 指定セクション数を取得. <BR>
     * <BR>
     * 指定セクション数が返されます. <BR>
     * 
     * @return int セクション数が返されます.
     */
    public int size() {
        return params.size();
    }

    /**
     * 指定キー群を取得. <BR>
     * <BR>
     * 指定キー群を取得します. <BR>
     * 
     * @param section
     *            対象のセクション名を設定します.
     * @return String[] 指定キー群が返されます.
     */
    public String[] getKeys(String section) {
        if (section == null) {
            return null;
        }
        section = section.trim();
        ReadAtomicList keys = keyByNum.get(section);
        String[] ret = null ;
        if( keys != null && keys.size() > 0 ) {
            int len = keys.size() ;
            ret = new String[ len ] ;
            for( int i = 0 ; i < len ; i ++ ) {
                ret[ i ] = ( String )keys.get( i ) ;
            }
        }
        return ret ;
    }

    /**
     * セクション名一覧を取得. <BR>
     * <BR>
     * セクション名一覧を取得します. <BR>
     * 
     * @return Object[] セクション名一覧が返されます.
     */
    public Object[] getSections() {
        if (params != null && params.size() > 0) {
            return params.keySet().toArray();
        }
        return null;
    }

    /**
     * 指定キー名が存在するかチェック. <BR>
     * <BR>
     * 指定キー名が存在するかチェックします. <BR>
     * 
     * @param section
     *            対象のセクション名を設定します.
     * @param key
     *            対象のキー名を設定します.
     * @return boolean [true]の場合は存在します.
     */
    public boolean isKeys(String section,String key) {
        if (section == null) {
            return false;
        }
        section = section.trim();
        Map<String, ReadAtomicList> sectionValue = params.get(section);
        if (sectionValue != null) {
            return ( sectionValue.get( key ) != null ) ? true : false ;
        }
        return false;
    }

    /**
     * セクション名が存在するかチェック.
     * <BR><BR>
     * セクション名が存在するかチェックします.
     * <BR>
     * @param name セクション名を設定します.
     * @return boolean [true]の場合は存在します.
     */
    public boolean isSection( String section ) {
        section = section.trim();
        return ( params.get( section ) != null ) ? true : false ;
    }

    /**
     * 文字列に変換.
     */
    public String toString() {
        if (params != null && params.size() > 0) {
            StringBuilder buf = new StringBuilder();
            Object[] objs = params.keySet().toArray();
            int len = objs.length;
            for (int i = 0; i < len; i++) {
                buf.append("[").append(objs[i]).append("]\n");
                Map<String, ReadAtomicList> key = params.get(objs[i]);
                if (key != null && key.size() > 0) {
                    Object[] sObjs = key.keySet().toArray();
                    int lenJ = sObjs.length;
                    for (int j = 0; j < lenJ; j++) {
                        buf.append("  ").append(sObjs[j]).append("\n");
                        ReadAtomicList emt = key.get(sObjs[j]);
                        if (emt != null && emt.size() > 0) {
                            int lenK = emt.size();
                            buf.append("    ");
                            for (int k = 0; k < lenK; k++) {
                                if (k != 0) {
                                    buf.append(",");
                                }
                                buf.append(emt.get(k));
                            }
                            buf.append("\n");
                        }
                    }
                }
            }
            return buf.toString();
        }
        return "null";
    }
}

