package org.maachang.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;

import org.maachang.util.FileUtil;
import org.maachang.util.atomic.TargetLock;

/**
 * ファイル操作処理.
 * 
 * @version 2009/12/18
 * @author masahito suzuki
 * @since MaachangBase 1.0B
 */
public abstract class FileOp {
    
    /**
     * ターゲット単位での同期.
     */
    private static final TargetLock sync = new TargetLock() ;
    
    public FileOp() {
    }
    
    /** ターゲットロック用のフルパス情報を取得 **/
    protected static final String fullPath( String path ) {
        try {
            return new File( path ).getCanonicalPath() ;
        } catch( Exception e ) {
        }
        return path ;
    }
    
    /**
     * ディレクトリ作成.
     * @param name ディレクトリを作成します.
     * @return boolean [true]の場合は処理が成功.
     */
    public static boolean mkdir( String name ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return false ;
        }
        boolean ret = false ;
        name = fullPath( name ) ;
        sync.lock( name ) ;
        try {
            FileUtil.getFsCache().clean( name ) ;
            if (FileUtil.isDirExists(name) == true) {
                return false ;
            } else {
                ret = new File(name).mkdirs() ;
            }
        } catch (Exception e) {
        } finally {
            sync.unlock( name ) ;
        }
        return ret ;
    }
    
    /**
     * ファイルを削除.
     * @param name 削除対象のファイル名を設定します.
     * @return boolean [true]の場合は処理が成功.
     */
    public static boolean removeFile( String name ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return false ;
        }
        name = fullPath( name ) ;
        sync.lock( name ) ;
        try {
            FileUtil.getFsCache().clean( name ) ;
            return FileUtil.removeFile( name ) ;
        } finally {
            sync.unlock( name ) ;
        }
    }
    
    /**
     * ディレクトリを削除.
     * @param name 削除対象のディレクトリ名を設定します.
     * @return boolean [true]の場合は処理が成功.
     */
    public static boolean removeDirectory( String name ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return false ;
        }
        name = fullPath( name ) ;
        sync.lock( name ) ;
        try {
            FileUtil.getFsCache().clean( name ) ;
            return FileUtil.removeDirectory( name ) ;
        } finally {
            sync.unlock( name ) ;
        }
    }
    
    /**
     * ファイル／ディレクトリの移動.
     * @param src 移動元の名前を設定します.
     * @param dest 移動先の名前を設定します.
     * @return boolean [true]の場合は処理が成功.
     */
    public static boolean move( String src,String dest ) {
        if( src == null || ( src = src.trim() ).length() <= 0 ||
            dest == null || ( dest = dest.trim() ).length() <= 0 ) {
            return false ;
        }
        src = fullPath( src ) ;
        sync.lock( src ) ;
        try {
            FileUtil.getFsCache().clean( src ) ;
            FileUtil.getFsCache().clean( dest ) ;
            return FileUtil.moveFile( src,dest ) ;
        } finally {
            sync.unlock( src ) ;
        }
    }
    
    /**
     * ファイルコピー.
     * @param src 移動元のファイル名を設定します.
     * @param dest 移動先のファイル名を設定します.
     * @return boolean [true]の場合は処理が成功.
     */
    public static boolean copyFile( String src,String dest ) {
        if( src == null || ( src = src.trim() ).length() <= 0 ||
            dest == null || ( dest = dest.trim() ).length() <= 0 ) {
            return false ;
        }
        src = fullPath( src ) ;
        sync.lock( src ) ;
        try {
            FileUtil.getFsCache().clean( dest ) ;
            return FileUtil.copyFile( src,dest ) ;
        } finally {
            sync.unlock( src ) ;
        }
    }
    
    /**
     * ディレクトリコピー.
     * @param src 移動元のディレクトリ名を設定します.
     * @param dest 移動先のディレクトリ名を設定します.
     * @return boolean [true]の場合は処理が成功.
     */
    public static boolean copyDirectory( String src,String dest ) {
        if( src == null || ( src = src.trim() ).length() <= 0 ||
            dest == null || ( dest = dest.trim() ).length() <= 0 ) {
            return false ;
        }
        src = fullPath( src ) ;
        sync.lock( src ) ;
        try {
            FileUtil.getFsCache().clean( src ) ;
            FileUtil.getFsCache().clean( dest ) ;
            return FileUtil.copyDirectory( src,dest ) ;
        } catch( Exception e ) {
        } finally {
            sync.unlock( src ) ;
        }
        return false ;
    }
    
    /**
     * ファイル長を取得.
     * @param name 対象のファイル名を設定します.
     * @return long ファイル長が返されます.
     */
    public static long getLength( String name ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return -1L ;
        }
        name = fullPath( name ) ;
        sync.lock( name ) ;
        try {
            return FileUtil.getLength( name ) ;
        } finally {
            sync.unlock( name ) ;
        }
    }
    
    /**
     * ファイル日付を取得.
     * @param name 対象のファイル名を設定します.
     * @return long ファイル日付が返されます.
     */
    public static long getLastTime( String name ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return -1L ;
        }
        name = fullPath( name ) ;
        sync.lock( name ) ;
        try {
            return FileUtil.getLastTime( name ) ;
        } finally {
            sync.unlock( name ) ;
        }
    }
    
    /**
     * 指定パス名のフルパスを取得.
     * @param name 対象のパスを設定します.
     * @return String フルパス情報が返されます.
     */
    public static String getFullPath( String name ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return "" ;
        }
        try {
            return FileUtil.getFullPath( name ) ;
        } catch( Exception e ) {
        }
        return name ;
    }
    
    /**
     * 指定パスのディレクトリ名を取得.
     * @param name 対象のパスを設定します.
     * @return String ディレクトリ名情報が返されます.
     */
    public static String getDirName( String name ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return "" ;
        }
        return FileUtil.getDirName( name ) ;
    }
    
    /**
     * 指定パスのファイル名を取得.
     * @param name 対象のパスを設定します.
     * @return String ファイル名情報が返されます.
     */
    public static String getFileName( String name ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return "" ;
        }
        return FileUtil.getFileName( name ) ;
    }
    
    /**
     * 指定ファイルが存在するかチェック.
     * @param name 対象のパスを設定します.
     * @return boolean [true]の場合、存在します.
     */
    public static boolean isFileExists( String name ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return false ;
        }
        name = fullPath( name ) ;
        sync.lock( name ) ;
        try {
            return FileUtil.isFileExists( name ) ;
        } finally {
            sync.unlock( name ) ;
        }
    }
    
    /**
     * 指定ディレクトリが存在するかチェック.
     * @param name 対象のパスを設定します.
     * @return boolean [true]の場合、存在します.
     */
    public static boolean isDirExists( String name ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return false ;
        }
        name = fullPath( name ) ;
        sync.lock( name ) ;
        try {
            return FileUtil.isDirExists( name ) ;
        } finally {
            sync.unlock( name ) ;
        }
    }
    
    /**
     * 指定パスが読み込み可能が存在するかチェック.
     * @param name 対象のパスを設定します.
     * @return boolean [true]の場合、可能です.
     */
    public static boolean isRead( String name ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return false ;
        }
        name = fullPath( name ) ;
        sync.lock( name ) ;
        try {
            return FileUtil.isRead( name ) ;
        } finally {
            sync.unlock( name ) ;
        }
    }
    
    /**
     * 指定パスが書き込み可能が存在するかチェック.
     * @param name 対象のパスを設定します.
     * @return boolean [true]の場合、可能です.
     */
    public static boolean isWrite( String name ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return false ;
        }
        name = fullPath( name ) ;
        sync.lock( name ) ;
        try {
            return FileUtil.isWrite( name ) ;
        } finally {
            sync.unlock( name ) ;
        }
    }
    
    /**
     * 指定ディレクトリ以下の一覧を取得.
     * @param name 対象のパスを設定します.
     * @return List<String> 指定ディレクトリ以下の情報が返されます.
     */
    public static List<String> getListByArray( String name ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return new ArrayList<String>() ;
        }
        name = fullPath( name ) ;
        sync.lock( name ) ;
        try {
            if( FileUtil.isDirExists( name ) ) {
                return FileUtil.getListByArray( name ) ;
            }
            return new ArrayList<String>() ;
        } finally {
            sync.unlock( name ) ;
        }
    }
    
    /**
     * 指定ディレクトリ以下の一覧を全て取得.
     * @param name 対象のパスを設定します.
     * @return List<String> 指定ディレクトリ以下の全ての情報が返されます.
     */
    public static List<String> getListAll( String name ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return new ArrayList<String>() ;
        }
        name = fullPath( name ) ;
        sync.lock( name ) ;
        try {
            if( FileUtil.isDirExists( name ) ) {
                return FileUtil.getListAll( name ) ;
            }
        } catch( Exception e ) {
        } finally {
            sync.unlock( name ) ;
        }
        return new ArrayList<String>() ;
    }
    
    /**
     * 指定ディレクトリ以下のファイルだけを取得.
     * @param name 対象のパスを設定します.
     * @return List<String> 指定ディレクトリ以下のファイルだけの情報が返されます.
     */
    public static List<String> getOnlyFileToArray( String name ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return new ArrayList<String>() ;
        }
        name = fullPath( name ) ;
        sync.lock( name ) ;
        try {
            if( FileUtil.isDirExists( name ) ) {
                return FileUtil.getOnlyFileToArray( name ) ;
            }
        } finally {
            sync.unlock( name ) ;
        }
        return new ArrayList<String>() ;
    }
    
    /**
     * 指定ファイル内容をバイナリで取得.
     * @param name 対象のファイル名を設定します.
     * @return byte[] バイナリ情報が返されます.
     */
    public static byte[] getFile( String name ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return null ;
        }
        name = fullPath( name ) ;
        sync.lock( name ) ;
        try {
            if( FileUtil.isFileExists( name ) ) {
                return FileUtil.getFile( name ) ;
            }
        } catch( Exception e ) {
        } finally {
            sync.unlock( name ) ;
        }
        return null ;
    }
    
    /**
     * 指定ファイル内容を文字列で取得.
     * @param name 対象のファイル名を設定します.
     * @param charset 対象のキャラクタセットを設定します.
     * @return String 文字列情報が返されます.
     */
    public static String getFileByString( String name,String charset ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return null ;
        }
        name = fullPath( name ) ;
        sync.lock( name ) ;
        try {
            if( FileUtil.isFileExists( name ) ) {
                if( charset == null || charset.length() <= 0 ) {
                    return FileUtil.getFileByString( name ) ;
                } else {
                    return FileUtil.getFileByString( name,charset ) ;
                }
            }
        } catch( Exception e ) {
        } finally {
            sync.unlock( name ) ;
        }
        return null ;
    }
    
    /**
     * 指定ファイル内容にバイナリで出力.
     * @param name 対象のファイル名を設定します.
     * @param mode [true]の場合、上書き、[false]の場合は、追加書き込みで処理します.
     * @param binary 書き込み対象のバイナリを設定します.
     * @return boolean [true]の場合、書き込み成功しました.
     */
    public static boolean setFile( String name,boolean mode,byte[] binary ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return false ;
        }
        name = fullPath( name ) ;
        sync.lock( name ) ;
        try {
            FileUtil.setFile( name,mode,binary ) ;
            return true ;
        } catch( Exception e ) {
        } finally {
            sync.unlock( name ) ;
        }
        return false ;
    }
    
    /**
     * 指定ファイル内容に文字列で出力.
     * @param name 対象のファイル名を設定します.
     * @param mode [true]の場合、上書き、[false]の場合は、追加書き込みで処理します.
     * @param string 書き込み対象の文字列を設定します.
     * @param charset 対象のキャラクタセットを設定します.
     * @return boolean [true]の場合、書き込み成功しました.
     */
    public static boolean setFileByString( String name,boolean mode,String string,String charset ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return false ;
        }
        name = fullPath( name ) ;
        sync.lock( name ) ;
        try {
            if( charset == null || charset.length() <= 0 ) {
                FileUtil.setFileByString( name,mode,string ) ;
            } else {
                FileUtil.setFileByString( name,mode,string,charset ) ;
            }
            return true ;
        } catch( Exception e ) {
        } finally {
            sync.unlock( name ) ;
        }
        return false ;
    }
    
    /**
     * 空のファイルを作成.
     * @param name 対象のファイル名を設定します.
     * @return boolean [true]の場合、成功しました.
     */
    public static boolean createFile( String name ) {
        if (name == null || ( name = name.trim() ).length() <= 0 ) {
            return false ;
        }
        name = fullPath( name ) ;
        sync.lock( name ) ;
        try {
            FileUtil.createFile( name ) ;
            return true ;
        } catch( Exception e ) {
        } finally {
            sync.unlock( name ) ;
        }
        return false ;
    }
    
    /**
     * 指定ファイル内容を行指定で取得.
     * @param name 対象のファイル名を設定します.
     * @param charset 対象の文字コードを設定します.
     * @param start 開始行位置を設定します.
     * @param size 取得行サイズを取得します.
     * @return String 取得結果が返されます.
     * @exception Exception 例外.
     */
    public static String getLines( String name,String charset,int start,int size )
        throws Exception {
        if( name == null || ( name = name.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        if( size <= 0 ) {
            return "" ;
        }
        if( charset == null || ( charset = charset.trim() ).length() <= 0 ) {
            charset = "UTF8" ;
        }
        if( start < 0 ) {
            start = 0 ;
        }
        String ret = null ;
        BufferedReader r = null ;
        name = fullPath( name ) ;
        sync.lock( name ) ;
        try {
            r = new BufferedReader( new InputStreamReader( new FileInputStream( name ),charset ) ) ;
            for( int i = 0 ; i < start ; i ++ ) {
                if( r.readLine() == null ) {
                    return "" ;
                }
            }
            String s ;
            StringBuilder buf = new StringBuilder() ;
            for( int i = 0 ; i < size ; i ++ ) {
                s = r.readLine() ;
                if( s == null ) {
                    break ;
                }
                buf.append( s ).append( "\n" ) ;
            }
            r.close() ;
            r = null ;
            ret = buf.toString() ;
            buf = null ;
        } finally {
            if( r != null ) {
                r.close() ;
            }
            sync.unlock( name ) ;
        }
        return ret ;
    }
    
    /**
     * ターゲットロック.
     * @param name ファイル名を指定します.
     */
    public static void lock( String name ) {
        name = fullPath( name ) ;
        sync.lock( name ) ;
    }
    
    /**
     * ターゲットアンロック.
     * @param name ファイル名を指定します.
     */
    public static void unlock( String name ) {
        name = fullPath( name ) ;
        sync.unlock( name ) ;
    }
    
    /**
     * ターゲットロックを取得.
     * @return TargetLock ターゲットロックが返されます.
     */
    public static TargetLock getTargetLock() {
        return sync ;
    }
    
}
