package org.maachang.util;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel.MapMode;
import java.util.concurrent.locks.ReentrantLock;

import org.maachang.util.atomic.AtomicBOOL;
import org.maachang.util.atomic.AtomicINT;
import org.maachang.util.atomic.AtomicOBJECT;

/**
 * ファイルで管理するシーケンス情報.
 * 
 * @version 2010/01/14
 * @author masahito suzuki
 * @since MaachangBase 1.00
 */
public class FileSequence {
    private RandomAccessFile fp = null ;
    private MappedByteBuffer buf = null ;
    private final AtomicOBJECT<String> name = new AtomicOBJECT<String>() ;
    private final AtomicINT size = new AtomicINT( 0 ) ;
    private final AtomicBOOL openFlag = new AtomicBOOL() ;
    private final ReentrantLock sync = new ReentrantLock() ;
    
    /**
     * コンストラクタ.
     */
    public FileSequence() {
        
    }
    
    /**
     * 新しく生成.
     * @param name ファイル名を設定します.
     * @param size シーケンス数を設定します.
     * @exception Exception 例外.
     */
    public void create( String name,int size ) throws Exception {
        if( name == null || ( name = name.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "ファイル名が設定されていません" ) ;
        }
        if( size <= 0 ) {
            size = 1 ;
        }
        else if( size >= 65535 ) {
            size = 65535 ;
        }
        sync.lock() ;
        try {
            close() ;
            RandomAccessFile f = null ;
            MappedByteBuffer m = null ;
            int len = size * 8 ;
            try {
                f = new RandomAccessFile( name,"rw" ) ;
                f.setLength( len ) ;
                m = f.getChannel().map( MapMode.READ_WRITE,0,len ) ;
            } catch( Exception e ) {
                if( f != null ) {
                    try {
                        f.close() ;
                    } catch( Exception ee ) {
                    }
                    f = null ;
                    m = null ;
                }
                throw e ;
            }
            this.fp = f ;
            this.buf = m ;
            this.name.set( name ) ;
            this.size.set( size ) ;
            this.openFlag.set( true ) ;
        } finally {
            sync.unlock() ;
        }
    }
    
    /**
     * 以前の構成をロード.
     * @param name ロード対象のファイル名を設定します.
     * @exception Exception 例外.
     */
    public void load( String name ) throws Exception {
        if( name == null || ( name = name.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "ファイル名が設定されていません" ) ;
        }
        if( ( new File( name ) ).isFile() == false ) {
            throw new IOException( "ファイル名[" + name + "]は存在しません" ) ;
        }
        sync.lock() ;
        try {
            close() ;
            RandomAccessFile f = null ;
            MappedByteBuffer m = null ;
            int len = -1 ;
            try {
                f = new RandomAccessFile( name,"rw" ) ;
                len = (int)( f.length() / 8L ) ;
                m = f.getChannel().map( MapMode.READ_WRITE,0,len * 8 ) ;
            } catch( Exception e ) {
                if( f != null ) {
                    try {
                        f.close() ;
                    } catch( Exception ee ) {
                    }
                    f = null ;
                    m = null ;
                }
                throw e ;
            }
            this.fp = f ;
            this.buf = m ;
            this.name.set( name ) ;
            this.size.set( len ) ;
            this.openFlag.set( true ) ;
        } finally {
            sync.unlock() ;
        }
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        close() ;
    }
    
    /**
     * クローズ.
     */
    public void close() {
        this.openFlag.set( false ) ;
        sync.lock() ;
        try {
            if( this.fp != null ) {
                try {
                    this.fp.close() ;
                } catch( Exception e ) {
                }
            }
            this.fp = null ;
            this.buf = null ;
            this.name.set( null ) ;
            this.size.set( -1 ) ;
        } finally {
            sync.unlock() ;
        }
    }
    
    /**
     * シーケンス取得.
     * @param no 項番を設定します.
     * @return long シーケンス番号が返されます.
     * @exception Exception 例外.
     */
    public long get( int no ) throws Exception {
        if( !openFlag.get() ) {
            throw new IOException( "オブジェクトは既に閉じられています" ) ;
        }
        if( no < 0 || no >= size.get() ) {
            throw new IllegalArgumentException( "指定番号は範囲外です:" + no ) ;
        }
        int n = no * 8 ;
        long ret ;
        sync.lock() ;
        try {
            ret = buf.getLong( n ) ;
            if( ret + 1L >= Long.MAX_VALUE ) {
                ret = 0L ;
            }
            else {
                ret ++ ;
            }
            buf.putLong( n,ret ) ;
        } finally {
            sync.unlock() ;
        }
        return ret ;
    }
    
    /**
     * 配列数取得.
     * @return int シーケンス利用可能配列数が返されます.
     * @exception Exception 例外.
     */
    public int size() throws Exception {
        if( !openFlag.get() ) {
            throw new IOException( "オブジェクトは既に閉じられています" ) ;
        }
        return size.get() ;
    }
    
    /**
     * オープンファイル名を取得.
     * @return String オープンファイル名が返されます.
     * @exception Exception 例外.
     */
    public String getName() throws Exception {
        if( !openFlag.get() ) {
            throw new IOException( "オブジェクトは既に閉じられています" ) ;
        }
        return name.get() ;
    }
    
    /**
     * オブジェクト有効チェック.
     * @return boolean [true]の場合、オブジェクトは有効です.
     */
    public boolean isOpen() {
        return openFlag.get() ;
    }
}
