package org.maachang.util;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayOutputStream;
import java.io.CharArrayWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * ファイルユーティリティ.
 * 
 * @version 2007/10/18
 * @author masahito suzuki
 * @since MaachangBase 1.00
 */
public class FileUtil {

    /**
     * デフォルト読み込みキャラクタセット.
     */
    private static final String CHARSET = "UTF8";
    
    /**
     * OSキャラクタセット.
     */
    public static final String OS_CHARSET =
        ( "MS932".equals( System.getProperty("file.encoding") ) ?
            "Windows-31J" : System.getProperty("file.encoding") ) ;

    /**
     * OSファイルスペース.
     */
    public static final String FILE_SPACE = System.getProperty("file.separator");
    
    /**
     * ファイルキャッシュオブジェクト.
     */
    private static final FsCache fsCache ;
    static {
        fsCache = FsCache.getInstance() ;
    }
    
    /**
     * ファイルキャッシュオブジェクトを取得.
     * @return FsCache ファイルキャッシュオブジェクトが返されます.
     */
    public static final FsCache getFsCache() {
        return fsCache ;
    }
    
    /**
     * 指定パスを結合.
     * <BR><BR>
     * 指定パスを結合します.
     * <BR>
     * @param ctx 基本パスを設定します.
     * @param path 結合対象のパスを設定します.
     * @return String 結合されたパスが返されます.
     */
    public static String convertFullPath( String ctx,String path ) {
        if( ctx == null || ( ctx = ctx.trim() ).length() <= 0 ||
            path == null || ( path = path.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数が不正です" ) ;
        }
        if( path.startsWith( "/" ) || path.startsWith( "\\" ) ) {
            ctx = "/" ;
        }
        List<String> ctxList = StringUtil.cutString( ctx,"/\\" ) ;
        List<String> pathList = StringUtil.cutString( path,"/\\" ) ;
        if( pathList != null ) {
            int len = pathList.size() ;
            for( int i = 0 ; i < len ; i ++ ) {
                String c = pathList.get( i ) ;
                if( ".".equals( c ) ) {
                    continue ;
                }
                if( "..".equals( c ) ) {
                    ctxList.remove( ctxList.size()-1 ) ;
                }
                else {
                    ctxList.add( c ) ;
                }
            }
            pathList = null ;
        }
        StringBuilder buf = new StringBuilder() ;
        int len = ctxList.size() ;
        for( int i = 0 ; i < len ; i ++ ) {
            buf.append( "/" ).append( ctxList.get( i ) ) ;
        }
        return buf.toString() ;
    }
    
    /**
     * リアルパスと指定パスをマージ.
     * <BR><BR>
     * リアルパスと指定パスをマージします.
     * <BR>
     * @param real 対象のリアルパスを設定します.
     * @param path 対象のパスを設定します.
     * @return String マージされたフルパスが返されます.
     * @exception Exception 例外.
     */
    public static final String marge( String real,String path )
        throws Exception {
        if( real == null || ( real = real.trim() ).length() <= 0 ||
            path == null || ( path = path.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数が不正です" ) ;
        }
        if( real.endsWith( "/" ) || real.endsWith( "\\" ) ) {
            real = real.substring( 0,real.length()-1 ) ;
        }
        if( path.startsWith( "/" ) || path.startsWith( "\\" ) ) {
            path = new StringBuilder().append( real ).
                append( path ).toString() ;
        }
        else {
            path = new StringBuilder().append( real ).
                append( FileUtil.FILE_SPACE ).
                append( path ).toString() ;
        }
        return FileUtil.getFullPath( path ) ;
    }
    
    /**
     * 対象のディレクトリを生成. <BR>
     * <BR>
     * 対象のディレクトリ情報を生成します. <BR>
     * 
     * @param dirName
     *            生成対象のディレクトリ名を設定します.
     * @return String 結果のファイル名が返されます.
     * @exception IOException
     *                I/O例外.
     */
    public static final String mkdirs(String dirName) throws IOException {
        File fp = null;
        StringBuilder buf = null;
        String tmp = null;
        String ret = null;

        try {

            if (dirName == null) {
                throw new IllegalArgumentException("引数は不正です");
            }
            
            fsCache.clean( dirName ) ;
            if (isDirExists(dirName) == true) {

                fp = new File(dirName);
                tmp = fp.getCanonicalPath();
                buf = new StringBuilder();

                buf.append(tmp);
                buf.append(FILE_SPACE);
                ret = buf.toString();

            } else {

                fp = new File(dirName);
                tmp = fp.getCanonicalPath();
                buf = new StringBuilder();

                buf.append(tmp);
                buf.append(FILE_SPACE);
                ret = buf.toString();

                if (fp.mkdirs() == false) {
                    throw new IOException("ディレクトリ(" + dirName + ")の作成に失敗しました");
                }

            }

        } catch (IOException e) {
            throw e;
        } finally {
            fp = null;
            buf = null;
            tmp = null;
        }

        return ret;

    }

    /**
     * パス名をフルパス名に設定. <BR>
     * <BR>
     * パス名をフルパス名に設定します. <BR>
     * 
     * @param path
     *            対象のパス名を設定します.
     * @return String フルパス名が返されます.
     * @exception IOException
     *                I/O例外.
     */
    public static final String getFullPath(String path) throws IOException {
        File fp = new File(path);
        String s = fp.getCanonicalPath();
        s = StringUtil.changeString( s,"\\","/" ) ;
        if( !s.startsWith( "/" ) ) {
            s = "/" + s ;
        }
        return s ;
    }

    /**
     * 指定パスから、ディレクトリ名を取得. <br>
     * <br>
     * 指定パスから、ディレクトリ名を取得します.
     * 
     * @param path
     *            対象のパス名を設定します.
     * @return String 対象のディレクトリ名が返されます.
     */
    public static final String getDirName(String path) {
        int p = path.lastIndexOf("\\");
        int p2 = path.lastIndexOf("/");
        if (p == -1) {
            if (p2 != -1) {
                p = p2;
            }
        } else if (p2 != -1 && p < p2) {
            p = p2;
        }

        if (p == -1) {
            return path;
        }
        return path.substring(0,p);
    }

    /**
     * 指定パスから、ファイル名を取得. <BR>
     * <BR>
     * 指定パスから、ファイル名を取得します. <BR>
     * 
     * @param path
     *            対象のパス名を設定します.
     * @return String 対象のファイル名が返されます.
     */
    public static final String getFileName(String path) {
        int p = path.lastIndexOf("\\");
        int p2 = path.lastIndexOf("/");
        if (p == -1) {
            if (p2 != -1) {
                p = p2;
            }
        } else if (p2 != -1 && p < p2) {
            p = p2;
        }

        if (p == -1) {
            return path;
        }
        return path.substring(p + 1);
    }

    /**
     * 指定位置のファイル区切りを削除. <BR>
     * <BR>
     * 指定位置のファイル区切りを削除します. <BR>
     * 
     * @param mode
     *            対象の指定位置を設定します.<BR>
     *            [true]を設定した場合、先頭のファイル区切りを外します.
     * @param name
     *            対象のファイル名を設定します.
     * @return String 削除された情報が返されます.
     */
    public static final String cutSeparator(boolean mode, String name) {
        if (mode == true) {
            name = name.trim();
            if (name.startsWith("/") == true || name.startsWith("\\") == true) {
                return name.substring(1);
            }
        } else {
            name = name.trim();
            if (name.endsWith("/") == true || name.endsWith("\\") == true) {
                return name.substring(0, name.length() - 1);
            }
        }
        return name;
    }

    /**
     * 指定ディレクトリ以下の内容を全て抽出.
     * @param out 内容格納オブジェクトを設定します.
     * @param name 対象のディレクトリ名を設定します.
     * @return boolean [true]の場合、取得に成功しました.
     * @exception Exception 例外.
     */
    public static final boolean getInnerList( List<String> out,String name )
        throws Exception {
        if( out == null || name == null || ( name = name.trim() ).length() <= 0 ) {
            return false ;
        }
        if( !isDirExists( name ) ) {
            return false ;
        }
        name = getFullPath( name ) ;
        innerListTo( out,name ) ;
        return true ;
    }

    private static final void innerListTo( List<String> out,String name ) {
        if( !name.endsWith( "/" ) ) {
            name = name + "/" ;
        }
        int len ;
        String[] lst = getList( name ) ;
        if( lst == null || ( len = lst.length ) <= 0 ) {
            return ;
        }
        for( int i = 0 ; i < len ; i ++ ) {
            String f = name + lst[ i ] ;
            out.add( f ) ;
            if( isDirExists( f ) ) {
                innerListTo( out,f ) ;
            }
        }
    }

    /**
     * 指定ファイルを削除. <BR>
     * <BR>
     * 
     * @param name
     *            対象のファイル名を設定します.
     * @return boolean 削除結果が返されます.
     */
    public static final boolean removeFile(String name) {
        if( isFileExists( name ) && !isDirExists( name ) ) {
            return _remove( name ) ;
        }
        return false ;
    }

    /**
     * 指定ディレクトリを削除.
     * @param name 削除対象のディレクトリ名を設定します.
     * @return boolean 削除結果が返されます.
     */
    public static final boolean removeDirectory( String name ) {
        if( isDirExists( name ) ) {
            try {
                name = getFullPath( name ) ;
                if( !_remove( name ) ) {
                    if( _removeDirectoryTo( name ) ) {
                        return _remove( name ) ;
                    }
                    _remove( name ) ;
                }
                else {
                    return true ;
                }
            } catch( Exception e ) {
            }
        }
        return false ;
    }

    /** ファイル／ディレクトリ削除処理. **/
    private static final boolean _remove(String name) {
        if( name == null || ( name = name.trim() ).length() <= 0 ) {
            return false ;
        }
        boolean ret = false ;
        try {
            fsCache.clean( name ) ;
            File file = new File(name);
            ret = file.delete();
        } catch (Exception e) {
            ret = false ;
        }
        return ret ;
    }
    
    /** ディレクトリ内の内容を削除. **/
    private static final boolean _removeDirectoryTo( String name ) {
        if( !name.endsWith( "/" ) ) {
            name = name + "/" ;
        }
        int len ;
        String[] list = getList( name ) ;
        if( list != null && ( len = list.length ) > 0 ) {
            boolean ret = true ;
            for( int i = 0 ; i < len ; i ++ ) {
                String file = name + list[ i ] ;
                if( isDirExists( file ) ) {
                    if( !_removeDirectoryTo( file ) ) {
                        ret = false ;
                    }
                }
                if( !_remove( file ) ) {
                    ret = false ;
                }
            }
            return ret ;
        }
        return true ;
    }

    /**
     * ファイルコピー
     * @param src コピー元のファイル名を設定します.
     * @param dest コピー先のファイル名を設定します.
     * @return boolean [true]の場合、コピーに成功しました.
     */
    public static final boolean copyFile( String src,String dest ) {
        if( isFileExists( src ) && !isDirExists( src ) ) {
            return _copy( src,dest ) ;
        }
        return false ;
    }

    /**
     * ディレクトリコピー.
     * @param src コピー元のディレクトリ名を設定します.
     * @param dest コピー先のディレクトリ名を設定します.
     * @return boolean [true]の場合、コピーに成功しました.
     * @exception Exception 例外.
     */
    public static final boolean copyDirectory( String src,String dest )
        throws Exception {
        boolean ret = false ;
        if( isDirExists( src ) ) {
            ret = true ;
            if( !isDirExists( dest ) ) {
                mkdirs( dest ) ;
            }
            src = getFullPath( src ) ;
            dest = getFullPath( dest ) ;
            if( !src.endsWith( "/" ) ) {
                src += "/" ;
            }
            if( !dest.endsWith( "/" ) ) {
                dest += "/" ;
            }
            List<String> val = new ArrayList<String>() ;
            int len ;
            getInnerList( val,src ) ;
            if( ( len = val.size() ) > 0 ) {
                for( int i = 0 ; i < len ; i ++ ) {
                    String s = val.get( i ) ;
                    String v = s.substring( src.length() ) ;
                    String d = dest + v ;
                    if( !isDirExists( s ) ) {
                        if( isFileExists( d ) || !_copy( s,d ) ) {
                            ret = false ;
                        }
                    }
                    else if( isDirExists( s ) && !isDirExists( d ) ) {
                        mkdirs( d ) ;
                    }
                }
            }
        }
        return ret ;
    }

    /** ファイルのコピー処理 **/
    private static final boolean _copy( String src,String dest ) {
        if( src == null || ( src = src.trim() ).length() <= 0 ||
            dest == null || ( dest = dest.trim() ).length() <= 0 ) {
            return false ;
        }
        if( isFileExists( src ) && isFileExists( dest ) ) {
            return false ;
        }
        fsCache.clean( dest ) ;
        byte[] b = new byte[ 2048 ] ;
        InputStream in = null ;
        OutputStream out = null ;
        try {
            in = new BufferedInputStream( new FileInputStream( src ) ) ;
            out = new BufferedOutputStream( new FileOutputStream( dest ) ) ;
            int len = 0 ;
            while( true ) {
                if( ( len = in.read( b ) ) <= -1 ) {
                    break ;
                }
                out.write( b,0,len ) ;
            }
            out.flush() ;
            out.close() ;
            in.close() ;
            out = null ;
            in = null ;
            b = null ;
            return true ;
        } catch( Exception e ) {
        } finally {
            if( out != null ) {
                try {
                    out.close() ;
                } catch( Exception e ) {
                }
                _remove( dest ) ;
            }
            if( in != null ) {
                try {
                    in.close() ;
                } catch( Exception e ) {
                }
            }
        }
        return false ;
    }

    /**
     * 指定ファイルを移動. <BR>
     * <BR>
     * 
     * @param src
     *            移動元のファイル名を設定します.
     * @param dest
     *            移動先のファイル名を設定します.
     * @return boolean 成功した場合[true]が返されます.
     */
    public static final boolean moveFile(String src, String dest) {
        try {
            fsCache.clean( src ) ;
            fsCache.clean( dest ) ;
            File fileSrc = new File(src);
            File fileDest = new File(dest);
            return fileSrc.renameTo(fileDest);
        } catch (Exception e) {
        }
        return false ;
    }

    /**
     * 指定ファイル長を取得. <BR>
     * <BR>
     * 
     * @param name
     *            対象のファイル名を設定します.
     * @return long 対象のファイル長が返されます.
     */
    public static final long getLength(String name) {
        long ret = -1L;

        try {
            //File file = new File(name);
            //ret = (file.exists() == true) ? file.length() : -1L;
            ret = fsCache.getLength( name ) ;
        } catch (Exception e) {
            ret = -1L;
        }

        return ret;
    }

    /**
     * 指定ファイル日付を取得. <BR>
     * <BR>
     * 
     * @param name
     *            対象のファイル名を設定します.
     * @return long 対象のファイル日付が返されます.
     */
    public static final long getLastTime(String name) {
        long ret = 0L;

        try {
            //File file = new File(name);
            //ret = (file.exists() == true) ? file.lastModified() : 0L;
            ret = fsCache.getLastTime( name ) ;
        } catch (Exception e) {
            ret = 0L;
        }

        return ret;
    }

    /**
     * 指定ファイル日付を設定. <BR>
     * <BR>
     * 
     * @param name
     *            対象のファイル名を設定します.
     * @param lastTime
     *            対象の更新日付を設定します.
     * @return boolean 設定結果が返されます.
     */
    public static final boolean setLastTime(String name, long lastTime) {
        boolean ret = false;
        try {
            File file = new File(name);
            if (file.exists() == true) {
                ret = file.setLastModified(lastTime);
            }
        } catch (Exception e) {
            ret = false;
        }

        return ret;
    }

    /**
     * 指定ファイル名が存在するかチェック. <BR>
     * <BR>
     * 指定ファイル名が存在するかチェックします.<BR>
     * <BR>
     * 
     * @param name
     *            ファイル名を指定します.
     * @return boolean ファイル存在フラグが返されます.<BR>
     *         [true]が返された場合、指定したファイル名は存在します.<BR>
     *         [false]が返された場合、指定したファイル名は存在しません.
     */
    public static final boolean isFileExists(String name) {
        boolean ret;

        if (name == null || name.length() <= 0) {
            return false;
        }

        try {
            //File file = new File(name);
            //ret = (file.exists() == true) ? ((file.isDirectory() == false) ? file
            //        .canRead()
            //        : false)
            //        : false;
            ret = fsCache.isFileExists( name ) ;
        } catch (Exception e) {
            ret = false;
        }

        return ret;
    }

    /**
     * 指定ディレクトリ名が存在するかチェック. <BR>
     * <BR>
     * 指定ディレクトリ名が存在するかチェックします. <BR>
     * 
     * @param dirName
     *            ディレクトリ名を指定します.
     * @return boolean ディレクトリ存在フラグが返されます.<BR>
     *         [true]が返された場合、指定したディレクトリ名は存在します.<BR>
     *         [false]が返された場合、指定したディレクトリ名は存在しません.
     */
    public static final boolean isDirExists(String dirName) {
        boolean ret;

        if (dirName == null) {
            return false;
        }

        try {
            //File check = new File(dirName);
            //ret = check.isDirectory();
            ret = fsCache.isDirExists( dirName ) ;
        } catch( Exception e ) {
            ret = false ;
        }

        return ret;
    }

    /**
     * 指定ファイル名が読み取り可能かチェック. <BR>
     * <BR>
     * 指定ファイル名が読み取り可能かチェックします.<BR>
     * <BR>
     * 
     * @param name
     *            読み取り可能チェック対象のディレクトリ／ファイル名を 指定します.
     * @return boolean 読み取り可能か否かが返されます.<BR>
     *         [true]が返された場合、読み取り可能です.<BR>
     *         [false]が返された場合、読み取り不可能か存在しません.
     */
    public static final boolean isRead(String name) {
        boolean ret;
        if (name == null || name.length() <= 0) {
            return false;
        }
        try {
            //File file = new File(name);
            //ret = file.canRead();
            ret = fsCache.isRead( name ) ;
        } catch (Exception e) {
            ret = false;
        }
        return ret;
    }

    /**
     * 指定ファイル名が書き込み可能かチェック. <BR>
     * <BR>
     * 指定ファイル名が書き込み可能かチェックします. <BR>
     * 
     * @param name
     *            書き込み可能チェック対象のディレクトリ／ファイル名を 指定します.
     * @return boolean 書き込み可能か否かが返されます.<BR>
     *         [true]が返された場合、書き込み可能です.<BR>
     *         [false]が返された場合、書き込み不可能か存在しません.
     */
    public static final boolean isWrite(String name) {
        boolean ret;

        if (name == null) {
            return false;
        }

        try {
            //File check = new File(name);
            //ret = check.canWrite();
            ret = fsCache.isWrite( name ) ;
        } catch( Exception e ) {
            ret = false ;
        }

        return ret;
    }

    /**
     * リスト条件に対して文字内容が一致するかチェック.
     * @param start [true]の場合、startsWithでチェックします.
     * @param w チェック対象ワードを設定します.
     * @param lst チェックするリスト条件を設定します.
     * @return boolean [true]の場合、一致します.
     */
    public static final boolean isListWith( boolean start,String w,List<String> lst ) {
        if( lst == null ) {
            return false ;
        }
        int len = lst.size() ;
        if( start ) {
            for( int i = 0 ; i < len ; i ++ ) {
                if( w.startsWith( lst.get( i ) ) ) {
                    return true ;
                }
            }
        }
        else {
            for( int i = 0 ; i < len ; i ++ ) {
                if( w.endsWith( lst.get( i ) ) ) {
                    return true ;
                }
            }
        }
        return false ;
    }

    /**
     * リスト条件に対して文字内容が一致するかチェック.
     * @param w チェック対象ワードを設定します.
     * @param lst チェックするリスト条件を設定します.
     * @return boolean [true]の場合、一致します.
     */
    public static final boolean isListEquals( String w,List<String> lst ) {
        if( lst == null ) {
            return false ;
        }
        int len = lst.size() ;
        for( int i = 0 ; i < len ; i ++ ) {
            if( w.equals( lst.get( i ) ) ) {
                return true ;
            }
        }
        return false ;
    }

    /**
     * 指定ディレクトリ以下のファイル/ディレクトリ名一覧を取得. <BR>
     * <BR>
     * 指定ディレクトリ以下のファイル/ディレクトリ名一覧を取得します. <BR>
     * 
     * @param dir
     *            対象ディレクトリ名を設定します.
     * @param head
     *            マスクヘッダ名を設定します.
     * @param foot
     *            マスクフッダ名を設定します.
     * @param full
     *            除外フルネームを設定します.
     * @return int ファイル/ディレクトリ名数が返されます.
     */
    public static final int getListLength(String dir,List<String> head,List<String> foot,List<String> full) {
        if (dir == null || dir.length() <= 0) {
            return 0;
        }
        if (dir.endsWith("/") == true || dir.endsWith("\\") == true) {
            dir.substring(0, dir.length() - 1);
        }
        dir += FILE_SPACE;
        try {
            File fp = new File(dir);
            String[] lst = fp.list() ;
            if( lst == null ) {
                return 0 ;
            }
            else {
                int len = lst.length ;
                if( head == null && foot == null && full == null ) {
                    return len ;
                }
                int cnt = 0 ;
                if( head != null && foot != null && full != null ) {
                    for( int i = 0 ; i < len ; i ++ ) {
                        if( !isListWith( true,lst[ i ],head ) && !isListWith( false,lst[ i ],foot ) && !isListEquals( lst[ i ],full ) ) {
                            cnt ++ ;
                        }
                    }
                    return cnt ;
                }
                if( head == null ) {
                    if( full != null ) {
                        for( int i = 0 ; i < len ; i ++ ) {
                            if( !isListWith( false,lst[ i ],foot ) && !isListEquals( lst[ i ],full ) ) {
                                cnt ++ ;
                            }
                        }
                    }
                    else {
                        for( int i = 0 ; i < len ; i ++ ) {
                            if( !isListWith( false,lst[ i ],foot ) ) {
                                cnt ++ ;
                            }
                        }
                    }
                    return cnt ;
                }
                else {
                    if( full != null ) {
                        for( int i = 0 ; i < len ; i ++ ) {
                            if( !isListWith( true,lst[ i ],head ) && !isListEquals( lst[ i ],full ) ) {
                                cnt ++ ;
                            }
                        }
                    }
                    else {
                        for( int i = 0 ; i < len ; i ++ ) {
                            if( !isListWith( true,lst[ i ],head ) ) {
                                cnt ++ ;
                            }
                        }
                    }
                    return cnt ;
                }
            }
        } catch (Exception e) {
        }
        return 0;
    }

    /**
     * 指定ディレクトリ以下のファイル/ディレクトリ名一覧を取得. <BR>
     * <BR>
     * 指定ディレクトリ以下のファイル/ディレクトリ名一覧を取得します. <BR>
     * 
     * @param dir
     *            対象ディレクトリ名を設定します.
     * @return String[] ファイル/ディレクトリ名一覧が返されます.
     */
    public static final String[] getList(String dir) {
        if (dir == null || dir.length() <= 0) {
            return null;
        }
        if (dir.endsWith("/") == true || dir.endsWith("\\") == true) {
            dir.substring(0, dir.length() - 1);
        }
        dir += FILE_SPACE;
        try {
            File fp = new File(dir);
            return fp.list();
        } catch (Exception e) {
        }
        return null;
    }

    /**
     * 指定ディレクトリ以下のファイル/ディレクトリ名一覧を取得. <BR>
     * <BR>
     * 指定ディレクトリ以下のファイル/ディレクトリ名一覧を取得します. <BR>
     * 
     * @param dir
     *            対象ディレクトリ名を設定します.
     * @return ArrayList<String> ファイル/ディレクトリ名一覧が返されます.
     */
    public static final ArrayList<String> getListByArray(String dir) {
        String[] lst = getList( dir ) ;
        if( lst != null && lst.length > 0 ) {
            int len = lst.length ;
            ArrayList<String> ret = new ArrayList<String>() ;
            for( int i = 0 ; i < len ; i ++ ) {
                ret.add( lst[ i ] ) ;
            }
            return ret ;
        }
        return null ;
    }

    /**
     * 指定ディレクトリ以下のファイル名一覧を取得. <BR>
     * <BR>
     * 指定ディレクトリ以下のファイル名一覧を取得します. <BR>
     * 
     * @param dir
     *            対象ディレクトリ名を設定します.
     * @return String[] ファイル名一覧が返されます.
     */
    public static final String[] getFileList(String dir) {
        if (dir == null || dir.length() <= 0) {
            return null;
        }
        if (dir.endsWith("/") == true || dir.endsWith("\\") == true) {
            dir.substring(0, dir.length() - 1);
        }
        dir += FILE_SPACE;
        try {
            File fp = new File(dir);
            String[] names = fp.list();
            int len;
            if (names != null && (len = names.length) > 0) {
                for (int i = 0; i < len; i++) {
                    File f = new File(dir + names[i]);
                    if (f.exists() == false || f.isDirectory() == true) {
                        names[i] = null;
                    }
                }
            }
            return names;
        } catch (Exception e) {
        }
        return null;
    }

    /**
     * 指定ディレクトリ以下のファイル名一覧を取得. <BR>
     * <BR>
     * 指定ディレクトリ以下のファイル名一覧を取得します. <BR>
     * 
     * @param dir
     *            対象ディレクトリ名を設定します.
     * @return String[] ファイル名一覧が返されます.
     */
    public static final String[] getFileDirList(String dir) {
        if (dir == null || dir.length() <= 0) {
            return null;
        }
        if (dir.endsWith("/") == true || dir.endsWith("\\") == true) {
            dir.substring(0, dir.length() - 1);
        }
        dir += FILE_SPACE;
        try {
            File fp = new File(dir);
            String[] names = fp.list();
            int len;
            if (names != null && (len = names.length) > 0) {
                for (int i = 0; i < len; i++) {
                    File f = new File(dir + names[i]);
                    if (f.canRead() == false) {
                        names[i] = null;
                    }
                }
            }
            return names;
        } catch (Exception e) {
        }
        return null;
    }

    /**
     * 指定ディレクトリ以下のファイル名一覧を取得. <BR>
     * <BR>
     * 指定ディレクトリ以下のファイル名一覧を取得します. <BR>
     * 
     * @param dir
     *            対象ディレクトリ名を設定します.
     * @return ArrayList<String> ファイル名一覧が返されます.
     */
    public static final ArrayList<String> getFileListByArray(String dir) {
        String[] lst = useStringArray( getFileList( dir ) ) ;
        if( lst != null && lst.length > 0 ) {
            int len = lst.length ;
            ArrayList<String> ret = new ArrayList<String>() ;
            for( int i = 0 ; i < len ; i ++ ) {
                ret.add( lst[ i ] ) ;
            }
            return ret ;
        }
        return new ArrayList<String>() ;
    }
    
    /**
     * 指定ディレクトリ以下のファイルのみ一覧で取得. <BR>
     * 
     * @param dir
     *            対象ディレクトリ名を設定します.
     * @return ArrayList<String> ファイル名一覧が返されます.
     */
    public static final ArrayList<String> getOnlyFileToArray(String dir) {
        String[] lst = useStringArray( getFileList( dir ) ) ;
        if( lst != null && lst.length > 0 ) {
            Arrays.sort( lst ) ;
            int len = lst.length ;
            ArrayList<String> ret = new ArrayList<String>() ;
            for( int i = 0 ; i < len ; i ++ ) {
                if( lst[ i ] != null ) {
                    ret.add( lst[ i ] ) ;
                }
            }
            return ret ;
        }
        return new ArrayList<String>() ;
    }

    /** 有効な文字配列のみを取得. **/
    private static final String[] useStringArray( String[] lst ) {
        int len ;
        if( lst == null || ( len = lst.length ) <= 0 ) {
            return null ;
        }
        int cnt = 0 ;
        for( int i = 0 ; i < len ; i ++ ) {
            if( lst[ i ] != null ) {
                cnt ++ ;
            }
        }
        String[] ret = new String[ cnt ] ;
        cnt = 0 ;
        for( int i = 0 ; i < len ; i ++ ) {
            if( lst[ i ] != null ) {
                ret[ cnt ] = lst[ i ] ;
                cnt ++ ;
            }
        }
        return ret ;
    }

    
    /**
     * 指定ディレクトリ以下のすべての内容を取得.<BR>
     * <BR>
     * 指定ディレクトリ以下のすべての内容を取得します.<BR>
     *
     * @param dir 対象のディレクトリを設定します.
     * @return List<String> すべてのファイル名一覧が返されます.
     * @exception Exception 例外.
     */
    public static final List<String> getListAll(String dir)
        throws Exception {
        if(dir == null || ( dir = dir.trim() ).length() <= 0 ) {
            return new ArrayList<String>() ;
        }
        dir = getFullPath( dir ) ;
        ArrayList<String> ret = new ArrayList<String>() ;
        privateAllList( ret,dir,dir ) ;
        return ret ;
    }
    private static final void privateAllList( ArrayList<String> out,String base,String dir )
        throws Exception {
        if (dir.endsWith("/") == true || dir.endsWith("\\") == true) {
            dir.substring(0, dir.length() - 1);
        }
        dir += FILE_SPACE;
        File fp = new File(dir);
        String[] names = fp.list();
        int len;
        if (names != null && (len = names.length) > 0) {
            for (int i = 0; i < len; i++) {
                String s = dir + names[i] ;
                File f = new File( s );
                if (f.exists() == false) {
                    continue ;
                }
                if (f.isDirectory() == true) {
                    privateAllList( out,base,s ) ;
                }
                else {
                    out.add( s.substring( base.length() ) ) ;
                }
            }
        }
    }

    /**
     * 指定ディレクトリ内の最終更新日付を取得. <BR>
     * <BR>
     * 指定ディレクトリ内の最終更新日付を取得します. <BR>
     * 
     * @param name
     *            対象のディレクトリ名を設定します.
     * @return long ディレクトリ内の情報に対する最終更新日が返されます.
     */
    public static final long getDirectoryToLastTime(String name) {
        fsCache.clean( name ) ;
        if (isDirExists(name) == false) {
            return -1L;
        }
        return lastUpdateByDirectory(name, -1L);
    }

    /**
     * 指定ディレクトリ内の指定時間より後の更新日付のファイル名を取得. <BR>
     * <BR>
     * 指定ディレクトリ内の指定時間より後の更新日付のファイル名を取得します. <BR>
     * 
     * @param out
     *            ファイル名を格納するオブジェクトを設定します.
     * @param name
     *            対象のディレクトリ名を設定します.
     * @param targetTime
     *            ターゲットの日付を設定します.
     */
    public static final void getDirectoryToTargetNextTime(
            ArrayList<String> out, String name, long targetTime) {
        fsCache.clean( name ) ;
        if (isDirExists(name) == false) {
            return;
        }
        targetNextUpdateByDirectory(out, name, targetTime);
    }

    /**
     * 指定ファイル内容をバイナリ情報として取得. <BR>
     * <BR>
     * 指定されたファイルの内容をバイナリ情報として取得します.<BR>
     * <BR>
     * 
     * @param name
     *            取得対象のファイル名を設定します.
     * @return byte[] 指定ファイル内容が返されます.
     * @exception Exception
     *                例外.
     */
    public static final byte[] getFile(String name) throws Exception {
        int len;
        InputStream buf = null;
        ByteArrayOutputStream bo = null;
        byte[] b = new byte[ 512 ] ;
        try {
            bo = new ByteArrayOutputStream() ;
            buf = new BufferedInputStream( new FileInputStream( name ) ) ;
            while( true ) {
                if( ( len = buf.read(b) ) <= 0 ) {
                    if( len <= -1 ) {
                        break ;
                    }
                    continue ;
                }
                bo.write( b,0,len ) ;
            }
            buf.close() ;
            buf = null ;
        } finally {
            if( buf != null ) {
                try {
                    buf.close();
                } catch (Exception t) {
                }
            }
            buf = null;
        }
        return bo.toByteArray() ;
    }

    /**
     * 指定ファイル内容を文字列情報として取得. <BR>
     * <BR>
     * 指定されたファイルの内容を文字列情報として取得します.<BR>
     * <BR>
     * 
     * @param name
     *            取得対象のファイル名を設定します.
     * @return String 取得された情報が返されます.
     * @exception Exception
     *                例外.
     */
    public static final String getFileByString(String name) throws Exception {
        return FileUtil.getFileByString(name, CHARSET);
    }

    /**
     * 指定ファイル内容を文字列情報として取得. <BR>
     * <BR>
     * 指定されたファイルの内容を文字列情報として取得します. <BR>
     * 
     * @param name
     *            取得対象のファイル名を設定します.
     * @param charset
     *            取得対象のキャラクターセットを設定します.
     * @return String 取得された情報が返されます.
     * @exception Exception
     *                例外.
     */
    public static final String getFileByString(String name, String charset)
            throws Exception {
        int len;
        char[] tmp = new char[512];
        CharArrayWriter ca = null;
        Reader buf = null;
        String ret = null;
        try {
            ca = new CharArrayWriter();
            buf = new BufferedReader(
                new InputStreamReader( new FileInputStream( name ),charset) ) ;
            while ((len = buf.read(tmp, 0, 512)) > 0) {
                ca.write(tmp, 0, len);
            }
            ret = ca.toString();
            ca.close();
            ca = null;
            buf.close() ;
            buf = null ;
        } finally {
            if( buf != null ) {
                try {
                    buf.close();
                } catch (Exception t) {
                }
            }
            if( ca != null ) {
                try {
                    ca.close();
                } catch (Exception t) {
                }
            }
            buf = null;
            ca = null;
            tmp = null;
        }
        return ret;
    }

    /**
     * 指定ファイル名に対象のバイナリ情報を出力. <BR>
     * <BR>
     * 指定されたファイル名に対してバイナリ情報を出力します. <BR>
     * 
     * @param name
     *            出力対象のファイル名を設定します.
     * @param binary
     *            出力対象のバイナリ情報を設定します.
     * @exception Exception
     *                例外.
     */
    public static final void setFile(String name, byte[] binary)
            throws Exception {
        FileUtil.setFile(name, true, binary);
    }

    /**
     * 指定ファイル名に対象のバイナリ情報を出力. <BR>
     * <BR>
     * 指定されたファイル名に対してバイナリ情報を出力します. <BR>
     * 
     * @param name
     *            出力対象のファイル名を設定します.
     * @param mode
     *            書き込みモードを設定します.<BR>
     *            [true]を設定した場合、既に同一のファイルが存在しても上書きします.<BR>
     *            [false]を設定した場合、既に同一のファイルが存在した場合後に書き込みます.
     * @param binary
     *            出力対象のバイナリ情報を設定します.
     * @exception Exception
     *                例外.
     */
    public static final void setFile(String name, boolean mode, byte[] binary)
            throws Exception {
        OutputStream buf = null;
        if (name == null || binary == null) {
            throw new IllegalArgumentException("引数は不正です");
        }
        fsCache.clean( name ) ;
        FileOutputStream o = null ;
        try {
            if( mode ) {
                o = new FileOutputStream( name ) ;
            }
            else {
                o = new FileOutputStream( name,true ) ;
            }
            buf = new BufferedOutputStream( new FileOutputStream( name ) ) ;
            buf.write(binary);
            buf.flush() ;
            //o.flush() ;
            //o.getFD().sync() ;//書き込み同期を取る.
            buf.close() ;
            o.close() ;
            buf = null ;
            o = null ;
        } finally {
            if( buf != null ) {
                try {
                    buf.close();
                } catch (Exception t) {
                }
            }
            if( o != null ) {
                try {
                    o.close() ;
                } catch( Exception e ) {
                }
            }
        }
    }

    /**
     * 指定ファイル名に対象の文字列情報を出力. <BR>
     * <BR>
     * 指定されたファイル名に対して文字列情報を出力します. <BR>
     * 
     * @param name
     *            出力対象のファイル名を設定します.
     * @param value
     *            出力対象の文字列情報を設定します.
     * @exception Exception
     *                例外.
     */
    public static final void setFileByString(String name, String value)
            throws Exception {
        FileUtil.setFileByString(name, true, value, CHARSET);
    }

    /**
     * 指定ファイル名に対象の文字列情報を出力. <BR>
     * <BR>
     * 指定されたファイル名に対して文字列情報を出力します. <BR>
     * 
     * @param name
     *            出力対象のファイル名を設定します.
     * @param mode
     *            書き込みモードを設定します.<BR>
     *            [true]を設定した場合、既に同一のファイルが存在しても上書きします.<BR>
     *            [false]を設定した場合、既に同一のファイルが存在した場合後に書き込みます.
     * @param value
     *            出力対象の文字列情報を設定します.
     * @exception Exception
     *                例外.
     */
    public static final void setFileByString(String name, boolean mode,
            String value) throws Exception {
        FileUtil.setFileByString(name, mode, value, CHARSET);
    }

    /**
     * 指定ファイル名に対象の文字列情報を出力. <BR>
     * <BR>
     * 指定されたファイル名に対して文字列情報を出力します. <BR>
     * 
     * @param name
     *            出力対象のファイル名を設定します.
     * @param mode
     *            書き込みモードを設定します.<BR>
     *            [true]を設定した場合、既に同一のファイルが存在しても上書きします.<BR>
     *            [false]を設定した場合、既に同一のファイルが存在した場合後に書き込みます.
     * @param value
     *            出力対象の文字列情報を設定します.
     * @param charset
     *            出力対象のキャラクターセットを設定します.
     * @exception Exception
     *                例外.
     */
    public static final void setFileByString(String name, boolean mode,
            String value, String charset) throws Exception {
        BufferedWriter buf = null;
        if (name == null || value == null) {
            throw new IllegalArgumentException("引数は不正です");
        }
        fsCache.clean( name ) ;
        FileOutputStream o = null ;
        try {
            if( mode ) {
                o = new FileOutputStream( name ) ;
            }
            else {
                o = new FileOutputStream( name,true ) ;
            }
            buf = new BufferedWriter(new OutputStreamWriter(o, charset));
            buf.write(value);
            buf.flush() ;
            //o.flush() ;
            //o.getFD().sync() ;//書き込み同期を取る.
            buf.close() ;
            o.close() ;
            buf = null ;
            o = null ;
        } finally {
            if( buf != null ) {
                try {
                    buf.close();
                } catch (Exception t) {
                }
                buf = null;
            }
            if( o != null ) {
                try {
                    o.close() ;
                } catch( Exception e ) {
                }
            }
        }
    }

    /**
     * 空のファイルを作成.
     * @param name 対象のファイル名を設定します.
     * @exception Exception 例外.
     */
    public static final void createFile( String name )
        throws Exception {
        if (name == null) {
            throw new IllegalArgumentException("引数は不正です");
        }
        fsCache.clean( name ) ;
        FileOutputStream out = null ;
        try {
            out = new FileOutputStream( name ) ;
            out.close() ;
            out = null ;
        } finally {
            if( out != null ) {
                try {
                    out.close() ;
                } catch( Exception e ) {
                }
            }
        }
    }

    /**
     * 指定リソース内容をバイナリ情報として取得. <BR>
     * <BR>
     * 指定されたリソースの内容をバイナリ情報として取得します.<BR>
     * <BR>
     * 
     * @param name
     *            取得対象のリソース名を設定します.
     * @return byte[] 指定リソース内容が返されます.
     * @exception Exception
     *                例外.
     */
    public static final byte[] getFileByResource(String name) throws Exception {
        BufferedInputStream buf = null;
        byte[] ret = null;
        try {
            ClassLoader cl = Thread.currentThread().getContextClassLoader();
            buf = new BufferedInputStream(cl.getResourceAsStream(name));
            ByteArrayOutputStream bs = new ByteArrayOutputStream() ;
            for( ;; ) {
                int n = buf.read() ;
                if( n <= -1 ) {
                    break ;
                }
                bs.write( n ) ;
            }
            ret = bs.toByteArray() ;
            bs.close() ;
            bs = null ;
        } finally {
            try {
                buf.close();
            } catch (Exception t) {
            }
            buf = null;
        }
        return ret;
    }

    /**
     * 指定パスのスクリプトを取得. <BR>
     * <BR>
     * 指定パスのスクリプトを取得します. <BR>
     * 
     * @param path
     *            対象のスクリプトパスを設定します.
     * @return String スクリプト内容が返されます.
     * @exception Exception
     *                例外.
     */
    public static String getScriptByResource(String path, String charset)
            throws Exception {
        if (charset == null || charset.length() <= 0) {
            charset = "UTF8";
        }
        ClassLoader cl = Thread.currentThread().getContextClassLoader();
        BufferedReader br = null;
        try {
            StringBuilder buf = new StringBuilder();
            br = new BufferedReader( new InputStreamReader(cl
                    .getResourceAsStream(path), charset));
            while (true) {
                String s = br.readLine();
                if (s == null) {
                    break;
                }
                buf.append(s).append("\r\n");
            }
            br.close();
            br = null;
            return buf.toString();
        } catch (Exception e) {
            throw e;
        } finally {
            if (br != null) {
                try {
                    br.close();
                } catch (Exception e) {
                }
            }
        }
    }

    /**
     * オブジェクトをシリアライズ.
     * @param name ファイル名を設定します.
     * @param value シリアライズ対象オブジェクトを設定します.
     * @exception Exception 例外.
     */
    public static final void encodeSerializable( String name,Object value )
        throws Exception {
        if( name == null || ( name = name.trim() ).length() <= 0 || value == null ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        fsCache.clean( name ) ;
        ObjectOutputStream o = null ;
        try{
            o = new ObjectOutputStream( new FileOutputStream( name ) ) ;
            o.writeObject( value ) ;
            o.flush() ;
            o.close() ;
            o = null ;
        } finally {
            if( o != null ) {
                try {
                    o.close() ;
                } catch( Exception e ) {
                }
            }
        }
    }
    
    /**
     * オブジェクトを復元.
     * @param name ファイル名を設定します.
     * @exception Exception 例外.
     */
    public static final Object decodeSerializable( String name )
        throws Exception {
        if( name == null || ( name = name.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        ObjectInputStream in = null;
        Object ret = null ;
        try{
            in = new ObjectInputStream( new FileInputStream( name ) ) ;
            ret = in.readObject() ;
            in.close() ;
            in = null ;
        }finally{
            if( in != null ) {
                try{
                    in.close() ;
                }catch( Exception e ){
                }
                in = null ;
            }
        }
        return ret ;
    }

    /**
     * ディレクトリ内の最終時間を取得.
     */
    private static final long lastUpdateByDirectory(String name, long ret) {

        if (name.endsWith("\\") == false && name.endsWith("/") == false) {
            name = name + FILE_SPACE;
        }

        File file = new File(name);
        String[] names = file.list();
        file = null;
        int len;
        if (names != null && (len = names.length) > 0) {
            for (int i = 0; i < len; i++) {
                String target = name + names[i];
                names[i] = null;
                File fp = new File(target);

                if (fp.isDirectory() == true) {
                    fp = null;
                    ret = lastUpdateByDirectory(target, ret);
                } else {
                    long now = fp.lastModified();
                    fp = null;
                    if (now > ret) {
                        ret = now;
                    }
                }
            }
        }

        return ret;
    }

    /**
     * 指定ディレクトリ以下にある、ターゲット時間より後のファイル日付の名前を取得.
     */
    private static final void targetNextUpdateByDirectory(
            ArrayList<String> lst, String name, long targetTime) {

        if (name.endsWith("\\") == false && name.endsWith("/") == false) {
            name = name + FILE_SPACE;
        }

        File file = new File(name);
        String[] names = file.list();
        file = null;
        int len;
        if (names != null && (len = names.length) > 0) {
            for (int i = 0; i < len; i++) {
                String target = name + names[i];
                names[i] = null;
                File fp = new File(target);

                if (fp.isDirectory() == true) {
                    fp = null;
                    targetNextUpdateByDirectory(lst, target, targetTime);
                } else {
                    long now = fp.lastModified();
                    fp = null;
                    if (now > targetTime) {
                        lst.add(new String(target));
                    }
                }
            }
        }

    }

}
