package org.maachang.util ;

import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.concurrent.ConcurrentHashMap;

import org.maachang.util.atomic.AtomicBOOL;

/**
 * ファイルステータスキャッシュ.
 * 
 * @version 2009/01/17
 * @author masahito suzuki
 * @since MaachangBase 1.0A
 */
public class FsCache {
    
    private static class CacheFsChild {
        private long createCacheTime = -1L ;
        private boolean exists = false ;
        private boolean dir = false ;
        private boolean read = false ;
        private boolean write = false ;
        private long lastTime = -1L ;
        private long length = -1L ;
    }
    
    private final CacheFsChild load( CacheFsChild ch,String name ) throws IOException {
        File f = new File(name);
        if( f.exists() ) {
            CacheFsChild c = null ;
            if( ch != null ) {
                c = ch ;
            }
            else {
                c = new CacheFsChild() ;
            }
            c.exists = f.exists() ;
            if( c.exists ) {
                c.dir = f.isDirectory() ;
                c.read = f.canRead() ;
                c.write = f.canWrite() ;
                c.lastTime = f.lastModified() ;
                c.length = f.length() ;
            }
            else {
                c.dir = false ;
                c.read = false ;
                c.write = false ;
                c.lastTime = -1L ;
                c.length = -1L ;
            }
            c.createCacheTime = System.currentTimeMillis() ;
            return c ;
        }
        return null ;
    }
    
    private final long CACHE_TIMEOUT = 5000L ;
    private static final FsCache SNGL = new FsCache() ;
    
    private final ConcurrentHashMap<String,CacheFsChild> map = new ConcurrentHashMap<String,CacheFsChild>() ;
    private CleanFsCache mon = null ;
    
    public static final FsCache getInstance() {
        return SNGL ;
    }
    
    public FsCache() {
        mon = new CleanFsCache( map ) ;
    }
    
    protected void finalize() throws Exception {
        destroy() ;
    }
    
    public void destroy() {
        mon.stopThread() ;
        map.clear() ;
    }
    
    private final CacheFsChild getCache( String name ) throws IOException {
        CacheFsChild ch = map.get( name ) ;
        if( ch == null || ch.createCacheTime + CACHE_TIMEOUT < System.currentTimeMillis() ) {
            ch = load( ch,name ) ;
            if( ch != null ) {
                map.put( name,ch ) ;
            }
        }
        return ch ;
    }
    
    public final void clean( String name ) {
        map.remove( name ) ;
    }
    
    public final boolean isExists( String name ) throws IOException {
        CacheFsChild ch = getCache( name ) ;
        if( ch != null ) {
            return ch.exists ;
        }
        return false ;
    }
    
    public final boolean isFileExists( String name ) throws IOException {
        CacheFsChild ch = getCache( name ) ;
        if( ch != null ) {
            return ch.exists && !ch.dir && ch.read && ch.write ;
        }
        return false ;
    }
    
    public final boolean isDirExists( String name ) throws IOException {
        CacheFsChild ch = getCache( name ) ;
        if( ch != null ) {
            return ch.exists && ch.dir && ch.read && ch.write ;
        }
        return false ;
    }
    
    public final boolean isRead( String name ) throws IOException {
        CacheFsChild ch = getCache( name ) ;
        if( ch != null ) {
            return ch.read ;
        }
        return false ;
    }
    
    public final boolean isWrite( String name ) throws IOException {
        CacheFsChild ch = getCache( name ) ;
        if( ch != null ) {
            return ch.write ;
        }
        return false ;
    }
    
    public final long getLastTime( String name ) throws IOException {
        CacheFsChild ch = getCache( name ) ;
        if( ch != null ) {
            return ch.lastTime ;
        }
        return -1L ;
    }
    
    public final long getLength( String name ) throws IOException {
        CacheFsChild ch = getCache( name ) ;
        if( ch != null ) {
            return ch.length ;
        }
        return -1L ;
    }
    
    
    private static class CleanFsCache extends Thread {
        private static final long DELETE_TIME = 30000L ;
        private final AtomicBOOL stopFlag = new AtomicBOOL( false ) ;
        private ConcurrentHashMap<String,CacheFsChild> map = null ;
        
        private CleanFsCache() {
            
        }
        
        public CleanFsCache( ConcurrentHashMap<String,CacheFsChild> map ) {
            this.map = map ;
            super.setDaemon( true ) ;
            start() ;
        }
        
        public void stopThread() {
            stopFlag.set( true ) ;
        }
        
        public void run() {
            boolean endFlag = false ;
            ThreadDeath threadDeach = null ;
            while( true ) {
                if( endFlag || stopFlag.get() ) {
                    break ;
                }
                try {
                    sleep( WAIT_2500 ) ;
                    execution() ;
                } catch( InterruptedException ite ) {
                    endFlag = true ;
                } catch ( NullPointerException ne ) {
                    try {
                        Thread.sleep( 1L ) ;
                    }
                    catch( InterruptedException ite2 ) {
                        endFlag = true ;
                    }
                } catch( OutOfMemoryError mem ) {
                    try {
                        Thread.sleep( 1L ) ;
                    }
                    catch( InterruptedException ite2 ) {
                        endFlag = true ;
                    }
                } catch( Exception e ) {
                    try {
                        Thread.sleep( 1L ) ;
                    }
                    catch( InterruptedException ite2 ) {
                        endFlag = true ;
                    }
                } catch( ThreadDeath td ) {
                    endFlag = true ;
                    threadDeach = td ;
                }
            }
            if( threadDeach != null ) {
                throw threadDeach ;
            }
        }
        
        private void execution() throws Exception {
            long time = System.currentTimeMillis() ;
            Iterator it = map.keySet().iterator() ;
            while( it.hasNext() ) {
                try {
                    sleep( WAIT ) ;
                    time += WAIT ;
                    String key = ( String )it.next() ;
                    CacheFsChild ch = map.get( key ) ;
                    if( ch.createCacheTime + DELETE_TIME < time ) {
                        map.remove( key ) ;
                    }
                } catch( InterruptedException ie ) {
                    throw ie ;
                } catch( Exception e ) {
                    sleep( WAIT ) ;
                    time += WAIT ;
                }
            }
        }
        
        private static final int WAIT_2500 = 2500 ;
        private static final long WAIT = 50L ;
    }
}

