package org.maachang.util ;

import java.io.InputStream;
import java.util.List;

/**
 * コマンド実行処理.
 * 
 * @version 2008/11/01
 * @author masahito suzuki
 * @since MaachangBase 1.08
 */
public class RunProcess {
    
    /**
     * 実行コマンド.
     */
    private String[] cmd = null ;
    
    /**
     * 実行プロセス.
     */
    private Process process = null ;
    
    /**
     * プロセス実行されたかチェック.
     */
    private boolean executionFlag = false ;
    
    /**
     * コンストラクタ.
     */
    private RunProcess() {
        
    }
    
    /**
     * コンストラクタ.
     * @param cmd 実行コマンドを設定します.
     */
    public RunProcess( String cmd ) {
        if( cmd == null || cmd.length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        this.cmd = cutCommand( cmd ) ;
        if( this.cmd == null ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
    }
    
    /**
     * コンストラクタ.
     * @param cmd 実行コマンドを設定します.
     */
    public RunProcess( String[] cmd ) {
        if( cmd == null || cmd.length <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        this.cmd = cmd ;
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        destroy() ;
    }
    
    /**
     * オブジェクト破棄.
     */
    public void destroy() {
        if( process != null ) {
            try {
                process.getInputStream().close() ;
            } catch( Exception e ) {
            }
            try {
                process.getOutputStream().close() ;
            } catch( Exception e ) {
            }
            try {
                process.getErrorStream().close() ;
            } catch( Exception e ) {
            }
            try {
                process.destroy() ;
            } catch( Exception e ) {
            }
        }
        cmd = null ;
        process = null ;
        executionFlag = false ;
    }
    
    /**
     * コマンド分割.
     */
    private static final String[] cutCommand( String cmd ) {
        if( cmd == null || cmd.length() <= 0 ) {
            return null ;
        }
        List<String> c = StringUtil.cutString( cmd," \t" ) ;
        if( c != null && c.size() > 0 ) {
            int len = c.size() ;
            String[] ret = new String[ len ] ;
            for( int i = 0 ; i < len ; i ++ ) {
                ret[ i ] = c.get( i ) ;
            }
            return ret ;
        }
        return null ;
    }
    
    /**
     * コマンド設定.
     * @param cmd 対象のコマンドを設定します.
     */
    public void setCommand( String cmd ) {
        if( cmd == null || cmd.length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        this.destroy() ;
        this.cmd = cutCommand( cmd ) ;
        if( this.cmd == null ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
    }
    
    /**
     * コマンド設定.
     * @param cmd 対象のコマンドを設定します.
     */
    public void setCommand( String[] cmd ) {
        if( cmd == null || cmd.length <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        this.destroy() ;
        this.cmd = cmd ;
    }
    
    /**
     * コマンド取得.
     * @return String[] 設定されているコマンドが返されます.
     */
    public String[] getCommand() {
        return this.cmd ;
    }
    
    /**
     * コマンド実行.
     * @exception Exception 例外.
     */
    public void execution() throws Exception {
        if( cmd == null || cmd.length <= 0 ) {
            throw new IllegalArgumentException( "コマンドが設定されていません" ) ;
        }
        try {
            this.process = Runtime.getRuntime().exec(cmd) ;
            this.executionFlag = true ;
        } catch( Exception e ) {
            destroy() ;
            throw e ;
        }
    }
    
    /**
     * 正常ストリーム情報を取得.
     * @return InputStream 正常ストリーム情報が返されます.
     */
    public InputStream success() {
        if( this.process != null ) {
            return this.process.getInputStream() ;
        }
        return null ;
    }
    
    /**
     * エラーストリーム情報を取得.
     * @return InputStream エラーストリーム情報が返されます.
     */
    public InputStream error() {
        if( this.process != null ) {
            return this.process.getErrorStream() ;
        }
        return null ;
    }
    
    /**
     * コマンドが終了したかチェック.
     * @return boolean [true]の場合、コマンドは終了しています.
     */
    public boolean isExit() {
        if( this.process == null ) {
            return true ;
        }
        try {
            this.process.exitValue() ;
        } catch( IllegalThreadStateException i ) {
            return false ;
        }
        return true ;
    }
    
    /**
     * コマンドは実行されたかチェック.
     * @return boolean [true]の場合、既に実行されました.
     */
    public boolean isExecution() {
        return executionFlag ;
    }
    
    /**
     * コマンド終了まで、待機.
     * @return int 終了コマンドコードが返されます.
     * @exception Exception 例外.
     */
    public int waitFor() throws Exception {
        if( this.process != null ) {
            return this.process.waitFor() ;
        }
        return -1 ;
    }
}
