package org.maachang.util;

import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.locks.ReentrantLock;

import org.maachang.util.atomic.AtomicBOOL;

/**
 * シリアライズロールオブジェクト.
 *
 * @version 2009/12/14
 * @author masahito suzuki
 * @since MaachangBase 1.0B
 */
public class SerializeRoll {
    
    /**
     * デフォルト長.
     */
    private static final int DEFAULT_DATA_LENGTH = 512 ;
    
    /**
     * デフォルトロール長.
     */
    private static final int DEFAULT_ROLL_LENGTH = 256 ;
    
    /**
     * 設定タイムアウト.
     */
    private static final long DEFAULT_TIMEOUT = 1500L ;
    
    /**
     * ヘッダデータ長.
     */
    private static final int HEADER_LENGTH = 4+4+8 ;
    
    /**
     * データ長格納領域.
     */
    private static final int DATA_HEADER_LENGTH = 8+4 ;
    
    /**
     * 永続化ファイルI/Oオブジェクト.
     */
    private RandomAccessFile rand = null ;
    
    /**
     * メモリマップ.
     */
    private MappedByteBuffer map = null ;
    
    /**
     * １ロールデータ長.
     */
    private int dataLength = -1 ;
    
    /**
     * ロールデータ.
     */
    private int rollLength = -1 ;
    
    /**
     * タイムアウト値.
     */
    private long timeout = -1L ;
    
    /**
     * ロック
     */
    private final ReentrantLock sync = new ReentrantLock() ;
    
    /**
     * オープンチェック.
     */
    private final AtomicBOOL openFlag = new AtomicBOOL() ;
    
    
    /**
     * コンストラクタ.
     * @param name 対象のファイル名を設定します.
     * @exception Exception 例外.
     */
    public SerializeRoll( String name )
        throws Exception {
        this( name,-1,-1,DEFAULT_TIMEOUT ) ;
    }
    
    /**
     * コンストラクタ.
     * @param name 対象のファイル名を設定します.
     * @param length 対象のロール長を設定します.
     * @exception Exception 例外.
     */
    public SerializeRoll( String name,int length )
        throws Exception {
        this( name,length,-1,DEFAULT_TIMEOUT ) ;
    }
    
    /**
     * コンストラクタ.
     * @param name 対象のファイル名を設定します.
     * @param length 対象のロール長を設定します.
     * @param data 対象の１つのデータ長を設定します.
     * @exception Exception 例外.
     */
    public SerializeRoll( String name,int length,int data )
        throws Exception {
        this( name,length,data,DEFAULT_TIMEOUT ) ;
    }
    
    /**
     * コンストラクタ.
     * @param name 対象のファイル名を設定します.
     * @param length 対象のロール長を設定します.
     * @param data 対象の１つのデータ長を設定します.
     * @param timeout 対象のタイムアウト値をミリ秒単位で設定します.
     * @exception Exception 例外.
     */
    public SerializeRoll( String name,int length,int data,long timeout )
        throws Exception {
        if( name == null || ( name = name.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "ファイル名が設定されていません" ) ;
        }
        if( length <= 0 ) {
            length = DEFAULT_ROLL_LENGTH ;
        }
        if( data <= 0 ) {
            data = DEFAULT_DATA_LENGTH ;
        }
        if( timeout <= 0L ) {
            timeout = 0 ;
        }
        RandomAccessFile r = null ;
        MappedByteBuffer m = null ;
        try {
            r = new RandomAccessFile( name,"rw" ) ;
            r.getChannel().force( false ) ;
            m = r.getChannel().map( FileChannel.MapMode.READ_WRITE,0L,
                ( long )( ( length * ( data + DATA_HEADER_LENGTH ) ) + HEADER_LENGTH ) ) ;
        } catch( Exception e ) {
            if( r != null ) {
                try {
                    r.close() ;
                } catch( Exception ee ) {
                }
            }
            throw e ;
        }
        this.rand = r ;
        this.map = m ;
        this.dataLength = data ;
        this.rollLength = length ;
        this.timeout = timeout ;
        this.openFlag.set( true ) ;
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        this.destroy() ;
    }
    
    /** 次回書き込み可能な項番を取得. **/
    private int wpos() {
        int len = map.getInt( 4 ) + 1 ;
        if( len > rollLength ) {
            return -1 ;
        }
        int ret = map.getInt( 0 ) ;
        int pos = ret + 1 ;
        if( pos >= rollLength ) {
            pos = 0 ;
        }
        map.putInt( 0,pos ) ;
        map.putInt( 4,len ) ;
        return ret ;
    }
    
    /** 読み込み位置を取得. **/
    private int rpos() {
        int len = map.getInt( 4 ) ;
        if( len <= 0 ) {
            return -1 ;
        }
        int pos = map.getInt( 0 ) ;
        int ret = pos - len ;
        if( ret < 0 ) {
            ret = rollLength + ret ;
        }
        map.putInt( 4,len - 1 ) ;
        return ret ;
    }
    
    /** 読み取り位置を項番指定して取得(roll読み取り位置は移動させない). **/
    private int rpos( int no ) {
        int len = map.getInt( 4 ) - no ;
        if( len <= 0 ) {
            return -1 ;
        }
        int pos = map.getInt( 0 ) ;
        int ret = pos - len ;
        if( ret < 0 ) {
            ret = rollLength + ret ;
        }
        return ret ;
    }
    
    /** 現在の長さを取得. **/
    private int len() {
        return map.getInt( 4 ) ;
    }
    
    /** 新しいIDを取得. **/
    private long newId() {
        long ret = map.getLong( 8 ) ;
        long n = ret ;
        if( n >= Long.MAX_VALUE ) {
            n = 0L ;
        }
        else {
            n ++ ;
        }
        map.putLong( 8,n ) ;
        return ret ;
    }
    
    /** 指定位置に対して、バイナリを設定. **/
    private void putBinary( int pos,long id,byte[] bin ) {
        int len = bin.length ;
        int off = ( pos * ( dataLength + DATA_HEADER_LENGTH ) ) + HEADER_LENGTH ;
        map.putLong( off,id ) ;
        off += 8 ;
        map.putInt( off,len ) ;
        off += 4 ;
        for( int i = 0 ; i < len ; i ++ ) {
            map.put( off+i,bin[ i ] ) ;
        }
    }
    
    /** 指定位置のバイナリを取得. **/
    private boolean getBinary( SerializeRollElement em,int pos ) {
        int off = ( pos * ( dataLength + DATA_HEADER_LENGTH ) ) + HEADER_LENGTH ;
        long id = map.getLong( off ) ;
        off += 8 ;
        int len = map.getInt( off ) ;
        if( len <= 0 || dataLength <= len ) {
            return false ;
        }
        byte[] ret = new byte[ len ] ;
        off += 4 ;
        for( int i = 0 ; i < len ; i ++ ) {
            ret[ i ] = map.get( off+i ) ;
        }
        em.setBinary( ret ) ;
        em.setId( id ) ;
        em.setPosition( pos ) ;
        return true ;
    }
    
    /**
     * オブジェクト破棄.
     */
    public void destroy() {
        if( rand != null ) {
            sync.lock() ;
            try {
                try {
                    rand.close() ;
                } catch( Exception e ) {
                }
                map = null ;
                openFlag.set( false ) ;
            } finally {
                sync.unlock() ;
            }
        }
    }
    
    /**
     * オブジェクトが有効かチェック.
     * @return boolean [true]の場合、有効です.
     */
    public boolean isUse() {
        return openFlag.get() ;
    }
    
    /**
     * 新しい送信IDを取得.
     * @return long 送信IDが返されます.
     * @exception Exception 例外.
     */
    public long getId() throws Exception {
        long ret = -1L ;
        sync.lock() ;
        try {
            ret = newId() ;
        } finally {
            sync.unlock() ;
        }
        return ret ;
    }
    
    /**
     * データ設定.
     * @param out 以前のデータ内容が返されます.
     * @param id 対象のIDを設定します.
     * @param data 対象のデータを設定します.
     * @return boolean [false]の場合、前回のデータは存在しません
     * @exception TimeoutException データ設置場所が存在しない場合の例外.
     * @exception Exception 例外.
     */
    public boolean put( SerializeRollElement out,long id,byte[] data ) throws TimeoutException,Exception {
        if( id < 0L ) {
            throw new IllegalArgumentException( "IDは規定外の値です:" + id ) ;
        }
        else if( data == null || data.length <= 0 ) {
            throw new IllegalArgumentException( "入力データが存在しません" ) ;
        }
        else if( data.length >= dataLength ) {
            throw new IllegalArgumentException( "入力データは長すぎます(max:" + dataLength + ")" ) ;
        }
        boolean ret = false ;
        long time = -1L ;
        while( true ) {
            if( time != -1L ) {
                Thread.sleep( 100 ) ;
            }
            sync.lock() ;
            try {
                if( rand == null ) {
                    throw new IOException( "オブジェクトは既に閉じられています" ) ;
                }
                int pos = wpos() ;
                // 書き込み領域が存在しない場合.
                if( pos == -1 ) {
                    if( time == -1L ) {
                        if( timeout <= 0L ) {
                            throw new TimeoutException( "データの設定に失敗しました" ) ;
                        }
                        time = System.currentTimeMillis() + timeout ;
                    }
                    else if( time <= System.currentTimeMillis() ) {
                        throw new TimeoutException( "データの設定に失敗しました" ) ;
                    }
                    continue ;
                }
                // ターゲット位置の前回の情報を取得して後、
                // 今回のデータを設定する.
                if( !getBinary( out,pos ) ) {
                    out.clear() ;
                    ret = false ;
                }
                else {
                    ret = true ;
                }
                putBinary( pos,id,data ) ;
                break ;
            } finally {
                sync.unlock() ;
            }
        }
        return ret ;
    }
    
    /**
     * データの取得.
     * @param out 取得結果の内容が返されます.
     * @return boolean [true]の場合、情報の取得に成功しました.
     * @exception Exception 例外.
     */
    public boolean get( SerializeRollElement out ) throws Exception {
        boolean ret = false ;
        sync.lock() ;
        try {
            if( rand == null ) {
                throw new IOException( "オブジェクトは既に閉じられています" ) ;
            }
            int pos = rpos() ;
            if( pos != -1 ) {
                ret = getBinary( out,pos ) ;
            }
        } finally {
            sync.unlock() ;
        }
        return ret ;
    }
    
    /**
     * データの取得.
     * @param out 取得結果の内容が返されます.
     * @param no 対象の項番を設定します.
     * @return boolean [true]の場合、情報の取得に成功しました.
     * @exception Exception 例外.
     */
    public boolean get( SerializeRollElement out,int no ) throws Exception {
        if( no < 0 || no >= rollLength ) {
            return false ;
        }
        boolean ret = false ;
        sync.lock() ;
        try {
            if( rand == null ) {
                throw new IOException( "オブジェクトは既に閉じられています" ) ;
            }
            int pos = rpos( no ) ;
            if( pos != -1 ) {
                ret = getBinary( out,pos ) ;
            }
        } finally {
            sync.unlock() ;
        }
        return ret ;
    }
    
    /**
     * IDでデータを検索.
     * @param out 取得結果の内容が返されます.
     * @param id 検索対象のIDを設定します.
     * @return boolean [true]の場合、検索に成功しました.
     * @exception Exception 例外.
     */
    public boolean search( SerializeRollElement out,long id ) throws Exception {
        if( id < 0L ) {
            return false ;
        }
        boolean flg = false ;
        sync.lock() ;
        try {
            if( rand == null ) {
                throw new IOException( "オブジェクトは既に閉じられています" ) ;
            }
            int len = rollLength ;
            for( int i = 0 ; i < len ; i ++ ) {
                flg = getBinary( out,i ) ;
                if( flg && out.getId() == id ) {
                    return true ;
                }
            }
        } finally {
            sync.unlock() ;
        }
        return false ;
    }
    
    /**
     * データを直接書き換え.
     * @param pos 書き換え対象のデータ位置を設定します.
     * @param id 書き換え対象のIDを設定します.
     * @param data 書き換え対象のデータを設定します.
     * @return boolean [true]の場合、書き換え成功です.
     * @exception Exception 例外.
     */
    public boolean write( int pos,long id,byte[] data ) throws Exception {
        if( id < 0L ) {
            throw new IllegalArgumentException( "IDは規定外の値です:" + id ) ;
        }
        else if( data == null || data.length <= 0 ) {
            throw new IllegalArgumentException( "入力データが存在しません" ) ;
        }
        else if( data.length >= dataLength ) {
            throw new IllegalArgumentException( "入力データは長すぎます(max:" + dataLength + ")" ) ;
        }
        boolean ret = false ;
        SerializeRollElement out = new SerializeRollElement() ;
        sync.lock() ;
        try {
            if( getBinary( out,pos ) && id == out.getId() ) {
                putBinary( pos,id,data ) ;
                ret = true ;
            }
        } finally {
            sync.unlock() ;
        }
        return ret ;
    }
    
    /**
     * 現在格納されている長さを取得.
     * @return int 現在格納されている長さが返されます.
     * @exception Exception 例外.
     */
    public int length() throws Exception {
        int ret = -1 ;
        sync.lock() ;
        try {
            ret = len() ;
        } finally {
            sync.unlock() ;
        }
        return ret ;
    }
    
    /**
     * ロール最大長を取得.
     * @return int ロール最大長が返されます.
     */
    public int maxLength() {
        return rollLength ;
    }
    
    /**
     * 最大データ長を取得.
     * @return int データ最大長が返されます.
     */
    public int maxDataLength() {
        return dataLength ;
    }
    
    /**
     * 同期オブジェクトを取得.
     * @return ReentrantLock 同期オブジェクトが返されます.
     */
    public ReentrantLock getSync() {
        return sync ;
    }
}

