package org.maachang.util.atomic ;

import java.io.IOException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

/**
 * ブロッキングオブジェクト.
 * 
 * @version 2009/04/21
 * @author  masahito suzuki
 * @since   MaachangBase 1.0B
 */
public class Blocking {
    private int count = 0 ;
    private final ReentrantLock lock = new ReentrantLock() ;
    private final Condition condition = lock.newCondition() ;
    
    /**
     * コンストラクタ.
     */
    public Blocking() {
        
    }
    
    /**
     * ブロック処理.
     * @return boolean [true]の場合、指定時間に対して、タイムアウトを検出しました.
     * @exception InterruptedException 現在のスレッドで割り込みが発生した場合.
     */
    public boolean stop() throws InterruptedException {
        lock.lockInterruptibly() ;
        try {
            count ++ ;
            condition.await() ;
        } catch (InterruptedException e) {
            condition.signal() ;
            throw e ;
        } finally {
            count -- ;
            lock.unlock() ;
        }
        return false ;
    }
    
    /**
     * ブロック処理.
     * @param time 待機する最長時間を設定します.
     * @return boolean [true]の場合、指定時間に対して、タイムアウトを検出しました.
     * @exception InterruptedException 現在のスレッドで割り込みが発生した場合.
     */
    public boolean stop( long time ) throws InterruptedException {
        if( time <= 0 ) {
            return stop() ;
        }
        boolean ret = false ;
        lock.lockInterruptibly() ;
        try {
            count ++ ;
            ret = !condition.await( time,TimeUnit.MILLISECONDS ) ;
        } catch (InterruptedException e) {
            condition.signal() ;
            throw e ;
        } finally {
            count -- ;
            lock.unlock() ;
        }
        return ret ;
    }
    
    /**
     * ブロック処理.
     * @param time 待機する最長時間を設定します.
     * @param unit 引数の時間単位を設定します.
     * @return boolean [true]の場合、指定時間に対して、タイムアウトを検出しました.
     * @exception InterruptedException 現在のスレッドで割り込みが発生した場合.
     */
    public boolean stop( long time,TimeUnit unit ) throws InterruptedException {
        if( time <= 0 ) {
            return stop() ;
        }
        if( unit == null ) {
            unit = TimeUnit.MILLISECONDS ;
        }
        boolean ret = false ;
        lock.lockInterruptibly() ;
        try {
            count ++ ;
            ret = !condition.await( time,unit ) ;
        } catch (InterruptedException e) {
            condition.signal() ;
            throw e ;
        } finally {
            count -- ;
            lock.unlock() ;
        }
        return ret ;
    }
    
    /**
     * ブロック処理.
     * @param time 待機する最長時間を設定します.
     * @return boolean [true]の場合、指定時間に対して、タイムアウトを検出しました.
     * @exception IOException I/O例外.
     */
    public boolean waitTo( long time ) throws IOException {
        try {
            return stop( time ) ;
        } catch (InterruptedException e) {
            throw new IOException( "割り込みが発生しました" ) ;
        }
    }
    
    /**
     * ブロック処理.
     * @param time 待機する最長時間を設定します.
     * @param unit 引数の時間単位を設定します.
     * @return boolean [true]の場合、指定時間に対して、タイムアウトを検出しました.
     * @exception IOException I/O例外.
     */
    public boolean waitTo( long time,TimeUnit unit ) throws IOException {
        try {
            return stop( time,unit ) ;
        } catch (InterruptedException e) {
            throw new IOException( "割り込みが発生しました" ) ;
        }
    }
    
    /**
     * ブロック解除処理.
     */
    public void free() {
        lock.lock() ;
        try {
            if( count > 0 ) {
                condition.signal() ;
            }
        } finally {
            lock.unlock() ;
        }
    }
    
    /**
     * ブロックを全て解除.
     */
    public void freeAll() {
        lock.lock() ;
        try {
            if( count > 0 ) {
                condition.signalAll() ;
            }
        } finally {
            lock.unlock() ;
        }
    }
}
