package org.maachang.util.atomic;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.locks.ReentrantLock;

/**
 * Queue処理.
 * 
 * @version 2009/09/01
 * @author masahito suzuki
 * @since MaachangAtomic 1.00
 */
public class QueueToLock<E> {
    private static class _queueToLockChild<E> {
        public E object ;
        public _queueToLockChild<E> next = null ;
    }
    
    private final ReentrantLock sync = new ReentrantLock() ;
    private _queueToLockChild<E> head = null ;
    private _queueToLockChild<E> last = null ;
    private int size = 0 ;
    
    protected void finalize() throws Exception {
        clear() ;
    }
    
    /**
     * 情報をクリア.
     */
    public void clear() {
        sync.lock() ; try {
            head = null ;
            last = null ;
            size = 0 ;
        } finally {
            sync.unlock() ;
        }
    }
    
    /**
     * 先頭に情報を追加.
     * @param e 追加対象の情報を設定します.
     */
    public void addFirst( E e ) {
        if( e == null ) {
            return ;
        }
        _queueToLockChild<E> a = new _queueToLockChild<E>() ;
        a.object = e ;
        sync.lock() ; try {
            if( head == null ) {
                head = a ;
                last = a ;
            }
            else {
                a.next = head ;
                head = a ;
            }
            size ++ ;
        } finally {
            sync.unlock() ;
        }
    }
    
    /**
     * 最後に情報を追加.
     * @param e 追加対象の情報を設定します.
     */
    public void addLast( E e ) {
        offer( e ) ;
    }
    
    /**
     * 最後に情報を追加.
     * @param e 追加対象の情報を設定します.
     */
    public void offer( E e ) {
        if( e == null ) {
            return ;
        }
        _queueToLockChild<E> a = new _queueToLockChild<E>() ;
        a.object = e ;
        sync.lock() ; try {
            if( head == null ) {
                head = a ;
                last = a ;
            }
            else {
                last.next = a ;
                last = a ;
            }
            size ++ ;
        } finally {
            sync.unlock() ;
        }
    }
    
    /**
     * 先頭情報を削除して取得.
     * @return E 対象の情報が返されます.
     */
    public E poll() {
        E ret = null ;
        sync.lock() ; try {
            if( head != null ) {
                ret = head.object ;
                head = head.next ;
                if( head == null ) {
                    last = null ;
                }
                size -- ;
            }
        } finally {
            sync.unlock() ;
        }
        return ret ;
    }
    
    /**
     * 先頭情報を削除せず取得.
     * @return E 対象の情報が返されます.
     */
    public E peek() {
        E ret = null ;
        sync.lock() ; try {
            if( head != null ) {
                ret = head.object ;
            }
        } finally {
            sync.unlock() ;
        }
        return ret ;
    }
    
    /**
     * サイズを取得.
     * @return int サイズが返されます.
     */
    public int size() {
        int ret ;
        sync.lock() ; try {
            ret = size ;
        } finally {
            sync.unlock() ;
        }
        return ret ;
    }
    
    /**
     * 全内容を取得.
     * @return List<E> 全内容が返されます.
     */
    public List<E> getAll() {
        List<E> ret = new ArrayList<E>() ;
        _queueToLockChild<E> n ;
        sync.lock() ; try {
            for( n = head ; n != null ; n = n.next ) {
                ret.add( n.object ) ;
            }
        } finally {
            sync.unlock() ;
        }
        return ret ;
    }
}
