package org.maachang.util.atomic.lock ;

import java.util.concurrent.atomic.AtomicInteger;

/**
 * HashLock.
 *  
 * @version 2009/03/17
 * @author  masahito suzuki
 * @since MaachangBase 1.09
 */
public final class HashLock {
    private static final int DEF_SPIN_COUNT = 10000 ;
    private AtomicInteger[] list = null ;
    private int spin = 0 ;
    private int hashMask ;
    
    /**
     * コンストラクタ.
     */
    public HashLock() {
        this( 0,0 ) ;
    }
    
    /**
     * コンストラクタ.
     * @param hash ハッシュ値を設定します.
     */
    public HashLock( int hash ) {
        this( hash,0 ) ;
    }
    
    /**
     * コンストラクタ.
     * @param hash ハッシュ値を設定します.
     * @param spin spinロック値を設定します.
     */
    public HashLock( int hash,int spin ) {
        if( spin <= 0 ) {
            spin = DEF_SPIN_COUNT ;
        }
        hash = trimHash( hash ) ;
        int len = hash+1 ;
        AtomicInteger[] lst = new AtomicInteger[ len ] ;
        for( int i = 0 ; i < len ; i ++ ) {
            lst[ i ] = new AtomicInteger( 0 ) ;
        }
        this.hashMask = hash ;
        this.spin = spin ;
        this.list = lst ;
    }
    
    /**
     * ロック処理.
     */
    public void lock( int hc ) {
        AtomicInteger c = list[ hc & hashMask ] ;
        if( !c.compareAndSet( 0,1 ) ) {
            // spinlock.
            for( int i = 0 ; i < spin ; i ++ ) {
                if( c.compareAndSet( 0,1 ) ) {
                    return ;
                }
            }
            // spinlockが一定ループ以上の場合は、WaitしたLock解除待ち.
            try {
                while( true ) {
                    Thread.sleep( 1 ) ;
                    if( c.compareAndSet( 0,1 ) ) {
                        return ;
                    }
                }
            } catch( Exception e ) {
            }
        }
    }
    
    /**
     * アンロック処理.
     */
    public void unlock( int hc ) {
        AtomicInteger c = list[ hc & hashMask ] ;
        if( !c.compareAndSet( c.get(),0 ) ) {
            while( true ) {
                if( c.compareAndSet( c.get(),0 ) ) {
                    return ;
                }
            }
        }
    }
    
    /**
     * hashMask値を取得.
     * @return int hashMask値が返されます.
     */
    public int getHashMask() {
        return hashMask ;
    }
    
    /**
     * spinロック値を取得.
     * @return int spinロック値が返されます.
     */
    public int getSpin() {
        return spin ;
    }
    
    /** hash値の最適値を生成. **/
    protected static final int trimHash( int hash ) {
        if( hash <= 0 ) {
            return 0x000000ff ;
        }
        else if( hash >= 0x7fffffff ) {
            hash = 0x7fffffff ;
        }
        else if( hash >= 0x3fffffff ) {
            hash = 0x3fffffff ;
        }
        else if( hash >= 0x1fffffff ) {
            hash = 0x1fffffff ;
        }
        else if( hash >= 0x0fffffff ) {
            hash = 0x0fffffff ;
        }
        else if( hash >= 0x07ffffff ) {
            hash = 0x07ffffff ;
        }
        else if( hash >= 0x03ffffff ) {
            hash = 0x03ffffff ;
        }
        else if( hash >= 0x01ffffff ) {
            hash = 0x01ffffff ;
        }
        else if( hash >= 0x00ffffff ) {
            hash = 0x00ffffff ;
        }
        else if( hash >= 0x007fffff ) {
            hash = 0x007fffff ;
        }
        else if( hash >= 0x003fffff ) {
            hash = 0x003fffff ;
        }
        else if( hash >= 0x001fffff ) {
            hash = 0x001fffff ;
        }
        else if( hash >= 0x000fffff ) {
            hash = 0x000fffff ;
        }
        else if( hash >= 0x0007ffff ) {
            hash = 0x0007ffff ;
        }
        else if( hash >= 0x0003ffff ) {
            hash = 0x0003ffff ;
        }
        else if( hash >= 0x0001ffff ) {
            hash = 0x0001ffff ;
        }
        else if( hash >= 0x0000ffff ) {
            hash = 0x0000ffff ;
        }
        else if( hash >= 0x00007fff ) {
            hash = 0x00007fff ;
        }
        else if( hash >= 0x00003fff ) {
            hash = 0x00003fff ;
        }
        else if( hash >= 0x00001fff ) {
            hash = 0x00001fff ;
        }
        else if( hash >= 0x00000fff ) {
            hash = 0x00000fff ;
        }
        else if( hash >= 0x000007ff ) {
            hash = 0x000007ff ;
        }
        else if( hash >= 0x000003ff ) {
            hash = 0x000003ff ;
        }
        else if( hash >= 0x000001ff ) {
            hash = 0x000001ff ;
        }
        else {
            hash = 0x000000ff ;
        }
        return hash ;
    }
}

