package org.maachang.dao ;

import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.locks.ReentrantLock;

import org.maachang.dao.dbms.Record;
import org.maachang.dao.dbms.RecordFactory;
import org.maachang.dao.dbms.RecordSequenceId;
import org.maachang.util.FileUtil;
import org.maachang.util.atomic.AtomicOBJECT;

/**
 * DaoSessionFactory.
 *
 * @version 2007/11/03
 * @author  masahito suzuki
 * @since MaachangDao 1.00
 */
public class DaoSessionFactory {
    
    /**
     * Javaシーケンス出力先ディレクトリ.
     */
    private static final String OUTPUT_SEQ_DIR = "./conf/jseq/" ;
    
    /**
     * Javaシーケンスファイル名.
     */
    private static final String SEQ_FILE = "jseq.seq" ;
    
    /**
     * Javaシーケンスメタファイル名.
     */
    private static final String SEQ_META_FILE = "jseq.meta" ;
    
    /**
     * シングルトン.
     */
    private static final DaoSessionFactory SNGL = new DaoSessionFactory() ;
    
    /**
     * レコードファクトリ.
     */
    private final AtomicOBJECT<RecordFactory> recordFactory = new AtomicOBJECT<RecordFactory>() ;
    
    /**
     * Javaシーケンス.
     */
    private final AtomicOBJECT<RecordSequenceId> javaSeq = new AtomicOBJECT<RecordSequenceId>() ;
    
    /**
     * Daoセッション管理オブジェクト.
     */
    private final ConcurrentLinkedQueue<DaoSession> sessions = new ConcurrentLinkedQueue<DaoSession>() ;
    
    /**
     * 同期.
     */
    private final ReentrantLock sync = new ReentrantLock() ;
    //private final AtomicLock sync = new AtomicLock() ;
    
    /**
     * コンストラクタ.
     */
    private DaoSessionFactory() {
    }
    
    /**
     * オブジェクトを取得.
     * <BR><BR>
     * オブジェクトを取得します.
     * <BR>
     * @return DaoSessionFactory オブジェクト情報が返されます.
     */
    public static final DaoSessionFactory getInstance() {
        return SNGL ;
    }
    
    /**
     * レコードファクトリを設定.
     * <BR><BR>
     * レコードファクトリを設定します.
     * <BR>
     * @param recordFactory 対象のレコードファクトリを設定します.
     */
    protected void setRecordFactory( RecordFactory recordFactory ) {
        this.recordFactory.set( recordFactory ) ;
    }
    
    /**
     * レコードファクトリを取得.
     * @return RecordFactory レコードファクトリが返されます.
     */
    public RecordFactory getRecordFactory() {
        return recordFactory.get() ;
    }
    
    /**
     * 新しいDaoSessionを取得.
     * <BR><BR>
     * 新しいDaoSessionを取得します.
     * <BR>
     * @return DaoSession 新しいDaoSessionが返されます.
     * @exception MaachangDaoException MaachangDao例外.
     */
    public DaoSession getDaoSession()
        throws MaachangDaoException {
        RecordFactory rf = recordFactory.get() ;
        if( rf != null ) {
            Record record = null ;
            DaoSession ret = null ;
            try {
                record = rf.getRecord() ;
                ret = poolDaoSession( record ) ;
            } catch( Exception e ) {
                if( record != null ) {
                    try {
                        record.close() ;
                    } catch( Exception ee ) {
                    }
                }
                throw new MaachangDaoException( e ) ;
            }
            return ret ;
        }
        throw new MaachangDaoException(
            "初期化されていないか、条件が不正です" ) ;
    }
    
    /**
     * 新しいDaoSessionを取得.
     * <BR><BR>
     * 新しいDaoSessionを取得します.
     * <BR>
     * @param record 対象のレコードを設定します.
     * @return DaoSession 新しいDaoSessionが返されます.
     * @exception MaachangDaoException MaachangDao例外.
     */
    public DaoSession getDaoSession( Record record )
        throws MaachangDaoException {
        DaoSession s = poolDaoSession( record ) ;
        return s ;
    }
    
    /**
     * 新しいレコードを取得.
     * <BR><BR>
     * 新しいレコードを取得します.
     * <BR>
     * @return Record 新しいレコード情報が返されます.
     * @exception MaachangDaoException MaachangDao例外.
     */
    public Record getRecord()
        throws MaachangDaoException {
        RecordFactory rf = recordFactory.get() ;
        if( rf != null ) {
            Record ret = null ;
            try {
                ret = rf.getRecord() ;
            } catch( Exception e ) {
                if( ret != null ) {
                    try {
                        ret.close() ;
                    } catch( Exception ee ) {
                    }
                }
                throw new MaachangDaoException( e ) ;
            }
            return ret ;
        }
        throw new MaachangDaoException(
            "初期化されていないか、条件が不正です" ) ;
    }
    
    /**
     * Java側でシーケンス発行を行う場合の処理.
     * @exception Exception 例外.
     */
    public void createJavaSequence() throws Exception {
        RecordSequenceId sq = javaSeq.get() ;
        if( sq != null && sq.isUse() ) {
            return ;
        }
        sync.lock() ;
        try {
            if( FileUtil.isDirExists( OUTPUT_SEQ_DIR ) == false ) {
                FileUtil.mkdirs( OUTPUT_SEQ_DIR ) ;
            }
            RecordSequenceId rs = new RecordSequenceId( OUTPUT_SEQ_DIR+SEQ_FILE,OUTPUT_SEQ_DIR+SEQ_META_FILE ) ;
            this.javaSeq.set( rs ) ;
        } finally {
            sync.unlock() ;
        }
    }
    
    /**
     * シーケンス発行オブジェクトを破棄.
     */
    public void clearJavaSequence() {
        RecordSequenceId sq = javaSeq.get() ;
        if( sq != null ) {
            sq.destroy() ;
        }
        javaSeq.set( null ) ;
    }
    
    /**
     * シーケンス発行オブジェクト用ファイルを削除.
     */
    public void deleteJavaSequence() {
        RecordSequenceId sq = javaSeq.get() ;
        if( sq != null ) {
            try {
                sq.allClear() ;
            } catch( Exception e ) {
            }
            sq.destroy() ;
        }
        sync.lock() ;
        try {
            javaSeq.set( null ) ;
            try {
                FileUtil.removeFile( OUTPUT_SEQ_DIR+SEQ_META_FILE ) ;
            } catch( Exception e ) {
            }
        } finally {
            sync.unlock() ;
        }
    }
    
    /**
     * シーケンス発行オブジェクトを取得.
     * @return RecordSequenceId シーケンス発行オブジェクトが返されます.
     */
    public RecordSequenceId getJavaSequence() {
        return javaSeq.get() ;
    }
    
    /**
     * シーケンス発行オブジェクトが有効かチェック.
     * @return boolean [true]の場合、有効です.
     */
    public boolean isJavaSequence() {
        return ( javaSeq.get() != null ) ;
    }
    
    /**
     * DaoSession管理数を取得.
     * @return int DaoSession管理数が返されます.
     */
    public int poolingLength() {
        return sessions.size() ;
    }
    
    /**
     * １つのDaoSessionオブジェクトを取得.
     */
    private DaoSession poolDaoSession( Record record ) {
        DaoSession ret = sessions.poll() ;
        if( ret == null ) {
            ret = new DaoSession( this ) ;
        }
        ret.create( record ) ;
        return ret ;
    }
    
    /**
     * 利用終了したDaoSessionをSessionプーリングオブジェクトに返却.
     */
    protected void push( DaoSession session ) {
        if( session != null ) {
            sessions.offer( session ) ;
        }
    }
}
