package org.maachang.dao ;

import java.io.IOException;
import java.sql.ResultSet;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.maachang.dao.dbms.DbUtil;
import org.maachang.dao.dbms.MetaColumn;
import org.maachang.dao.dbms.Record;
import org.maachang.dao.dbms.RecordSequenceId;
import org.maachang.dao.dbms.kind.SupportKind;
import org.maachang.util.ObjectArray;
import org.maachang.util.StringUtil;

/**
 * MaachangDao.
 * 
 * @version 2007/10/18
 * @author masahito suzuki
 * @since MaachangDao 1.00
 */
public class MaachangDao {
    
    /**
     * 現在バージョン.
     */
    private final String VERSION = "1.14" ;
    
    /**
     * シングルトン.
     */
    private static final MaachangDao SNGL = new MaachangDao() ;
    
    /**
     * 楽観ロック名(Java名).
     */
    public static final String COLUMN_OPTIMISTIC_LOCK_JAVA_NAME ;
    static {
        COLUMN_OPTIMISTIC_LOCK_JAVA_NAME = DbUtil.convertDBNameByJavaName( false,ExecutionDao.OPTIMISTIC_LOCK_COLUMN ) ; ;
    }
    
    /**
     * コンストラクタ.
     */
    private MaachangDao() {
        
    }
    
    /**
     * オブジェクトを取得.
     * <BR><BR>
     * オブジェクトを取得します.
     * <BR>
     * @return MaachangDao オブジェクト情報が返されます.
     */
    public static final MaachangDao getInstance() {
        return SNGL ;
    }
    
    /**
     * 現在のバージョンを取得.
     * @return String 現在のバージョンが返されます.
     */
    public String getVersion() {
        return VERSION ;
    }
    
    /**
     * データ保存.
     * <BR><BR>
     * データを保存します.
     * <BR>
     * @param session 対象のセッションを設定します.
     * @param modelName 対象のモデル名を設定します.
     * @param sequence [true]の場合、IDがシーケンスで処理されます.
     * @param params 保存対象のモデル情報を設定します.
     * @return Map<String,Object> 保存された内容が返されます.
     * @exception MaachangDaoException MaachangDao例外.
     */
    public Map<String,Object> save( DaoSession session,String modelName,boolean sequence,Map<String,Object> params )
        throws MaachangDaoException {
        if( session == null || params == null || params.size() <= 0 ||
            modelName == null || ( modelName = modelName.trim() ).length() <= 0 ) {
            if( session == null ) {
                throw new MaachangDaoException( "セッション情報が設定されていません" ) ;
            }
            else if( params == null || params.size() <= 0 ) {
                throw new MaachangDaoException( "保存パラメータがありません" ) ;
            }
            throw new MaachangDaoException( "モデル名が設定されていません" ) ;
        }
        try {
            Record record = session.getRecord() ;
            MetaColumn meta = session.getMeta( modelName ) ;
            boolean idFlag ;
            if( sequence ) {
                idFlag = ExecutionDao.isSequence( meta ) ;
            }
            else if( !params.containsKey( ExecutionDao.SEQ_COLUMN ) ) {
                throw new MaachangDaoException( "IDが指定されていません" ) ;
            }
            else {
                idFlag = ( params.get( ExecutionDao.SEQ_COLUMN ) != null ) ;
            }
            Object optLock = params.get( COLUMN_OPTIMISTIC_LOCK_JAVA_NAME ) ;
            // データが存在するかチェック.
            // または、IDカラムが存在しない場合は、強制的に
            // insertにする.
            if( isEntry( record,idFlag,meta,params,modelName ) == true ) {
                // データが存在する場合は、update.
                StringBuilder buf = new StringBuilder() ;
                buf.append( "update " ).append( meta.getTable() ).append( " set " ) ;
                int len = meta.size() ;
                int cnt = 0 ;
                ObjectArray pms = new ObjectArray() ;
                for( int i = 0 ; i < len ; i ++ ) {
                    String column = meta.getColumnName( i ) ;
                    String javaName = DbUtil.convertDBNameByJavaName( false,column ) ;
                    if( ExecutionDao.SEQ_COLUMN.equals( column ) == false ) {
                        Object value = null ;
                        // 更新カラムが存在する場合.
                        if( ExecutionDao.UPDATE_TIME.equals( column ) ) {
                            value = new java.sql.Timestamp( System.currentTimeMillis() ) ;
                            params.put( javaName,value ) ;
                        }
                        else {
                            value = params.get( javaName ) ;
                        }
                        if( value != null ) {
                            if( cnt != 0 ) {
                                buf.append( "," ) ;
                            }
                            buf.append( column ).append( "=" ).
                                append( "?" ) ;
                            pms.add( value ) ;
                            cnt ++ ;
                        }
                    }
                }
                buf.append( " where id=?" ) ;
                if( pms.size() <= 0 ) {
                    throw new IllegalArgumentException( "保存対象のパラメータが全てNULLです" ) ;
                }
                String sql = buf.toString() ;
                buf = null ;
                Object id = params.get( ExecutionDao.SEQ_COLUMN ) ;
                pms.add( id ) ;
                ExecutionDao.executionSql( null,record,modelName,meta,sequence,sql,pms.getObjectArray(),pms.size() ) ;
            }
            // データが存在せず、楽観的ロックカラム情報が存在して、バージョン情報がセットされている場合.
            else if( optLock instanceof Long && ((Long)optLock) > 1L && ExecutionDao.isOptimisticLock( meta ) ) {
                // 楽観的ロック例外を出力.
                throw new OptimisticLockException( "[" + modelName + "] - use optimistic lock.(src:" +
                    optLock + " now:[delete])" ) ;
            }
            // データが存在しない場合Insert.
            else {
                StringBuilder buf = new StringBuilder() ;
                buf.append( "insert into " ).append( meta.getTable() ).append( " (" ) ;
                int len = meta.size() ;
                ObjectArray pms = new ObjectArray() ;
                boolean addId = false ;
                // テーブルがシーケンスIDが利用でき、シーケンス処理が可能な場合.
                if( sequence && idFlag ) {
                    RecordSequenceId reqId = DaoSessionFactory.getInstance().getJavaSequence() ;
                    Object id = params.get( ExecutionDao.SEQ_COLUMN ) ;
                    // 新しいIDが存在する場合.
                    if( id != null ) {
                        // IDがLongでない場合は、処理しない.
                        if( id instanceof Long ) {
                            // Javaシーケンスを利用している場合.
                            if( reqId != null ) {
                                // 指定IDを、Javaシーケンスに設定する.
                                reqId.setId( modelName,( ( Long )id ).longValue() ) ;
                                pms.add( id ) ;
                            }
                            addId = true ;
                        }
                        else {
                            // 新しいシーケンスを付与するための、空間をセット.
                            pms.add( new Object() ) ;
                            params.remove( ExecutionDao.SEQ_COLUMN ) ;
                            addId = true ;
                        }
                    }
                    // 新しいIDが設定されておらず、Javaシーケンスが利用可能.
                    else if( reqId != null ) {
                        pms.add( new Long( reqId.getId( modelName ) ) ) ;
                        addId = true ;
                    }
                    // シーケンス先付与の場合.
                    else if( record.isSequence() == true ) {
                        // 新しいシーケンスを付与するための、空間をセット.
                        pms.add( new Object() ) ;
                        addId = true ;
                    }
                    // IDが付与されている.
                    if( addId ) {
                        buf.append( ExecutionDao.SEQ_COLUMN ) ;
                        if( len != 1 ) {
                            buf.append( "," ) ;
                        }
                    }
                }
                // テーブルがシーケンスIDを利用できない場合.
                else {
                    Object id = params.get( ExecutionDao.SEQ_COLUMN ) ;
                    if( id != null && id instanceof Long ) {
                        pms.add( id ) ;
                    }
                    else {
                        throw new MaachangDaoException(
                            "非sequenceIDテーブル[" + modelName + "]に対して、IDが設定されていません" ) ;
                    }
                    buf.append( ExecutionDao.SEQ_COLUMN ) ;
                    if( len != 1 ) {
                        buf.append( "," ) ;
                    }
                    addId = true ;
                }
                int cnt = 0 ;
                for( int i = 0 ; i < len ; i ++ ) {
                    String column = meta.getColumnName( i ) ;
                    int type = meta.getColumnType( i ) ;
                    String lowerColumn = column.toLowerCase() ;
                    String javaName = DbUtil.convertDBNameByJavaName( false,column ) ;
                    if( ExecutionDao.SEQ_COLUMN.equals( column ) == false ) {
                        Object value = null ;
                        // 生成/更新カラムが存在する場合.
                        if( ExecutionDao.CREATE_TIME.equals( lowerColumn ) ||
                            ExecutionDao.UPDATE_TIME.equals( lowerColumn ) ) {
                            value = new java.sql.Timestamp( System.currentTimeMillis() ) ;
                            params.put( javaName,value ) ;
                        }
                        // 楽観的ロックカラムが存在する場合.
                        else if( ExecutionDao.OPTIMISTIC_LOCK_COLUMN.equals( lowerColumn ) &&
                            Types.BIGINT == type ) {
                            value = new Long( 1 ) ;
                            params.put( javaName,value ) ;
                        }
                        else {
                            value = params.get( javaName ) ;
                        }
                        if( cnt != 0 ) {
                            buf.append( "," ) ;
                        }
                        buf.append( column ) ;
                        if( value == null ) {
                            pms.add( null ) ;
                        }
                        else {
                            pms.add( value ) ;
                        }
                        cnt ++ ;
                    }
                }
                buf.append( ") values (" ) ;
                if( addId == true ) {
                    if( cnt == 0 ) {
                        buf.append( "?)" ) ;
                    }
                    else {
                        buf.append( "?," ) ;
                    }
                }
                for( int i = 0 ; i < cnt ; i ++ ) {
                    if( i != 0 ) {
                        buf.append( "," ) ;
                    }
                    buf.append( "?" ) ;
                }
                buf.append( ")" ) ;
                String sql = buf.toString() ;
                Long[] resId = new Long[ 1 ] ;
                // SQL実行.
                ExecutionDao.executionSql( resId,record,modelName,meta,sequence,sql,pms.getObjectArray(),pms.size() ) ;
                // シーケンスIDが発行されている場合.
                if( sequence && addId && resId[ 0 ] != null && resId[ 0 ] instanceof Long ) {
                    params.put( ExecutionDao.SEQ_COLUMN,resId[ 0 ] ) ;
                }
            }
            return params ;
        } catch( MaachangDaoException md ) {
            throw md ;
        } catch( Exception e ) {
            throw new MaachangDaoException( e ) ;
        }
    }
    
    /**
     * アップデート.
     * <BR><BR>
     * 指定条件のデータをアップデートします.
     * <BR>
     * @param session 対象のセッションを設定します.
     * @param modelName 対象のモデル名を設定します.
     * @param set update文の[set]以降の条件を設定します.
     * @param params 対象のパラメータ群を設定します.
     * @exception MaachangDaoException MaachangDao例外.
     */
    public void update( DaoSession session,String modelName,String set,Object... params )
        throws MaachangDaoException {
        if( session == null || set == null || modelName == null ||
            ( set = set.trim() ).length() <= 0 || ( modelName = modelName.trim() ).length() <= 0 ) {
            if( session == null ) {
                throw new MaachangDaoException( "セッション情報が設定されていません" ) ;
            }
            else if( set == null || set.length() <= 0 ) {
                throw new MaachangDaoException( "Set条件が不正です" ) ;
            }
            throw new MaachangDaoException( "モデル名が設定されていません" ) ;
        }
        try {
            Record record = session.getRecord() ;
            MetaColumn meta = session.getMeta( modelName ) ;
            StringBuilder buf = new StringBuilder() ;
            if( set.toLowerCase().startsWith( "set " ) == false ) {
                buf.append( "update " ).append( meta.getTable() ).
                    append( " set " ).append( set ) ;
            }
            else {
                buf.append( "update " ).append( meta.getTable() ).append( " " ).append( set ) ;
            }
            String sql = buf.toString() ;
            buf = null ;
            if( ExecutionDao.isUpdateTime( meta ) == true ) {
                boolean[] addFlag = new boolean[ 1 ] ;
                addFlag[ 0 ] = false ;
                sql = addUpdateTime( addFlag,sql ) ;
                if( addFlag[ 0 ] == true ) {
                    params = addUpdateTimeParam( params ) ;
                }
            }
            ExecutionDao.executionSql( null,record,modelName,meta,false,sql,params,
                ( ( params != null ) ? params.length : 0 ) ) ;
        } catch( MaachangDaoException md ) {
            throw md ;
        } catch( Exception e ) {
            throw new MaachangDaoException( e ) ;
        }
    }
    
    /**
     * 削除.
     * <BR><BR>
     * 指定Entryの情報を削除します.
     * <BR>
     * @param session 対象のセッションを設定します.
     * @param srcJoin 対象のsrcJoin条件を設定します.
     * @param modelName 対象のモデル名を設定します.
     * @param model 対象のモデル情報が返されます.
     * @exception MaachangDaoException MaachangDao例外.
     */
    public void delete( DaoSession session,Map<String,String> srcJoin,String modelName,Map<String,Object> model )
        throws MaachangDaoException {
        if( session == null || modelName == null || ( modelName = modelName.trim() ).length() <= 0 ) {
            if( session == null ) {
                throw new MaachangDaoException( "セッション情報が設定されていません" ) ;
            }
            throw new MaachangDaoException( "モデル名が設定されていません" ) ;
        }
        try {
            Record record = session.getRecord() ;
            MetaColumn meta = session.getMeta( modelName ) ;
            if( ExecutionDao.isSequence( meta ) == true ) {
                Object id = model.get( ExecutionDao.SEQ_COLUMN ) ;
                if( id == null ) {
                    throw new IllegalArgumentException( "IDが存在しません" ) ;
                }
                deleteBySrcJoin( session,srcJoin,modelName,"where id=?",id ) ;// テーブル結合.
                StringBuilder buf = new StringBuilder() ;
                buf.append( "delete from " ).append( meta.getTable() ).append( " where id=?;" ) ;
                String sql = buf.toString() ;
                buf = null ;
                ExecutionDao.executionSql( null,record,modelName,meta,false,sql,new Object[]{id},1 ) ;
            }
            else {
                StringBuilder buf = new StringBuilder() ;
                buf.append( "delete from " ).append( meta.getTable() ).append( " where " ) ;
                int len = meta.size() ;
                int cnt = 0 ;
                ObjectArray pms = new ObjectArray() ;
                for( int i = 0 ; i < len ; i ++ ) {
                    if( i != 0 ) {
                        buf.append( " and " ) ;
                    }
                    String column = meta.getColumnName( i ) ;
                    Object value = model.get( DbUtil.convertDBNameByJavaName( false,column ) ) ;
                    if( value == null ) {
                        buf.append( column ).append( " is null" ) ;
                    }
                    else {
                        buf.append( column ).append( "=?" ) ;
                        pms.add( value ) ;
                        cnt ++ ;
                    }
                }
                if( pms.size() <= 0 ) {
                    pms = null ;
                }
                String sql = buf.toString() ;
                buf = null ;
                ExecutionDao.executionSql( null,record,modelName,meta,false,sql,pms.getObjectArray(),pms.size() ) ;
            }
        } catch( MaachangDaoException md ) {
            throw md ;
        } catch( Exception e ) {
            throw new MaachangDaoException( e ) ;
        }
    }
    
    /**
     * 削除.
     * <BR><BR>
     * 指定条件のデータを削除します.
     * <BR>
     * @param session 対象のセッションを設定します.
     * @param srcJoin 対象のsrcJoin条件を設定します.
     * @param modelName 対象のモデル名を設定します.
     * @param where 条件文を設定します.
     * @param params 対照のパラメータ群を設定します.
     * @exception MaachangDaoException MaachangDao例外.
     */
    public void delete( DaoSession session,Map<String,String> srcJoin,String modelName,String where,Object... params )
        throws MaachangDaoException {
        if( session == null || modelName == null || ( modelName = modelName.trim() ).length() <= 0 ) {
            if( session == null ) {
                throw new MaachangDaoException( "セッション情報が設定されていません" ) ;
            }
            throw new MaachangDaoException( "モデル名が設定されていません" ) ;
        }
        try {
            deleteBySrcJoin( session,srcJoin,modelName,where,params ) ;// テーブル結合.
            Record record = session.getRecord() ;
            MetaColumn meta = session.getMeta( modelName ) ;
            StringBuilder buf = new StringBuilder() ;
            if( where == null || ( where = where.trim() ).length() <= 0 ) {
                buf.append( "delete from " ).append( meta.getTable() ) ;
                params = null ;
            }
            else {
                buf.append( "delete from " ).append( meta.getTable() ).
                    append( " " ).append( where ) ;
            }
            String sql = buf.toString() ;
            buf = null ;
            ExecutionDao.executionSql( null,record,modelName,meta,false,sql,params,
                ( ( params != null ) ? params.length : 0 ) ) ;
        } catch( MaachangDaoException md ) {
            throw md ;
        } catch( Exception e ) {
            throw new MaachangDaoException( e ) ;
        }
    }
    
    /**
     * 取得.
     * <BR><BR>
     * 指定条件のデータを取得します.
     * <BR>
     * @param session 対象のセッションを設定します.
     * @param join 対象のJoin条件を設定します.
     * @param srcJoin 対象のsrcJoin条件を設定します.
     * @param modelName 対象のモデル名を設定します.
     * @param where 条件文を設定します.
     * @param params 対照のパラメータ群を設定します.
     * @exception List<Map<String,Object>> 情報が返されます.
     * @exception MaachangDaoException MaachangDao例外.
     */
    public List<Map<String,Object>> find( DaoSession session,Map<String,Object> join,
        Map<String,Object> srcJoin,String modelName,String where,Object... params )
        throws MaachangDaoException {
        return find( session,join,srcJoin,modelName,where,-1,-1,params ) ;
    }
    
    /**
     * 取得.
     * <BR><BR>
     * 指定条件のデータを取得します.
     * <BR>
     * @param session 対象のセッションを設定します.
     * @param join 対象のJoin条件を設定します.
     * @param srcJoin 対象のsrcJoin条件を設定します.
     * @param modelName 対象のモデル名を設定します.
     * @param where 条件文を設定します.
     * @param offset 対象のオフセット値を設定します.
     * @param limit 対象のリミットを設定します.
     * @param params 対照のパラメータ群を設定します.
     * @exception List<Map<String,Object>> 情報が返されます.
     * @exception MaachangDaoException MaachangDao例外.
     */
    public List<Map<String,Object>> find( DaoSession session,Map<String,String> join,
        Map<String,Object> srcJoin,String modelName,String where,int offset,int limit,Object... params )
        throws MaachangDaoException {
        if( session == null || modelName == null || ( modelName = modelName.trim() ).length() <= 0 ) {
            if( session == null ) {
                throw new MaachangDaoException( "セッション情報が設定されていません" ) ;
            }
            throw new MaachangDaoException( "モデル名が設定されていません" ) ;
        }
        try {
            Record record = session.getRecord() ;
            SupportKind kind = record.getSupportKind() ;
            if( kind == null ) {
                throw new IOException( "サポート外のアダプタです" ) ;
            }
            MetaColumn meta = session.getMeta( modelName ) ;
            String sql = null ;
            List<String> joinNameList = null ;
            StringBuilder buf = null ;
            boolean joinMode = false ;
            // １対１テーブル結合.
            if( join != null && join.size() > 0 && srcJoin == null ) {
                joinNameList = new ArrayList<String>() ;
                buf = ExecutionDao.createJoinSQL( joinNameList,meta,session,join ) ;
                joinMode = false ;
            }
            // １対他テーブル結合.
            else if( srcJoin != null && srcJoin.size() > 0 && join == null ) {
                joinNameList = new ArrayList<String>() ;
                if( offset != -1 && limit != -1 ) {
                    // リミット付きの場合は、通常SQL文で処理.
                    buf = ExecutionDao.createSQL( meta,session ) ;
                }
                else {
                    // リミット無しの場合は、ManySQL文で処理.
                    buf = ExecutionDao.createManySQL( joinNameList,meta,session,srcJoin ) ;
                }
                joinMode = true ;
            }
            // 通常呼び出し.
            else {
                buf = ExecutionDao.createSQL( meta,session ) ;
            }
            if( where == null || ( where = where.trim() ).length() <= 0 ) {
                params = null ;
            }
            else {
                buf.append( ExecutionDao.addWhereByTableSimbol( where,meta ) ) ;
            }
            // リミットつきSQL発行.
            if( kind.isLimit() && offset != -1 && limit != -1 ) {
                buf.append( kind.getLimit( false,offset,limit ) ) ;
            }
            sql = buf.toString() ;
            buf = null ;
            List<Map<String,Object>> ret = null ;
            // joinテーブル結合結果を処理.
            if( joinNameList != null ) {
                // １対他テーブル結合結果で情報取得.
                if( joinMode == true ) {
                    if( offset != -1 && limit != -1 ) {
                        // 手動結合(1対他).
                        ret = ExecutionDao.executionHandMany( record,modelName,meta,sql,meta.size(),offset,limit,params,
                            ( ( params != null ) ? params.length : 0 ),srcJoin) ;
                    }
                    else {
                        // SQL結合(1対他).
                        ret = ExecutionDao.executionMany( record,modelName,meta,sql,params,
                            ( ( params != null ) ? params.length : 0 ),joinNameList ) ;
                    }
                }
                // １対１テーブル結合結果で情報取得.
                else {
                    ret = ExecutionDao.executionJoin( record,modelName,meta,sql,offset,limit,params,
                        ( ( params != null ) ? params.length : 0 ),join,joinNameList ) ;
                }
            }
            else {
                // 通常処理.
                ret = ExecutionDao.executionSelect( record,modelName,meta,sql,meta.size(),offset,limit,params,
                    ( ( params != null ) ? params.length : 0 )) ;
            }
            return ret ;
        } catch( MaachangDaoException md ) {
            throw md ;
        } catch( Exception e ) {
            throw new MaachangDaoException( e ) ;
        }
    }
    
    /**
     * １行単位取得.
     * @param out 格納対象の条件を設定します.
     * @param session 対象のセッションを設定します.
     * @param modelName 対象のモデル名を設定します.
     * @param where 条件文を設定します.
     * @param params 対照のパラメータ群を設定します.
     * @exception MaachangDaoException MaachangDao例外.
     */
    public void readLine( ResultLine out,DaoSession session,String modelName,String where,Object... params )
        throws MaachangDaoException {
        if( out == null ) {
            throw new MaachangDaoException( "戻り情報が設定されていません" ) ;
        }
        if( session == null || modelName == null || ( modelName = modelName.trim() ).length() <= 0 ) {
            if( session == null ) {
                throw new MaachangDaoException( "セッション情報が設定されていません" ) ;
            }
            throw new MaachangDaoException( "モデル名が設定されていません" ) ;
        }
        try {
            ResultSet rs = null ;
            Record record = session.getRecord() ;
            MetaColumn meta = session.getMeta( modelName ) ;
            StringBuilder buf = ExecutionDao.createSQL( meta,session ) ;
            String sql = null ;
            if( where == null || ( where = where.trim() ).length() <= 0 ) {
                sql = buf.toString() ;
                params = null ;
            }
            else {
                sql = buf.append( ExecutionDao.addWhereByTableSimbol( where,meta ) ).toString() ;
            }
            buf = null ;
            SupportKind kind = record.getSupportKind() ;
            if( kind == null ) {
                throw new MaachangDaoException( "サポート外のアダプタです" ) ;
            }
            if( params == null || params.length <= 0 ) {
                rs = record.executeQuery( sql ) ;
            }
            else {
                rs = record.executeQuery( false,sql,params,( ( params != null ) ? params.length : 0 ) ) ;
            }
            out.create( modelName,rs,meta ) ;
        } catch( MaachangDaoException md ) {
            throw md ;
        } catch( Exception e ) {
            throw new MaachangDaoException( e ) ;
        }
    }
    
    /**
     * 情報数を取得.
     * <BR><BR>
     * 指定条件の情報数を取得します.
     * <BR>
     * @param session 対象のセッションを設定します.
     * @param modelName 対象のモデル名を設定します.
     * @exception int 情報数が返されます.
     * @exception MaachangDaoException MaachangDao例外.
     */
    public int count( DaoSession session,String modelName )
        throws MaachangDaoException {
        return count( session,modelName,null,( Object[] )null ) ;
    }
    
    /**
     * 情報数を取得.
     * <BR><BR>
     * 指定条件の情報数を取得します.
     * <BR>
     * @param session 対象のセッションを設定します.
     * @param modelName 対象のモデル名を設定します.
     * @param where 条件文を設定します.
     * @param params 対照のパラメータ群を設定します.
     * @exception int 情報数が返されます.
     * @exception MaachangDaoException MaachangDao例外.
     */
    public int count( DaoSession session,String modelName,String where,Object... params )
        throws MaachangDaoException {
        if( session == null || modelName == null || ( modelName = modelName.trim() ).length() <= 0 ) {
            if( session == null ) {
                throw new MaachangDaoException( "セッション情報が設定されていません" ) ;
            }
            throw new MaachangDaoException( "モデル名が設定されていません" ) ;
        }
        try {
            Record record = session.getRecord() ;
            MetaColumn meta = session.getMeta( modelName ) ;
            String sql = null ;
            if( where == null || ( where = where.trim() ).length() <= 0 ) {
                sql = new StringBuilder().append( "select count(id) from " ).
                    append( meta.getTable() ).toString() ;
                params = null ;
            }
            else {
                StringBuilder buf = new StringBuilder() ;
                buf.append( "select count(id) from " ).append( meta.getTable() ).
                    append( " " ).append( where ) ;
                sql = buf.toString() ;
                buf = null ;
            }
            return ExecutionDao.executionCount( record,modelName,meta,sql,params,( ( params != null ) ? params.length : 0 ) ) ;
        } catch( MaachangDaoException md ) {
            throw md ;
        } catch( Exception e ) {
            throw new MaachangDaoException( e ) ;
        }
    }
    
    /**
     * Select系SQL文を直接実行.
     * @param session 対象のセッションを設定します.
     * @param modelName 対象のモデル名を設定します.
     * @param sql SQL文全文を設定します.
     * @param offset 対象のオフセット値を設定します.
     * @param limit 対象のリミットを設定します.
     * @param params 対照のパラメータ群を設定します.
     * @exception List<Map<String,Object>> 情報が返されます.
     * @exception MaachangDaoException MaachangDao例外.
     */
    public List<Map<String,Object>> selectSQL( DaoSession session,String modelName,String sql,int offset,int limit,Object... params )
        throws MaachangDaoException {
        if( session == null ) {
            throw new MaachangDaoException( "セッション情報が設定されていません" ) ;
        }
        else if( sql == null || sql.length() <= 0 ) {
            throw new MaachangDaoException( "実行対象のSQL文は設定されていません" ) ;
        }
        sql = StringUtil.trim( sql ) ;
        if( sql.endsWith( ";" ) ) {
            sql = StringUtil.trim( sql.substring( 0,sql.length()-1 ) ) ;
        }
        try {
            Record record = session.getRecord() ;
            SupportKind kind = record.getSupportKind() ;
            if( kind == null ) {
                throw new IOException( "サポート外のアダプタです" ) ;
            }
            StringBuilder buf = new StringBuilder().append( sql ).append( " " ) ;
            // リミットつきSQL発行.
            if( kind.isLimit() && offset != -1 && limit != -1 ) {
                buf.append( kind.getLimit( false,offset,limit ) ) ;
            }
            sql = buf.toString() ;
            buf = null ;
            // SQL実行処理.
            return ExecutionDao.executionNormalSelect( record,modelName,sql,offset,limit,params,
                ( ( params != null ) ? params.length : 0 )) ;
        } catch( MaachangDaoException md ) {
            throw md ;
        } catch( Exception e ) {
            throw new MaachangDaoException( e ) ;
        }
    }
    
    /**
     * 楽観的ロック条件を取得.
     * @param session 対象のセッションを設定します.
     * @param modelName 対象のモデル名を設定します.
     * @param where 他条件を設定します.
     * @param value チェック対象の情報を設定します.
     * @exception MaachangDaoException MaachangDao例外.
     */
    public void optimisticLock( DaoSession session,String modelName,Map<String,Object> where,Map<String,Object> value )
        throws MaachangDaoException {
        if( session == null || modelName == null || ( modelName = modelName.trim() ).length() <= 0 ) {
            if( session == null ) {
                throw new MaachangDaoException( "セッション情報が設定されていません" ) ;
            }
            throw new MaachangDaoException( "モデル名が設定されていません" ) ;
        }
        ResultSet rs = null ;
        try {
            Record record = session.getRecord() ;
            MetaColumn meta = session.getMeta( modelName ) ;
            int len = meta.size() ;
            // 元情報に楽観的ロックIDが存在しない場合.
            Object optLock = value.get( COLUMN_OPTIMISTIC_LOCK_JAVA_NAME ) ;
            if( optLock == null || ( optLock instanceof Long ) == false ) {
                return ;
            }
            boolean lockFlag = false ;
            // 楽観的ロックIDが存在するかチェック.
            for( int i = 0 ; i < len ; i ++ ) {
                if( ExecutionDao.OPTIMISTIC_LOCK_COLUMN.equals( meta.getColumnName( i ) ) ) {
                    lockFlag = true ;
                    break ;
                }
            }
            // 存在しない場合は、処理しない.
            if( !lockFlag ) {
                return ;
            }
            // 一番初めの楽観的ロック例外を取得.
            Long lockId = null ;
            StringBuilder buf = new StringBuilder().
                append( "select " ).append( ExecutionDao.OPTIMISTIC_LOCK_COLUMN ).append( " from " ).
                append( meta.getTable() ).append( " where id=?" ) ;
            if( where != null ) {
                len = where.size() ;
            }
            else {
                len = 0 ;
            }
            Object[] pms = new Object[ len+1 ] ;
            pms[ 0 ] = value.get( "id" ) ;
            if( where != null ) {
                int cnt = 1 ;
                Iterator<String> it = where.keySet().iterator() ;
                while( it.hasNext() ) {
                    String k = ( String )it.next() ;
                    buf.append( " and " ).append( DbUtil.convertDBNameByJavaName( false,k ) ).append( "=?" ) ;
                    pms[ cnt ] = where.get( k ) ;
                }
            }
            String sql = buf.toString() ;
            buf = null ;
            rs = record.executeQuery( false,sql,pms,1 ) ;
            if( rs != null && rs.next() == true ) {
                lockId = rs.getLong( ExecutionDao.OPTIMISTIC_LOCK_COLUMN ) ;
            }
            rs.close() ;
            rs = null ;
            if( lockId == null ) {
                throw new OptimisticLockException( "[" + modelName + "] - use optimistic lock.(src:" +
                    optLock + " now:[delete])" ) ;
            }
            else if( lockId != null && !lockId.equals( optLock ) ) {
                throw new OptimisticLockException( "[" + modelName + "] - use optimistic lock.(src:" +
                    optLock + " now:" + lockId + ")" ) ;
            }
        } catch( MaachangDaoException me ) {
            throw me ;
        } catch( Exception e ) {
            throw new MaachangDaoException( e ) ;
        } finally {
            if( rs != null ) {
                try {
                    rs.close() ;
                } catch( Exception e ) {
                }
            }
        }
    }
    
    /**
     * 情報存在時のチェック処理.
     */
    private boolean isEntry( Record record,boolean idFlag,MetaColumn meta,Map<String,Object> model,String modelName )
        throws Exception {
        if( !idFlag ) {
            return false ;
        }
        Object value = model.get( ExecutionDao.SEQ_COLUMN ) ;
        if( value == null ) {
            return false ;
        }
        // 楽観的ロック対象のカラムが存在する.
        if( ExecutionDao.isOptimisticLock( meta ) ) {
            // 現在の楽観的ロック情報を取得.
            StringBuilder buf = new StringBuilder() ;
            buf.append( "select " ).append( ExecutionDao.OPTIMISTIC_LOCK_COLUMN ).append( " from " ).
                append( DbUtil.convertJavaNameByDBName( modelName ) ).
                append( " where id=?" ) ;
            String sql = buf.toString() ;
            buf = null ;
            ResultSet rs = null ;
            Long optimisticLockValue = convertObjectByLong( model.get( COLUMN_OPTIMISTIC_LOCK_JAVA_NAME ) ) ;
            try {
                boolean ret = false ;
                rs = record.executeQuery( false,sql,new Object[]{value},1 ) ;
                if( rs != null ) {
                    if( rs.next() == true ) {
                        // 楽観的ロックカラム内容が存在する場合.
                        if( optimisticLockValue != null ) {
                            Long nowLock = rs.getLong( ExecutionDao.OPTIMISTIC_LOCK_COLUMN ) ;
                            // 楽観的ロックが一致しない場合は、例外を出力.
                            if( !optimisticLockValue.equals( nowLock ) ) {
                                throw new OptimisticLockException( "[" + modelName + "] - use optimistic lock.(src:" +
                                    optimisticLockValue + " now:" + nowLock + ")" ) ;
                            }
                            // 更新対象の楽観的ロックを１インクリメント.
                            optimisticLockValue = optimisticLockValue + 1L ;
                            model.put( COLUMN_OPTIMISTIC_LOCK_JAVA_NAME,optimisticLockValue ) ;
                        }
                        // 楽観的ロックカラム内容が存在しない場合は、初期値を付加.
                        else {
                            model.put( COLUMN_OPTIMISTIC_LOCK_JAVA_NAME,new Long( 1 ) ) ;
                        }
                        ret = true ;
                    }
                    rs.close() ;
                    rs = null ;
                }
                return ret ;
            } finally {
                if( rs != null ) {
                    try {
                        rs.close() ;
                    } catch( Exception e ) {
                    }
                }
            }
        }
        // 楽観的ロック対象のカラムが存在しない場合.
        else {
            StringBuilder buf = new StringBuilder() ;
            buf.append( "select count(id) from " ).
                append( DbUtil.convertJavaNameByDBName( modelName ) ).
                append( " where id=?" ) ;
            String sql = buf.toString() ;
            buf = null ;
            
            ResultSet rs = null ;
            try {
                boolean ret = false ;
                rs = record.executeQuery( false,sql,new Object[]{value},1 ) ;
                if( rs != null ) {
                    if( rs.next() ) {
                        if( rs.getInt( 1 ) > 0 ) {
                            ret = true ;
                        }
                    }
                    rs.close() ;
                    rs = null ;
                }
                return ret ;
            } finally {
                if( rs != null ) {
                    try {
                        rs.close() ;
                    } catch( Exception e ) {
                    }
                }
            }
        }
    }
    
    /**
     * アップロード時間をSQL文に追加.
     */
    private static final String addUpdateTime( boolean[] addFlag,String sql ) throws Exception {
        addFlag[0] = false ;
        String lower = sql.toLowerCase() ;
        int s = lower.indexOf( " set " ) ;
        int e = lower.indexOf( " where " ) ;
        if( s == -1 ) {
            return sql ;
        }
        else if( e == -1 ) {
            e = lower.length() ;
        }
        s += " set ".length() ;
        int x = lower.indexOf( ExecutionDao.UPDATE_TIME,s ) ;
        lower = null ;
        if( x == -1 || x > e ) {
            sql = new StringBuilder().append( sql.substring( 0,s ) ).
                append( ExecutionDao.UPDATE_TIME ).append( " =?, " ).
                append( sql.substring( s,sql.length() ) ).toString() ;
            addFlag[0] = true ;
        }
        return sql ;
    }
    
    /**
     * アップロード時間をパラメータに追加.
     */
    private static final Object[] addUpdateTimeParam( Object[] params ) throws Exception {
        Object[] ret = null ;
        if( params == null || params.length <= 0 ) {
            ret = new Object[ 1 ] ;
        }
        else {
            int len = params.length ;
            ret = new Object[ len+1 ] ;
            System.arraycopy( params,0,ret,1,len ) ;
        }
        ret[ 0 ] = new java.sql.Timestamp( System.currentTimeMillis() ) ;
        return ret ;
    }
    
    /**
     * delete条件に対して、結合テーブル情報を削除.
     */
    private static final void deleteBySrcJoin( DaoSession session,Map<String,String> srcJoin,String modelName,String where,Object... params )
        throws Exception {
        if( srcJoin == null || srcJoin.size() <= 0 ) {
            return ;
        }
        Record record = session.getRecord() ;
        MetaColumn meta = session.getMeta( modelName ) ;
        String sql = null ;
        if( where == null || ( where = where.trim() ).length() <= 0 ) {
            sql = new StringBuilder().append( "select id from " ).
                append( meta.getTable() ).toString() ;
            params = null ;
        }
        else {
            sql = new StringBuilder().append( "select id from " ).
                append( meta.getTable() ).append( " " ).append( where ).toString() ;
        }
        List<Map<String,Object>> out = ExecutionDao.executionSelect( record,modelName,meta,sql,1,0,Integer.MAX_VALUE,params,
            ( ( params != null ) ? params.length : 0 ) ) ;
        if( out != null && out.size() > 0 ) {
            int len = out.size() ;
            for( int i = 0 ; i < len ; i ++ ) {
                Map<String,Object> o = out.get( i ) ;
                Object id = o.get( "id" ) ;
                if( id != null ) {
                    Iterator it = srcJoin.keySet().iterator() ;
                    while( it.hasNext() ) {
                        String table = ( String )it.next() ;
                        String column = srcJoin.get( table ) ;
                        if( column != null && ( column = column.trim() ).length() > 0 ) {
                            removeTableById( session,DbUtil.convertJavaNameByDBName( table ),
                                DbUtil.convertJavaNameByDBName( column ),id ) ;
                        }
                    }
                }
            }
        }
    }
    
    /**
     * テーブルIDの内容を削除.
     */
    private static final void removeTableById( DaoSession session,String modelName,String column,Object id )
        throws Exception {
        Record record = session.getRecord() ;
        MetaColumn meta = session.getMeta( modelName ) ;
        String sql = new StringBuilder().append( "delete from " ).
            append( meta.getTable() ).append( " where " ).append( column ).append( "=?" ).toString() ;
        ExecutionDao.executionSql( null,record,modelName,meta,false,sql,new Object[]{ id },1 ) ;
    }
    
    /**
     * 指定パラメータをLongで取得.
     */
    private static final Long convertObjectByLong( Object o ) {
        if( o == null ) {
            return null ;
        }
        if( o instanceof String ) {
            return new Long( ( String )o ) ;
        }
        else if( o instanceof Integer ) {
            return new Long( ( long )( ( Integer )o ).intValue() ) ;
        }
        else if( o instanceof Long ) {
            return ( Long )o ;
        }
        return new Long( o.toString() ) ;
    }
}
