package org.maachang.dao.dbms;

import java.io.IOException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.dao.dbms.kind.SupportKind;

/**
 * DBアクセスオブジェクト実装.
 * 
 * @version 2007/10/18
 * @author masahito suzuki
 * @since MaachangDao 1.00
 */
public class BaseRecord implements Record {
    
    /**
     * LOG.
     */
    private static final Log LOG = LogFactory.getLog( BaseRecord.class ) ;
    
    /**
     * サポートDBMS種類.
     */
    private SupportKind kind = null ;

    /**
     * Booleanサポートフラグ.
     */
    private boolean supportBoolean = false ;

    /**
     * コネクション.
     */
    private Connection connection = null;

    /**
     * コミットロールバックフラグ.
     */
    private boolean commitRollbackFlag = true;

    /**
     * シーケンス取得条件.
     */
    private boolean sequence = false ;

    /**
     * ステートメント管理(read).
     */
    private Statement rstm = null ;
    
    /**
     * ステートメント管理(write).
     */
    private Statement wstm = null ;
    
    /**
     * コンパイル済みSQL文管理.
     */
    private final Map<String,PreparedStatement> preMap = new HashMap<String,PreparedStatement>() ;
    
    /**
     * デバッグモード.
     */
    private boolean debug = false ;

    /**
     * コンストラクタ.
     */
    private BaseRecord() {

    }

    /**
     * コンストラクタ. <BR>
     * <BR>
     * コネクションオブジェクトを設定してオブジェクトを生成します. <BR>
     * 
     * @param adapter
     *            対象のアダプタ名を設定します.
     * @param connection
     *            コネクションオブジェクト.
     * @exception Exception
     *                例外.
     */
    public BaseRecord(SupportKind kind,boolean debug,Connection connection) throws Exception {
        if (connection == null) {
            throw new IllegalArgumentException("コネクションオブジェクトは存在しません");
        }
        this.kind = kind ;
        this.debug = debug ;
        this.supportBoolean = kind.isBoolean() ;
        this.sequence = ( kind.getSequenceId( null ) != null ) ? true : false ;
        this.connection = connection;
    }

    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        this.close();
    }

    /**
     * クローズ処理. <BR>
     * <BR>
     * オブジェクトをクローズします.
     */
    public void close() {
        if (commitRollbackFlag == false) {
            if (connection != null) {
                try {
                    connection.rollback();
                } catch (Exception e) {
                }
            }
        }
        else {
            if (connection != null) {
                try {
                    connection.commit();
                } catch (Exception e) {
                }
            }
        }
        if (connection != null) {
            try {
                connection.clearWarnings() ;
            } catch( Exception e ) {
            }
            try {
                connection.close();
            } catch (Exception e) {
            }
        }
        clearPreStatements() ;
        clearStatements() ;
        commitRollbackFlag = true;
        connection = null;
    }

    /**
     * コネクションオブジェクトを取得. <BR>
     * <BR>
     * コネクションオブジェクトを取得します. <BR>
     * 
     * @return Connection コネクションオブジェクトが返されます.
     */
    public Connection getConnection() {
        return connection;
    }

    /**
     * コミット処理. <BR>
     * <BR>
     * コミット処理を実施します. <BR>
     * 
     * @return boolean [true]の場合、コミット処理が呼ばれました.
     * @exception Exception 例外.
     */
    public boolean commit() throws Exception {
        if (connection != null) {
            if( !connection.getAutoCommit() ) {
                if( debug && LOG.isDebugEnabled() ) {
                    LOG.debug( "# sql:commit" ) ;
                }
                if( commitRollbackFlag == false ) {
                    connection.commit();
                    commitRollbackFlag = true;
                    return true ;
                }
                else {
                    return false ;
                }
            }
        }
        return false ;
    }

    /**
     * ロールバック処理. <BR>
     * <BR>
     * ロールバック処理を実施します. <BR>
     * 
     * @return boolean [true]の場合、ロールバック処理が呼ばれました.
     * @exception Exception 例外.
     */
    public boolean rollback() throws Exception {
        if (connection != null) {
            if( !connection.getAutoCommit() ) {
                if( debug && LOG.isDebugEnabled()) {
                    LOG.debug( "# sql:rollback" ) ;
                }
                connection.rollback();
                commitRollbackFlag = true;
                //clearPreStatements() ;
                //clearStatements() ;
                return true ;
            }
        }
        return false ;
    }

    /**
     * 管理ステートメントをクリア.
     * <BR><BR>
     * 管理しているステートメントをクリアします.
     */
    public void clearStatements() {
        Statement st = rstm ;
        rstm = null ;
        if( st != null ) {
            try {
                st.close() ;
            } catch( Exception e ) {
            }
        }
        st = wstm ;
        wstm = null ;
        if( st != null ) {
            try {
                st.close() ;
            } catch( Exception e ) {
            }
        }
    }

    /**
     * 管理コンパイル済みステートメントをクリア.
     * <BR><BR>
     * 管理しているコンパイル済みステートメントをクリアします.
     */
    public void clearPreStatements() {
        Iterator it = preMap.keySet().iterator() ;
        while( it.hasNext() ) {
            try {
                preMap.get( ( String )it.next() ).close() ;
            } catch( Exception e ) {
            }
        }
        preMap.clear() ;
    }

    /**
     * メタカラムを取得. <BR>
     * <BR>
     * メタカラムを取得します. <BR>
     * 
     * @param table
     *            対象のテーブル名を設定します.
     * @return MetaColumn メタカラムが返されます.
     * @exception Exception
     *                例外.
     */
    public MetaColumn getMetaColumn(String table) throws Exception {
        return MetaFactory.getInstance().getMetaColumn( false,table,this ) ;
    }

    /**
     * メタカラムを取得. <BR>
     * <BR>
     * メタカラムを取得します. <BR>
     * 
     * @param mode
     *            [true]の場合、名前をDB名に変換して処理します.
     * @param table
     *            対象のテーブル名を設定します.
     * @return MetaColumn メタカラムが返されます.
     * @exception Exception
     *                例外.
     */
    public MetaColumn getMetaColumn(boolean mode, String table) throws Exception {
        return MetaFactory.getInstance().getMetaColumn( mode,table,this ) ;
    }

    /**
     * このオブジェクトが利用可能かチェック. <BR>
     * <BR>
     * このオブジェクトが利用可能かチェックします. <BR>
     * 
     * @return boolean [true]の場合、利用可能です.
     */
    public boolean isUse() {
        boolean ret = false;
        try {
            ret = (connection == null || connection.isClosed() == true) ? false
                    : true;
        } catch (Exception e) {
            ret = false;
        }
        return ret;
    }

    /**
     * データ取得用ステートメントを実行. <BR>
     * <BR>
     * データ取得用のステートメントを実行します. <BR>
     * 
     * @param sql
     *            対象のSQLを設定します.
     * @return ResultSet 対象の結果情報が返されます.
     * @exception Exception
     *                例外.
     */
    public ResultSet executeQuery(String sql) throws Exception {
        if (connection == null) {
            throw new IOException("コネクション情報は存在しません");
        }
        if (sql == null || sql.length() <= 0) {
            throw new IllegalArgumentException("引数は不正です");
        }
        if( debug && LOG.isDebugEnabled() ) {
            LOG.debug( "# sql:" + sql ) ;
        }
        long time = System.currentTimeMillis() ;
        ResultSet ret = getStatement(true).executeQuery(sql);
        if( debug && LOG.isDebugEnabled() ) {
            LOG.debug( "# sql:" + sql + "["+( System.currentTimeMillis()-time )+"ms]" ) ;
        }
        commitRollbackFlag = false;
        return ret;
    }

    /**
     * データ取得用ステートメントを実行. <BR>
     * <BR>
     * データ取得用のステートメントを実行します. <BR>
     * 
     * @param sql
     *            対象のSQLを設定します.
     * @return int 対象の結果情報が返されます.
     * @exception Exception
     *                例外.
     */
    public int executeUpdate(String sql) throws Exception {
        if (connection == null) {
            throw new IOException("コネクション情報は存在しません");
        }
        if (sql == null || sql.length() <= 0) {
            throw new IllegalArgumentException("引数は不正です");
        }
        if( debug && LOG.isDebugEnabled() ) {
            LOG.debug( "# sql:" + sql ) ;
        }
        long time = System.currentTimeMillis() ;
        int ret = getStatement(false).executeUpdate(sql);
        if( debug && LOG.isDebugEnabled() ) {
            LOG.debug( "# sql:" + sql + "["+( System.currentTimeMillis()-time )+"ms]" ) ;
        }
        commitRollbackFlag = false;
        return ret;
    }

    /**
     * ステートメントを取得.
     */
    private Statement getStatement( boolean mode )
        throws Exception {
        if( mode ) {
            Statement st = rstm ;
            if( st == null ) {
                st = connection.createStatement(
                    ResultSet.TYPE_SCROLL_INSENSITIVE,
                    ResultSet.CONCUR_READ_ONLY ) ;
                rstm = st ;
            }
            return st ;
        }
        else {
            Statement st = wstm ;
            if( st == null ) {
                st = connection.createStatement(
                    ResultSet.TYPE_FORWARD_ONLY,
                    ResultSet.CONCUR_READ_ONLY ) ;
                wstm = st ;
            }
            return st ;
        }
    }

    /**
     * データ取得用ステートメントを実行. <BR>
     * <BR>
     * データ取得用のステートメントを実行します. <BR>
     * 
     * @param notCache
     *            対象SQLでキャッシュをしたくない場合は、[true]を設定します.
     * @param sql
     *            対象のパースされたSQLを設定します.
     * @param meta
     *            対象のメタデータを設定します.
     * @param params
     *            対象のパラメータを設定します.
     * @param size
     *            対象のパラメータ長を設定します.
     * @return ResultSet 対象の結果情報が返されます.
     * @exception Exception
     *                例外.
     */
    public ResultSet executeQuery(boolean notCache,String sql, Object[] params,int size)
            throws Exception {
        if (connection == null) {
            throw new IOException("コネクション情報は存在しません");
        }
        if (sql == null || sql.length() <= 0) {
            throw new IllegalArgumentException("引数は不正です");
        }
        if( debug && LOG.isDebugEnabled() ) {
            LOG.debug( "# sql:" + sql ) ;
        }
        long time = System.currentTimeMillis() ;
        PreparedStatement pre = getPreparedStatement(true,notCache,sql) ;
        putParams(supportBoolean,pre, params,size);
        ResultSet ret = pre.executeQuery();
        commitRollbackFlag = false;
        if( debug && LOG.isDebugEnabled() ) {
            LOG.debug( "# sql:" + sql + "["+( System.currentTimeMillis()-time )+"ms]" ) ;
        }
        return ret;
    }

    /**
     * データ取得用ステートメントを実行. <BR>
     * <BR>
     * データ取得用のステートメントを実行します. <BR>
     * 
     * @param sql
     *            対象のパースされたSQLを設定します.
     * @param meta
     *            対象のメタデータを設定します.
     * @param params
     *            対象のパラメータを設定します.
     * @param size
     *            対象のパラメータ長を設定します.
     * @return int 対象の結果情報が返されます.
     * @exception Exception
     *                例外.
     */
    public int executeUpdate(String sql, Object[] params,int size)
            throws Exception {
        if (connection == null) {
            throw new IOException("コネクション情報は存在しません");
        }
        if (sql == null || sql.length() <= 0) {
            throw new IllegalArgumentException("引数は不正です");
        }
        if( debug && LOG.isDebugEnabled() ) {
            LOG.debug( "# sql:" + sql ) ;
        }
        long time = System.currentTimeMillis() ;
        PreparedStatement pre = getPreparedStatement(false,false,sql);
        putParams(supportBoolean,pre, params,size);
        int ret = pre.executeUpdate();
        commitRollbackFlag = false;
        if( debug && LOG.isDebugEnabled() ) {
            LOG.debug( "# sql:" + sql + "["+( System.currentTimeMillis()-time )+"ms]" ) ;
        }
        return ret;
    }

    /**
     * PreparedStatementを取得.
     */
    private PreparedStatement getPreparedStatement( boolean readOnly,boolean notCache,String sql )
        throws Exception {
        if( sql == null || sql.length() <= 0 ||
            connection == null ) {
            return null ;
        }
        PreparedStatement ret = null ;
        if( notCache ) {
            if( readOnly == true ) {
                ret = connection.prepareStatement(sql,
                    ResultSet.TYPE_SCROLL_INSENSITIVE,
                    ResultSet.CONCUR_READ_ONLY) ;
            }
            else {
                ret = connection.prepareStatement(sql,
                    ResultSet.TYPE_FORWARD_ONLY,
                    ResultSet.CONCUR_READ_ONLY ) ;
            }
        }
        else {
            ret = preMap.get( sql ) ;
            if( ret == null ) {
                if( readOnly == true ) {
                    ret = connection.prepareStatement(sql,
                        ResultSet.TYPE_SCROLL_INSENSITIVE,
                        ResultSet.CONCUR_READ_ONLY) ;
                }
                else {
                    ret = connection.prepareStatement(sql,
                        ResultSet.TYPE_FORWARD_ONLY,
                        ResultSet.CONCUR_READ_ONLY ) ;
                }
                preMap.put( sql,ret ) ;
            }
        }
        return ret ;
    }

    /**
     * サポートDBMS種類オブジェクトを取得.
     * <BR><BR>
     * サポートDBMS種類オブジェクトを取得します.
     * <BR>
     * @return SupportKind サポートDBMS種類が返されます.
     */
    public SupportKind getSupportKind() {
        return kind ;
    }
    
    /**
     * オートコミットチェック.
     * <BR><BR>
     * オートコミットチェックします.
     * <BR>
     * @return boolean [true]の場合、オートコミットがONです.
     */
    public boolean isAutoCommit() {
        boolean ret = false ;
        try {
            if( connection == null ) {
                ret = false ;
            }
            ret = connection.getAutoCommit() ;
        } catch( Exception e ) {
            ret = false ;
        }
        return ret ;
    }

    /**
     * シーケンス取得条件を取得.
     * <BR><BR>
     * シーケンス取得条件を取得します.
     * <BR>
     * @return boolean [true]の場合は、前に取得できます.
     */
    public boolean isSequence() {
        return sequence ;
    }

    /**
     * デバッグモードを取得.
     * <BR>
     * @return [true]の場合、デバッグは有効です.
     */
    public boolean isDebug() {
        return debug ;
    }

    /**
     * 各パラメータを設定.
     */
    private static final void putParams(boolean supportBoolean,PreparedStatement pre,
            Object[] params,int size) throws Exception {
        if( params != null && params.length >= size ) {
            for (int i = 0; i < size; i++) {
                putOneParam(supportBoolean,i, pre, params[i]);
            }
        }
    }

    /**
     * １つの条件に対して、条件を設定.
     */
    private static final void putOneParam(boolean supportBoolean,int no, PreparedStatement pre,
        Object one) throws Exception {
        if( one == null ) {
            pre.setObject( no + 1,null );
        }
        else if (one instanceof Boolean) {
            boolean b = ((Boolean) one).booleanValue() ;
            if( supportBoolean == false ) {
                if( b == true ) {
                    pre.setString(no + 1, "1");
                }
                else {
                    pre.setString(no + 1, "0");
                }
            }
            else {
                pre.setBoolean(no + 1, b);
            }
        } else if (one instanceof String) {
            pre.setString(no + 1, (String) one);
        } else if (one instanceof Integer) {
            pre.setInt(no + 1, ((Integer) one).intValue());
        } else if (one instanceof Long) {
            pre.setLong(no + 1, ((Long) one).longValue());
        } else if (one instanceof Float) {
            pre.setFloat(no + 1, ((Float) one).floatValue());
        } else if (one instanceof Double) {
            pre.setDouble(no + 1, ((Double) one).doubleValue());
        } else if (one instanceof java.util.Date) {
            if (one instanceof java.sql.Timestamp) {
                pre.setTimestamp(no + 1, (java.sql.Timestamp) one);
            } else if (one instanceof java.sql.Time) {
                pre.setTime(no + 1, (java.sql.Time) one);
            } else if (one instanceof java.sql.Date) {
                pre.setDate(no + 1, (java.sql.Date) one);
            } else {
                pre.setTimestamp(no + 1, new java.sql.Timestamp(
                        ((java.util.Date) one).getTime()));
            }
        } else if (one instanceof byte[]) {
            //pre.setBlob(no + 1, new ByteArrayInputStream((byte[]) one));
            pre.setBytes(no + 1,( byte[] )one ) ;
        } else {
            throw new IllegalArgumentException("指定パラメータ[" + (no + 1)
                    + "]の条件はオブジェクトが一致しません:" + one.getClass().getName());
        }
    }
}
