package org.maachang.dao.dbms;

import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.sql.ResultSet;
import java.text.SimpleDateFormat;
import java.util.HashMap;
import java.util.Map;

import org.maachang.conf.Config;
import org.maachang.dao.ExecutionDao;
import org.maachang.dao.dbms.ctbl.CreateTable;
import org.maachang.dao.dbms.ctbl.InsertTable;
import org.maachang.dao.dbms.ctbl.Table;
import org.maachang.dao.dbms.ctbl.TableMaster;
import org.maachang.dao.dbms.kind.SupportKind;
import org.maachang.util.Base64;

/**
 * テーブル操作用オブジェクト.
 * 
 * @version 2010/08/28
 * @author masahito suzuki
 * @since MaachangDao 1.14
 */
public abstract class OperationTable {
    OperationTable() {}
    
    /**
     * 指定テーブル情報を再構築.
     * @param factory 対象のレコードファクトリを設定します.
     * @param confDir 対象のコンフィグディレクトリを設定します.
     * @param model 対象のモデル名を設定します.
     * @exception Exception 例外.
     */
    public static final void recreateTable( RecordFactory factory,String confDir,String model )
        throws Exception {
        SupportKind dbms = factory.getSupportKind() ;
        if( dbms == null ) {
            throw new IOException( "アダプタ名が非サポートのデータベースです." ) ;
        }
        Table table = TableMaster.get( dbms.getAdapter() ) ;
        Record record = factory.getRecord() ;
        // テーブルコンフィグ一覧を取得.
        Config conf = CreateTable.readTableConfig( true,true,confDir ) ;
        // テーブル再構築.
        CreateTable.recreateTable( true,model,table,conf,record ) ;
    }
    
    /**
     * 指定テーブル情報に対象CSVをImport.
     * @param factory 対象のレコードファクトリを設定します.
     * @param record 対象のレコードオブジェクトを設定します.
     * @param confDir 対象のコンフィグディレクトリを設定します.
     * @param model 対象のモデル名を設定します.
     * @param reader インポート対象のInputStreamReaderを設定します.
     * @return int インポート件数が返されます.
     * @exception Exception 例外.
     */
    public static final int importCsv( RecordFactory factory,Record record,String confDir,String model,Reader reader )
        throws Exception {
        if( reader == null ) {
            throw new IllegalArgumentException( "Reader情報が設定されていません" ) ;
        }
        SupportKind dbms = factory.getSupportKind() ;
        if( dbms == null ) {
            throw new IOException( "アダプタ名が非サポートのデータベースです." ) ;
        }
        Table table = TableMaster.get( dbms.getAdapter() ) ;
        // テーブルコンフィグ一覧を取得.
        Config conf = CreateTable.readTableConfig( true,true,confDir ) ;
        // テーブル再構築.
        return InsertTable.insertCSV( true,model+".csv",table,record,conf,reader ) ;
    }
    
    /**
     * 指定テーブル情報をExport.
     * @param writer 出力先のオブジェクトを設定します.
     * @param record 対象のレコードオブジェクトを設定します.
     * @param model 対象のモデル名を設定します.
     * @return int 件数が返されます.
     * @exception Exception 例外.
     */
    public static final int exportCsv( Writer writer,Record record,String model )
        throws Exception {
        model = DbUtil.convertJavaNameByDBName( model ) ;
        MetaColumn meta = MetaFactory.getInstance().getMetaColumn( model ) ;
        Map<String,Object> m = new HashMap<String,Object>() ;
        int columnLen = meta.size() ;
        char[] ENTER = "\r\n".toCharArray() ;
        SimpleDateFormat TIMESTAMP = new SimpleDateFormat( "yyyy/MM/dd HH:mm:ss" ) ;
        SimpleDateFormat DATE = new SimpleDateFormat( "yyyy/MM/dd" ) ;
        SimpleDateFormat TIME = new SimpleDateFormat( "HH:mm:ss" ) ;
        // ヘッダ名を出力.
        for( int i = 0 ; i < columnLen ; i ++ ) {
            if( i != 0 ) {
                writer.write( ',' ) ;
            }
            writer.write( meta.getColumnName( i ).toCharArray() ) ;
        }
        int cnt = 0 ;
        ResultSet rs = null ;
        try {
            rs = record.executeQuery( "select * from " + model ) ;
            while( rs.next() ) {
                writer.write( ENTER ) ;
                ExecutionDao.convertResultByObject( m,meta,rs,columnLen ) ;
                for( int i = 0 ; i < columnLen ; i ++ ) {
                    if( i != 0 ) {
                        writer.write( ',' ) ;
                    }
                    Object o = m.get( meta.getColumnName( true,i ) ) ;
                    if( o == null ) {
                        continue ;
                    }
                    else if( o instanceof String ) {
                        writer.write( '\"' ) ;
                        writer.write( ((String)o).toCharArray() ) ;
                        writer.write( '\"' ) ;
                    }
                    else if( o instanceof java.sql.Timestamp ) {
                        writer.write( '\"' ) ;
                        writer.write( TIMESTAMP.format( (java.util.Date)o ).toCharArray() ) ;
                        writer.write( '\"' ) ;
                    }
                    else if( o instanceof java.sql.Time ) {
                        writer.write( '\"' ) ;
                        writer.write( TIME.format( (java.util.Date)o ).toCharArray() ) ;
                        writer.write( '\"' ) ;
                    }
                    else if( o instanceof java.util.Date ) {
                        writer.write( '\"' ) ;
                        writer.write( DATE.format( (java.util.Date)o ).toCharArray() ) ;
                        writer.write( '\"' ) ;
                    }
                    else if( o instanceof byte[] ) {
                        writer.write( '\"' ) ;
                        writer.write( Base64.encode( (byte[])o ).toCharArray() ) ;
                        writer.write( '\"' ) ;
                    }
                    else {
                        writer.write( String.valueOf(o).toCharArray() ) ;
                    }
                }
                m.clear() ;
                cnt ++ ;
            }
            rs.close() ;
            rs = null ;
        } finally {
            if( rs != null ) {
                try {
                    rs.close() ;
                } catch( Exception e ) {
                }
            }
            rs = null ;
        }
        return cnt ;
    }
    
}
