package org.maachang.dao.dbms.pool.atomic;

import java.sql.Array;
import java.sql.Blob;
import java.sql.CallableStatement;
import java.sql.Clob;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.NClob;
import java.sql.PreparedStatement;
import java.sql.SQLClientInfoException;
import java.sql.SQLException;
import java.sql.SQLWarning;
import java.sql.SQLXML;
import java.sql.Savepoint;
import java.sql.Statement;
import java.sql.Struct;
import java.util.Properties;

import org.maachang.util.atomic.AtomicBOOL;
import org.maachang.util.atomic.AtomicLONG;
import org.maachang.util.atomic.AtomicOBJECT;

/**
 * RDBMSConnection実体(java1.6版).
 * 
 * @version 2009/05/30
 * @author masahito suzuki
 * @since MaachangDao 1.13
 */
class RdbmsConnection1_6 implements RdbmsConnection {
    private RdbmsPooling pooling = null ;
    private final AtomicOBJECT<Connection> conn = new AtomicOBJECT<Connection>() ;
    private final AtomicLONG accessTime = new AtomicLONG( -1 ) ;
    private final AtomicBOOL destroyFlag = new AtomicBOOL( false ) ;
    private final AtomicBOOL closeFlag = new AtomicBOOL( true ) ;
    
    /**
     * コンストラクタ.
     */
    private RdbmsConnection1_6() {}
    
    /**
     * コンストラクタ.
     * @param pooling 対象のプーリングオブジェクトを設定します.
     * @param conn 生成されたコネクションオブジェクトを設定します.
     */
    public RdbmsConnection1_6( RdbmsPooling pooling,Connection conn ) {
        this.pooling = pooling ;
        this.conn.set( conn ) ;
        this.accessTime.set( System.currentTimeMillis() ) ;
        this.closeFlag.set( false ) ;
        pooling.addCount() ;
    }
    
    /** finalize. **/
    protected final void finalize() throws Exception {
        if( conn.get() != null ) {
            try {
                close() ;
            } catch( Exception e ) {
            }
        }
    }
    
    /** オブジェクトを破棄. **/
    public final void destroy() {
        destroyFlag.set( true ) ;
    }
    
    /** 非アクティブ時間を取得. **/
    public final long noActiveTime() {
        return accessTime.get() ;
    }
    
    /** オブジェクトアクティブ. **/
    public final void toActive() {
        closeFlag.set( false ) ;
        this.accessTime.set( -1L ) ;
    }
    
    /**
     * クローズ処理.
     * @exception SQLException SQL例外.
     */
    public void close() throws SQLException {
        Connection cn = conn.get() ;
        if( cn != null ) {
            if( destroyFlag.get() ) {
                pooling.removeCount() ;
                conn.set( null ) ;
                closeFlag.set( true ) ;
                try {
                    cn.close() ;
                } catch( Exception e ) {
                }
                accessTime.set( 0L ) ;
            }
            else if( !closeFlag.get() ) {
                try {
                    commit() ;
                } catch( Exception e ) {
                }
                closeFlag.set( true ) ;
                // Poolingにコネクションオブジェクトを返す.
                accessTime.set( System.currentTimeMillis() ) ;
                pooling.releaseConnection( this ) ;
            }
        }
    }
    
    /**
     * 管理コネクションがクローズされているかチェック.
     */
    public boolean isConnectionClose() {
        boolean ret = false ;
        Connection cn = conn.get() ;
        if( cn != null ) {
            try {
                ret = cn.isClosed() ;
            } catch( Exception e ) {
                ret = true ;
            }
        }
        else {
            ret = false ;
        }
        return ret ;
    }
    
    /**
     * コミット処理.
     * @exception SQLException SQL例外.
     */
    public void commit() throws SQLException {
        conn.get().commit() ;
    }
    
    /**
     * ロールバック処理.
     * @exception SQLException SQL例外.
     */
    public void rollback() throws SQLException {
        conn.get().rollback() ;
    }
    
    // /////////////////////////
    // 以下Wapperオブジェクト.
    // /////////////////////////

    public Statement createStatement() throws SQLException {
        return conn.get().createStatement();
    }

    public PreparedStatement prepareStatement(String sql) throws SQLException {
        return conn.get().prepareStatement(sql);
    }

    public CallableStatement prepareCall(String sql) throws SQLException {
        return conn.get().prepareCall(sql);
    }

    public String nativeSQL(String sql) throws SQLException {
        return conn.get().nativeSQL(sql);
    }

    public void setAutoCommit(boolean autoCommit) throws SQLException {
        conn.get().setAutoCommit(autoCommit);
    }

    public boolean getAutoCommit() throws SQLException {
        return conn.get().getAutoCommit();
    }

    public boolean isClosed() throws SQLException {
        return closeFlag.get() ;
    }

    // ======================================================================
    // Advanced features:

    public DatabaseMetaData getMetaData() throws SQLException {
        return conn.get().getMetaData();
    }

    public void setReadOnly(boolean readOnly) throws SQLException {
        conn.get().setReadOnly(readOnly);
    }

    public boolean isReadOnly() throws SQLException {
        return conn.get().isReadOnly();
    }

    public void setCatalog(String catalog) throws SQLException {
        conn.get().setCatalog(catalog);
    }

    public String getCatalog() throws SQLException {
        return conn.get().getCatalog();
    }

    public void setTransactionIsolation(int level) throws SQLException {
        conn.get().setTransactionIsolation(level);
    }

    public int getTransactionIsolation() throws SQLException {
        return conn.get().getTransactionIsolation();
    }

    public SQLWarning getWarnings() throws SQLException {
        return conn.get().getWarnings();
    }

    public void clearWarnings() throws SQLException {
        conn.get().clearWarnings();
    }

    // --------------------------JDBC 2.0-----------------------------

    public Statement createStatement(int resultSetType, int resultSetConcurrency)
        throws SQLException {
        return conn.get().createStatement(resultSetType,
                resultSetConcurrency);
    }

    public PreparedStatement prepareStatement(String sql, int resultSetType,
        int resultSetConcurrency) throws SQLException {
        return conn.get().prepareStatement(sql,
                resultSetType, resultSetConcurrency);
    }

    public CallableStatement prepareCall(String sql, int resultSetType,
        int resultSetConcurrency) throws SQLException {
        return conn.get().prepareCall(sql, resultSetType,
                resultSetConcurrency);
    }

    public java.util.Map<String, Class<?>> getTypeMap() throws SQLException {
        return conn.get().getTypeMap();
    }

    public void setTypeMap(java.util.Map<String, Class<?>> map)
        throws SQLException {
        conn.get().setTypeMap(map);
    }

    // --------------------------JDBC 3.0-----------------------------

    public void setHoldability(int holdability) throws SQLException {
        conn.get().setHoldability(holdability);
    }

    public int getHoldability() throws SQLException {
        return conn.get().getHoldability();
    }

    public Savepoint setSavepoint() throws SQLException {
        return conn.get().setSavepoint();
    }

    public Savepoint setSavepoint(String name) throws SQLException {
        return conn.get().setSavepoint(name);
    }

    public void rollback(Savepoint savepoint) throws SQLException {
        conn.get().rollback(savepoint);
    }

    public void releaseSavepoint(Savepoint savepoint) throws SQLException {
        conn.get().releaseSavepoint(savepoint);
    }

    public Statement createStatement(int resultSetType,
        int resultSetConcurrency, int resultSetHoldability)
        throws SQLException {
        return conn.get().createStatement(resultSetType,
                resultSetConcurrency, resultSetHoldability);
    }

    public PreparedStatement prepareStatement(String sql, int resultSetType,
        int resultSetConcurrency, int resultSetHoldability)
        throws SQLException {
        return conn.get().prepareStatement(sql,
                resultSetType, resultSetConcurrency, resultSetHoldability);
    }

    public CallableStatement prepareCall(String sql, int resultSetType,
        int resultSetConcurrency, int resultSetHoldability)
        throws SQLException {
        return conn.get().prepareCall(sql, resultSetType,
                resultSetConcurrency, resultSetHoldability);
    }

    public PreparedStatement prepareStatement(String sql, int autoGeneratedKeys)
        throws SQLException {
        return conn.get().prepareStatement(sql,
                autoGeneratedKeys);
    }

    public PreparedStatement prepareStatement(String sql, int columnIndexes[])
        throws SQLException {
        return conn.get().prepareStatement(sql, columnIndexes);
    }

    public PreparedStatement prepareStatement(String sql, String columnNames[])
        throws SQLException {
        return conn.get().prepareStatement(sql, columnNames);
    }

    public Clob createClob() throws SQLException {
        return conn.get().createClob();
    }

    public Blob createBlob() throws SQLException {
        return conn.get().createBlob();
    }

    public NClob createNClob() throws SQLException {
        return conn.get().createNClob();
    }

    public SQLXML createSQLXML() throws SQLException {
        return conn.get().createSQLXML();
    }

    public boolean isValid(int timeout) throws SQLException {
        return conn.get().isValid(timeout);
    }

    public void setClientInfo(String name, String value)
        throws SQLClientInfoException {
        conn.get().setClientInfo(name, value);
    }

    public void setClientInfo(Properties properties)
        throws SQLClientInfoException {
        conn.get().setClientInfo(properties);
    }

    public String getClientInfo(String name) throws SQLException {
        return conn.get().getClientInfo(name);
    }

    public Properties getClientInfo() throws SQLException {
        return conn.get().getClientInfo();
    }

    public Array createArrayOf(String typeName, Object[] elements)
        throws SQLException {
        return conn.get().createArrayOf(typeName, elements);
    }

    public Struct createStruct(String typeName, Object[] attributes)
        throws SQLException {
        return conn.get().createStruct(typeName, attributes);
    }

    // Wrapper.java
    public <T> T unwrap(java.lang.Class<T> iface) throws java.sql.SQLException {
        return conn.get().unwrap(iface);
    }

    public boolean isWrapperFor(java.lang.Class<?> iface)
        throws java.sql.SQLException {
        return conn.get().isWrapperFor(iface);
    }

}
