package org.maachang.dao.dbms.pool.atomic;

import java.sql.Connection;
import java.util.concurrent.ConcurrentLinkedQueue;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.dao.dbms.pool.ConnectionPooling;
import org.maachang.util.atomic.AtomicBOOL;
import org.maachang.util.atomic.AtomicINT;

/**
 * RDBMSプーリングマネージャ.
 * 
 * @version 2009/05/30
 * @author masahito suzuki
 * @since MaachangDao 1.13
 */
public class RdbmsPooling implements ConnectionPooling {
    private static final Log LOG = LogFactory.getLog( RdbmsPooling.class ) ;
    private final ConcurrentLinkedQueue<RdbmsConnection> queue = new ConcurrentLinkedQueue<RdbmsConnection>() ;
    private RdbmsDefine define = null ;
    private RdbmsMonitor mon = null ;
    private int maxLength ;
    private final AtomicINT allLength = new AtomicINT( 0 ) ;
    
    private static final String VERSION ;
    static {
        String s = System.getProperty( "java.version" ) ;
        if( s.startsWith( "1.6" ) ) {
            VERSION = "1.6" ;
        }
        else if( s.startsWith( "1.7" ) ) {
            VERSION = "1.7" ;
        }
        else {
            VERSION = "unknown" ;
        }
    }
    
    /**
     * コンストラクタ.
     */
    private RdbmsPooling() {}
    
    /**
     * コンストラクタ.
     * @param define コネクション定義情報を設定します.
     * @param keepAlive コネクションキープアライブ時間を設定します.
     * @param maxLength コネクション最大数を設定します.
     * @exception Exception 例外.
     */
    public RdbmsPooling( String driver,RdbmsDefine define,long keepAlive,int maxLength )
        throws Exception {
        Class.forName( driver ) ;
        this.define = define ;
        this.mon = new RdbmsMonitor( queue,keepAlive ) ;
        this.maxLength = maxLength ;
    }
    
    protected void finalize() throws Exception {
        destroy() ;
    }
    
    /**
     * オブジェクト破棄.
     */
    public void destroy() {
        if( mon != null ) {
            mon.destroy() ;
        }
        mon = null ;
        queue.clear() ;
    }
    
    /**
     * コネクションオブジェクトを取得.
     * @return Connection コネクションオブジェクトが返されます.
     * @exception Exception 例外.
     */
    public Connection getConnection() throws Exception {
        Connection conn = queue.poll() ;
        if( conn == null || ((RdbmsConnection)conn).isConnectionClose() ) {
            if( conn != null ) {
                try {
                    ((RdbmsConnection)conn).destroy() ;
                } catch( Exception e ) {
                }
                try {
                    conn.close() ;
                } catch( Exception e ) {
                }
            }
            conn = RdbmsDriverManager.driverConnection( define ) ;
            if( maxLength >= allLength.get() ) {
                conn = createConnection( this,conn ) ;
            }
        }
        else {
            ((RdbmsConnection)conn).toActive() ;
            if( LOG.isDebugEnabled() ) {
                LOG.debug( "## activeConnection!!" ) ;
            }
        }
        return conn ;
    }
    
    /**
     * 最大コネクション数を取得.
     * @return int 最大コネクション数が返されます.
     */
    public int max() {
        return maxLength ;
    }
    
    /**
     * 使用中のコネクション数を取得.
     * @return int 使用中のコネクション数が返されます.
     */
    public int active() {
        return allLength.get() - queue.size() ;
    }
    
    /**
     * 未使用コネクション数を取得.
     * @return int 未使用中のコネクション数が返されます.
     */
    public int idle() {
        return queue.size() ;
    }
    
    /**
     * 全体コネクション数を取得.
     * @return int 全体コネクション数が返されます.
     */
    public int getAllLength() {
        return allLength.get() ;
    }
    
    /** 全体コネクション数に１インクリメント. **/
    protected void addCount() {
        allLength.inc() ;
    }
    
    /** 全体コネクション数から１デクリメント. **/
    protected void removeCount() {
        allLength.dec() ;
    }
    
    /** 利用完了のコネクションをPoolingに設定. **/
    protected void releaseConnection( RdbmsConnection conn ) {
        if( LOG.isDebugEnabled() ) {
            LOG.debug( "## releaseConnection!!" ) ;
        }
        queue.offer( conn ) ;
    }
    
    /**
     * コネクションバージョンを取得.
     * @return String コネクションバージョンが返されます.
     */
    public static final String getVersion() {
        return VERSION ;
    }
    
    /** Javaバージョンにあわせたコネクションプーリングを取得. **/
    private static final Connection createConnection( RdbmsPooling pool,Connection conn ) {
        if( LOG.isDebugEnabled() ) {
            LOG.debug( "## createConnection!!" ) ;
        }
        RdbmsConnection ret = null ;
        if( VERSION.equals( "1.6" ) ) {
            ret = new RdbmsConnection1_6( pool,conn ) ;
        }
        else {
            ret = new RdbmsConnection1_6( pool,conn ) ;
        }
        ret.toActive() ;
        return ret ;
    }
    
    
}

/** Poolingモニター **/
class RdbmsMonitor extends Thread {
    private static final Log LOG = LogFactory.getLog( RdbmsMonitor.class ) ;
    private static final long WAIT = 50L ;
    private ConcurrentLinkedQueue<RdbmsConnection> queue = null ;
    private long maxKeepAlive = -1 ;
    private AtomicBOOL exitFlag = new AtomicBOOL( false ) ;
    
    private RdbmsMonitor(){}
    
    protected RdbmsMonitor( ConcurrentLinkedQueue<RdbmsConnection> queue,long maxKeepAlive ) {
        this.queue = queue ;
        this.maxKeepAlive = maxKeepAlive ;
        this.setDaemon( true ) ;
        this.start() ;
    }
    
    public void destroy() {
        exitFlag.set( true ) ;
    }
    
    public void run() {
        RdbmsConnection conn ;
        while( true ) {
            if( exitFlag.get() ) {
                break ;
            }
            try {
                Thread.sleep( WAIT ) ;
                if( ( conn = queue.poll() ) != null ) {
                    if( conn.noActiveTime() != -1L &&
                        conn.noActiveTime()+maxKeepAlive <= System.currentTimeMillis() ) {
                        if( LOG.isDebugEnabled() ) {
                            LOG.debug( "## dropConnection!!" ) ;
                        }
                        try {
                            conn.destroy() ;
                        } catch( Exception ee ) {
                        }
                        try {
                            conn.close() ;
                        } catch( Exception ee ) {
                        }
                    }
                    else {
                        queue.offer( conn ) ;
                    }
                }
            } catch( InterruptedException ie ) {
                exitFlag.set( true ) ;
                break ;
            } catch( Exception e ) {
            } catch( OutOfMemoryError e ) {
            }
        }
    }
}
