package org.maachang.report ;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Iterator;

import org.apache.poi.hssf.usermodel.HSSFCell;
import org.apache.poi.hssf.usermodel.HSSFCellStyle;
import org.apache.poi.hssf.usermodel.HSSFDateUtil;
import org.apache.poi.hssf.usermodel.HSSFRichTextString;
import org.apache.poi.hssf.usermodel.HSSFRow;
import org.apache.poi.hssf.usermodel.HSSFSheet;
import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.poifs.filesystem.POIFSFileSystem;
import org.maachang.util.FileUtil;
import org.maachang.util.StringUtil;

/**
 * １つのExcelシートを表すオブジェクト.
 * 
 * @version 2008/09/29
 * @author masahito suzuki
 * @since MaachangReport 1.00
 */
public class ExcelSheet {
    
    /**
     * セルタイプ : ブランク(なし).
     */
    public static final int CELL_BLANK = 0 ;
    
    /**
     * セルタイプ : 数字.
     */
    public static final int CELL_NUMBER = 1 ;
    
    /**
     * セルタイプ : 文字.
     */
    public static final int CELL_STRING = 2 ;
    
    /**
     * セルタイプ : Boolean.
     */
    public static final int CELL_BOOLEAN = 3 ;
    
    /**
     * セルタイプ : 日付.
     */
    public static final int CELL_DATE = 4 ;
    
    /**
     * POIワークブック(１つのExcelファイルを表す).
     */
    private HSSFWorkbook work = null ;
    
    /**
     * シート内容.
     */
    private HSSFSheet sheet = null ;
    
    /**
     * シート名.
     */
    private String sheetName = null ;
    
    /**
     * コンストラクタ.
     */
    public ExcelSheet() {
        work = new HSSFWorkbook() ;// 空のワークブックを生成.
        sheet = null ;
        sheetName = null ;
    }
    
    /**
     * コンストラクタ.
     * @param name Excelファイル名を設定します.
     * @param sheetName 利用対象のExcelシート名を設定します.
     * @exception Exception 例外.
     */
    public ExcelSheet( String name,String sheetName ) throws Exception {
        this.open( name,sheetName ) ;
    }
    
    /**
     * コンストラクタ.
     * @param in 対象のInputStreamを設定します.
     * @param sheetName 利用対象のExcelシート名を設定します.
     * @exception Exception 例外.
     */
    public ExcelSheet( InputStream in,String sheetName ) throws Exception {
        this.open( in,sheetName ) ;
    }
    
    /**
     * コンストラクタ.
     * @param binary 対象のバイナリを設定します.
     * @param sheetName 利用対象のExcelシート名を設定します.
     * @exception Exception 例外.
     */
    public ExcelSheet( byte[] binary,String sheetName ) throws Exception {
        this.open( binary,sheetName ) ;
    }
    
    /**
     * 帳票雛型のExcelファイルをオープン.
     * @param name Excelファイル名を設定します.
     * @param sheetName 利用対象のExcelシート名を設定します.
     * @exception Exception 例外.
     */
    public void open( String name,String sheetName ) throws Exception {
        if( name == null || ( name = name.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "ファイル名が設定されていません" ) ;
        }
        if( name.endsWith( ".xls" ) == false ) {
            name += ".xls" ;
        }
        if( FileUtil.isFileExists( name ) == false ) {
            throw new IOException( "指定ファイル[" + name + "]は存在しません" ) ;
        }
        else if( FileUtil.isRead( name ) == false ) {
            throw new IOException( "指定ファイル[" + name + "]は読み込み権限が存在しません" ) ;
        }
        _open( new BufferedInputStream( new FileInputStream( name ) ),sheetName ) ;
    }
    
    /**
     * 帳票雛形のExcelファイルをオープン.
     * @param in 対象のInputStreamを設定します.
     * @param sheetName 利用対象のExcelシート名を設定します.
     * @exception Exception 例外.
     */
    public void open( InputStream in,String sheetName ) throws Exception {
        if( in == null ) {
            throw new IllegalArgumentException( "InputStreamが設定されていません" ) ;
        }
        _open( in,sheetName ) ;
    }
    
    /**
     * 帳票雛形のExcelファイルをオープン.
     * @param binary 対象のバイナリを設定します.
     * @param sheetName 利用対象のExcelシート名を設定します.
     * @exception Exception 例外.
     */
    public void open( byte[] binary,String sheetName ) throws Exception {
        if( binary == null || binary.length <= 0 ) {
            throw new IllegalArgumentException( "バイナリが設定されていません" ) ;
        }
        _open( new ByteArrayInputStream( binary ),sheetName ) ;
    }
    
    private void _open( InputStream in,String sheetName ) throws Exception {
        if( sheetName == null || ( sheetName = sheetName.trim() ).length() <= 0 ) {
            sheetName = "Sheet1" ;
        }
        destroy() ;
        try {
            POIFSFileSystem fs = new POIFSFileSystem( in ) ;
            this.work = new HSSFWorkbook( fs ) ;
            this.sheetName = sheetName ;
            this.sheet = work.getSheet( sheetName ) ;
            in.close() ;
            in = null ;
        } catch( Exception e ) {
            destroy() ;
            throw e ;
        } finally {
            if( in != null ) {
                try {
                    in.close() ;
                } catch( Exception e ) {
                }
            }
            in = null ;
        }
    }
    
    /**
     * 帳票破棄.
     */
    public void destroy() {
        work = null ;
        sheet = null ;
        sheetName = null ;
    }
    
    /**
     * ワークシートを追加.
     * @param name ワークシート名を設定します.
     * @param sheet ワークシート情報を設定します.
     * @exception Exception 例外.
     */
    public void createSheet( String name ) {
        work.createSheet( name ) ;
    }
    
    /**
     * ワークシートを変更.
     * @param name ワークシート名を設定します.
     * @return boolean [true]の場合、処理成功.
     */
    public boolean changeSheet( String name ) {
        HSSFSheet s = work.getSheet( name ) ;
        if( s != null ) {
            this.sheet = s ;
            this.sheetName = name ;
            return true ;
        }
        return false ;
    }
    
    /**
     * 指定ワークシートを削除.
     * @param name ワークシート名を設定します.
     * @return boolean [true]の場合、処理成功.
     */
    public boolean removeSheet( String name ) {
        int index = work.getSheetIndex( name ) ;
        if( index <= -1 ) {
            return false ;
        }
        work.removeSheetAt( index ) ;
        return true ;
    }
    
    /**
     * シート情報をコピー.
     * @param name コピー対象のシート名を設定します.
     * @param newName コピーされたシート名に対する新しい名前を設定します.
     * @return boolean [true]の場合、処理成功.
     * @exception Exception 例外.
     */
    public boolean copySheet( String name,String newName )
        throws Exception {
        int index = work.getSheetIndex( name ) ;
        if( index <= -1 ) {
            return false ;
        }
        if( work.cloneSheet( index ) != null ) {
            int i = 9 ;
            while( true ) {
                if( i <= 0 ) {
                    break ;
                }
                String indexStr = String.valueOf( i -- ) ;
                String nm = name ;
                if (nm.length() + 3 < 31) {
                    nm = nm + " ("+indexStr+")";
                } else {
                    nm = nm.substring(0, 31 - 3) + " ("+indexStr+")";
                }
                index = work.getSheetIndex( nm ) ;
                if( index >= 0 ) {
                    work.setSheetName( index,newName ) ;
                    return true ;
                }
            }
        }
        return false ;
    }
    
    /**
     * ワークシート名を取得.
     * @return String ワークシート名が返されます.
     */
    public String getSheetName() {
        return sheetName ;
    }
    
    /**
     * シートオブジェクトを取得.
     * @return HSSFWorkbook シートオブジェクトが返されます.
     */
    public HSSFSheet getSheet() {
        return sheet ;
    }
    
    /**
     * 現在のExcel情報をバイナリで取得.
     * @return byte[] 現在のExcel情報が返されます.
     * @exception Exception 例外.
     */
    public byte[] getBytes() throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        ByteArrayOutputStream bo = null ;
        try {
            bo = new ByteArrayOutputStream() ;
            work.write( bo ) ;
            bo.flush() ;
            byte[] ret = bo.toByteArray() ;
            bo.close() ;
            bo = null ;
            return ret ;
        } finally {
            if( bo != null ) {
                try {
                    bo.close() ;
                } catch( Exception e ) {
                }
            }
        }
    }
    
    /**
     * 現在のExcel情報をファイル出力.
     * @param name 出力先のファイル名を設定します.
     * @exception Exception 例外.
     */
    public void output( String name ) throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        OutputStream bo = null ;
        try {
            bo = new BufferedOutputStream( new FileOutputStream( name ) ) ;
            work.write( bo ) ;
            bo.flush() ;
            bo.close() ;
            bo = null ;
        } finally {
            if( bo != null ) {
                try {
                    bo.close() ;
                } catch( Exception e ) {
                }
            }
        }
    }
    
    /**
     * １つのセル情報に設定.
     * @param x 位置X値を設定します.
     * @param y 位置Y値を設定します.
     * @param data 設定情報をセットします.
     * @exception Exception 例外.
     */
    public void setCell( int x,int y,Object data )
        throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( x <= -1 || x > 32767 || y <= -1 || y > 65535 ) {
            throw new IllegalArgumentException( "指定座標は不正です[x:" + x + " y:" + y + "]" ) ;
        }
        if( data == null ) {
            data = "" ;
        }
        HSSFRow row = sheet.getRow( y ) ;
        if( row == null ) {
            row = sheet.createRow( y ) ;
        }
        HSSFCell cell = row.getCell( x );
        if( cell == null ) {
            cell = row.createCell( x ) ;
        }
        if( data instanceof String ) {
            //cell.setCellType( HSSFCell.CELL_TYPE_STRING ) ;
            cell.setCellValue( new HSSFRichTextString( ( String )data ) ) ;
        }
        else if( data instanceof Boolean ) {
            //cell.setCellType( HSSFCell.CELL_TYPE_BOOLEAN ) ;
            cell.setCellValue( ( ( Boolean )data ).booleanValue() ) ;
        }
        else if( data instanceof Integer ) {
            //cell.setCellType( HSSFCell.CELL_TYPE_NUMERIC ) ;
            cell.setCellValue( ( ( Integer )data ).intValue() ) ;
        }
        else if( data instanceof Double ) {
            //cell.setCellType( HSSFCell.CELL_TYPE_NUMERIC ) ;
            cell.setCellValue( ( ( Double )data ).doubleValue() ) ;
        }
        else if( data instanceof java.util.Date ) {
            //cell.setCellType( HSSFCell.CELL_TYPE_NUMERIC ) ;
            cell.setCellValue( ( ( java.util.Date )data ) ) ;
        }
        else {
            throw new IllegalArgumentException(
                "指定オブジェクトタイプ[" + data.getClass().getName() + "]は対応していません" ) ;
        }
    }
    
    /**
     * １つのセル情報に設定.
     * @param x 位置X値(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @param data 設定情報をセットします.
     * @exception Exception 例外.
     */
    public void setCell( String x,int y,Object data )
        throws Exception {
        setCell( ExcelUtil.getPosAz( x ),y,data ) ;
    }
    
    /**
     * １つのセル内容を取得.
     * @param x 位置X値を設定します.
     * @param y 位置Y値を設定します.
     * @return Object セル内容が返されます.
     * @exception Exception 例外.
     */
    public Object getCell( int x,int y )
        throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( x <= -1 || x > 32767 || y <= -1 || y > 65535 ) {
            throw new IllegalArgumentException( "指定座標は不正です[x:" + x + " y:" + y + "]" ) ;
        }
        HSSFRow row = sheet.getRow( y ) ;
        if( row == null ) {
            throw new IllegalArgumentException( "Y座標[" + y + "]の取得に失敗しました" ) ;
        }
        HSSFCell cell = row.getCell( x );
        if( cell == null ) {
            throw new IllegalArgumentException( "座標[x:" + x + " y:" + y + "]の取得に失敗しました" ) ;
        }
        Object ret = null ;
        int type = cell.getCellType() ;
        switch( type ) {
            case HSSFCell.CELL_TYPE_BLANK  :
                ret = "" ;
                break ;
            case HSSFCell.CELL_TYPE_NUMERIC :
                if( HSSFDateUtil.isCellDateFormatted( cell ) ) {
                    ret = cell.getDateCellValue() ;
                }
                else {
                    ret = cell.getNumericCellValue() ;
                }
                break ;
            case HSSFCell.CELL_TYPE_STRING :
                HSSFRichTextString tx = cell.getRichStringCellValue() ;
                if( tx == null ) {
                    ret = "" ;
                }
                else {
                    ret = tx.getString() ;
                }
                break ;
            case HSSFCell.CELL_TYPE_BOOLEAN :
                ret = cell.getBooleanCellValue() ;
                break ;
        }
        return ret ;
    }
    
    /**
     * １つのセル内容を取得.
     * @param x 位置X(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @return Object セル内容が返されます.
     * @exception Exception 例外.
     */
    public Object getCell( String x,int y )
        throws Exception {
        return getCell( ExcelUtil.getPosAz( x ),y ) ;
    }
    
    /**
     * １つのセルタイプを取得.
     * @param x 位置X値を設定します.
     * @param y 位置Y値を設定します.
     * @return int セルタイプが返されます.
     *             [-1]の場合、取得に失敗したか、不明なタイプです.
     */
    public int getCellType( int x,int y ) {
        if( isUse() == false ) {
            return -1 ;
        }
        if( x <= -1 || x > 32767 || y <= -1 || y > 65535 ) {
            return -1 ;
        }
        HSSFRow row = sheet.getRow( y ) ;
        if( row == null ) {
            return -1 ;
        }
        HSSFCell cell = row.getCell( x );
        if( cell == null ) {
            return -1 ;
        }
        int type = cell.getCellType() ;
        switch( type ) {
            case HSSFCell.CELL_TYPE_BLANK  :
                return CELL_BLANK ;
            case HSSFCell.CELL_TYPE_NUMERIC :
                if( HSSFDateUtil.isCellDateFormatted( cell ) ) {
                    return CELL_DATE ;
                }
                else {
                    return CELL_NUMBER ;
                }
            case HSSFCell.CELL_TYPE_STRING :
                return CELL_STRING ;
            case HSSFCell.CELL_TYPE_BOOLEAN :
                return CELL_BOOLEAN ;
        }
        return -1 ;
    }
    
    /**
     * １つのセルタイプを取得.
     * @param x 位置X(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @return int セルタイプが返されます.
     *             [-1]の場合、取得に失敗したか、不明なタイプです.
     */
    public int getCellType( String x,int y ) {
        return getCellType( ExcelUtil.getPosAz( x ),y ) ;
    }
    
    /**
     * 新しいセルスタイルを生成.
     * @return HSSFCellStyle 空のセルスタイルを作成します.
     * @exception Exception 例外.
     */
    public HSSFCellStyle createCellStyle() throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        return work.createCellStyle() ;
    }
    
    /**
     * 指定セルにスタイルオブジェクトを設定.
     * @param style セルスタイルオブジェクトを設定します.
     * @param x 位置X値を設定します.
     * @param y 位置Y値を設定します.
     * @exception Exception 例外.
     */
    public void setCellStyle( HSSFCellStyle style,int x,int y )
        throws Exception {
        if( style == null ) {
            throw new IllegalArgumentException( "スタイルオブジェクトが設定されていません" ) ;
        }
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( x <= -1 || x > 32767 || y <= -1 || y > 65535 ) {
            throw new IllegalArgumentException( "指定座標は不正です[x:" + x + " y:" + y + "]" ) ;
        }
        HSSFRow row = sheet.getRow( y ) ;
        if( row == null ) {
            throw new IllegalArgumentException( "Y座標[" + y + "]の取得に失敗しました" ) ;
        }
        HSSFCell cell = row.getCell( x );
        if( cell == null ) {
            throw new IllegalArgumentException( "座標[x:" + x + " y:" + y + "]の取得に失敗しました" ) ;
        }
        cell.setCellStyle( style ) ;
    }
    
    /**
     * 指定セルのスタイルオブジェクトを取得.
     * @param style セルスタイルオブジェクトを設定します.
     * @param x 位置X(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @exception Exception 例外.
     */
    public void setCellStyle( HSSFCellStyle style,String x,int y )
        throws Exception {
        setCellStyle( style,ExcelUtil.getPosAz( x ),y ) ;
    }
    
    /**
     * 指定セルのスタイルオブジェクトを取得.
     * @param x 位置X値を設定します.
     * @param y 位置Y値を設定します.
     * @return HSSFCellStyle セルスタイルオブジェクトが返されます.
     * @exception Exception 例外.
     */
    public HSSFCellStyle getCellStyle( int x,int y )
        throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( x <= -1 || x > 32767 || y <= -1 || y > 65535 ) {
            throw new IllegalArgumentException( "指定座標は不正です[x:" + x + " y:" + y + "]" ) ;
        }
        HSSFRow row = sheet.getRow( y ) ;
        if( row == null ) {
            throw new IllegalArgumentException( "Y座標[" + y + "]の取得に失敗しました" ) ;
        }
        HSSFCell cell = row.getCell( x );
        if( cell == null ) {
            throw new IllegalArgumentException( "座標[x:" + x + " y:" + y + "]の取得に失敗しました" ) ;
        }
        return cell.getCellStyle() ;
    }
    
    /**
     * 指定セルのスタイルオブジェクトを取得.
     * @param x 位置X(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @return HSSFCellStyle セルスタイルオブジェクトが返されます.
     * @exception Exception 例外.
     */
    public HSSFCellStyle getCellStyle( String x,int y )
        throws Exception {
        return getCellStyle( ExcelUtil.getPosAz( x ),y ) ;
    }
    
    /**
     * 指定行を削除.
     * @param y 削除行位置を設定します.
     * @exception Exception 例外.
     */
    public void removeRow( int y ) throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( y <= -1 || y > 65535 ) {
            throw new IllegalArgumentException( "指定座標は不正です[y:" + y + "]" ) ;
        }
        HSSFRow row = sheet.getRow( y ) ;
        if( row == null ) {
            throw new IllegalArgumentException( "Y座標[" + y + "]の取得に失敗しました" ) ;
        }
        sheet.removeRow( row ) ;
    }
    
    /**
     * 指定セルを削除.
     * @param x 位置Xを設定します.
     * @param y 位置Y値を設定します.
     * @exception Exception 例外.
     */
    public void removeCell( int x,int y ) throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( x <= -1 || x > 32767 || y <= -1 || y > 65535 ) {
            throw new IllegalArgumentException( "指定座標は不正です[x:" + x + " y:" + y + "]" ) ;
        }
        HSSFRow row = sheet.getRow( y ) ;
        if( row == null ) {
            throw new IllegalArgumentException( "Y座標[" + y + "]の取得に失敗しました" ) ;
        }
        HSSFCell cell = row.getCell( x );
        if( cell == null ) {
            return ;
        }
        row.removeCell( cell ) ;
    }
    
    /**
     * 指定セルを削除.
     * @param x 位置X(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @exception Exception 例外.
     */
    public void removeCell( String x,int y ) throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        removeCell( ExcelUtil.getPosAz( x ),y ) ;
    }
    
    /**
     * セルのロック条件を設定.
     * @param boolean mode ロックをONにする場合は、[true]を設定します.
     * @param x 位置X値を設定します.
     * @param y 位置Y値を設定します.
     * @exception Exception 例外.
     */
    public void cellLock( boolean mode,int x,int y )
        throws Exception {
        HSSFCellStyle style = getCellStyle( x,y ) ;
        style.setLocked( mode ) ;
    }
    
    /**
     * セルロック条件を設定.
     * @param boolean mode ロックをONにする場合は、[true]を設定します.
     * @param x 位置X(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @exception Exception 例外.
     */
    public void cellLock( boolean mode,String x,int y )
        throws Exception {
        cellLock( mode,ExcelUtil.getPosAz( x ),y ) ;
    }
    
    /**
     * セルのロック条件を取得.
     * @param x 位置X値を設定します.
     * @param y 位置Y値を設定します.
     * @return boolean [true]の場合、セルロック条件が返されます.
     * @exception Exception 例外.
     */
    public boolean isCellLock( int x,int y )
        throws Exception {
        HSSFCellStyle style = getCellStyle( x,y ) ;
        return style.getLocked() ;
    }
    
    /**
     * セルのロック条件を取得.
     * @param x 位置X(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @return boolean [true]の場合、セルロック条件が返されます.
     * @exception Exception 例外.
     */
    public boolean isCellLock( String x,int y )
        throws Exception {
        return isCellLock( ExcelUtil.getPosAz( x ),y ) ;
    }
    
    /**
     * 指定Row対して、セル追加.
     * @param y 追加対象のY軸を設定します.
     * @param n 追加個数を設定します.
     * @exception Exception 例外.
     */
    public void pushRow( int y,int n ) throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( y <= -1 || y >= 65536 ) {
            throw new IllegalArgumentException( "指定座標は不正です[y:" + y + "]" ) ;
        }
        int last = sheet.getLastRowNum() ;
        // シート領域を越している情報追加の場合.
        if( last < (y+n)-1 ) {
            for( int i = 0 ; i < n ; i ++ ) {
                sheet.createRow( y+n ) ;
            }
        }
        // シート領域内の場合.
        else {
            sheet.shiftRows( y,last,n ) ;
        }
    }
    
    /**
     * Excelをロックする.
     * ※アンロック自体はできないので注意.
     * @param passwd ロックパスワードを設定します.
     * @exception Exception 例外.
     */
    public void setLock( String passwd ) throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        sheet.protectSheet( passwd ) ;
    }
    
    /**
     * Excelのロック状態を取得.
     * @return boolean [true]の場合、ロックされています.
     */
    public boolean isLock() throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        return sheet.getProtect() ;
    }
    
    /**
     * 挿入可能なセル情報一覧を取得.
     * @param csList セルスタイルリストを設定します.
     * @return CellPositionList 挿入可能なセルポジションリストが返されます.
     * @exception Exception 例外.
     */
    public CellPositionList getCellPositionList( CellStyleList csList )
        throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        int p,e,x,y ;
        CellPositionList ret = new CellPositionList( csList ) ;
        Iterator it = sheet.rowIterator() ;
        while( it.hasNext() ) {
            HSSFRow row = ( HSSFRow )it.next() ;
            if( row != null ) {
                Iterator it2 = row.iterator() ;
                while( it2.hasNext() ) {
                    HSSFCell cell = ( HSSFCell )it2.next() ;
                    if( cell != null ) {
                        try {
                            //System.out.println( "############## cell:" + cell ) ;
                            HSSFRichTextString tx = cell.getRichStringCellValue() ;
                            if( tx != null ) {
                                String s = tx.toString() ;
                                if( s != null && ( s = StringUtil.trim( s ) ).length() > 0 ) {
                                    // １つのセルに、セルポジション条件だけが存在する.
                                    if( ( s.startsWith( "${" ) && s.endsWith( "}" ) ) ||
                                        ( s.startsWith( "$(" ) && s.endsWith( ")" ) ) ) {
                                        s = StringUtil.trim( s.substring( 2,s.length()-1 ) ) ;
                                        x = ( int )( cell.getColumnIndex() & 0x0000ffff ) ;
                                        y = row.getRowNum() ;
                                        CellPosition pos = new CellPosition( x,y,s ) ;
                                        ret.put( pos ) ;
                                        cell.setCellValue( new HSSFRichTextString("") ) ;
                                    }
                                    // １つのセルに、セルポジション条件以外に存在する.
                                    else {
                                        p = -1 ;
                                        e = -1 ;
                                        p = s.indexOf( "${" ) ;
                                        if( p <= -1 ) {
                                            p = s.indexOf( "$(" ) ;
                                            if( p <= -1 ) {
                                                continue ;
                                            }
                                            e = s.indexOf( ")",p ) ;
                                        }
                                        else {
                                            e = s.indexOf( "}",e ) ;
                                        }
                                        if( e <= -1 ) {
                                            continue ;
                                        }
                                        String k = StringUtil.trim( s.substring( p+2,e ) ) ;
                                        String etc = new StringBuilder().
                                            append( s.substring( 0,p ) ).
                                            append( "_${" ).append( k ).append( "}$_" ).
                                            append( s.substring( e+1 ) ).toString() ;
                                        x = ( int )( cell.getColumnIndex() & 0x0000ffff ) ;
                                        y = row.getRowNum() ;
                                        CellPosition pos = new CellPosition( x,y,k,etc ) ;
                                        ret.put( pos ) ;
                                        cell.setCellValue(
                                            new HSSFRichTextString(s.substring( 0,p )+" "+s.substring( e )) ) ;
                                    }
                                }
                            }
                        } catch( IllegalStateException is ) {
                            // Cannot get a text value from a numeric cell
                            // この条件は無視.
                        }
                    }
                }
            }
        }
        return ret ;
    }
    
    /**
     * オブジェクトが有効かチェック.
     * @return boolean [true]の場合、有効です.
     */
    public boolean isUse() {
        return ( work != null && sheet != null ) ;
    }
    
}

