package org.maachang.report ;

import java.io.IOException;
import java.io.InputStream;

import org.apache.poi.hssf.usermodel.HSSFCellStyle;

/**
 * Excel読み込み用.
 * 
 * @version 2009/08/12
 * @author masahito suzuki
 * @since MaachangReport 1.00
 */
public class ReadExcelSheet {
    private ReadExcelSheet(){}
    private ExcelSheet sheet = null ;
    
    /**
     * セルタイプ : ブランク(なし).
     */
    public static final int CELL_BLANK = ExcelSheet.CELL_BLANK ;
    
    /**
     * セルタイプ : 数字.
     */
    public static final int CELL_NUMBER = ExcelSheet.CELL_NUMBER ;
    
    /**
     * セルタイプ : 文字.
     */
    public static final int CELL_STRING = ExcelSheet.CELL_STRING ;
    
    /**
     * セルタイプ : Boolean.
     */
    public static final int CELL_BOOLEAN = ExcelSheet.CELL_BOOLEAN ;
    
    /**
     * セルタイプ : 日付.
     */
    public static final int CELL_DATE = ExcelSheet.CELL_DATE ;
    
    /**
     * Excel読み込み用オブジェクトを取得.
     * @param name Excelファイル名を設定します.
     * @param sheetName 利用対象のExcelシート名を設定します.
     * @return ReadExcelSheet Excel読み込み用オブジェクトが返されます.
     * @exception Exception 例外.
     */
    public static final ReadExcelSheet read( String name,String sheetName )
        throws Exception {
        ReadExcelSheet ret = null ;
        ret = new ReadExcelSheet() ;
        ret.open( name,sheetName ) ;
        return ret ;
    }
    
    /**
     * Excel読み込み用オブジェクトを取得.
     * @param in InputStreamを設定します.
     * @param sheetName 利用対象のExcelシート名を設定します.
     * @return ReadExcelSheet Excel読み込み用オブジェクトが返されます.
     * @exception Exception 例外.
     */
    public static final ReadExcelSheet readStream( InputStream in,String sheetName )
        throws Exception {
        ReadExcelSheet ret = null ;
        ret = new ReadExcelSheet() ;
        ret.open( in,sheetName ) ;
        return ret ;
    }
    
    /**
     * Excel読み込み用オブジェクトを取得.
     * @param binary バイナリを設定します.
     * @param sheetName 利用対象のExcelシート名を設定します.
     * @return ReadExcelSheet Excel読み込み用オブジェクトが返されます.
     * @exception Exception 例外.
     */
    public static final ReadExcelSheet readBinary( byte[] binary,String sheetName )
        throws Exception {
        ReadExcelSheet ret = null ;
        ret = new ReadExcelSheet() ;
        ret.open( binary,sheetName ) ;
        return ret ;
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        close() ;
    }
    
    /**
     * 帳票雛型のExcelファイルをオープン.
     * @param name Excelファイル名を設定します.
     * @param sheetName 利用対象のExcelシート名を設定します.
     * @exception Exception 例外.
     */
    public void open( String name,String sheetName ) throws Exception {
        if( sheet == null ) {
            sheet = new ExcelSheet( name,sheetName ) ;
        }
        else {
            sheet.open( name,sheetName ) ;
        }
    }
    
    /**
     * 帳票雛型のExcelファイルをオープン.
     * @param in InputStreamを設定します.
     * @param sheetName 利用対象のExcelシート名を設定します.
     * @exception Exception 例外.
     */
    public void open( InputStream in,String sheetName ) throws Exception {
        if( sheet == null ) {
            sheet = new ExcelSheet( in,sheetName ) ;
        }
        else {
            sheet.open( in,sheetName ) ;
        }
    }
    
    /**
     * 帳票雛型のExcelファイルをオープン.
     * @param binary バイナリを設定します.
     * @param sheetName 利用対象のExcelシート名を設定します.
     * @exception Exception 例外.
     */
    public void open( byte[] binary,String sheetName ) throws Exception {
        if( sheet == null ) {
            sheet = new ExcelSheet( binary,sheetName ) ;
        }
        else {
            sheet.open( binary,sheetName ) ;
        }
    }
    
    /**
     * オブジェクトクローズ.
     */
    public void close() {
        if( sheet != null ) {
            sheet.destroy() ;
        }
        sheet = null ;
    }
    
    /**
     * １つのセル内容を取得.
     * @param x 位置X値を設定します.
     * @param y 位置Y値を設定します.
     * @return Object セル内容が返されます.
     * @exception Exception 例外.
     */
    public Object getCell( int x,int y )
        throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        return sheet.getCell( x-1,y-1 ) ;
    }
    
    /**
     * １つのセル内容を取得.
     * @param x 位置X(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @return Object セル内容が返されます.
     * @exception Exception 例外.
     */
    public Object getCell( String x,int y )
        throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        return sheet.getCell( x,y-1 ) ;
    }
    
    /**
     * １つのセルタイプを取得.
     * @param x 位置X値を設定します.
     * @param y 位置Y値を設定します.
     * @return int セルタイプが返されます.
     *             [-1]の場合、取得に失敗したか、不明なタイプです.
     */
    public int getCellType( int x,int y ) {
        if( sheet == null ) {
            return -1 ;
        }
        return sheet.getCellType( x-1,y-1 ) ;
    }
    
    /**
     * １つのセルタイプを取得.
     * @param x 位置X(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @return int セルタイプが返されます.
     *             [-1]の場合、取得に失敗したか、不明なタイプです.
     */
    public int getCellType( String x,int y ) {
        if( sheet == null ) {
            return -1 ;
        }
        return sheet.getCellType( x,y-1 ) ;
    }
    
    /**
     * 指定セルのスタイルオブジェクトを取得.
     * @param x 位置X値を設定します.
     * @param y 位置Y値を設定します.
     * @return HSSFCellStyle セルスタイルオブジェクトが返されます.
     * @exception Exception 例外.
     */
    public HSSFCellStyle getCellStyle( int x,int y )
        throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        return sheet.getCellStyle( x-1,y-1 ) ;
    }
    
    /**
     * 指定セルのスタイルオブジェクトを取得.
     * @param x 位置X(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @return HSSFCellStyle セルスタイルオブジェクトが返されます.
     * @exception Exception 例外.
     */
    public HSSFCellStyle getCellStyle( String x,int y )
        throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        return sheet.getCellStyle( x,y-1 ) ;
    }
    
    /**
     * セルのロック条件を取得.
     * @param x 位置X値を設定します.
     * @param y 位置Y値を設定します.
     * @return boolean [true]の場合、セルロック条件が返されます.
     * @exception Exception 例外.
     */
    public boolean isCellLock( int x,int y )
        throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        return sheet.isCellLock( x-1,y-1 ) ;
    }
    
    /**
     * セルのロック条件を取得.
     * @param x 位置X(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @return boolean [true]の場合、セルロック条件が返されます.
     * @exception Exception 例外.
     */
    public boolean isCellLock( String x,int y )
        throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        return sheet.isCellLock( x,y-1 ) ;
    }
    
    /**
     * Excelのロック状態を取得.
     * @return boolean [true]の場合、ロックされています.
     * @exception Exception 例外.
     */
    public boolean isLock() throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        return sheet.isLock() ;
    }
    
    /**
     * 情報が存在するかチェック.
     * @return boolean [true]の場合、情報が利用できます.
     */
    public boolean isUse() {
        return sheet != null && sheet.isUse() ;
    }
}

