package org.maachang.report ;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.poi.hssf.usermodel.HSSFSheet;
import org.maachang.conf.Config;
import org.maachang.conf.ConvIniParam;
import org.maachang.util.FileUtil;
import org.maachang.util.StringUtil;

/**
 * 帳票エンジン.
 * 
 * @version 2008/10/01
 * @author masahito suzuki
 * @since MaachangReport 1.00
 */
public class ReportEngine {
    
    /**
     * ExcelSheetオブジェクト.
     */
    private ExcelSheet excelSheet = null ;
    
    /**
     * 帳票コアオブジェクト.
     */
    private final Map<String,ReportCore> coreMap = new HashMap<String,ReportCore>() ;
    
    /**
     * 帳票テンプレートシート名群.
     */
    private final List<String> templateSheetName = new ArrayList<String>() ;
    
    /**
     * 現在のアクティブシート名.
     */
    private String activeSheetName = null ;
    
    /**
     * 出力ファイル名.
     */
    private String outName = null ;
    
    /**
     * コンストラクタ.
     */
    private ReportEngine() {
    }
    
    /**
     * コンストラクタ.
     * @param name 対象のコンフィグファイル名を設定します.
     * @exception Exception 例外.
     */
    public ReportEngine( String name ) throws Exception {
        this( name,null,Config.read( name ) ) ;
    }
    
    /**
     * コンストラクタ.
     * @param name 対象のコンフィグファイル名を設定します.
     * @param binary 対象のExcel実体となるバイナリ情報を設定します.<br>
     *               [null]の場合は、nameのファイル名をExcel実体情報として採用します.
     * @exception Exception 例外.
     */
    public ReportEngine( String name,byte[] binary ) throws Exception {
        this( name,binary,Config.read( name ) ) ;
    }
    
    /**
     * コンストラクタ.
     * @param name オープン対象のファイル名を設定します.
     * @param binary 対象のExcel実体となるバイナリ情報を設定します.<br>
     *               [null]の場合は、nameのファイル名をExcel実体情報として採用します.
     * @param conf 対象のコンフィグ情報を設定します.
     * @exception Exception 例外.
     */
    public ReportEngine( String name,byte[] binary,Config conf ) throws Exception {
        _create( name,binary,conf ) ;
    }
    
    /**
     * コンストラクタ.
     * @param name オープン対象のファイル名を設定します.
     * @param sheetName 対象のシート名を設定します.
     * @param outputName 出力先のファイル名を設定します.
     * @exception Exception 例外.
     */
    public ReportEngine( String name,String sheetName,String outputName ) throws Exception {
        if( name == null || ( name = name.trim() ).length() <= 0 ||
            sheetName == null || ( sheetName = sheetName.trim() ).length() <= 0 ||
            outputName == null || ( outputName = outputName.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        Config conf = new Config() ;
        conf.put( "excel","name",name ) ;
        conf.put( "excel","sheet",sheetName ) ;
        conf.put( "excel","outName",outputName ) ;
        byte[] b = FileUtil.getFile( name ) ;
        _create( name,b,conf ) ;
    }
    
    /**
     * 生成処理.
     * @param name オープン対象のファイル名を設定します.
     * @param binary 対象のExcel実体となるバイナリ情報を設定します.<br>
     *               [null]の場合は、nameのファイル名をExcel実体情報として採用します.
     * @param conf 対象のコンフィグ情報を設定します.
     * @exception Exception 例外.
     */
    private void _create( String name,byte[] binary,Config conf ) throws Exception {
        if( name == null || ( name = name.trim() ).length() <= 0 || conf == null ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        String excelFile = conf.get( "excel","name",0 ) ;
        if( binary == null || binary.length <= 0 ) {
            if( excelFile == null || ( excelFile = excelFile.trim() ).length() <= 0 ) {
                if( name.endsWith( ".conf" ) ) {
                    excelFile = name.substring( 0,name.length()-".conf".length() ) + ".xls" ;
                }
                else {
                    excelFile = null ;
                }
            }
            else if( excelFile.indexOf( "/" ) == -1 ) {
                if( name.indexOf( "/" ) != -1 ) {
                    excelFile = name.substring( 0,name.lastIndexOf( "/" )+1 ) + excelFile ;
                }
            }
        }
        else {
            excelFile = null ;
        }
        ExcelSheet excelSheet = null ;
        // テンプレート対象シート名群を取得.
        int sheetLength = conf.size( "excel","sheet" ) ;
        if( sheetLength <= 0 ) {
            throw new IOException( "コンフィグ設定にシート名が設定されていません" ) ;
        }
        // リスト定義を取得.
        List<String> sheetList = new ArrayList<String>() ;
        List<CellStyleBean[]> engineList = new ArrayList<CellStyleBean[]>() ;
        for( int ii = 0 ; ii < sheetLength ; ii ++ ) {
            String sheetName = conf.get( "excel","sheet",ii ) ;
            CellStyleBean[] cbList = null ;
            String cellStyleName ;
            String[] keys ;
            if( sheetLength == 1 ) {
                cellStyleName = "cellStyle" ;
                keys = conf.getKeys( cellStyleName ) ;
                if( keys == null || keys.length <= 0 ) {
                    cellStyleName = "cellStyle-"+sheetName ;
                    keys = conf.getKeys( cellStyleName ) ;
                }
            }
            else {
                cellStyleName = "cellStyle-"+sheetName ;
                keys = conf.getKeys( cellStyleName ) ;
            }
            if( keys != null && keys.length > 0 ) {
                ArrayList<CellStyleBean> lst = new ArrayList<CellStyleBean>() ;
                int lenJ = keys.length ;
                for( int j = 0 ; j < lenJ ; j ++ ) {
                    String v = conf.get( cellStyleName,keys[ j ],0 ) ;
                    if( v != null && v.length() > 0 ) {
                        List<String> csv = StringUtil.cutString( v,"," ) ;
                        if( csv.size() != 4 ) {
                            continue ;
                        }
                        if( StringUtil.checkType( StringUtil.CHECK_NUMBER,csv.get( 1 ) ) == false ||
                            StringUtil.checkType( StringUtil.CHECK_NUMBER,csv.get( 3 ) ) == false ) {
                            continue ;
                        }
                        if( StringUtil.checkType( StringUtil.CHECK_ALPHABET,csv.get( 0 ) ) ) {
                            if( StringUtil.checkType( StringUtil.CHECK_ALPHABET,csv.get( 2 ) ) ) {
                                CellStyleBean bean = new CellStyleBean( keys[ j ],
                                    csv.get( 0 ),
                                    ConvIniParam.getInt( csv.get( 1 ) ),
                                    csv.get( 2 ),
                                    ConvIniParam.getInt( csv.get( 3 ) ) ) ;
                                lst.add( bean ) ;
                            }
                            else {
                                CellStyleBean bean = new CellStyleBean( keys[ j ],
                                    csv.get( 0 ),
                                    ConvIniParam.getInt( csv.get( 1 ) ),
                                    ConvIniParam.getInt( csv.get( 2 ) ),
                                    ConvIniParam.getInt( csv.get( 3 ) ) ) ;
                                lst.add( bean ) ;
                            }
                        }
                        else if( StringUtil.checkType( StringUtil.CHECK_NUMBER,csv.get( 0 ) ) ) {
                            if( StringUtil.checkType( StringUtil.CHECK_ALPHABET,csv.get( 2 ) ) ) {
                                CellStyleBean bean = new CellStyleBean( keys[ j ],
                                    ConvIniParam.getInt( csv.get( 0 ) ),
                                    ConvIniParam.getInt( csv.get( 1 ) ),
                                    csv.get( 2 ),
                                    ConvIniParam.getInt( csv.get( 3 ) ) ) ;
                                lst.add( bean ) ;
                            }
                            else {
                                CellStyleBean bean = new CellStyleBean( keys[ j ],
                                    ConvIniParam.getInt( csv.get( 0 ) ),
                                    ConvIniParam.getInt( csv.get( 1 ) ),
                                    ConvIniParam.getInt( csv.get( 2 ) ),
                                    ConvIniParam.getInt( csv.get( 3 ) ) ) ;
                                lst.add( bean ) ;
                            }
                        }
                        else {
                            continue ;
                        }
                    }
                }
                if( lst.size() != 0 ) {
                    lenJ = lst.size() ;
                    cbList = new CellStyleBean[ lenJ ] ;
                    for( int j = 0 ; j < lenJ ; j ++ ) {
                        cbList[ j ] = lst.get( j ) ;
                    }
                    lst.clear() ;
                }
                lst = null ;
            }
            if( excelSheet == null ) {
                if( binary != null && binary.length > 0 ) {
                    excelSheet = new ExcelSheet( binary,sheetName ) ;
                }
                else {
                    excelSheet = new ExcelSheet( excelFile,sheetName ) ;
                }
            }
            engineList.add( cbList ) ;
            sheetList.add( sheetName ) ;
        }
        // セル定義条件を処理.
        createCellDefine( conf,sheetList,excelSheet ) ;
        // Engineを生成.
        int len = sheetList.size() ;
        for( int i = 0 ; i < len ; i ++ ) {
            this.createEngine( excelSheet,sheetList.get( i ),engineList.get( i ) ) ;
        }
        // アクティブなExcelSheetにセット.
        excelSheet.changeSheet( activeSheetName ) ;
        String outName = conf.get( "excel","outName",0 ) ;
        if( outName == null || ( outName = outName.trim() ).length() <= 0 ) {
            outName = "report.xls" ;
        }
        this.outName = outName ;
        this.excelSheet = excelSheet ;
    }
    
    /** エンジン生成. **/
    private void createEngine( ExcelSheet excelSheet,String sheetName,CellStyleBean[] cbList ) throws Exception {
        excelSheet.changeSheet( sheetName ) ;
        ReportCore pc = new ReportCore( excelSheet ) ;
        HSSFSheet hs = excelSheet.getSheet() ;
        int len = 0 ;
        if( cbList != null ) {
            len = cbList.length ;
        }
        CellStyle cs = null ;
        for( int i = 0 ; i < len ; i ++ ) {
            cs = new CellStyle( cbList[ i ],hs ) ;
            pc.addCellStyle( cs ) ;
        }
        pc.fix() ;
        coreMap.put( sheetName,pc ) ;
        templateSheetName.add( sheetName ) ;
        // 最初のシート名がアクティブシートとする.
        if( activeSheetName == null ) {
            activeSheetName = sheetName ;
        }
    }
    
    /** セル定義処理. **/
    private void createCellDefine( Config conf,List<String> sheetList,ExcelSheet excelSheet ) throws Exception {
        int len = sheetList.size() ;
        Object o ;
        String s ;
        Map<Integer,String> target = new HashMap<Integer,String>() ;
        for( int i = 0 ; i < len ; i ++ ) {
            String sheetName = sheetList.get( i ) ;
            String section = "staticList-"+sheetName ;
            String[] keys = conf.getKeys( section ) ;
            if( keys == null || keys.length <= 0 ) {
                continue ;
            }
            // 対象シートをアクティブ.
            excelSheet.changeSheet( sheetName ) ;
            int lenJ = keys.length ;
            for( int j = 0 ; j < lenJ ; j ++ ) {
                target.clear() ;
                String v = conf.get( section,keys[ j ],0 ) ;
                if( v == null ) {
                    continue ;
                }
                List<String> csv = StringUtil.cutString( v,"," ) ;
                if( csv.size() != 4 ) {
                    continue ;
                }
                if( !StringUtil.checkType( StringUtil.CHECK_NUMBER,csv.get( 0 ) ) ||
                    !StringUtil.checkType( StringUtil.CHECK_NUMBER,csv.get( 1 ) ) ) {
                    continue ;
                }
                if( StringUtil.checkType( StringUtil.CHECK_ALPHABET,csv.get( 2 ) ) ) {
                    csv.set( 2,String.valueOf( ExcelUtil.getPosAz( csv.get( 2 ) ) ) ) ;
                }
                if( StringUtil.checkType( StringUtil.CHECK_ALPHABET,csv.get( 3 ) ) ) {
                    csv.set( 3,String.valueOf( ExcelUtil.getPosAz( csv.get( 3 ) ) ) ) ;
                }
                int sx,ex ;
                int sy,ey ;
                sy = Integer.parseInt( csv.get( 0 ) ) - 1 ;
                ey = ( Integer.parseInt( csv.get( 1 ) ) + sy ) - 1 ;
                sx = Integer.parseInt( csv.get( 2 ) ) ;
                ex = Integer.parseInt( csv.get( 3 ) ) ;
                if( sy < 0 ) {
                    continue ;
                }
                // 範囲内の情報に対して、固定帳票用処理を実装.
                boolean firstFlag = false ;
                int count = 1 ;
                for( int y = sy ; y <= ey ; y ++ ) {
                    for( int x = sx ; x <= ex ; x ++ ) {
                        // 最初の行は、文字条件を検索して取得.
                        if( firstFlag == false ) {
                            o = excelSheet.getCell( x,y ) ;
                            if( o instanceof String && (s=StringUtil.trim( (String)o )).length() > 0 &&
                                ( ( s.startsWith( "${" ) || s.startsWith( "$(" ) ) && ( s.endsWith( "}" ) || s.endsWith( ")" ) ) ) ) {
                                if( s.indexOf( "#" ) == -1 ) {
                                    s = StringUtil.trim( s.substring( 2,s.length()-1 ) ) ;
                                    s = "${"+s+"#}" ;
                                }
                                target.put( x,s ) ;
                                excelSheet.setCell( x,y,StringUtil.changeString( s,"#",String.valueOf( count ) ) ) ;
                            }
                        }
                        else if( ( s = target.get( x ) ) != null ) {
                            excelSheet.setCell( x,y,StringUtil.changeString( s,"#",String.valueOf( count ) ) ) ;
                        }
                    }
                    firstFlag = true ;
                    count ++ ;
                }
            }
        }
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        destroy() ;
    }
    
    /**
     * オブジェクトを破棄.
     */
    public void destroy() {
        if( coreMap.size() > 0 ) {
            Iterator it = coreMap.keySet().iterator() ;
            while( it.hasNext() ) {
                String k = (String)it.next() ;
                ReportCore c = coreMap.get( k ) ;
                if( c != null ) {
                    c.destroy() ;
                }
            }
            coreMap.clear() ;
        }
        templateSheetName.clear() ;
        outName = null ;
        excelSheet = null ;
        activeSheetName = null ;
    }
    
    /**
     * 現在のアクティブシート名に対するReportCoreを取得.
     */
    protected ReportCore _reportCore() {
        if( activeSheetName == null ) {
            return null ;
        }
        return coreMap.get( activeSheetName ) ;
    }
    
    /**
     * ExcelSheetオブジェクトを取得.
     * @return ExcelSheet ExcelSheetオブジェクトが返されます.
     * @exception Exception 例外.
     */
    public ExcelSheet getExcelSheet()
        throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        return excelSheet ;
    }
    
    /**
     * アクティブなシート名を取得.
     * @return String アクティブなシート名が返されます.
     */
    public String getActiveSheetName() {
        return activeSheetName ;
    }
    
    /**
     * ExcelSheetを変更.
     * @param name シート名を設定します.
     * @return boolean [true]の場合、処理成功.
     * @exception Exception 例外.
     */
    public boolean changeSheet( String name )
        throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( name == null || name.length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        ExcelSheet sheet = excelSheet ;
        if( sheet != null ) {
            boolean ret = sheet.changeSheet( name ) ;
            if( ret ) {
                activeSheetName = name ;
            }
        }
        return false ;
    }
    
    /**
     * ExcelSheetを削除.
     * @param name シート名を設定します.
     * @return boolean [true]の場合、処理成功.
     * @exception Exception 例外.
     */
    public boolean removeSheet( String name )
        throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( name == null || name.length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        if( activeSheetName.equals( name ) ) {
            throw new IOException( "アクティブなシートは削除できません" ) ;
        }
        ExcelSheet sheet = excelSheet ;
        if( sheet != null ) {
            boolean ret = sheet.removeSheet( name ) ;
            if( ret ) {
                coreMap.remove( name ) ;
            }
            return ret ;
        }
        return false ;
    }
    
    /**
     * シート情報をコピー.
     * @param name コピー対象のシート名を設定します.
     * @param newName コピーされたシート名に対する新しい名前を設定します.
     * @return boolean [true]の場合、処理成功.
     * @exception Exception 例外.
     */
    public boolean copySheet( String name,String newName )
        throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( name == null || name.length() <= 0 || newName == null || newName.length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        if( coreMap.containsKey( newName ) ) {
            throw new IOException( "既に同一名のシートが存在します:" + newName ) ;
        }
        ExcelSheet sheet = excelSheet ;
        if( sheet != null ) {
            boolean ret = sheet.copySheet( name,newName ) ;
            if( ret ) {
                ReportCore c = coreMap.get( name ) ;
                if( c != null ) {
                    coreMap.put( newName,c ) ;
                }
            }
            return ret ;
        }
        return false ;
    }
    
    /**
     * レポートテンプレート用のシートを削除.
     * @exception Exception 例外.
     */
    public void removeTemplate() throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        int len = templateSheetName.size() ;
        for( int i = 0 ; i < len ; i ++ ) {
            removeSheet( templateSheetName.get( i ) ) ;
        }
    }
    
    /**
     * レポートコアでのセル移動をリセット.
     * @exception Exception 例外.
     */
    public void resetCellStyle() throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        ReportCore core = _reportCore() ;
        if( core != null ) {
            core.resetCellStyle() ;
        }
    }
    
    /**
     * Fixを解除.
     * @exception Exception 例外.
     */
    public void resetFix() throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        ReportCore core = _reportCore() ;
        if( core != null ) {
            core.resetFix() ;
        }
    }
    
    /**
     * 新しいリストを作成.
     * <p>※この処理は「セルスタイル」を定義していないと、意味を持ちません.</p>
     * <p>利用方法として、リスト条件に情報をセットした後、次のリストに対して
     * 情報をセットする時に、この処理を呼び出します.</p>
     * @exception Exception 例外.
     */
    public void nextList() throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        ReportCore core = _reportCore() ;
        Set<String> csMan = core.getCellManager() ;
        Iterator it = csMan.iterator() ;
        while( it.hasNext() ) {
            String name = ( String )it.next() ;
            CellStyle cs = core.getCellStyleList().get( name ) ;
            if( cs != null ) {
                cs.insert() ;
            }
        }
        csMan.clear() ;
    }
    
    /**
     * 変数に対して、情報を追加.
     * @param name 設定名を設定します.
     * @param value 設定オブジェクトを設定します.
     * @exception Exception 例外.
     */
    public void putValue( String name,Object value )
        throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        ReportCore core = _reportCore() ;
        if( core.isFix() == false ) {
            throw new IOException( "この処理はオブジェクトがフィックスしないと、利用できません" ) ;
        }
        if( name == null || ( name = name.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        if( name.startsWith( "seq:" ) ) {
            // シーケンス項番の場合は、無視.
            return ;
        }
        if( value == null ) {
            value = "" ;
        }
        CellPositionList cpList = core.getCellPositionList() ;
        ExcelSheet element = excelSheet ;
        CellPosition[] cp = cpList.get( name ) ;
        if( cp == null ) {
            throw new IllegalArgumentException( "指定名[" + name + "]の要素は存在しません" ) ;
        }
        int len = cp.length ;
        Set<String> csMan = core.getCellManager() ;
        for( int i = 0 ; i < len ; i ++ ) {
            if( cp[ i ].getCellStyleName() != null ) {
                String csName = cp[ i ].getCellStyleName() ;
                if( csName != null && csMan.contains( csName ) == false ) {
                    CellStyle cs = core.getCellStyleList().get( csName ) ;
                    if( cs != null && cs.isInsert() ) {
                        core.insertCellStyle( csName ) ;
                    }
                    core.pushSequence( csName ) ;
                    csMan.add( csName ) ;
                }
            }
            element.setCell( cp[ i ].getX(),cp[ i ].getY(),cp[ i ].output( value ) ) ;
        }
    }
    
    /**
     * 指定名が利用可能かチェック.
     * @param name 対象の名前を設定します.
     * @return boolean [true]の場合、利用可能です.
     * @exception Exception 例外.
     */
    public boolean containsKey( String name )
        throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        ReportCore core = _reportCore() ;
        if( core.isFix() == false ) {
            throw new IOException( "この処理はオブジェクトがフィックスしないと、利用できません" ) ;
        }
        return ( core.getCellPositionList().get( name ) != null ) ;
    }
    
    /**
     * 処理結果のExcel情報をバイナリで取得.
     * @return byte[] 処理結果のExcel情報が返されます.
     * @exception Exception 例外.
     */
    public byte[] getBytes() throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        return excelSheet.getBytes() ;
    }
    
    /**
     * 処理結果のExcel情報をファイル出力.
     * @exception Exception 例外.
     */
    public void output() throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        excelSheet.output( this.outName ) ;
    }
    
    /**
     * 処理結果のExcel情報をファイル出力.
     * @param name 出力先のファイル名を設定します.
     * @exception Exception 例外.
     */
    public void output( String name ) throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        excelSheet.output( name ) ;
    }
    
    /**
     * ReportElementオブジェクトを取得.
     * <p>このメソッドは、オープン中のExcelSheetを直接操作したい場合に利用します.</p>
     * @return ReportElement ReportElementオブジェクトが返されます.
     * @exception Exception 例外.
     */
    public ReportElement getElement()
        throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        return new ReportElement( excelSheet ) ;
    }
    
    /**
     * 出力名を取得.
     * @return 出力名が返されます.
     */
    public String getOutName() {
        return outName ;
    }
    
    /**
     * オブジェクトが有効かチェック.
     * @return boolean [true]の場合、有効です.
     */
    public boolean isUse() {
        return ( excelSheet != null ) ;
    }
    
}

