package org.maachang.report ;

import java.io.IOException;
import java.io.InputStream;

import org.apache.poi.hssf.usermodel.HSSFCellStyle;

/**
 * Excel読み書き用.
 * 
 * @version 2009/08/26
 * @author masahito suzuki
 * @since MaachangReport 1.00
 */
public class WriteExcelSheet {
    private WriteExcelSheet(){}
    private ExcelSheet sheet = null ;
    
    /**
     * セルタイプ : ブランク(なし).
     */
    public static final int CELL_BLANK = ExcelSheet.CELL_BLANK ;
    
    /**
     * セルタイプ : 数字.
     */
    public static final int CELL_NUMBER = ExcelSheet.CELL_NUMBER ;
    
    /**
     * セルタイプ : 文字.
     */
    public static final int CELL_STRING = ExcelSheet.CELL_STRING ;
    
    /**
     * セルタイプ : Boolean.
     */
    public static final int CELL_BOOLEAN = ExcelSheet.CELL_BOOLEAN ;
    
    /**
     * セルタイプ : 日付.
     */
    public static final int CELL_DATE = ExcelSheet.CELL_DATE ;
    
    /**
     * Excel読み書き用オブジェクトを取得.
     * @param name Excelファイル名を設定します.
     * @param sheetName 利用対象のExcelシート名を設定します.
     * @return WriteExcelSheet Excel読み書き用オブジェクトが返されます.
     * @exception Exception 例外.
     */
    public static final WriteExcelSheet read( String name,String sheetName )
        throws Exception {
        WriteExcelSheet ret = null ;
        try {
            ret = new WriteExcelSheet() ;
            ret.open( name,sheetName ) ;
        } catch( Exception e ) {
            ret = null ;
        }
        return ret ;
    }
    
    /**
     * Excel読み書き用オブジェクトを取得.
     * @param in InputStreamを設定します.
     * @param sheetName 利用対象のExcelシート名を設定します.
     * @return WriteExcelSheet Excel読み書き用オブジェクトが返されます.
     * @exception Exception 例外.
     */
    public static final WriteExcelSheet readStream( InputStream in,String sheetName )
        throws Exception {
        WriteExcelSheet ret = null ;
        try {
            ret = new WriteExcelSheet() ;
            ret.open( in,sheetName ) ;
        } catch( Exception e ) {
            ret = null ;
        }
        return ret ;
    }
    
    /**
     * Excel読み書き用オブジェクトを取得.
     * @param binary バイナリを設定します.
     * @param sheetName 利用対象のExcelシート名を設定します.
     * @return WriteExcelSheet Excel読み書き用オブジェクトが返されます.
     * @exception Exception 例外.
     */
    public static final WriteExcelSheet readBinary( byte[] binary,String sheetName )
        throws Exception {
        WriteExcelSheet ret = null ;
        try {
            ret = new WriteExcelSheet() ;
            ret.open( binary,sheetName ) ;
        } catch( Exception e ) {
            ret = null ;
        }
        return ret ;
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        close() ;
    }
    
    /**
     * 帳票雛型のExcelファイルをオープン.
     * @param name Excelファイル名を設定します.
     * @param sheetName 利用対象のExcelシート名を設定します.
     * @exception Exception 例外.
     */
    public void open( String name,String sheetName ) throws Exception {
        if( sheet == null ) {
            sheet = new ExcelSheet( name,sheetName ) ;
        }
        else {
            sheet.open( name,sheetName ) ;
        }
    }
    
    /**
     * 帳票雛型のExcelファイルをオープン.
     * @param in InputStreamを設定します.
     * @param sheetName 利用対象のExcelシート名を設定します.
     * @exception Exception 例外.
     */
    public void open( InputStream in,String sheetName ) throws Exception {
        if( sheet == null ) {
            sheet = new ExcelSheet( in,sheetName ) ;
        }
        else {
            sheet.open( in,sheetName ) ;
        }
    }
    
    /**
     * 帳票雛型のExcelファイルをオープン.
     * @param binary バイナリを設定します.
     * @param sheetName 利用対象のExcelシート名を設定します.
     * @exception Exception 例外.
     */
    public void open( byte[] binary,String sheetName ) throws Exception {
        if( sheet == null ) {
            sheet = new ExcelSheet( binary,sheetName ) ;
        }
        else {
            sheet.open( binary,sheetName ) ;
        }
    }
    
    /**
     * オブジェクトクローズ.
     */
    public void close() {
        if( sheet != null ) {
            sheet.destroy() ;
        }
        sheet = null ;
    }
    
    /**
     * 現在のExcel情報をファイル出力.
     * @param name 出力先のファイル名を設定します.
     * @exception Exception 例外.
     */
    public void output( String name ) throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        sheet.output( name ) ;
    }
    
    /**
     * 現在のExcel情報をバイナリで取得.
     * @return byte[] 現在のExcel情報が返されます.
     * @exception Exception 例外.
     */
    public byte[] getBytes() throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        return sheet.getBytes() ;
    }
    
    /**
     * １つのセル情報に設定.
     * @param x 位置X値を設定します.
     * @param y 位置Y値を設定します.
     * @param data 設定情報をセットします.
     * @exception Exception 例外.
     */
    public void setCell( int x,int y,Object data )
        throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        sheet.setCell( x-1,y-1,data ) ;
    }
    
    /**
     * １つのセル情報に設定.
     * @param x 位置X値(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @param data 設定情報をセットします.
     * @exception Exception 例外.
     */
    public void setCell( String x,int y,Object data )
        throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        sheet.setCell( x,y-1,data ) ;
    }
    
    /**
     * 指定Row対して、セル追加.
     * @param y 追加対象のY軸を設定します.
     * @param n 追加個数を設定します.
     * @exception Exception 例外.
     */
    public void pushRow( int y,int n ) throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        sheet.pushRow( y-1,n ) ;
    }
    
    /**
     * １つのセル内容を取得.
     * @param x 位置X値を設定します.
     * @param y 位置Y値を設定します.
     * @return Object セル内容が返されます.
     * @exception Exception 例外.
     */
    public Object getCell( int x,int y )
        throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        return sheet.getCell( x-1,y-1 ) ;
    }
    
    /**
     * １つのセル内容を取得.
     * @param x 位置X(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @return Object セル内容が返されます.
     * @exception Exception 例外.
     */
    public Object getCell( String x,int y )
        throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        return sheet.getCell( x,y-1 ) ;
    }
    
    /**
     * １つのセルタイプを取得.
     * @param x 位置X値を設定します.
     * @param y 位置Y値を設定します.
     * @return int セルタイプが返されます.
     *             [-1]の場合、取得に失敗したか、不明なタイプです.
     */
    public int getCellType( int x,int y ) {
        if( sheet == null ) {
            return -1 ;
        }
        return sheet.getCellType( x-1,y-1 ) ;
    }
    
    /**
     * １つのセルタイプを取得.
     * @param x 位置X(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @return int セルタイプが返されます.
     *             [-1]の場合、取得に失敗したか、不明なタイプです.
     */
    public int getCellType( String x,int y ) {
        if( sheet == null ) {
            return -1 ;
        }
        return sheet.getCellType( x,y-1 ) ;
    }
    
    /**
     * 新しいセルスタイルを生成.
     * @return HSSFCellStyle 空のセルスタイルを作成します.
     * @exception Exception 例外.
     */
    public HSSFCellStyle createCellStyle() throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        return sheet.createCellStyle() ;
    }
    
    /**
     * 指定セルにスタイルオブジェクトを設定.
     * @param style セルスタイルオブジェクトを設定します.
     * @param x 位置X値を設定します.
     * @param y 位置Y値を設定します.
     * @exception Exception 例外.
     */
    public void setCellStyle( HSSFCellStyle style,int x,int y )
        throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        sheet.setCellStyle( style,x-1,y-1 ) ;
    }
    
    /**
     * 指定セルのスタイルオブジェクトを取得.
     * @param style セルスタイルオブジェクトを設定します.
     * @param x 位置X(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @exception Exception 例外.
     */
    public void setCellStyle( HSSFCellStyle style,String x,int y )
        throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        sheet.setCellStyle( style,x,y-1 ) ;
    }
    
    /**
     * 指定セルのスタイルオブジェクトを取得.
     * @param x 位置X値を設定します.
     * @param y 位置Y値を設定します.
     * @return HSSFCellStyle セルスタイルオブジェクトが返されます.
     * @exception Exception 例外.
     */
    public HSSFCellStyle getCellStyle( int x,int y )
        throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        return sheet.getCellStyle( x-1,y-1 ) ;
    }
    
    /**
     * 指定セルのスタイルオブジェクトを取得.
     * @param x 位置X(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @return HSSFCellStyle セルスタイルオブジェクトが返されます.
     * @exception Exception 例外.
     */
    public HSSFCellStyle getCellStyle( String x,int y )
        throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        return sheet.getCellStyle( x,y-1 ) ;
    }
    
    /**
     * 指定行を削除.
     * @param y 削除行位置を設定します.
     * @exception Exception 例外.
     */
    public void removeRow( int y ) throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        sheet.removeRow( y-1 ) ;
    }
    
    /**
     * 指定セルを削除.
     * @param x 位置Xを設定します.
     * @param y 位置Y値を設定します.
     * @exception Exception 例外.
     */
    public void removeCell( int x,int y ) throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        sheet.removeCell( x,y-1 ) ;
    }
    
    /**
     * 指定セルを削除.
     * @param x 位置X(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @exception Exception 例外.
     */
    public void removeCell( String x,int y ) throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        sheet.removeCell( x,y-1 ) ;
    }
    
    /**
     * セルのロック条件を設定.
     * @param boolean mode ロックをONにする場合は、[true]を設定します.
     * @param x 位置X値を設定します.
     * @param y 位置Y値を設定します.
     * @exception Exception 例外.
     */
    public void cellLock( boolean mode,int x,int y )
        throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        sheet.cellLock( mode,x-1,y-1 ) ;
    }
    
    /**
     * セルロック条件を設定.
     * @param boolean mode ロックをONにする場合は、[true]を設定します.
     * @param x 位置X(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @exception Exception 例外.
     */
    public void cellLock( boolean mode,String x,int y )
        throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        sheet.cellLock( mode,x,y-1 ) ;
    }
    
    /**
     * セルのロック条件を取得.
     * @param x 位置X値を設定します.
     * @param y 位置Y値を設定します.
     * @return boolean [true]の場合、セルロック条件が返されます.
     * @exception Exception 例外.
     */
    public boolean isCellLock( int x,int y )
        throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        return sheet.isCellLock( x-1,y-1 ) ;
    }
    
    /**
     * セルのロック条件を取得.
     * @param x 位置X(ExcelA-Z座標)を設定します.
     * @param y 位置Y値を設定します.
     * @return boolean [true]の場合、セルロック条件が返されます.
     * @exception Exception 例外.
     */
    public boolean isCellLock( String x,int y )
        throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        return sheet.isCellLock( x,y-1 ) ;
    }
    
    /**
     * Excelをロックする.
     * ※アンロック自体はできないので注意.
     * @param passwd ロックパスワードを設定します.
     * @exception Exception 例外.
     */
    public void setLock( String passwd ) throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        sheet.setLock( passwd ) ;
    }
    
    /**
     * Excelのロック状態を取得.
     * @return boolean [true]の場合、ロックされています.
     * @exception Exception 例外.
     */
    public boolean isLock() throws Exception {
        if( sheet == null ) {
            throw new IOException( "ExcelSheetは既に閉じています" ) ;
        }
        return sheet.isLock() ;
    }
    
    /**
     * 情報が存在するかチェック.
     * @return boolean [true]の場合、情報が利用できます.
     */
    public boolean isUse() {
        return sheet != null && sheet.isUse() ;
    }
}

