package org.maachang.seabass.conf ;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;

import org.maachang.conf.Config;
import org.maachang.conf.ReadIni;
import org.maachang.seabass.SeaBassDef;
import org.maachang.util.FileUtil;
import org.maachang.util.atomic.AtomicINT;
import org.maachang.util.atomic.AtomicLONG;
import org.maachang.util.atomic.AtomicOBJECT;

/**
 * javascript内でデフォルトインポートされるパッケージ名管理.
 *
 * @version 2007/08/29
 * @author  masahito suzuki
 * @since   SeaBass 1.00
 */
public class PackagesConfig {
    
    /**
     * ターゲットセクション名.
     */
    private static final String BASE_SECTION = "import-package" ;
    
    /**
     * キャラクタセット.
     */
    private static final String CHARSET = "UTF8" ;
    
    /**
     * 読み込みファイル名.
     */
    private static final String READ_FILE = SeaBassDef.DIRECTORY_CONFIG+"packages.conf" ;
    
    /**
     * デフォルトパッケージ.
     */
    private static final String[] DEFAULT_PACKAGE = new String[]{
        "java.util.*",
        "java.lang.*",
        "java.io.*",
        "java.text.*",
        "java.math.*",
        "org.maachang.seabass.mod.object.*" } ;
    
    /**
     * 対象ファイル.
     */
    private String fileName = null ;
    
    /**
     * 読み込み対象データ.
     */
    private final AtomicOBJECT<Config> iniParams = new AtomicOBJECT<Config>() ;
    
    /**
     * 管理パッケージ名群.
     */
    private final AtomicOBJECT<String[]> packages = new AtomicOBJECT<String[]>() ;
    
    /**
     * 最終更新日.
     */
    private final AtomicLONG lastUpdate = new AtomicLONG( -1L ) ;
    
    /**
     * 更新ID.
     */
    private final AtomicINT updateId = new AtomicINT( 0 ) ;
    
    /**
     * オープン.
     * <BR><BR>
     * オープンします.
     * <BR>
     * @exception Exception 例外.
     */
    public void open() throws Exception {
        this.open( READ_FILE ) ;
    }
    
    /**
     * オープン.
     * <BR><BR>
     * オープンします.
     * <BR>
     * @param name 対象のファイル名を設定します.
     * @exception Exception 例外.
     */
    public void open( String name ) throws Exception {
        this.fileName = FileUtil.getFullPath( name ) ;
        try {
            this.reload() ;
        } catch( Exception e ) {
           this.close() ;
           throw e ;
        } finally {
        }
    }
    
    /**
     * クローズ.
     * <BR><BR>
     * クローズ処理.
     */
    public void close() {
        iniParams.set( null ) ;
        packages.set( null ) ;
        lastUpdate.set( -1L ) ;
        updateId.set( 0 ) ;
    }
    
    /**
     * パッケージ名群を取得.
     * <BR><BR>
     * 格納されているパッケージ名群を取得します.
     * <BR>
     * @return String[] パッケージ名群が返されます.
     * @exception Exception 例外.
     */
    public String[] getPackages()
        throws Exception {
        updateByReload() ;
        if( iniParams.get() == null ) {
            return null ;
        }
        if( packages.get() != null ) {
            return packages.get() ;
        }
        String[] ret = iniParams.get().getAll( BASE_SECTION,"package" ) ;
        if( ret != null && ret.length > 0 ) {
            int len = ret.length ;
            for( int i = 0 ; i < len ; i ++ ) {
                if( ret[i] == null || ( ret[i] = ret[i].trim() ).length() <= 0 ) {
                    ret[i] = null ;
                    continue ;
                }
            }
            int defLen = DEFAULT_PACKAGE.length ;
            List<String> tmp = new ArrayList<String>() ;
            for( int i = 0 ; i < len ; i ++ ) {
                if( ret[i] != null ) {
                    tmp.add( ret[ i ] ) ;
                }
            }
            ret = null ;
            len = tmp.size() ;
            for( int def = 0 ; def < defLen ; def ++ ) {
                boolean flg = false ;
                for( int i = 0 ; i < len ; i ++ ) {
                    if( tmp.get( i ).indexOf( DEFAULT_PACKAGE[def] ) != -1 ) {
                        flg = true ;
                        break ;
                    }
                }
                if( flg == false ) {
                    tmp.add( DEFAULT_PACKAGE[def] ) ;
                }
            }
            len = tmp.size() ;
            ret = new String[ len ] ;
            for( int i = 0 ; i < len ; i ++ ) {
                ret[ i ] = tmp.get( i ) ;
            }
            tmp = null ;
        }
        else {
            int len = DEFAULT_PACKAGE.length ;
            ret = new String[ len ] ;
            for( int i = 0 ; i < len ; i ++ ) {
                ret[ i ] = DEFAULT_PACKAGE[ i ] ;
            }
        }
        int len = ret.length ;
        for( int i = 0 ; i < len ; i ++ ) {
            if( ret[i] != null && ret[i].startsWith( "import" ) == false ) {
                ret[i] = "import " + ret[i] ;
            }
        }
        packages.set( ret ) ;
        return ret ;
    }
    
    /**
     * 更新IDを取得.
     * <BR><BR>
     * 更新IDを取得します.
     * <BR>
     * @return int 更新IDが返されます.
     */
    public int getUpdateId() {
        updateByReload() ;
        return updateId.get() ;
    }
    
    /**
     * 再読み込み.
     */
    private final void reload()
        throws Exception {
        BufferedReader buf = null ;
        if( FileUtil.isFileExists( fileName ) == true ) {
            try {
                buf = new BufferedReader(
                    new InputStreamReader(
                        new FileInputStream( fileName ),CHARSET ) ) ;
                Config iniParams = new Config() ;
                ReadIni.analisys( iniParams,buf ) ;
                buf.close() ;
                buf = null ;
                long last = FileUtil.getLastTime( fileName ) ;
                // 読み込みデータを設定.
                this.iniParams.set( iniParams ) ;
                this.lastUpdate.set( last ) ;
            } catch( Exception e ) {
                throw e ;
            } finally {
                if( buf != null ) {
                    try {
                        buf.close() ;
                    } catch( Exception e ) {
                    }
                }
            }
        }
    }
    
    /**
     * ファイルが更新されている場合は、再読み込み.
     */
    private void updateByReload() {
        try {
            long tm = FileUtil.getLastTime( fileName ) ;
            if( lastUpdate.get() != tm ) {
                this.reload() ;
                packages.set( null ) ;
                updateId.inc() ;
                if( updateId.get() >= 10000 ) {
                    updateId.set( 0 ) ;
                }
            }
        } catch( Exception e ) {
        }
    }
}
