package org.maachang.seabass.conf ;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.InputStreamReader;

import org.maachang.ModUtil;
import org.maachang.conf.Config;
import org.maachang.conf.ReadIni;
import org.maachang.seabass.SeaBassDef;
import org.maachang.seabass.httpd.HttpdVersionDef;
import org.maachang.seabass.httpd.engine.PublicCache;
import org.maachang.util.Digest;
import org.maachang.util.FileUtil;
import org.maachang.util.RandomUtil;
import org.maachang.util.atomic.AtomicBOOL;
import org.maachang.util.atomic.AtomicOBJECT;

/**
 * Seabass32暗号基本キー定義.
 *
 * @version 2009/08/30
 * @author  masahito suzuki
 * @since   SeaBass 1.00
 */
public class Seabass32Key {
    
    /** デフォルトキー発行用. **/
    public static final String DEFAULT_CREATE_KEY_INFO = "_@seabass32("+
        HttpdVersionDef.getName()+")-" + HttpdVersionDef.getVersion() + "_" + ModUtil.version() + "_@@" ;
    
    /** デフォルトキー保管オブジェクト. **/
    private final AtomicOBJECT<String> defaultKey = new AtomicOBJECT<String>() ;
    
    /**
     * ターゲットセクション名.
     */
    private static final String BASE_SECTION = "seabass32" ;
    
    /**
     * キャラクタセット.
     */
    private static final String CHARSET = "UTF8" ;
    
    /**
     * 読み込みファイル名.
     */
    private static final String READ_FILE = SeaBassDef.DIRECTORY_CONFIG+"seabass32.conf" ;
    
    /**
     * 読み込み対象データ.
     */
    private Config iniParams = null ;
    
    /**
     * 対象ファイル.
     */
    private String fileName = null ;
    
    /**
     * 最終更新日.
     */
    private long lastUpdate = -1L ;
    
    /**
     * Seabass32キー情報.
     */
    private final AtomicOBJECT<String> key = new AtomicOBJECT<String>( null ) ;
    
    /**
     * ランダムデータ発行フラグ.
     */
    private final AtomicBOOL randomFlag = new AtomicBOOL( false ) ;
    
    /**
     * オープン.
     * <BR><BR>
     * オープンします.
     * <BR>
     * @exception Exception 例外.
     */
    public synchronized void open() throws Exception {
        this.open( READ_FILE ) ;
    }
    
    /**
     * オープン.
     * <BR><BR>
     * オープンします.
     * <BR>
     * @param name 対象のファイル名を設定します.
     * @exception Exception 例外.
     */
    public synchronized void open( String name ) throws Exception {
        this.fileName = FileUtil.getFullPath( name ) ;
        try {
            this.reload() ;
        } catch( Exception e ) {
            this.close() ;
            throw e ;
        } finally {
        }
    }
    
    /**
     * クローズ.
     * <BR><BR>
     * クローズ処理.
     */
    public synchronized void close() {
        iniParams = null ;
        fileName = null ;
        lastUpdate = -1L ;
        randomFlag.set( false ) ;
        key.set( null ) ;
    }
    
    /**
     * Seabass32暗号コードを取得.
     * @return String Seabass32暗号コードが返されます..
     */
    public String getSeabass32() {
        String ret = null ;
        if( updateByReload() ) {
            PublicCache.getInstance().clear() ;
        }
        if( ( ret = key.get() ) == null ) {
            if( iniParams.isKeys( BASE_SECTION,"key" ) ) {
                try {
                    ret = iniParams.get( BASE_SECTION,"key",0 ) ;
                } catch( Exception e ) {
                    ret = null ;
                }
            }
            else {
                ret = null ;
            }
            if( ret == null || ( ret = ret.trim() ).length() <= 0 ) {
                ret = RandomUtil.randomString( 32,true,true,true ) ;
                randomFlag.set( true ) ;
                PublicCache.getInstance().clear() ;
            }
            key.set( ret ) ;
        }
        return ret ;
    }
    
    /**
     * デフォルトキーを生成.
     * @return String 生成されたデフォルトキーが返されます.
     * @exception Exception 例外.
     */
    public String createDefaultKey() throws Exception {
        String key = Digest.convert( "SHA1",Seabass32Key.DEFAULT_CREATE_KEY_INFO.getBytes( "UTF8" ) ) ;
        if( defaultKey.get() == null ) {
        	defaultKey.set( key ) ;
        }
        return key ;
    }
    
    /**
     * 現在のSeabass32暗号コードがランダム条件から取得されたデータかチェック.
     * @return boolean [true]の場合、ランダム発行です.
     */
    public boolean isRandom() {
        return randomFlag.get() ;
    }
    
    /**
     * デフォルトキーを設定.
     * @param key デフォルトキーを設定します.
     */
    public void setDefaultKey( String key ) {
        defaultKey.set( key ) ;
    }
    
    /**
     * デフォルトキーを取得.
     * @return String デフォルトキーが返されます.
     */
    public String getDefaultKey() {
        return defaultKey.get() ;
    }
    
    /**
     * 再読み込み.
     */
    private final boolean reload()
        throws Exception {
        boolean ret = false ;
        BufferedReader buf = null ;
        if( FileUtil.isFileExists( fileName ) == true ) {
            try {
                buf = new BufferedReader(
                    new InputStreamReader(
                        new FileInputStream( fileName ),CHARSET ) ) ;
                Config iniParams = new Config() ;
                ReadIni.analisys( iniParams,buf ) ;
                buf.close() ;
                buf = null ;
                long last = FileUtil.getLastTime( fileName ) ;
                // 読み込みデータを設定.
                this.iniParams = iniParams ;
                this.lastUpdate = last ;
                ret = true ;
                key.set( null ) ;
                randomFlag.set( false ) ;
            } catch( Exception e ) {
                throw e ;
            } finally {
                if( buf != null ) {
                    try {
                        buf.close() ;
                    } catch( Exception e ) {
                    }
                }
            }
        }
        return ret ;
    }
    
    /**
     * ファイルが更新されている場合は、再読み込み.
     */
    private boolean updateByReload() {
        boolean ret = false ;
        try {
            long tm = FileUtil.getLastTime( fileName ) ;
            if( lastUpdate != tm ) {
                ret = this.reload() ;
            }
        } catch( Exception e ) {
            ret = false ;
        }
        return ret ;
    }
}
