package org.maachang.seabass.dbm;

import java.io.IOException;

import org.maachang.bdb.BQueue;
import org.maachang.bdb.BdbKeyValue;
import org.maachang.bdb.BdbManager;
import org.maachang.bdb.TransactionQueue;

/**
 * BQueueファクトリ.
 * 
 * @version 2010/01/21
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
public abstract class BQueueFactory {
    private BQueueFactory(){}
    
    /**
     * BdbKeyValueオブジェクトを取得.
     * <p>※このオブジェクトは、ThreadLocalから取得されているので、Threadセーフです.</p>
     * @return BdbKeyValue BdbKeyValueオブジェクトが返されます.
     */
    public static final BdbKeyValue getBdbKeyValue() {
        return BdbManager.getBdbKeyValue() ;
    }
    
    /**
     * 指定されたBQueueオブジェクトを取得.
     * @param name 対象のBQueue名を設定します.
     * @return BQueue 取得されたBQueueオブジェクトが返されます.
     * @exception Exception 例外.
     */
    public static final BQueue get( String name )
        throws Exception {
        if( name == null || ( name = name.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        BdbService s = BdbService.getInstance() ;
        if( !s.isUse() ) {
            throw new IOException( "BDBサービスが起動していません" ) ;
        }
        return s.get().getQueue( name ) ;
    }
    
    /**
     * 指定されたBQueueオブジェクトを取得.
     * <p>情報が存在しない場合のみ、max,warningパラメータは有効になります.
     * 既に存在する場合は、その情報が返されます.</p>
     * @param name 対象のBQueue名を設定します.
     * @param max キュー格納最大数を設定します.
     * @param warning 警告値を設定します.<br>たとえば、75％の場合は、0.75dで設定します.
     * @return BQueue 取得されたBQueueオブジェクトが返されます.
     * @exception Exception 例外.
     */
    public static final BQueue get( String name,long max,double warning )
        throws Exception {
        if( name == null || ( name = name.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        BdbService s = BdbService.getInstance() ;
        if( !s.isUse() ) {
            throw new IOException( "BDBサービスが起動していません" ) ;
        }
        return s.get().getQueue( name,max,warning ) ;
    }
    
    /**
     * 指定されたTransactionQueueオブジェクトを取得.
     * @param keyValue 対象のKeyValueオブジェクトを設定します.
     * @param name 対象のBQueue名を設定します.
     * @return TransactionQueue 取得されたTransactionQueueオブジェクトが返されます.
     * @exception Exception 例外.
     */
    public static final TransactionQueue getTransaction( BdbKeyValue keyValue,String name )
        throws Exception {
        return new TransactionQueue( keyValue,BdbFactory.get( name ) ) ;
    }
    
    /**
     * 指定されたTransactionQueueオブジェクトを取得.
     * <p>情報が存在しない場合のみ、max,warningパラメータは有効になります.
     * 既に存在する場合は、その情報が返されます.</p>
     * @param keyValue 対象のKeyValueオブジェクトを設定します.
     * @param name 対象のBQueue名を設定します.
     * @param max キュー格納最大数を設定します.
     * @param warning 警告値を設定します.<br>たとえば、75％の場合は、0.75dで設定します.
     * @return TransactionQueue 取得されたTransactionQueueオブジェクトが返されます.
     * @exception Exception 例外.
     */
    public static final TransactionQueue getTransaction( BdbKeyValue keyValue,String name,long max,double warning )
        throws Exception {
        return new TransactionQueue( keyValue,BdbFactory.get( name ),max,warning ) ;
    }
    
    /**
     * Bdbサービス状態を取得.
     * @return boolean [true]の場合、サービスは実行中です.
     */
    public static final boolean isUse() {
        return BdbService.getInstance().isUse() ;
    }
}
