package org.maachang.seabass.httpd.engine ;

import java.io.IOException;
import java.io.OutputStream;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.conf.ConvIniParam;
import org.maachang.dao.OptimisticLockException;
import org.maachang.manager.GlobalManager;
import org.maachang.seabass.SeaBassDef;
import org.maachang.seabass.ServiceDef;
import org.maachang.seabass.State;
import org.maachang.seabass.conf.HttpdAccessConfig;
import org.maachang.seabass.conf.IniFile;
import org.maachang.seabass.conf.Seabass32Key;
import org.maachang.seabass.httpd.HttpdErrorDef;
import org.maachang.seabass.httpd.HttpdExecutionManager;
import org.maachang.seabass.httpd.HttpdHeaders;
import org.maachang.seabass.httpd.HttpdParams;
import org.maachang.seabass.httpd.HttpdRequest;
import org.maachang.seabass.httpd.HttpdResponse;
import org.maachang.seabass.httpd.HttpdStateException;
import org.maachang.seabass.httpd.engine.auth.HttpdAuthBasic;
import org.maachang.seabass.httpd.engine.auth.HttpdAuthDigest;
import org.maachang.seabass.httpd.engine.auth.HttpdAuthElement;
import org.maachang.seabass.httpd.engine.auth.HttpdAuthManager;
import org.maachang.seabass.httpd.engine.auth.HttpdAuthUsers;
import org.maachang.seabass.httpd.engine.filter.HttpdIOFilterManager;
import org.maachang.seabass.httpd.engine.script.BaseScriptException;
import org.maachang.seabass.httpd.engine.script.scripts.ValidateScript;
import org.maachang.seabass.httpd.engine.script.validate.ServerValidateCache;
import org.maachang.seabass.net.HttpReceiveCallback;
import org.maachang.seabass.net.HttpdHeadersImpl;
import org.maachang.seabass.net.HttpdRequestImpl;
import org.maachang.seabass.net.NioNotConnectException;
import org.maachang.seabass.net.nio.ConnectionInfo;
import org.maachang.util.Digest;
import org.maachang.util.FileUtil;
import org.maachang.util.StringUtil;
import org.maachang.util.URLDecode;
import org.maachang.util.atomic.AtomicOBJECT;

/**
 * HTTPD受信実行.
 * 
 * @version 2007/08/29
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
class HttpdExecutionReceive implements HttpReceiveCallback {
    
    /**
     * LOG.
     */
    private static final Log LOG = LogFactory.getLog( HttpdExecutionReceive.class ) ;
    
    /**
     * アクセスコンフィグ.
     */
    private HttpdAccessConfig accessControll = null ;
    
    /**
     * IOFilter管理.
     */
    private final AtomicOBJECT<HttpdIOFilterManager> filterMan = new AtomicOBJECT<HttpdIOFilterManager>() ;
    
    /**
     * コンストラクタ.
     */
    public HttpdExecutionReceive() throws Exception {
        this.accessControll = new HttpdAccessConfig() ;
        try {
            this.accessControll.open() ;
        } catch( Exception e ) {
            LOG.error( "アクセスコンフィグオープンエラー",e ) ;
            throw e ;
        }
    }
    
    /**
     * 実行処理.
     * <BR><BR>
     * 受信実行処理を行います.
     * <BR>
     * @param binary 一時的に利用するバイナリを設定します.
     * @param conn コネクションオブジェクトを設定します.
     * @param seqId シーケンスIDを設定します.
     * @param request 対象のリクエスト情報を設定します.
     * @param response 対象のレスポンス情報を設定します.
     * @exception Exception 例外.
     */
    public void execution( final byte[] binary,final ConnectionInfo conn,final int seqId,final HttpdRequest request,final HttpdResponse response )
        throws Exception {
        if( conn == null || conn.isUse() == false ) {
            return ;
        }
        try {
            // Seabass32暗号変換.
            convertSeabass32( request ) ;
            // URLを取得.
            String url = request.getUrlPath() ;
            // 指定URLがディレクトリ指定である場合の処理.
            requestURLByWindowDirector( request ) ;
            
            // URLとリモートアドレスを取得.
            ( ( HttpdRequestImpl )request ).setRemote( conn.getInetAddress().getHostAddress(),conn.getPort() ) ;
            
            // アクセス条件に対するログ出力.
            if( LOG.isInfoEnabled() ) {
                LOG.info( new StringBuilder().append( "id:" ).append( seqId ).append( "[" ).
                    append( request.getRemoteAddress() ).append( ":" ).append( request.getRemotePort() ).
                    append( "]-" ).
                    append( "(" ).append( request.getMethod() ).
                    append( ")-url:" ).append( url ).
                    append( " [" ).append( request.getHeader().getHeader( HttpdDef.VALUE_CLIENT_NAME ) ).append( "]" ).
                    toString() ) ;
            }
            // GZIP許可フラグをリクエストに設定.
            putGzip( request ) ;
            // URL変換.
            if( (url = pushUrl( seqId,request,response,url )) == null ) {
                // nullの場合は、リダイレクト.
                return ;
            }
            // 不正なURLが指定された場合.
            if( HttpdDef.isSeePath( url ) == false ) {
                throw new HttpdStateException( HttpdErrorDef.HTTP11_403,"指定URLは不正です" ) ;
            }
            // アクセス制御.
            else if( accessControll.isMask( url,request.getRemoteAddress() ) == false ) {
                LOG.warn( new StringBuilder().append( "id:" ).append( seqId ).
                    append( "アクセス違反を検知" ).toString() ) ;
                // アクセス制御エラーの場合.
                throw new HttpdStateException( HttpdErrorDef.HTTP11_403,"許可されていないURLです" ) ;
            }
            // 正常処理.
            else {
                // 現在のステータスを取得.
                int nowState = State.getInstance().state() ;
                // ステータスが、正常以外の場合.
                if( nowState != State.SUCCESS ) {
                    // 計画停止の場合は、メンテナンス中にする.
                    if( nowState == State.PLAN_STOP ) {
                        if( !(url.startsWith( "/~seabass/" ) ||
                            url.startsWith( "/img/" ) ||
                            url.startsWith( "/js/" ) ||
                            url.startsWith( "/css/" ) ||
                            url.endsWith( ".js" ) ||
                            url.endsWith( ".css" ) ) ) {
                            throw new HttpdStateException( HttpdErrorDef.HTTP11_503,"ただいま、メンテナンス中です" ) ;
                        }
                    }
                    // 計画停止以外の場合は、何も返さない.
                    else {
                        return ;
                    }
                }
                // 特殊Javascript呼び出しの場合.
                if( url.startsWith( "/@js/" ) ) {
                    // validate用Javascriptの場合.
                    if( url.startsWith( "/@js/validate/" ) && url.endsWith( ".js" ) ) {
                        ServerValidateCache validateCache = ( ServerValidateCache )GlobalManager.getInstance().get(
                            ServiceDef.VALIDATE_CACHE_MANAGER ) ;
                        ValidateScript script = validateCache.get( url.substring( "/@js/validate".length(),url.length()-".js".length() ) ) ;
                        if( script == null ) {
                            throw new HttpdStateException( HttpdErrorDef.HTTP11_404,"指定ValidateJavascriptは存在しません" ) ;
                        }
                        // Validate用Javascriptデータ送信処理.
                        response.setHttpCache( false ) ;
                        sendTo( script.toJavascript(),script.getCreateTime(),"text/javascript",request,response ) ;
                        return ;
                    }
                    else if( url.equals( "/@js/ref" ) ) {
                        String s = request.getQuery().getParam( "ref" ) ;
                        String m = request.getQuery().getParam( "mime" ) ;
                        if( s == null ) {
                            throw new HttpdStateException( HttpdErrorDef.HTTP11_403,"パラメータ[ref]が存在しません" ) ;
                        }
                        else {
                            if( m == null || m.length() <= 0 ) {
                                m = "application/octet-stream" ;
                            }
                            if( ( s = s.trim() ).length() <= 0 ) {
                                sendZero( request,response,m ) ;
                            }
                            else {
                                response.setHttpCache( false ) ;
                                sendTo( s.getBytes( "utf8" ),System.currentTimeMillis(),m,request,response ) ;
                            }
                        }
                        return ;
                    }
                }
                // アクセス制御正常処理.
                String method = request.getMethod() ;
                // POST及びGETの場合.
                if( HttpdDef.METHOD_GET.equals( method ) || HttpdDef.METHOD_POST.equals( method ) ) {
                    // IO-FilterManagerを取得.
                    HttpdIOFilterManager fman ;
                    if( ( fman = filterMan.get() ) == null ) {
                        fman = ( HttpdIOFilterManager )GlobalManager.getInstance().get( ServiceDef.IO_FILTER_MANAGER ) ;
                        filterMan.set( fman ) ;
                    }
                    if( fman.size() <= 0 ) {
                        // コンテンツ実行.
                        executionTargetURL( binary,method,request,response,seqId ) ;
                    }
                    else {
                        HttpdResponse dummy = response ;
                        // コンテンツ実行前処理.
                        dummy = fman.executionBeforeFilter( request,dummy ) ;
                        
                        // コンテンツ実行.
                        executionTargetURL( binary,method,request,response,seqId ) ;
                        
                        // コンテンツ実行後処理.
                        if( dummy == response ) {
                            dummy = null ;
                        }
                        fman.executionEndFilter( request,response,dummy ) ;
                    }
                }
                // それ以外のメソッド.
                else {
                    execNotPostAndGet( method,request,response ) ;
                }
            }
        } catch( Exception e ) {
            outputBaseError( request,response,seqId,e ) ;
        } catch( Error er ) {
            outputBaseError( request,response,seqId,er ) ;
        }
    }
    
    /**
     * POST.GET以外の処理.
     */
    private void execNotPostAndGet( final String method,final HttpdRequest request,final HttpdResponse response )
        throws Exception {
        if( HttpdDef.METHOD_HEAD.equals( method ) ) {
            executionHead( request,response ) ;
        }
        else if( HttpdDef.METHOD_OPTIONS.equals( method ) ) {
            executionOptions( request,response ) ;
        }
        else {
            executionNoSupprtMethd( request,response ) ;
        }
    }
    
    /**
     * Head処理.
     */
    private void executionHead( final HttpdRequest request,final HttpdResponse response )
        throws Exception {
        int state = -1 ;
        String authHeader = null ;
        HttpdAuthManager auth = ( HttpdAuthManager )GlobalManager.getValue(
            ServiceDef.MANAGER_BY_AUTH ) ;
        HttpdAuthElement emt = null ;
        // 認証条件が存在する場合.
        if( auth != null && ( emt = auth.getElement( request.getUrlPath() ) ) != null ) {
            HttpdAuthUsers users = ( HttpdAuthUsers )GlobalManager.getValue(
                ServiceDef.MANAGER_BY_USERS ) ;
            // Basic認証の場合.
            if( emt.getAuthType() == HttpdAuthElement.AUTH_BY_BASIC ) {
                // 既にBasic認証内容が存在し、間違った結果の場合.
                if( HttpdAuthBasic.isAuth( request,emt,users,request.getHeader() ) == false ) {
                    authHeader = HttpdAuthBasic.getSendAuth( emt ) ;
                    state = HttpdErrorDef.HTTP11_401 ;
                }
            }
            // Digest認証の場合.
            else if( emt.getAuthType() == HttpdAuthElement.AUTH_BY_DIGEST ) {
                // 既にDigest認証内容が存在し、間違った結果の場合.
                if( HttpdAuthDigest.isAuth( request,emt,users,request.getHeader() ) == false ) {
                    authHeader = HttpdAuthDigest.getSendAuth( emt ) ;
                    state = HttpdErrorDef.HTTP11_401 ;
                }
            }
        }
        long last = -1L ;
        if( state == -1 ) {
            boolean pub = false ;
            boolean sc = false ;
            HttpdExecutionManager pubMan = ( HttpdExecutionManager )GlobalManager.getValue(
                ServiceDef.MANAGER_BY_PUBLIC_PAGE ) ;
            HttpdExecutionManager scMan = ( HttpdExecutionManager )GlobalManager.getValue(
                ServiceDef.MANAGER_BY_APPLICATON_PAGE ) ;
            // publicが存在するかチェック.
            last = pubMan.getFileTime( request.getUrlPath() ) ;
            if( last == -1L ) {
                // applicationが存在するかチェック.
                last = scMan.getFileTime( request.getUrlPath() ) ;
                if( last != -1L ) {
                    sc = true ;
                }
            }
            else {
                pub = true ;
            }
            // 指定URLの内容が存在しない場合.
            if( last == -1L ) {
                state = HttpdErrorDef.HTTP11_404 ;
            }
            // 読み込み可能かチェック.
            if( ( pub == true && pubMan.isRead( request.getUrlPath() ) == false ) ||
                ( sc == true && scMan.isRead( request.getUrlPath() ) == false ) ) {
                state = HttpdErrorDef.HTTP11_403 ;
            }
            else {
                state = HttpdErrorDef.HTTP11_200 ;
                // キャッシュ範囲内の場合.
                if( HttpdUtil.isClientCache( request.getHeader(),last ) == true ) {
                    state = HttpdErrorDef.HTTP11_304 ;
                }
            }
        }
        // ヘッダ送信.
        HttpdHeaders header = null ;
        if( authHeader != null ) {
            header = new HttpdHeadersImpl() ;
            header.addHeader( HttpdDef.VALUE_START_AUTH,authHeader ) ;
        }
        returnHeader( request,response,state,last,-1L,-1,header ) ;
    }
    
    /**
     * Options処理.
     */
    private void executionOptions( final HttpdRequest request,final HttpdResponse response )
        throws Exception {
        HttpdResponse res = response.create( request,request.getUrlPath(),
            request.getKeepAliveTimeout(),request.getKeepAliveCount() ) ;
        if( res.flushAndClose() == false ) {
            throw new IOException( "レスポンス返信処理に失敗しました" ) ;
        }
    }
    
    /**
     * サポート外のmethod処理.
     */
    private void executionNoSupprtMethd( final HttpdRequest request,final HttpdResponse response )
        throws Exception {
        WriteError.output( request,response,null,HttpdErrorDef.HTTP11_501,null ) ;
    }
    
    /**
     * ターゲットURL内容を処理.
     */
    private void executionTargetURL( final byte[] binary,final String method,final HttpdRequest request,final HttpdResponse response,final int seqId )
        throws Exception {
        boolean errorFlag = false ;
        try {
            HttpdExecutionManager pubMan = ( HttpdExecutionManager )GlobalManager.getValue(
                ServiceDef.MANAGER_BY_PUBLIC_PAGE ) ;
            HttpdExecutionManager scMan = ( HttpdExecutionManager )GlobalManager.getValue(
                ServiceDef.MANAGER_BY_APPLICATON_PAGE ) ;
            
            HttpdExecutionManager man = null ;
            
            long last = -1L ;
            // 拡張子がSeaBassスクリプト実行用拡張子の場合.
            if( HttpdDef.isApplication( request ) ) {
                last = scMan.checkToLastUpdateTime( request ) ;
                if( last != -1L ) {
                    man = scMan ;
                }
            }
            else {
                // 指定パスを取得.
                last = pubMan.checkToLastUpdateTime( request ) ;
                if( last == -1L ) {
                    last = scMan.checkToLastUpdateTime( request ) ;
                    if( last != -1L ) {
                        man = scMan ;
                    }
                }
                else {
                    man = pubMan ;
                }
            }
            
            if( man == null ) {
                throw new HttpdStateException( HttpdErrorDef.HTTP11_404,
                    "指定パス[" + request.getUrlPath() + "]の内容は存在しません" ) ;
            }
            
            //////////////////
            // レスポンス処理.
            //////////////////
            
            // リクエストでキャッシュヘッダが付加されている場合.
            if( HttpdUtil.isClientCache( request.getHeader(),last ) == true ) {
                returnHeader( request,response,HttpdErrorDef.HTTP11_304,
                    last,request.getKeepAliveTimeout(),request.getKeepAliveCount(),null ) ;
            }
            // レスポンスが送信可能な場合.
            else {
                man.get( binary,request,response ) ;
            }
        } catch( Exception e ) {
            errorFlag = true ;
            try {
                errorScript( request,response,seqId,e ) ;
            } catch( Exception ee ) {
                LOG.error( "output-error",ee ) ;
            }
        } finally {
            // コメット処理でない場合.
            if( !request.getConnectionInfo().isComet() ) {
                // レスポンスを更新/破棄.
                response.flushAndClose() ;
            }
            else {
                if( LOG.isDebugEnabled() ) {
                    LOG.debug( "## comet処理" ) ;
                }
            }
        }
    }
    
    /**
     * スクリプトエラー処理.
     */
    private void errorScript( final HttpdRequest request,final HttpdResponse response,final int seqId,Throwable e )
        throws Exception {
        if( e instanceof BaseScriptException ) {
            BaseScriptException se = ( BaseScriptException )e ;
            // スクリプト処理エラー.
            LOG.error( new StringBuilder().append( "## error(id:" ).
                append( seqId ).
                append( ")-(url:" ).
                append( request.getUrlPath() ).
                append( ")" ).
                toString(),se ) ;
            StringBuilder buf = new StringBuilder() ;
            buf.append( se.getMessage() ).append( "<BR><BR>" ) ;
            if( se.getErrorScriptInfo() != null ) {
                buf.append( se.getErrorScriptInfo() ) ;
            }
            //buf.append( "&nbsp;行:" ).append( se.getLine() ).append( "<BR>" ) ;
            //buf.append( "&nbsp;列:" ).append( se.getColumnNumber() ).append( "<BR>" ) ;
            //buf.append( "&nbsp;file:" ).append( se.getFileName() ) ;
            String msg = buf.toString() ;
            buf = null ;
            // HTTPエラーページを渡す.
            try {
                WriteError.output( request,response,null,HttpdErrorDef.HTTP11_500,msg ) ;
            } catch( Exception ee ) {
                LOG.error( "output-error",ee ) ;
                throw ee ;
            }
        }
        else {
            outputBaseError( request,response,seqId,e ) ;
        }
    }
    
    /**
     * HTTPエラー／その他エラーの出力処理.
     */
    private void outputBaseError( final HttpdRequest request,final HttpdResponse response,final int seqId,Throwable e )
        throws Exception {
        if( e instanceof NioNotConnectException ) {
            return ;
        }
        if( e instanceof HttpdStateException ) {
            HttpdStateException hs = ( HttpdStateException )e ;
            // 処理エラー(400系).
            if( hs.getState() <= 499 ) {
                if( request != null ) {
                    LOG.warn( new StringBuilder().append( "## url:(id:" ).
                        append( seqId ).
                        append( "):" ).
                        append( request.getUrlPath() ).
                        append( " - status(" ).
                        append( hs.getState() ).
                        append( ") warning[" ).
                        append( hs.getErrorMessage() ).
                        append( "]" ).
                        toString() ) ;
                }
            }
            // 処理エラー(500系-その他).
            else {
                if( request != null ) {
                    LOG.error( new StringBuilder().append( "## url:(id:" ).
                        append( seqId ).
                        append( "):" ).
                        append( request.getUrlPath() ).
                        append( " - status(" ).
                        append( hs.getState() ).
                        append( ") error" ).
                        toString(),e ) ;
                }
            }
            // HTTPエラーページを渡す.
            try {
                WriteError.output( request,response,null,hs.getState(),hs.getErrorMessage() ) ;
            } catch( Exception ee ) {
                LOG.error( "output-error",ee ) ;
                throw ee ;
            }
        }
        // 楽観的エラーが発生した場合.
        else if( e instanceof OptimisticLockException ) {
            // 処理エラー.
            LOG.error( "## error(id:" + seqId + "):",e ) ;
            // リソース衝突ステータスを返す.
            try {
                WriteError.output( request,response,null,HttpdErrorDef.HTTP11_409,e.getMessage() ) ;
            } catch( Exception ee ) {
                LOG.error( "output-error",ee ) ;
                throw ee ;
            }
        }
        else {
            // 処理エラー.
            LOG.error( "## error(id:" + seqId + "):",e ) ;
            // その他エラー内容を渡す.
            try {
                WriteError.output( request,response,null,HttpdErrorDef.HTTP11_500,e.getMessage() ) ;
            } catch( Exception ee ) {
                LOG.error( "output-error",ee ) ;
                throw ee ;
            }
        }
    }
    
    /**
     * ヘッダ条件のみを送付.
     */
    private void returnHeader( final HttpdRequest request,final HttpdResponse response,
        int state,long date,long timeout,int count,final HttpdHeaders header )
        throws Exception {
        HttpdResponse res = response.create( request,request.getUrlPath(),state,
            request.getKeepAliveTimeout(),request.getKeepAliveCount() ) ;
        if( date == -1L ) {
            res.getHeader().setHeader( "Date",HttpdTimestamp.getTimestamp( 0L ) ) ;
        }
        else {
            res.getHeader().setHeader( "Date",HttpdTimestamp.getTimestamp( date ) ) ;
        }
        res.getHeader().setHeader( HttpdDef.VALUE_CONTENT_LENGTH,"0" ) ;
        res.setHttpClose( false ) ;
        if( header != null ) {
            res.getHeader().add( header ) ;
        }
        if( res.flushAndClose() == false ) {
            throw new IOException( "レスポンス返信処理に失敗しました" ) ;
        }
    }
    
    /**
     * 認証処理.
     */
    private boolean isAuth( final HttpdRequest request,final HttpdResponse response )
        throws Exception {
        String authHeader = null ;
        HttpdAuthManager auth = ( HttpdAuthManager )GlobalManager.getValue(
            ServiceDef.MANAGER_BY_AUTH ) ;
        HttpdAuthElement emt = null ;
        // 認証条件が存在する場合.
        if( auth != null && ( emt = auth.getElement( request.getUrlPath() ) ) != null ) {
            HttpdAuthUsers users = ( HttpdAuthUsers )GlobalManager.getValue(
                ServiceDef.MANAGER_BY_USERS ) ;
            // Basic認証の場合.
            if( emt.getAuthType() == HttpdAuthElement.AUTH_BY_BASIC ) {
                // 既にBasic認証内容が存在し、間違った結果の場合.
                if( HttpdAuthBasic.isAuth( request,emt,users,request.getHeader() ) == false ) {
                    authHeader = HttpdAuthBasic.getSendAuth( emt ) ;
                }
            }
            // Digest認証の場合.
            else if( emt.getAuthType() == HttpdAuthElement.AUTH_BY_DIGEST ) {
                // 既にDigest認証内容が存在し、間違った結果の場合.
                if( HttpdAuthDigest.isAuth( request,emt,users,request.getHeader() ) == false ) {
                    authHeader = HttpdAuthDigest.getSendAuth( emt ) ;
                }
            }
        }
        // 認証条件が存在する場合.
        if( authHeader != null  ) {
            HttpdHeaders addHeader = new HttpdHeadersImpl() ;
            addHeader.addHeader( HttpdDef.VALUE_START_AUTH,authHeader ) ;
            WriteError.output( request,response,addHeader,HttpdErrorDef.HTTP11_401,null ) ;
            return true ;
        }
        else {
            return false ;
        }
    }
    
    /**
     * GZIP許可フラグを設定.
     */
    private static final void putGzip( final HttpdRequest req ) {
        IniFile config = ( IniFile )GlobalManager.getValue( ServiceDef.MANAGER_BY_CONFIG ) ;
        boolean flag = ConvIniParam.getBoolean( config.get( "server","gzip",0 ) ) ;
        req.setGzip( flag ) ;
    }
    
    /**
     * リクエストURLがディレクトリ指定である場合の付加処理.
     */
    private void requestURLByWindowDirector( final HttpdRequest request ) throws Exception {
        String path = request.getUrlPath() ;
        if( path.endsWith( "/" ) ) {
            return ;
        }
        HttpdExecutionManager man ;
        man = ( HttpdExecutionManager )GlobalManager.getValue(
            ServiceDef.MANAGER_BY_APPLICATON_PAGE ) ;
        if( man.isDirectory( path ) ) {
            request.setUrlPath( path+"/" ) ;
            return ;
        }
        man = ( HttpdExecutionManager )GlobalManager.getValue(
            ServiceDef.MANAGER_BY_PUBLIC_PAGE ) ;
        if( man.isDirectory( path ) ) {
            request.setUrlPath( path+"/" ) ;
            return ;
        }
    }
    
    /**
     * データ送信処理.
     */
    private void sendTo( byte[] data,long time,String mimeType,final HttpdRequest request,final HttpdResponse response )
        throws Exception {
        if( data != null ) {
            response.create( request,request.getUrlPath(),
                request.getKeepAliveTimeout(),request.getKeepAliveCount() ) ;
            response.setCookieSession( request ) ;
            response.getHeader().setHeader( HttpdDef.VALUE_LAST_UPDATE,
                HttpdTimestamp.getTimestamp( time ) ) ;
            response.setHttpClose( false ) ;
            if( mimeType != null && mimeType.length() > 0 ) {
                response.getHeader().setHeader( HttpdDef.VALUE_CONTENT_TYPE,mimeType ) ;
            }
            if( !response.getGzip() || data.length < HttpdPublicManager.NOT_GZIP_LENGTH ) {
                response.setGzip( false ) ;
                response.getHeader().setHeader( HttpdDef.VALUE_CONTENT_LENGTH,String.valueOf( data.length ) ) ;
            }
            OutputStream out = null ;
            try {
                out = response.getOutput() ;
                out.write( data ) ;
                out.flush() ;
                out.close() ;
                out = null ;
            } finally {
                if( out != null ) {
                    try {
                        out.close() ;
                    } catch( Exception ee ) {
                    }
                }
            }
        }
    }
    
    /**
     * 対象パスを整形してURLとして設定.
     */
    private final String pushUrl( final int seqId,final HttpdRequest request,final HttpdResponse response,String path )
        throws Exception {
        path = URLDecode.convert( path ) ;
        path = path.substring( 1 ) ;
        // URL内容がコーテーションで囲まれている.
        if( ( path.startsWith( "\"" ) && path.endsWith( "\"" ) ) ||
            path.startsWith( "\'" ) && path.endsWith( "\'" ) ) {
            path = path.substring( 1,path.length()-1 ) ;
        }
        // URL内容に\\コードが存在する.
        if( path.indexOf( "\\" ) != -1 ) {
            path = StringUtil.changeString( path,"\\","/" ) ;
        }
        String fPath = path ;
        if( !path.startsWith( "/" ) ) {
            fPath = "/" + path ;
        }
        // fullPathチェックを行い、一致する場合は、フルパス内容を解析して正しいURLにリダイレクト.
        if( fPath.startsWith( SeaBassDef.NOW_BASE_PATH ) ) {
            if( fPath.startsWith( SeaBassDef.NOW_APPLICATION_PATH ) ) {
                fPath = fPath.substring( SeaBassDef.NOW_APPLICATION_PATH.length()-1 ) ;
            }
            else if( fPath.startsWith( SeaBassDef.NOW_PUBLIC_PATH ) ) {
                fPath = fPath.substring( SeaBassDef.NOW_PUBLIC_PATH.length()-1 ) ;
            }
            else {
                // 不正なURL.
                throw new HttpdStateException( HttpdErrorDef.HTTP11_403,"指定URLは不正です" ) ;
            }
            // リダイレクト処理.
            redirect( seqId,request,response,fPath ) ;
            return null ;
        }
        fPath = null ;
        String base = path ;
        path = FileUtil.getFullPath( path ) ;
        // URLが一致しない場合.
        if( !path.startsWith( SeaBassDef.NOW_FULL_PATH ) ) {
            // 不正なURL.
            throw new HttpdStateException( HttpdErrorDef.HTTP11_403,"指定URLは不正です" ) ;
        }
        // ROOTパスの場合.
        if( path.equals( SeaBassDef.NOW_FULL_PATH ) ) {
            request.setUrlPath( "/" ) ;
            return "/" ;
        }
        boolean dirFlag = ( base.endsWith( "/" ) ) ;
        path = path.substring( SeaBassDef.NOW_FULL_PATH.length(),path.length() ) ;
        // ディレクトリ条件の場合.
        if( dirFlag ) {
            path += "/" ;
            ((HttpdRequestImpl)request).toDirectory() ;
        }
        // ファイル条件の場合、その条件がディレクトリ条件であるかチェックして、
        // ディレクトリ条件の場合は、リダイレクトする.
        else if( FileUtil.isDirExists( SeaBassDef.NOW_PUBLIC_PATH + base ) ||
            FileUtil.isDirExists( SeaBassDef.NOW_APPLICATION_PATH + base ) ) {
            path += "/" ;
            redirect( seqId,request,response,path ) ;
            return null ;
        }
        request.setUrlPath( path ) ;
        return path ;
    }
    
    /**
     * リダイレクト処理.
     */
    private final void redirect( final int seqId,final HttpdRequest request,final HttpdResponse response,String path )
        throws Exception {
        if( LOG.isDebugEnabled() ) {
            LOG.debug( "## id:" + seqId + " url:" + request.getUrlPath() + " message: パス " + path + " にリダイレクトします" ) ;
        }
        // リダイレクト処理を行う.
        HttpdHeadersImpl header = new HttpdHeadersImpl() ;
        header.addHeader( HttpdDef.VALUE_LOCATION,path ) ;
        returnHeader( request,response,HttpdErrorDef.HTTP11_303,-1,-1L,-1,header ) ;
    }
    
    /**
     * ゼロバイトのデータを返信.
     */
    private final void sendZero( final HttpdRequest request,final HttpdResponse response,String mime )
        throws Exception {
        HttpdHeadersImpl header = new HttpdHeadersImpl() ;
        if( mime == null || mime.length() <= 0 ) {
            mime = "application/octet-stream" ;
        }
        header.setHeader( HttpdDef.VALUE_CONTENT_TYPE,mime ) ;
        returnHeader( request,response,HttpdErrorDef.HTTP11_200,-1,-1L,-1,header ) ;
    }
    
    /**
     * Seabass32暗号で、暗号化されたパラメータを解析.
     */
    private static final void convertSeabass32( final HttpdRequest req )
        throws Exception {
        HttpdParams p = req.getQuery() ;
        // Seabass32暗号パラメータの内容が存在する場合.
        if( p.isKey( HttpdDef.SEABASS32_PARAM_NAME ) && p.size( HttpdDef.SEABASS32_PARAM_NAME ) == 1 &&
            p.isKey( HttpdDef.SEABASS32_SID ) && p.size( HttpdDef.SEABASS32_SID ) == 1 &&
            p.getParam( HttpdDef.SEABASS32_PARAM_NAME,0 ).startsWith( "@sb32" ) ) {
            String s = p.getParam( HttpdDef.SEABASS32_PARAM_NAME,0 ) ;
            //String key = req.getHeader().getHeader( "Host" ) ;
            String reqURL = req.getClientAccessURL() ;
            int pp = reqURL.indexOf( "?" ) ;
            if( pp >= 0 ) {
                reqURL = reqURL.substring( 0,pp ) ;
            }
            String secKey = Digest.convert( "MD5",reqURL.getBytes( "UTF8" ) ) +
                p.getParam( HttpdDef.SEABASS32_SID ) ;
            reqURL = null ;
            Seabass32Key seabass32Key = ( Seabass32Key )GlobalManager.getInstance().get( ServiceDef.MANAGER_BY_SEABASS32 ) ;
            String key ;
            // キー情報が直接指定されている場合.
            if( p.isKey( HttpdDef.SEABASS32_KEY ) ) {
                key = p.getParam( HttpdDef.SEABASS32_KEY,0 ) ;
                // キー情報がアンダーバー１文字の場合.
                if( "_".equals( key ) ) {
                    // デフォルト発行されたキーを取得.
                    if( (key = seabass32Key.getDefaultKey()) == null ) {
                        key = seabass32Key.createDefaultKey() ;
                    }
                }
                int h = key.length() / 2 ;
                key = Digest.convert( "MD5",( key.substring( h ) + key.substring( 0,h ) ).getBytes( "UTF8" ) ) ;
            }
            else {
                key = seabass32Key.getSeabass32() ;
            }
            (( HttpdRequestImpl )req).setSeabass32( true,key,secKey ) ;
            s = Seabass32.decToString( s,key,secKey ) ;
            HttpdUtil.convertQueryByParams( req.getQuery(),s,"UTF8" ) ;
            p.removeParam( HttpdDef.SEABASS32_PARAM_NAME ) ;
            p.removeParam( HttpdDef.SEABASS32_SID ) ;
        }
    }
}
