package org.maachang.seabass.httpd.engine;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.OutputStream;

import org.maachang.seabass.SeaBassDef;
import org.maachang.seabass.httpd.HttpdExecutionManager;
import org.maachang.seabass.httpd.HttpdMode;
import org.maachang.seabass.httpd.HttpdRequest;
import org.maachang.seabass.httpd.HttpdResponse;
import org.maachang.seabass.httpd.engine.script.scripts.IncludeScript;
import org.maachang.seabass.net.HttpdRequestImpl;
import org.maachang.util.BinaryUtil;
import org.maachang.util.FileUtil;

/**
 * Httpdパブリックディレクトリ管理.
 * 
 * @version 2007/08/29
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
public class HttpdPublicManager implements HttpdExecutionManager {
    
    /**
     * ZIPを行わないデータサイズ.
     */
    protected static final long NOT_GZIP_LENGTH = 1024 ;
    
    /**
     * Include対応拡張子キャラクタセット.
     */
    protected static final String INCLUDE_FILE_CHARSET = "UTF8" ;
    
    /**
     * Javascript特殊変換条件.
     */
    protected static final byte[] JS_SPECIAL_CODE = { (byte)'$',(byte)'@',(byte)'@' } ;
    
    /**
     * コンストラクタ.
     */
    public HttpdPublicManager() {
    }
    
    /**
     * 指定内容を取得.
     * <BR><BR>
     * 指定された内容を取得します.
     * <BR>
     * @param b 一時的に利用するバイナリを設定します.
     * @param request 対象のリクエストを設定します.
     * @param response 対象のレスポンスを設定します.
     * @exception Exception 例外.
     */
    public void get( final byte[] b,final HttpdRequest request,final HttpdResponse response )
        throws Exception {
        String path ;
        if( ( path = request.getFullPath() ) == null ) {
            path = getUsePath( request,null ) ;
        }
        // JSファイル、CSSファイル、ICOファイルは、キャッシュから取得.
        if( isJsOrCssOrIco( path ) ) {
            byte[] data = PublicCache.getInstance().getData( path ) ;
            if( data != null ) {
                long time = PublicCache.getInstance().getLastTime( path ) ;
                response.create( request,request.getUrlPath(),
                    request.getKeepAliveTimeout(),request.getKeepAliveCount() ) ;
                response.setCookieSession( request ) ;
                if( !HttpdMode.cacheMode ) {
                    response.getHeader().setHeader( HttpdDef.VALUE_LAST_UPDATE,
                        HttpdTimestamp.getTimestamp( time ) ) ;
                    response.setHttpCache( true ) ;
                }
                else {
                    response.setHttpCache( false ) ;
                }
                response.setHttpClose( false ) ;
                if( path.toLowerCase().endsWith( ".js" ) ) {
                    // 特殊変換条件が存在する場合は、処理する.
                    if( BinaryUtil.binaryIndexOf( data,JS_SPECIAL_CODE,0 ) != -1 ) {
                        String js = new String( data,"UTF8" ) ;
                        js = EnvJavascript.convert( request,js ) ;
                        data = js.getBytes( "UTF8" ) ;
                        js = null ;
                    }
                }
                // 圧縮フラグがOFF.
                // データサイズが一定以下の場合は圧縮しない.
                if( !response.getGzip() || data.length < HttpdPublicManager.NOT_GZIP_LENGTH ) {
                    response.setGzip( false ) ;
                    response.getHeader().setHeader( HttpdDef.VALUE_CONTENT_LENGTH,String.valueOf( data.length ) ) ;
                }
                OutputStream out = null ;
                try {
                    out = response.getOutput() ;
                    out.write( data ) ;
                    out.flush() ;
                    out.close() ;
                    out = null ;
                } finally {
                    if( out != null ) {
                        try {
                            out.close() ;
                        } catch( Exception ee ) {
                        }
                    }
                }
            }
        }
        // Include可能HTMLファイルの場合.
        else if( includeFileType( path ) != -1 ) {
            String val = IncludeScript.readScript( path,INCLUDE_FILE_CHARSET ) ;
            byte[] data = val.getBytes( "UTF8" ) ;
            val = null ;
            response.create( request,request.getUrlPath(),
                request.getKeepAliveTimeout(),request.getKeepAliveCount() ) ;
            response.setCookieSession( request ) ;
            if( !HttpdMode.cacheMode ) {
                response.getHeader().setHeader( HttpdDef.VALUE_LAST_UPDATE,
                    HttpdTimestamp.getTimestamp( FileUtil.getLastTime( path ) ) ) ;
                response.setHttpCache( true ) ;
            }
            else {
                response.setHttpCache( false ) ;
            }
            // 圧縮フラグがOFF.
            // データサイズが一定以下の場合は圧縮しない.
            if( !response.getGzip() || data.length < HttpdPublicManager.NOT_GZIP_LENGTH ) {
                response.setGzip( false ) ;
                response.getHeader().setHeader( HttpdDef.VALUE_CONTENT_LENGTH,String.valueOf( data.length ) ) ;
            }
            response.setHttpClose( false ) ;
            OutputStream out = null ;
            try {
                out = response.getOutput() ;
                out.write( data ) ;
                out.flush() ;
                out.close() ;
                out = null ;
            } finally {
                if( out != null ) {
                    try {
                        out.close() ;
                    } catch( Exception ee ) {
                    }
                }
            }
        }
        // 通常ファイル処理.
        else {
            response.create( request,request.getUrlPath(),
                request.getKeepAliveTimeout(),request.getKeepAliveCount() ) ;
            response.setCookieSession( request ) ;
            if( !HttpdMode.cacheMode ) {
                response.getHeader().setHeader( HttpdDef.VALUE_LAST_UPDATE,
                    HttpdTimestamp.getTimestamp( FileUtil.getLastTime( path ) ) ) ;
                response.setHttpCache( true ) ;
            }
            else {
                response.setHttpCache( false ) ;
            }
            long allLen = FileUtil.getLength( path ) ;
            // 圧縮フラグがOFFか、GZIP圧縮条件に満たないデータ長の場合.
            if( !response.getGzip() || allLen < (long)HttpdPublicManager.NOT_GZIP_LENGTH ) {
                response.setGzip( false ) ;
                response.getHeader().setHeader( HttpdDef.VALUE_CONTENT_LENGTH,String.valueOf( allLen ) ) ;
            }
            response.setHttpClose( false ) ;
            InputStream in = null ;
            OutputStream out = null ;
            try {
                in = new BufferedInputStream( new FileInputStream( path ) ) ;
                out = response.getOutput() ;
                int len ;
                while( true ) {
                    len = in.read( b ) ;
                    if( len <= 0 ) {
                        if( len <= -1 ) {
                            break ;
                        }
                    }
                    else {
                        out.write( b,0,len ) ;
                    }
                }
                out.flush() ;
                out.close() ;
                out = null ;
                in.close() ;
                in = null ;
            } finally {
                if( in != null ) {
                    try {
                        in.close() ;
                    } catch( Exception ee ) {
                    }
                }
                if( out != null ) {
                    try {
                        out.close() ;
                    } catch( Exception ee ) {
                    }
                }
            }
        }
    }
    
    /**
     * 実行可能かチェックを行い、実行可能な場合は、最終更新日を取得.
     * <BR><BR>
     * 実行可能かチェックを行い、実行可能な場合は、最終更新日を取得します.
     * <BR>
     * @param request 対象のリクエストを設定します.
     * @return long 最終更新日が返されます.
     * @exception Exception 例外.
     */
    public long checkToLastUpdateTime( HttpdRequest request ) throws Exception {
        long ret = -1L ;
        String path = getUsePath( request,null ) ;
        if( path != null ) {
            if( isJsOrCssOrIco( path ) ) {
                try {
                    ret = PublicCache.getInstance().getLastTime( path ) ;
                } catch( Exception e ) {
                }
            }
            else if( FileUtil.isFileExists( path ) ) {
                ret = FileUtil.getLastTime( path ) ;
            }
        }
        if( ret != -1L ) {
            ((HttpdRequestImpl)request).setFullPath( path ) ;
        }
        return ret ;
    }
    
    /**
     * 指定パスのファイル日付を取得.
     * <BR><BR>
     * 指定パスのファイル日付を取得します.
     * <BR>
     * @parma path 対象のパスを設定します.
     * @return long ファイル日付が返されます.
     */
    public long getFileTime( String path ) {
        path = getUsePath( null,path ) ;
        if( isJsOrCssOrIco( path ) ) {
            try {
                return PublicCache.getInstance().getLastTime( path ) ;
            } catch( Exception e ) {
            }
        }
        else if( FileUtil.isFileExists( path ) ) {
            return FileUtil.getLastTime( path ) ;
        }
        return -1L ;
    }
    
    /**
     * 指定パスの内容が存在するかチェック.
     * <BR><BR>
     * 指定パスの内容が存在するかチェックします.
     * <BR>
     * @param path 対象のパスを設定します.
     * @return boolean [true]の場合、存在します.
     */
    public boolean isPath( String path ) {
        path = getUsePath( null,path ) ;
        if( isJsOrCssOrIco( path ) ) {
            try {
                return ( PublicCache.getInstance().getLastTime( path ) > 0L ) ;
            } catch( Exception e ) {
            }
        }
        else if( FileUtil.isFileExists( path ) ) {
            return true ;
        }
        return false ;
    }
    
    /**
     * 指定パスのディレクトリが存在するかチェック.
     * <BR><BR>
     * 指定パスのディレクトリが存在するかチェックします.
     * <BR>
     * @param path 対象のパスを設定します.
     * @return boolean [true]の場合、ディレクトリとして存在します.
     */
    public boolean isDirectory( String path ) {
        if( path.startsWith( "/" ) == true ) {
            path = path.substring( 1,path.length() ) ;
        }
        if( FileUtil.isDirExists( SeaBassDef.NOW_PUBLIC_PATH + path ) == true ) {
            return true ;
        }
        return false ;
    }
    
    /**
     * 指定パスの内容が読み込み可能かチェック.
     * <BR><BR>
     * 指定パスの内容が読み込み可能かチェックします.
     * <BR>
     * @param path 対象のパスを設定します.
     * @return boolean [true]の場合、読み込み可能です.
     */
    public boolean isRead( String path ) {
        path = getUsePath( null,path ) ;
        if( isJsOrCssOrIco( path ) ) {
            boolean ret = false ;
            try {
                ret = ( PublicCache.getInstance().getLastTime( path ) > 0L ) ;
            } catch( Exception e ) {
                ret = false ;
            }
            return ret ;
        }
        return FileUtil.isRead( path ) ;
    }
    
    /**
     * 有効パスを取得.
     */
    public static final String getUsePath( HttpdRequest request,String path ) {
        if( path == null ) {
            path = request.getUrlPath() ;
        }
        String lower = path.toLowerCase() ;
        if( lower.endsWith( HttpdDef.ADDRESS_BAR_ICON ) ) {
            String ret = SeaBassDef.NOW_PUBLIC_PATH+path ;
            if( FileUtil.isFileExists( ret ) ) {
                return ret ;
            }
            ret = SeaBassDef.NOW_APPLICATION_PATH+path ;
            if( FileUtil.isFileExists( ret ) ) {
                return ret ;
            }
        }
        if( lower.equals( HttpdDef.ADDRESS_BAR_ICON ) ) {
            String ret = new StringBuilder().
                append( SeaBassDef.NOW_PUBLIC_PATH ).append( SeaBassDef.DIRECTORY_IMAGE ).
                append( HttpdDef.ADDRESS_BAR_ICON.substring( 1,HttpdDef.ADDRESS_BAR_ICON.length() ) ).
                toString() ;
            if( FileUtil.isFileExists( ret ) ) {
                return ret ;
            }
            ret = new StringBuilder().
                append( SeaBassDef.NOW_PUBLIC_PATH ).
                append( HttpdDef.ADDRESS_BAR_ICON.substring( 1,HttpdDef.ADDRESS_BAR_ICON.length() ) ).
                toString() ;
            if( FileUtil.isFileExists( ret ) ) {
                return ret ;
            }
            ret = new StringBuilder().
                append( SeaBassDef.NOW_APPLICATION_PATH ).append( SeaBassDef.DIRECTORY_IMAGE ).
                append( HttpdDef.ADDRESS_BAR_ICON.substring( 1,HttpdDef.ADDRESS_BAR_ICON.length() ) ).
                toString() ;
            if( FileUtil.isFileExists( ret ) ) {
                return ret ;
            }
            ret = new StringBuilder().
                append( SeaBassDef.NOW_APPLICATION_PATH ).
                append( HttpdDef.ADDRESS_BAR_ICON.substring( 1,HttpdDef.ADDRESS_BAR_ICON.length() ) ).
                toString() ;
            if( FileUtil.isFileExists( ret ) ) {
                return ret ;
            }
        }
        else if( lower.endsWith( ".html" ) || lower.endsWith( ".htm" ) || includeFileType( lower ) != -1 ) {
            if( path.startsWith( "/" ) ) {
                path = path.substring( 1,path.length() ) ;
            }
            String ph = SeaBassDef.NOW_PUBLIC_PATH + path ;
            if( FileUtil.isFileExists( ph ) ) {
                return ph ;
            }
            ph = SeaBassDef.NOW_APPLICATION_PATH + path ;
            if( FileUtil.isFileExists( ph ) ) {
                return ph ;
            }
            return null ;
        }
        else {
            if( path.startsWith( "/" ) && ( path = path.trim() ).length() != 1 ) {
                path = path.substring( 1,path.length() ) ;
            }
            if( path.endsWith( "/" ) ) {
                int len = HttpdDef.INDEX_HTML.length ;
                String pp = SeaBassDef.NOW_PUBLIC_PATH + path ;
                for( int i = 0 ; i < len ; i ++ ) {
                    String ph = pp + HttpdDef.INDEX_HTML[ i ] ;
                    if( FileUtil.isFileExists( ph ) ) {
                        if( request != null ) {
                            request.setUrlPath( request.getUrlPath() + HttpdDef.INDEX_HTML[ i ] ) ;
                        }
                        return ph ;
                    }
                }
                pp = SeaBassDef.NOW_APPLICATION_PATH + path ;
                for( int i = 0 ; i < len ; i ++ ) {
                    String ph = pp + HttpdDef.INDEX_HTML[ i ] ;
                    if( FileUtil.isFileExists( ph ) ) {
                        if( request != null ) {
                            request.setUrlPath( request.getUrlPath() + HttpdDef.INDEX_HTML[ i ] ) ;
                        }
                        return ph ;
                    }
                }
            }
            else {
                String ph = SeaBassDef.NOW_PUBLIC_PATH + path ;
                if( FileUtil.isFileExists( ph ) ) {
                    return ph ;
                }
                ph = SeaBassDef.NOW_APPLICATION_PATH + path ;
                if( FileUtil.isFileExists( ph ) ) {
                    return ph ;
                }
            }
        }
        return null ;
    }
    
    /**
     * 指定パスがJSファイルか、CSSファイルであるかチェック.
     */
    private static final boolean isJsOrCssOrIco( String path ) {
        path = path.toLowerCase() ;
        return ( path.endsWith( ".js" ) || path.endsWith( ".css" ) || path.endsWith( ".ico" ) ) ;
    }
    
    /**
     * 指定パスがInclude対応ファイルであるかチェック.
     */
    private static final int includeFileType( String path ) {
        path = path.toLowerCase() ;
        if( path.endsWith( ".ihtm" ) || path.endsWith( ".ihtml" ) ) {
            return 0 ;// html系.
        }
        else if( path.endsWith( ".ijs" ) ) {
            return 1 ;// javascript系.
        }
        else if( path.endsWith( ".icss" ) ) {
            return 2 ;// stylesheet系.
        }
        return -1 ;//なし.
    }
    
}

