package org.maachang.seabass.httpd.engine;

import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.seabass.httpd.HttpdErrorDef;
import org.maachang.seabass.httpd.HttpdStateException;
import org.maachang.util.FileUtil;
import org.maachang.util.atomic.AtomicLONG;
import org.maachang.util.thread.LoopThread;


/**
 * Httpdパブリックキャッシュ.
 * 
 * @version 2008/08/09
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
public class PublicCache {
    
    /**
     * LOG.
     */
    private static final Log LOG = LogFactory.getLog( PublicCache.class ) ;
    
    /**
     * シングルトン.
     */
    private static final PublicCache SNGL = new PublicCache() ;
    
    /**
     * キャッシュ情報格納オブジェクト.
     */
    private Map<String,PublicCacheChild> manager = null ;
    
    /**
     * スレッド監視用.
     */
    private PublicCacheChildThread thread = null ;
    
    /**
     * コンストラクタ.
     */
    private PublicCache() {
        try {
            manager = new ConcurrentHashMap<String,PublicCacheChild>() ;
            thread = new PublicCacheChildThread( manager ) ;
        } catch( Exception e ) {
            LOG.error( "error",e ) ;
        }
    }
    
    /**
     * オブジェクトを取得.
     * @return PublicCache オブジェクトが返されます.
     */
    public static final PublicCache getInstance() {
        return SNGL ;
    }
    
    /**
     * キャッシュ情報を全て破棄.
     */
    public void clear() {
        if( manager != null ) {
            manager.clear() ;
        }
    }
    
    /**
     * 指定パスのデータを取得.
     * @param path 対象のパスを設定します.
     * @return byte[] データが返されます.
     * @exception Exception 例外.
     */
    public byte[] getData( String path ) throws Exception {
        PublicCacheChild ch = get( path ) ;
        if( ch == null ) {
            return null ;
        }
        return ch.getData() ;
    }
    
    /**
     * 指定パスのファイル時間を取得.
     * @param path 対象のパスを設定します.
     * @return long 時間が返されます.
     * @exception Exception 例外.
     */
    public long getLastTime( String path ) throws Exception {
        PublicCacheChild ch = get( path ) ;
        if( ch == null ) {
            return -1L ;
        }
        return ch.getLastUpdateTime() ;
    }
    
    private PublicCacheChild get( String path ) throws Exception {
        if( path == null || ( path = path.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        PublicCacheChild ret = null ;
        if( manager.containsKey( path ) == false ) {
            if( FileUtil.isFileExists( path ) == true ) {
                if( FileUtil.isRead( path ) == false ) {
                    throw new HttpdStateException( HttpdErrorDef.HTTP11_403,
                        "指定パス[" + path + "]は読み込み権限がありません" ) ;
                }
                ret = new PublicCacheChild(
                    path,
                    FileUtil.getFile( path ),
                    FileUtil.getLastTime( path ) ) ;
                if( LOG.isDebugEnabled() ) {
                    LOG.debug( "## 指定ファイル[" + path + "]をキャッシュしました" ) ;
                }
                manager.put( path,ret ) ;
            }
        }
        else {
            ret = manager.get( path ) ;
            if( ret != null ) {
                ret.update() ;
            }
        }
        return ret ;
    }
    
    /**
     * オブジェクトが有効かチェック.
     * @return boolean [true]の場合は有効です.
     */
    public boolean isUse() {
        boolean ret = true ;
        ret = ( thread.isStop() == false ) ? true : false ;
        return ret ;
    }
}

/**
 * キャッシュ子要素.
 */
class PublicCacheChild {
    private static final long DELETE_TIME = 900000L ;
    private byte[] data = null ;
    private long lastUpdateTime = -1L ;
    private String path = null ;
    private final AtomicLONG lastAccessTime = new AtomicLONG( -1L ) ;
    
    private PublicCacheChild() {
        
    }
    public PublicCacheChild( String path,byte[] data,long lastUpdateTime ) {
        this.path = path ;
        this.data = data ;
        this.lastUpdateTime = lastUpdateTime ;
        this.lastAccessTime.set( System.currentTimeMillis() + DELETE_TIME ) ;
    }
    public void set( byte[] data,long lastUpdateTime ) {
        this.data = data ;
        this.lastUpdateTime = lastUpdateTime ;
    }
    public String getPath() {
        return path ;
    }
    public byte[] getData() {
        return data ;
    }
    public long getLastUpdateTime() {
        return lastUpdateTime ;
    }
    public void update() {
        this.lastAccessTime.set( System.currentTimeMillis() + DELETE_TIME ) ;
    }
    public long getUpdateTime() {
        return lastAccessTime.get() ;
    }
}

/**
 * データ確認用.
 */
class PublicCacheChildThread extends LoopThread {
    private Map<String,PublicCacheChild> manager = null ;
    public PublicCacheChildThread( Map<String,PublicCacheChild> manager )
        throws Exception {
        this.manager = manager ;
        super.startThread() ;
    }
    protected void finalize() throws Exception {
        this.destroy() ;
    }
    public void destroy() {
        super.stopThread() ;
    }
    protected void clear() {
        manager = null ;
    }
    protected boolean execution() throws Exception {
        Iterator itr = manager.keySet().iterator() ;
        while(itr.hasNext()) {
            if( super.isStop() ) {
                break ;
            }
            Thread.sleep( 100L ) ;
            String key = ( String )itr.next() ;
            PublicCacheChild ch = manager.get( key ) ;
            if( ch == null ) {
                continue ;
            }
            long time = -1L ;
            if( FileUtil.isFileExists( ch.getPath() ) == false ) {
                itr.remove() ;
                continue ;
            }
            else if( ch.getUpdateTime() <= System.currentTimeMillis() ) {
                itr.remove() ;
                continue ;
            }
            else if( ( time = FileUtil.getLastTime( ch.getPath() ) ) != ch.getLastUpdateTime() ) {
                byte[] bin = FileUtil.getFile( ch.getPath() ) ;
                ch.set( bin,time ) ;
            }
        }
        return false ;
    }
}
