package org.maachang.seabass.httpd.engine;

import java.io.IOException;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.security.MessageDigest;

import org.maachang.util.Base64;
import org.maachang.util.BinaryUtil;
import org.maachang.util.Xor128Random;

/**
 * SeaBass32暗号/復号処理.
 *
 * @version 2009/03/10
 * @author masahito suzuki
 * @since Seabass 1.0.0
 */
public abstract class Seabass32 {
    private Seabass32() {}
    private static final Xor128Random rand = new Xor128Random() ;
    
    /**
     * プライベートキーを発行.
     * @param domain 対象のドメイン名を設定します.
     * @param sessionId 対象のセッションIDを設定します.
     * @return byte[] プライベートキーが返されます.
     * @exception IOException I/O例外.
     */
    public static final byte[] createPrivateKey( String domain,String sessionId )
        throws IOException {
        if( domain == null || sessionId == null ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        byte[] domainBin = Seabass32._md5( domain ) ;
        byte[] sessionIdBin = Seabass32._md5( sessionId ) ;
        byte[] ret = new byte[32] ;
        System.arraycopy( domainBin,0,ret,0,16 ) ;
        System.arraycopy( sessionIdBin,0,ret,16,16 ) ;
        for( int i = 0 ; i < 16 ; i ++ ) {
            ret[ i ] = Seabass32._convert( ret,i,sessionIdBin[i] ) ;
        }
        for( int i = 15,j = 0 ; i >= 0 ; i --,j ++ ) {
            ret[ i+16 ] = Seabass32._convert( ret,i+16,domainBin[j] ) ;
        }
        return ret ;
    }
    
    /**
     * 暗号化.
     * @param binary 暗号化対象の文字列を設定します.
     * @param domain 対象のドメイン名を設定します.
     * @param sessionId 対象のセッションIDを設定します.
     * @return String 暗号化された内容が返されます.
     * @exception IOException I/O例外.
     */
    public static final String encToString( String s,String domain,String sessionId )
        throws IOException {
        return Seabass32.enc( URLEncoder.encode( s,"UTF8" ).getBytes( "UTF8" ),
            Seabass32.createPrivateKey( domain,sessionId ) ) ;
    }
    
    /**
     * 暗号化.
     * @param binary 暗号化対象のバイナリを設定します.
     * @param domain 対象のドメイン名を設定します.
     * @param sessionId 対象のセッションIDを設定します.
     * @return String 暗号化された内容が返されます.
     * @exception IOException I/O例外.
     */
    public static final String enc( byte[] binary,String domain,String sessionId )
        throws IOException {
        return Seabass32.enc( binary,Seabass32.createPrivateKey( domain,sessionId ) ) ;
    }
    
    /**
     * 暗号化.
     * @param s 暗号化対象の文字列を設定します.
     * @param pKey 対象のプライベートキーを設定します.
     * @return String 暗号化された内容が返されます.
     * @exception IOException I/O例外.
     */
    public static final String encToString( String s,byte[] pKey )
        throws IOException {
        return Seabass32.enc( URLEncoder.encode( s,"UTF8" ).getBytes( "UTF8" ),pKey ) ;
    }
    
    /**
     * 暗号化.
     * @param binary 暗号化対象のバイナリを設定します.
     * @param pKey 対象のプライベートキーを設定します.
     * @return String 暗号化された内容が返されます.
     * @exception IOException I/O例外.
     */
    public static final String enc( byte[] binary,byte[] pKey )
        throws IOException {
        if( binary == null || binary.length <= 0 || pKey == null || pKey.length != 32 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        byte[] pubKey = Seabass32._randKey() ;
        byte[] key32 = Seabass32._convertKey( pKey,pubKey ) ;
        byte[] key256 = Seabass32._key256( key32 ) ;
        key32 = null ;
        int stepNo = ( Seabass32._getStepNo( pKey,binary ) & 0x0000007f ) ;
        int nowStep = Seabass32._convert256To( key256,pubKey,stepNo ) ;
        Seabass32._ed( true,binary,key256,nowStep ) ;
        byte[] eb = new byte[ 33+binary.length ] ;
        eb[ 0 ] = ( byte )(~stepNo) ;
        System.arraycopy( pubKey,0,eb,1,32 ) ;
        System.arraycopy( binary,0,eb,33,binary.length ) ;
        return "@sb32" + Base64.encode( eb ) ;
    }
    
    /**
     * 復号化.
     * @param encString 暗号化された内容を設定します.
     * @param domain 対象のドメイン名を設定します.
     * @param sessionId 対象のセッションIDを設定します.
     * @return String 復号化された文字列が返されます.
     * @exception IOException I/O例外.
     */
    public static final String decToString( String encString,String domain,String sessionId )
        throws IOException {
        byte[] b = Seabass32.dec( encString,Seabass32.createPrivateKey( domain,sessionId ) ) ;
        return URLDecoder.decode( new String( b,"UTF8" ),"UTF8" ) ;
    }
    
    /**
     * 復号化.
     * @param encString 暗号化された内容を設定します.
     * @param domain 対象のドメイン名を設定します.
     * @param sessionId 対象のセッションIDを設定します.
     * @return byte[] 復号化されたバイナリ情報が返されます.
     * @exception IOException I/O例外.
     */
    public static final byte[] dec( String encString,String domain,String sessionId )
        throws IOException {
        return Seabass32.dec( encString,Seabass32.createPrivateKey( domain,sessionId ) ) ;
    }
    
    /**
     * 復号化.
     * @param encString 暗号化された内容を設定します.
     * @param pKey 対象のプライベートキーを設定します.
     * @return String 復号化された文字列が返されます.
     * @exception IOException I/O例外.
     */
    public static final String decToString( String encString,byte[] pKey )
        throws IOException {
        byte[] b = Seabass32.dec( encString,pKey ) ;
        return URLDecoder.decode( new String( b,"UTF8" ),"UTF8" ) ;
    }
    
    /**
     * 復号化.
     * @param encString 暗号化された内容を設定します.
     * @param pKey 対象のプライベートキーを設定します.
     * @return byte[] 復号化されたバイナリ情報が返されます.
     * @exception IOException I/O例外.
     */
    public static final byte[] dec( String encString,byte[] pKey )
        throws IOException {
        if( encString == null || !encString.startsWith( "@sb32" ) ||
            pKey == null || pKey.length != 32 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        encString = encString.substring( 5 ) ;
        byte[] bin = Base64.decode( encString ) ;
        encString = null ;
        if( bin.length <= 33 ) {
            throw new IOException( "不正なバイナリ長です" ) ;
        }
        int stepNo = (int)((~bin[ 0 ])&0x0000007f) ;
        byte[] pubKey = new byte[ 32 ] ;
        System.arraycopy( bin,1,pubKey,0,32 ) ;
        int bodyLen = bin.length - 33 ;
        byte[] body = new byte[ bodyLen ] ;
        System.arraycopy( bin,33,body,0,bodyLen ) ;
        bin = null ;
        byte[] key32 = Seabass32._convertKey( pKey,pubKey ) ;
        byte[] key256 = Seabass32._key256( key32 ) ;
        
        key32 = null ;
        int nowStep = Seabass32._convert256To( key256,pubKey,stepNo ) ;
        Seabass32._ed( false,body,key256,nowStep ) ;
        if( ( Seabass32._getStepNo( pKey,body ) & 0x0000007f ) != stepNo ) {
            throw new IOException(
                new StringBuilder().append( "Seabass32復号処理に失敗しました" ).
                    append( "\n    pubKey:\n" ).append( BinaryUtil.toString( pubKey,0,pubKey.length ) ).
                    append( "\n    key:\n" ).append( BinaryUtil.toString( pKey,0,pKey.length ) ).
                    toString()
            ) ;
        }
        return body ;
    }
    
    /**
     * コンバート処理.
     */
    private static final byte _convert(byte[] key, int no, byte pause) {
        switch ((no & 0x00000001)) {
            case 0:
                return (byte) ((~(pause ^ key[no])) & 0x000000ff) ;
            case 1:
                return (byte) ((pause ^ key[no]) & 0x000000ff) ;
        }
        return (byte)0 ;
    }
    
    /** ランダムキーコードを発行. **/
    private static final byte[] _randKey() {
        byte[] bin = new byte[ 32 ] ;
        rand.nextBytes( bin ) ;
        return bin ;
    }
    
    /** digestキーを発行. **/
    private static final byte[] _md5( String s )
        throws IOException {
        try {
            MessageDigest md = MessageDigest.getInstance("MD5");
            byte[] b = s.getBytes( "UTF8" ) ;
            md.update(b,0,b.length);
            return md.digest();
        } catch( IOException io ) {
            throw io ;
        } catch( Exception e ) {
            throw new IOException( e.getMessage() ) ;
        }
    }
    
    /** digestキーを発行. **/
    private static final byte[] _md5( byte[] b )
        throws IOException {
        try {
            MessageDigest md = MessageDigest.getInstance("MD5");
            md.update(b,0,b.length);
            return md.digest();
        } catch( Exception e ) {
            throw new IOException( e.getMessage() ) ;
        }
    }
    
    /** 公開キーとプライベートキーを用いて、本来の暗号キーを生成. **/
    private static final byte[] _convertKey( byte[] pKey,byte[] key )
        throws IOException {
        byte[] low = Seabass32._md5(pKey);
        byte[] hight = Seabass32._md5(key);
        byte[] ret = new byte[32];
        for (int i = 0,j = 0,k = 15; i < 16; i++, j += 2, k--) {
            ret[j] = Seabass32._convert(low, i, key[j]);
            ret[j + 1] = Seabass32._convert(hight, i, low[k]);
        }
        return ret;
    }
    
    /** key32をkey256に変換. **/
    private static final byte[] _key256( byte[] key32 )
        throws IOException {
        byte[] ret = new byte[ 256 ] ;
        byte[] b = new byte[ 4 ] ;
        int n = 0 ;
        for( int i = 0,j = 0 ; i < 31 ; i += 2,j += 16 ) {
            int s = (int)( key32[i] & 0x000000ff ) ;
            int e = (int)( key32[i+1] & 0x000000ff ) ;
            if( ( n & 0x00000001 ) != 0 ) {
                n += s ^ (~ e ) ;
            }
            else {
                n -= (~s) ^ e ;
            }
            b[0] = (byte)(n & 0x000000ff) ;
            b[1] = (byte)(((n & 0x0000ff00)>>8)&0x000000ff) ;
            b[2] = (byte)(((n & 0x00ff0000)>>16)&0x000000ff) ;
            b[3] = (byte)(((n & 0xff000000)>>24)&0x000000ff) ;
            byte[] o = Seabass32._md5( b ) ;
            System.arraycopy( o,0,ret,j,16 ) ;
        }
        return ret ;
    }
    
    /** stepCodeを生成. **/
    private static final int _getStepNo( byte[] pubKey,byte[] binary ) {
        int i, j;
        int bin;
        int ret = 0;
        int len = binary.length ;
        int addCd = (int) ((pubKey[(int) (binary[(int)(len / 2)] & 0x0000001f)] & 0x00000003) + 1);
        for (i = 0, j = 0; i < len; i += addCd, j += addCd) {
            bin = (int) ((~binary[i]) & 0x000000ff);
            ret = ((bin & 0x00000001) + ((bin & 0x00000002) >> 1)
                    + ((bin & 0x00000004) >> 2) + ((bin & 0x00000008) >> 3)
                    + ((bin & 0x00000010) >> 4) + ((bin & 0x00000020) >> 5)
                    + ((bin & 0x00000040) >> 6) + ((bin & 0x00000080) >> 7))
                    + (j & 0x000000ff) + ret;
        }
        if ((ret & 0x00000001) == 0) {
            for (i = 0; i <32; i++) {
                bin = (int) (((pubKey[i] & 0x00000001) == 0) ? ((~pubKey[i]) & 0x000000ff)
                        : (pubKey[i] & 0x000000ff));
                ret += ((bin & 0x00000001) + ((bin & 0x00000002) >> 1)
                        + ((bin & 0x00000004) >> 2) + ((bin & 0x00000008) >> 3)
                        + ((bin & 0x00000010) >> 4) + ((bin & 0x00000020) >> 5)
                        + ((bin & 0x00000040) >> 6) + ((bin & 0x00000080) >> 7));
            }
        } else {
            for (i = 0; i < 32; i++) {
                bin = (int) (((pubKey[i] & 0x00000001) == 0) ? ((~pubKey[i]) & 0x000000ff)
                        : (pubKey[i] & 0x000000ff));
                ret -= ((bin & 0x00000001) + ((bin & 0x00000002) >> 1)
                        + ((bin & 0x00000004) >> 2) + ((bin & 0x00000008) >> 3)
                        + ((bin & 0x00000010) >> 4) + ((bin & 0x00000020) >> 5)
                        + ((bin & 0x00000040) >> 6) + ((bin & 0x00000080) >> 7));
            }
        }
        return (int) ((~ret) & 0x000000ff);
    }
    
    /** Flip. **/
    private static final byte _flip(int pause, int step) {
        switch (step & 0x00000007) {
        case 1:
            return (byte)(((((pause & 0x00000003) << 6) & 0x000000c0) | (((pause & 0x000000fc) >> 2) & 0x0000003f)) & 0x000000ff) ;
        case 2:
            return (byte)(((((pause & 0x0000003f) << 2) & 0x000000fc) | (((pause & 0x000000c0) >> 6) & 0x00000003)) & 0x000000ff) ;
        case 3:
            return (byte)(((((pause & 0x00000001) << 7) & 0x00000080) | (((pause & 0x000000fe) >> 1) & 0x0000007f)) & 0x000000ff) ;
        case 4:
            return (byte)(((((pause & 0x0000000f) << 4) & 0x000000f0) | (((pause & 0x000000f0) >> 4) & 0x0000000f)) & 0x000000ff) ;
        case 5:
            return (byte)(((((pause & 0x0000007f) << 1) & 0x000000fe) | (((pause & 0x00000080) >> 7) & 0x00000001)) & 0x000000ff) ;
        case 6:
            return (byte)(((((pause & 0x00000007) << 5) & 0x000000e0) | (((pause & 0x000000f8) >> 3) & 0x0000001f)) & 0x000000ff) ;
        case 7:
            return (byte)(((((pause & 0x0000001f) << 3) & 0x000000f8) | (((pause & 0x000000e0) >> 5) & 0x00000007)) & 0x000000ff) ;
        }
        return (byte)(pause & 0x000000ff) ;
    }
    
    /** 逆Flip. **/
    private static final byte _nflip(int pause, int step) {
        switch (step & 0x00000007) {
        case 1:
            return (byte)(((((pause & 0x0000003f) << 2) & 0x000000fc) | (((pause & 0x000000c0) >> 6) & 0x00000003)) & 0x000000ff) ;
        case 2:
            return (byte)(((((pause & 0x00000003) << 6) & 0x000000c0) | (((pause & 0x000000fc) >> 2) & 0x0000003f)) & 0x000000ff) ;
        case 3:
            return (byte)(((((pause & 0x0000007f) << 1) & 0x000000fe) | (((pause & 0x00000080) >> 7) & 0x00000001)) & 0x000000ff) ;
        case 4:
            return (byte)(((((pause & 0x0000000f) << 4) & 0x000000f0) | (((pause & 0x000000f0) >> 4) & 0x0000000f)) & 0x000000ff) ;
        case 5:
            return (byte)(((((pause & 0x00000001) << 7) & 0x00000080) | (((pause & 0x000000fe) >> 1) & 0x0000007f)) & 0x000000ff) ;
        case 6:
            return (byte)(((((pause & 0x0000001f) << 3) & 0x000000f8) | (((pause & 0x000000e0) >> 5) & 0x00000007)) & 0x000000ff) ;
        case 7:
            return (byte)(((((pause & 0x00000007) << 5) & 0x000000e0) | (((pause & 0x000000f8) >> 3) & 0x0000001f)) & 0x000000ff) ;
        }
        return (byte)(pause & 0x000000ff) ;
    }
    
    /** Key256を、StepCodeとPrivateKeyで変換. **/
    private static final int _convert256To(byte[] key256, byte[] pKey, int step) {
        int ns = step ;
        for (int i = 0, j = 0; i < 256; i++, j = ((j + 1) & 0x0000001f)) {
            ns = (ns ^ (~((int) key256[i]))) ;
            if( (ns & 0x00000001 ) == 0 ) {
                ns = ~ns ;
            }
            key256[i] = Seabass32._convert(pKey, j, key256[i]);
            key256[i] = Seabass32._flip((int) key256[i], ns);
        }
        return ns;
    }
    
    /** 暗号化／復号化. **/
    private static final void _ed(boolean mode,byte[] binary,byte[] key256, int step) {
        int len = binary.length ;
        int ns = step ;
        if( mode ) {
            for (int i = 0, j = 0; i < len; i++, j = ((j + 1) & 0x000000ff)) {
                ns = (ns ^ (~((int) key256[j]))) ;
                if( (ns & 0x00000001 ) != 0 ) {
                    ns = ~ns ;
                }
                binary[i] = Seabass32._convert(key256, j, (byte) binary[i]);
                binary[i] = Seabass32._flip( binary[ i ],ns ) ;
            }
        }
        else {
            for (int i = 0, j = 0; i < len; i++, j = ((j + 1) & 0x000000ff)) {
                ns = (ns ^ (~((int) key256[j]))) ;
                if( (ns & 0x00000001 ) != 0 ) {
                    ns = ~ns ;
                }
                binary[i] = Seabass32._nflip( binary[ i ],ns ) ;
                binary[i] = Seabass32._convert(key256, j, (byte) binary[i]);
            }
        }
    }
}
