package org.maachang.seabass.httpd.engine.comet ;

import java.io.IOException;
import java.util.Iterator;
import java.util.concurrent.ConcurrentLinkedQueue;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.manager.GlobalManager;
import org.maachang.seabass.SeaBassContext;
import org.maachang.seabass.ServiceDef;
import org.maachang.seabass.httpd.HttpdErrorDef;
import org.maachang.seabass.httpd.HttpdRequest;
import org.maachang.seabass.httpd.HttpdResponse;
import org.maachang.seabass.httpd.engine.HttpdDef;
import org.maachang.seabass.httpd.engine.script.BaseModel;
import org.maachang.seabass.httpd.engine.script.BaseModelImpl;
import org.maachang.seabass.httpd.engine.script.EndScript;
import org.maachang.seabass.httpd.engine.script.Script;
import org.maachang.seabass.httpd.engine.script.ScriptDef;
import org.maachang.seabass.httpd.engine.script.cache.CacheScriptManager;
import org.maachang.seabass.httpd.engine.script.scripts.ApplicationScriptFactory;
import org.maachang.seabass.httpd.engine.session.HttpdSession;
import org.maachang.seabass.net.HttpdResponseImpl;
import org.maachang.seabass.net.nio.ConnectionInfo;
import org.maachang.seabass.net.nio.KeepAliveFactory;
import org.maachang.util.atomic.AtomicLONG;

/**
 * 1つのコメット情報.
 *
 * @version 2007/08/24
 * @author  masahito suzuki
 * @since   SeaBass 1.00
 */
public class Comet {
    
    /**
     * LOG.
     */
    private static final Log LOG = LogFactory.getLog( Comet.class ) ;
    
    /**
     * グループID.
     */
    private String groupId = null ;
    
    /**
     * コメットスクリプトパス.
     */
    private String scriptPath = null ;
    
    /**
     * このコメットに接続している情報群.
     */
    private final CometConnectQueue<ConnectComet> connects = new CometConnectQueue<ConnectComet>() ;
    
    /**
     * コメット生成時間.
     */
    private final AtomicLONG createTime = new AtomicLONG(-1L) ;
    
    /**
     * コメット更新時間.
     */
    private final AtomicLONG updateTime = new AtomicLONG(-1L) ;
    
    /**
     * レスポンス用.
     */
    private final HttpdResponseImpl response = new HttpdResponseImpl() ;
    
    /**
     * コンストラクタ.
     */
    private Comet() {
        
    }
    
    /**
     * コンストラクタ.
     * <BR><BR>
     * 条件を設定してコメット接続状態を確立します.
     * <BR>
     * @param groupId 対象のグループIDを設定します.
     * @param scriptPath コメットスクリプトパスを設定します.
     * @exception Exception 例外.
     */
    public Comet( String groupId,String scriptPath )
        throws Exception {
        if( groupId == null || ( groupId = groupId.trim() ).length() <= 0 ||
            scriptPath == null || ( scriptPath = scriptPath.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        this.groupId = groupId ;
        this.scriptPath = scriptPath ;
        this.createTime.set(System.currentTimeMillis()) ;
        this.updateTime.set(System.currentTimeMillis()) ;
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        this.clear() ;
    }
    
    /**
     * 情報クリア.
     * <BR><BR>
     * 情報をクリアします.
     */
    public void clear() {
        destroyConnection( connects ) ;
        this.createTime.set(-1L) ;
        this.updateTime.set(-1L) ;
    }
    
    /**
     * 新しいコネクションを追加.
     * <BR><BR>
     * 新しいコネクションを追加します.
     * <BR>
     * @param request リクエストを追加します.
     * @exception Exception 例外.
     */
    public void putConnection( HttpdRequest request )
        throws Exception {
        if( request == null || request.isUse() == false ) {
            return ;
        }
        this.connects.add( new ConnectComet( request ) ) ;
        this.updateTime.set(System.currentTimeMillis()) ;
    }
    
    /**
     * 接続先に情報を送信.
     * <BR><BR>
     * 接続先に情報を送信します.
     * <BR>
     * @param args コメット実行処理に渡したい、パラメータを設定します.
     * @param groupId 対象のグループIDを設定します.
     * @exception Exception 例外.
     */
    public void send( Object args,String groupId )
        throws Exception {
        try {
            ConcurrentLinkedQueue<ConnectComet> conns = getNowCometConnects() ;
            if( conns == null || conns.size() <= 0 ) {
                return ;
            }
            // Comet実行.
            sendAll( response,conns,scriptPath,args,groupId ) ;
            if( LOG.isDebugEnabled() ) {
                LOG.debug( "... exit - comet:" + groupId ) ;
            }
        } catch( Exception e ) {
            throw e ;
        }
    }
    
    /**
     * このコメットの接続数を取得.
     * <BR><BR>
     * このコメットの接続数を取得します.
     * <BR>
     * @return int 接続されている接続数が返されます.
     */
    public int getConnectSize() {
        return connects.size() ;
    }
    
    /**
     * グループIDを取得.
     * <BR><BR>
     * グループIDを取得します.
     * <BR>
     * @return String グループIDが返されます.
     */
    public String getGroupId() {
        return this.groupId ;
    }
    
    /**
     * スクリプトパスを取得.
     * <BR><BR>
     * スクリプトパスを取得します.
     * <BR>
     * @return String スクリプトパスが返されます.
     */
    public String getScriptPath() {
        return this.scriptPath ;
    }
    
    /**
     * コメット生成時間を取得.
     * <BR><BR>
     * コメット生成時間を取得します.
     * <BR>
     * @return long コメット生成時間が返されます.
     */
    public long getCreateTime() {
        return this.createTime.get() ;
    }
    
    /**
     * コメット更新時間を取得.
     * <BR><BR>
     * コメット更新時間を取得します.
     * <BR>
     * @return long コメット更新時間が返されます.
     */
    public long getUpdateTime() {
        return this.updateTime.get();
    }
    
    /**
     * １つのリクエスト情報を処理.
     * <BR><BR>
     * １つのリクエスト情報を処理します.
     * <BR>
     * @param res 出力先を設定します.
     * @param request 対象のリクエスト情報を設定します.
     * @param scriptPath 対象のスクリプトパスを設定します.
     * @param groupId 対象のグループIDを設定します.
     * @return Object 戻り情報が返されます.
     * @exception Exception 例外.
     */
    public static final Object oneSend( HttpdResponse res,BaseModel baseModel,
        HttpdRequest request,String scriptPath,String groupId )
        throws Exception {
        return oneSend( res,baseModel,request,scriptPath,null,groupId ) ;
    }
    
    /**
     * １つのリクエスト情報を処理.
     * <BR><BR>
     * １つのリクエスト情報を処理します.
     * <BR>
     * @param res 出力先を設定します.
     * @param request 対象のリクエスト情報を設定します.
     * @param scriptPath 対象のスクリプトパスを設定します.
     * @param args コメット実行処理に渡したい、パラメータを設定します.
     * @param groupId 対象のグループIDを設定します.
     * @return Object 戻り情報が返されます.
     * @exception Exception 例外.
     */
    public static final Object oneSend( HttpdResponse res,BaseModel baseModel,
        HttpdRequest request,String scriptPath,Object args,String groupId )
        throws Exception {
        ApplicationScriptFactory webapp = ( ApplicationScriptFactory )GlobalManager.getValue(
            ServiceDef.MANAGER_BY_WEB_APP_FACTORY ) ;
        if( webapp == null ) {
            throw new IOException( "WebAppScriptFactoryの取得に失敗しました" ) ;
        }
        // コメットスクリプトを取得.
        Script comet = webapp.getApplication( scriptPath ) ;
        if( comet == null ) {
            throw new IOException( "["+scriptPath+"]の取得に失敗しました" ) ;
        }
        Object ret = null ;
        boolean errorFlag = false ;
        SeaBassContext context = null ;
        try {
            // 実行コンテキスト生成.
            context = CacheScriptManager.getInstance().script( null ) ;
            
            // ライブラリスクリプトを読み込む.
            ScriptDef.setDirectoryByBindings( context ) ;
            
            // 各パラメータを設定.
            context.put( ScriptDef.SCRIPT_BY_QUERY,ScriptDef.getQuery( request.getQuery() ) ) ;
            context.put( ScriptDef.SCRIPT_MODE,ScriptDef.MODE_COMET ) ;
            context.put( ScriptDef.SCRIPT_BY_MODEL,baseModel ) ;
            context.put( ScriptDef.SCRIPT_BY_HEADER,request.getHeader() ) ;
            context.put( ScriptDef.SCRIPT_BY_COMET_ARGS,args ) ;
            context.put( ScriptDef.SCRIPT_BY_GROUP_ID,groupId ) ;
            HttpdSession session = request.getSession() ;
            if( session != null ) {
                context.put( ScriptDef.SCRIPT_BY_SESSION,session ) ;
            }
            
            if( res != null ) {
                // HTTPレスポンスを生成.
                res.setHttpCache( true ) ;
                res.setHttpClose( false ) ;//keep-alive.
                res.setCookieSession( request ) ;
                res.getHeader().setHeader( HttpdDef.VALUE_CONTENT_TYPE,HttpdDef.AJAX_MIME_TYPE ) ;
                res.setHttpCache( true ) ;
                
                // リクエスト/レスポンスを登録.
                context.put( ScriptDef.SCRIPT_BY_REQUEST,request ) ;
                context.put( ScriptDef.SCRIPT_BY_RESPONSE,res ) ;
            }
            
            // 実行.
            ret = comet.getScript().execution( context,null ) ;
            if( EndScript.isEndScript( ret ) ) {
                if( EndScript.executionExitByError( context,request,res ) ) {
                    errorFlag = true ;
                }
            }
            if( res != null ) {
                res.flush() ;
            }
        } catch( Exception e ) {
            LOG.warn( "error",e ) ;
            errorFlag = true ;
            throw e ;
        } finally {
            if( baseModel.isCreate() ) {
                try {
                    if( errorFlag ) {
                        baseModel.rollback() ;
                    }
                    else {
                        baseModel.commit() ;
                    }
                } catch( Exception ee ) {
                    try {
                        baseModel.rollback() ;
                    } catch( Exception eee ) {
                    }
                }
            }
            CacheScriptManager.getInstance().exitScript( context ) ;
        }
        return ret ;
    }
    
    /**
     * 現在までのコネクションオブジェクト群を取得.
     * <p>この処理で、現在までのCometコネクションだけを、コールバック対象と
     * するようにします</p>
     */
    private ConcurrentLinkedQueue<ConnectComet> getNowCometConnects() {
        ConcurrentLinkedQueue<ConnectComet> ret = this.connects.nowConnection() ;
        this.updateTime.set( System.currentTimeMillis() ) ;
        return ret ;
    }
    
    /**
     * 全ての接続先に情報を送信.
     */
    private static void sendAll( HttpdResponseImpl response,ConcurrentLinkedQueue<ConnectComet> conns,
        String scriptPath,Object args,String groupId )
        throws Exception {
        BaseModel baseModel = new BaseModelImpl() ;
        Iterator it = conns.iterator() ;
        int cnt = 0 ;
        while( it.hasNext() ) {
            cnt ++ ;
            ConnectComet conn = (ConnectComet)it.next() ;
            if( conn == null || !conn.isConnection() ) {
                if( conn != null ) {
                    conn.destroy() ;
                }
                continue ;
            }
            try {
                // レスポンス情報を生成.
                HttpdRequest req = conn.getRequest() ;
                response.create(
                    req,req.getUrlPath(),HttpdErrorDef.HTTP11_200,
                    req.getKeepAliveTimeout(),req.getKeepAliveCount() ) ;
                // 送信処理.
                oneSend( response,baseModel,req,scriptPath,args,groupId ) ;
                // 送信結果を反映.
                response.flushAndClose() ;
                ConnectionInfo info = conn.getConnectionInfo() ;
                try {
                    if( info != null && info.isUse() == true ) {
                        if( info.isCloseFlag() == true ) {
                            info.getElement().setKeepAliveTime( 0 ) ;
                        }
                        else {
                            if( info.recyclingConnection() ) {
                                conn.cancel() ;
                            }
                        }
                        // KeepAlive監視処理.
                        //KeepAliveFactory.getInstance().addKeepAlive( info.getElement() ) ;
                    }
                    else {
                        if( info != null ) {
                            info.destroy() ;
                        }
                    }
                } catch( Exception ee ) {
                }
                // コネクションキャンセル.
                conn.cancel() ;
            } catch( Exception e ) {
                if( conn != null ) {
                    conn.destroy() ;
                }
                LOG.error( "comet-error",e ) ;
            } finally {
                response.clear() ;
            }
        }
    }
    
    /**
     * 各接続先をクローズ.
     */
    private static final void destroyConnection( CometConnectQueue<ConnectComet> conns ) {
        Iterator it = conns.iterator() ;
        while( it.hasNext() ) {
            ConnectComet con = (ConnectComet)it.next() ;
            if( con != null ) {
                con.destroy() ;
            }
        }
        conns.clear() ;
    }
    
}

