package org.maachang.seabass.httpd.engine.filter ;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.seabass.httpd.HttpdIOFilter;
import org.maachang.seabass.httpd.HttpdRequest;
import org.maachang.seabass.httpd.HttpdResponse;
import org.maachang.util.ObjectList;

/**
 * HTTPD-入出力フィルタ管理.
 *
 * @version 2008/12/14
 * @author  masahito suzuki
 * @since   SeaBass 1.2B
 */
public class HttpdIOFilterManager {
    private static final Log LOG = LogFactory.getLog( HttpdIOFilterManager.class ) ;
    private final ObjectList list = new ObjectList() ;
    
    /**
     * コンストラクタ.
     */
    public HttpdIOFilterManager() {
        this.clear() ;
    }
    
    protected void finalize() throws Exception {
        list.clear() ;
    }
    
    /**
     * 情報クリア.
     */
    public void clear() {
        list.clear() ;
    }
    
    /**
     * フィルタ処理を破棄.
     */
    public void destroyAll() {
        int len = list.size() ;
        for( int i = len-1 ; i >= 0 ; i -- ) {
            try {
                Object[] os = ( Object[] )list.get( i ) ;
                HttpdIOFilter filter = ( HttpdIOFilter )os[ 1 ] ;
                filter.destroy() ;
            } catch( Exception e ) {
                LOG.error( "destroyAll-exception(" + i + ")",e ) ;
            } catch( Error er ) {
                LOG.error( "destroyAll-error(" + i + ")",er ) ;
            }
        }
    }
    
    /**
     * Filter追加.
     * @param url マスク対象のURLを設定します.
     * @param filter 対象のフィルタを設定します.
     * @exception Exception 例外.
     */
    public void add( String url,HttpdIOFilter filter )
        throws Exception {
        if( filter == null ) {
            return ;
        }
        // URL整形.
        if( url == null || ( url = url.trim() ).length() <= 0 || url.equals( "/" ) ) {
            url = null ;
        }
        else if( url.startsWith( "/" ) == false ) {
            url = "/" + url ;
        }
        list.add( new Object[]{ url,filter } ) ;
        filter.init() ;
    }
    
    /**
     * コンテンツ実行前フィルタ実行.
     * @param request 対象のHttpdRequestを設定します.
     * @param response 対象のレスポンスを設定します.
     * @return HttpdResponse 対象のレスポンスを設定します.
     * @exception Exception 例外.
     */
    public HttpdResponse executionBeforeFilter( HttpdRequest request,HttpdResponse response )
        throws Exception {
        HttpdResponse ret = response ;
        String url = request.getUrlPath() ;
        int len = list.size() ;
        for( int i = 0 ; i < len ; i ++ ) {
            Object[] os = ( Object[] )list.get( i ) ;
            String urlMask = ( String )os[ 0 ] ;
            HttpdIOFilter filter = ( HttpdIOFilter )os[ 1 ] ;
            if( urlMask == null && url.startsWith( urlMask ) ) {
                HttpdResponse res = filter.beforePointContents( request,ret ) ;
                if( res != null && res != response ) {
                    ret = res ;
                }
            }
        }
        return ret ;
    }
    
    /**
     * コンテンツ実行前フィルタ実行.
     * @param request 対象のリクエストを設定します.
     * @param baseResponse 元のレスポンス情報を設定します.
     * @param dummyResponse 仮のレスポンス処理を行った条件が渡されます.
     * @exception Exception 例外.
     */
    public void executionEndFilter( HttpdRequest request,HttpdResponse baseResponse,HttpdResponse dummyResponse )
        throws Exception {
        String url = request.getUrlPath() ;
        int len = list.size() ;
        for( int i = len-1 ; i >= 0 ; i -- ) {
            Object[] os = ( Object[] )list.get( i ) ;
            String urlMask = ( String )os[ 0 ] ;
            HttpdIOFilter filter = ( HttpdIOFilter )os[ 1 ] ;
            if( urlMask == null && url.startsWith( urlMask ) ) {
                filter.endPointContents( request,baseResponse,dummyResponse ) ;
            }
        }
    }
    
    /**
     * 登録フィルタ数を取得.
     * @return int 登録フィルタ数が返されます.
     */
    public int size() {
        return list.size() ;
    }
    
    /**
     * 現在の構成を文字列で取得.
     * @return String 現在の構成内容が返されます.
     */
    public String toString() {
        StringBuilder buf = new StringBuilder() ;
        int len = list.size() ;
        for( int i = 0 ; i < len ; i ++ ) {
            if( i != 0 ) {
                buf.append( "\n" ) ;
            }
            Object[] os = ( Object[] )list.get( i ) ;
            String urlMask = ( String )os[ 0 ] ;
            HttpdIOFilter filter = ( HttpdIOFilter )os[ 1 ] ;
            buf.append( "url:" ).append( urlMask ).append( " class:" ).append( filter.getClass().getName() ) ;
        }
        return buf.toString() ;
    }
}

