package org.maachang.seabass.httpd.engine.script;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.maachang.dao.ResultLine;
import org.maachang.dao.dbms.DbUtil;
import org.maachang.dao.dbms.MetaColumn;
import org.maachang.dao.dbms.MetaFactory;
import org.maachang.dao.dbms.Record;
import org.maachang.manager.GlobalManager;
import org.maachang.seabass.SeaBassContext;
import org.maachang.seabass.ServiceDef;
import org.maachang.seabass.httpd.engine.script.dao.Dao;
import org.maachang.seabass.httpd.engine.script.dao.DaoFactory;
import org.maachang.seabass.httpd.engine.script.dao.DaoUtil;
import org.maachang.seabass.httpd.engine.script.dao.MasterCache;
import org.maachang.util.StringUtil;

/**
 * モデル実行処理格納用.
 * 
 * @version 2007/08/27
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
public class BaseModelImpl implements BaseModel {
    
    /**
     * SQL実行処理用.
     */
    private Record record = null ;
    
    /**
     * スクリプトコンテキスト.
     */
    private SeaBassContext context = null ;
    
    /**
     * 更新/削除テーブル管理.
     */
    private List<String> updateTables = null ;
    
    /**
     * 生成チェック用.
     */
    private boolean createFlag = false ;
    
    /**
     * コンストラクタ.
     */
    public BaseModelImpl() {
        
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        this.clear() ;
    }
    
    /**
     * 生成処理.
     * <BR><BR>
     * 条件を設定して生成します.
     * <BR>
     * @param ctx 対象のスクリプトコンテキストを設定します.
     * @param record 対象のレコードオブジェクトを設定します.
     * @exception Exception 例外.
     */
    public void create( SeaBassContext ctx,Record record )
        throws Exception {
        if( record == null ) {
            throw new IOException( "Recordは不正です" ) ;
        }
        this.record = record ;
        this.context = ctx ;
        this.updateTables = null ;
        this.createFlag = true ;
    }
    
    /**
     * クリア処理.
     */
    public void clear() {
        if( record != null ) {
            try {
                record.commit() ;
            } catch( Exception e ) {
            }
            try {
                record.close() ;
            } catch( Exception e ) {
            }
        }
        if( updateTables != null ) {
            updateTables.clear() ;
        }
        record = null ;
        context = null ;
        updateTables = null ;
        createFlag = false ;
        
    }
    
    /**
     * コミット処理.
     * <BR><BR>
     * コミット処理を実施します.
     * <BR>
     * @exception Exception 例外.
     */
    public void commit() throws Exception {
        if( createFlag == true ) {
            checkCommit() ;
            record.commit() ;
            // マスターキャッシュに対する更新テーブルが存在する場合は、
            // その内容を破棄する.
            if( updateTables != null && updateTables.size() >= 0 ) {
                MasterCache cache = ( MasterCache )GlobalManager.getInstance().get( ServiceDef.MASTER_DB_CACHE_MANAGER ) ;
                if( cache != null ) {
                    int len = updateTables.size() ;
                    for( int i = 0 ; i < len ; i ++ ) {
                        cache.clearCache( updateTables.get( i ) ) ;
                    }
                }
                updateTables.clear() ;
                updateTables = null ;
            }
        }
    }
    
    /**
     * ロールバック処理.
     * <BR><BR>
     * ロールバック処理を実施します.
     * <BR>
     * @exception Exception 例外.
     */
    public void rollback() throws Exception {
        if( createFlag == true ) {
            if( record != null ) {
                record.rollback() ;
            }
        }
    }
    
    /**
     * 更新／削除テーブルを追加.
     * <BR><BR>
     * @param name 更新／削除テーブルを追加します.
     */
    public void setUpdateOrDeleteTable( String name ) {
        if( name == null || ( name = name.trim() ).length() <= 0 ) {
            return ;
        }
        if( updateTables == null ) {
            updateTables = new ArrayList<String>() ;
        }
        updateTables.add( name ) ;
    }
    
    /**
     * 生成チェック.
     * <BR><BR>
     * 生成チェック処理を行います.
     * <BR>
     * @return boolean [true]の場合生成されています.
     */
    public boolean isCreate() {
        return createFlag ;
    }
    
    /**
     * オブジェクトタイプを取得.
     * <BR><BR>
     * オブジェクトタイプを取得します.
     * <BR>
     * @return boolean [true]の場合、実行可能オブジェクトです.
     */
    public boolean isType() {
        return true ;
    }
    
    /**
     * スクリプトコンテキストを取得.
     * <BR><BR>
     * スクリプトコンテキストを取得します.
     * <BR>
     * @return SeaBassContext 対象のスクリプトコンテキストが返されます.
     */
    public SeaBassContext getScriptContext() {
        return context ;
    }
    
    /**
     * コネクションオブジェクトを取得.
     * <BR><BR>
     * コネクションオブジェクトを取得します.
     * <BR>
     * @return Record コネクションオブジェクトが返されます.
     */
    public Record getRecord() {
        return record ;
    }
    
    /**
     * 指定モデル名から、メタデータを取得.
     * <BR><BR>
     * 指定モデル名から、メタデータを取得します.
     * <BR>
     * @param model 対象のモデル名を設定します.
     * @return MetaColumn 対象のメタデータが返されます.
     * @exception Exception 例外.
     */
    public MetaColumn getMeta( String model )
        throws Exception {
        checkCreate() ;
        return MetaFactory.getInstance().getMetaColumn( true,model ) ;
    }
    
    /**
     * データを保存.
     * <BR><BR>
     * スクリプトから渡されたデータを保存します.
     * <BR>
     * @param model 対象のモデル名を設定します.
     * @param sequence [true]の場合、IDがシーケンスで処理されます.
     * @param args 対象のデータを設定します.
     * @return Long 更新されたIDが返されます.
     * @exception Exception 例外.
     */
    public Long save( String model,boolean sequence,Map<String,Object> args )
        throws Exception {
        checkCreate() ;
        DaoFactory factory = getDaoFactory() ;
        Dao dao = factory.getDao( record,model ) ;
        if( !dao.save( record,sequence,args ) ) {
            return new Long( -1L ) ;
        }
        Object val = args.get( "id" ) ;
        if( val instanceof Long ) {
            return (Long)val ;
        }
        else if( val instanceof Short ||
            val instanceof Integer ||
            val instanceof Float ||
            val instanceof Double ) {
            return DaoUtil.convertId( val ) ;
        }
        else if( val instanceof String && StringUtil.isNumber( (String)val ) ) {
            return DaoUtil.convertId( val ) ;
        }
        return null ;
    }
    
    /**
     * データをアップデート.
     * <BR><BR>
     * スクリプトから渡されたデータを更新します.
     * <BR>
     * @param model 対象のモデル名を設定します.
     * @param where 対象の条件を設定します.
     * @param args 対象のデータを設定します.
     * @exception Exception 例外.
     */
    public void update( String model,String where,List<Object> args )
        throws Exception {
        checkCreate() ;
        DaoFactory factory = getDaoFactory() ;
        Dao dao = factory.getDao( record,model ) ;
        dao.update( record,where,args ) ;
    }
    
    /**
     * データを削除.
     * <BR><BR>
     * スクリプトから渡されたデータを削除します.
     * <BR>
     * @param model 対象のモデル名を設定します.
     * @param srcJoin 連結テーブル内容を設定します.
     * @param where 対象の条件を設定します.
     * @param args 対象のデータを設定します.
     * @exception Exception 例外.
     */
    public void remove( String model,Map<String,String> srcJoin,String where,List<Object> args )
        throws Exception {
        checkCreate() ;
        DaoFactory factory = getDaoFactory() ;
        Dao dao = factory.getDao( record,model ) ;
        dao.remove( record,srcJoin,where,args ) ;
    }
    
    /**
     * データを削除.
     * <BR><BR>
     * スクリプトから渡されたデータを削除します.
     * <BR>
     * @param model 対象のモデル名を設定します.
     * @param srcJoin 連結テーブル内容を設定します.
     * @param args 対象のデータを設定します.
     * @exception Exception 例外.
     */
    public void removeData( String model,Map<String,String> srcJoin,Map<String,Object> args )
        throws Exception {
        checkCreate() ;
        DaoFactory factory = getDaoFactory() ;
        Dao dao = factory.getDao( record,model ) ;
        dao.remove( record,srcJoin,args ) ;
    }
    
    /**
     * データを取得.
     * <BR><BR>
     * データを取得します.
     * <BR>
     * @param model 対象のモデル名を設定します.
     * @param join 連結テーブル内容を設定します.
     * @param srcJoin 対象のsrcJoin条件を設定します.
     * @param where 対象の条件を設定します.
     * @param args 条件に対するパラメータを設定します.
     * @return List<Map<String,Object>> 処理結果が返されます.
     * @exception Exception 例外.
     */
    public List<Map<String,Object>> find( String model,Map<String,String> join,
        Map<String,Object> srcJoin,String where,List<Object> args )
        throws Exception {
        checkCreate() ;
        DaoFactory factory = getDaoFactory() ;
        Dao dao = factory.getDao( record,model ) ;
        return dao.find( record,join,srcJoin,where,args ) ;
    }
    
    /**
     * データを取得.
     * <BR><BR>
     * データを取得します.
     * <BR>
     * @param model 対象のモデル名を設定します.
     * @param join 連結テーブル内容を設定します.
     * @param srcJoin 対象のsrcJoin条件を設定します.
     * @param offset 対象のオフセット値を設定します.
     * @param limit 対象のリミット値を設定します.
     * @param where 対象の条件を設定します.
     * @param args 条件に対するパラメータを設定します.
     * @param options 条件に対するオプションを設定します.
     * @return List<Map<String,Object>> 処理結果が返されます.
     * @exception Exception 例外.
     */
    public List<Map<String,Object>> limit( String model,Map<String,String> join,
        Map<String,Object> srcJoin,int offset,int limit,String where,List<Object> args )
        throws Exception {
        checkCreate() ;
        DaoFactory factory = getDaoFactory() ;
        Dao dao = factory.getDao( record,model ) ;
        return dao.limit( record,join,srcJoin,offset,limit,where,args ) ;
    }
    
    /**
     * データを取得.
     * <BR><BR>
     * データを取得します.
     * <BR>
     * @param model 対象のモデル名を設定します.
     * @param join 連結テーブル内容を設定します.
     * @param srcJoin 対象のsrcJoin条件を設定します.
     * @param offset 対象のオフセット値を設定します.
     * @param limit 対象のリミット値を設定します.
     * @param where 対象の条件を設定します.
     * @param args 条件に対するパラメータを設定します.
     * @param options 条件に対するオプションを設定します.
     * @return List<Map<String,Object>> 処理結果が返されます.
     * @exception Exception 例外.
     */
    public List<Map<String,Object>> first( String model,Map<String,String> join,
        Map<String,Object> srcJoin,String where,List<Object> args )
        throws Exception {
        checkCreate() ;
        DaoFactory factory = getDaoFactory() ;
        Dao dao = factory.getDao( record,model ) ;
        return dao.first( record,join,srcJoin,where,args ) ;
    }
    
    /**
     * １行データを取得.
     * @param model 対象のモデル名を設定します.
     * @param where 対象の条件を設定します.
     * @param args 条件に対するパラメータを設定します.
     * @return ResultLine 1行取得結果が返されます.
     * @exception Exception 例外.
     */
    public ResultLine readLine( String model,String where,List<Object> args )
        throws Exception {
        DaoFactory factory = getDaoFactory() ;
        Dao dao = factory.getDao( record,model ) ;
        return dao.readLine( record,where,args ) ;
    }
    
    /**
     * データ数を取得.
     * <BR><BR>
     * データ数を取得します.
     * <BR>
     * @param model 対象のモデル名を設定します.
     * @param where 対象の条件を設定します.
     * @param args 条件に対するパラメータを設定します.
     * @return int 処理結果が返されます.
     * @exception Exception 例外.
     */
    public int count( String model,String where,List<Object> args )
        throws Exception {
        checkCreate() ;
        DaoFactory factory = getDaoFactory() ;
        Dao dao = factory.getDao( record,model ) ;
        return dao.count( record,where,args ) ;
    }
    
    /**
     * 参照系SQL実行.
     * <BR><BR>
     * 参照系SQL実行します.
     * <BR>
     * @param sql 対象の条件を設定します.
     * @param args 条件に対するパラメータを設定します.
     * @param offset 対象のオフセット値を設定します.
     * @param limit 対象のリミット値を設定します.
     * @return List<Map<String,Object>> 処理結果が返されます.
     * @exception Exception 例外.
     */
    public List<Map<String,Object>> selectSQL( String sql,List<Object> args,int offset,int limit )
        throws Exception {
        String modelName = DbUtil.getSelectSqlByModelName( sql ) ;
        checkCreate() ;
        DaoFactory factory = getDaoFactory() ;
        Dao dao = factory.getDao( record,modelName ) ;
        return dao.selectSQL( record,modelName,sql,offset,limit,args ) ;
    }
    
    /**
     * 更新系SQL実行.
     * <BR><BR>
     * 更新系のSQL文を実行します.
     * <BR>
     * @param sql 対象の更新系SQL文を設定します.
     * @return int 件数が返されます.
     * @exception Exception 例外.
     */
    public int executionUpdate( String sql ) throws Exception {
        checkCreate() ;
        return record.executeUpdate( sql ) ;
    }
    
    /**
     * 楽観的ロックチェック.
     * <br><br>
     * 楽観的ロック条件を取得します.
     * @param model 対象のモデル名を設定します.
     * @param whereParam ロック条件を取得する条件をパラメータ指定します.
     * @param value 比較対象の１テーブル行情報を設定します.
     * @exception Exception 例外.
     */
    public void optimisticLock( String model,Map<String,Object> whereParam,Map<String,Object> value )
        throws Exception {
        checkCreate() ;
        DaoFactory factory = getDaoFactory() ;
        Dao dao = factory.getDao( record,model ) ;
        dao.optimisticLock( record,model,whereParam,value ) ;
    }
    
    /**
     * デバッグモードを取得.
     * <BR>
     * @return [true]の場合、デバッグは有効です.
     */
    public boolean isDebug() {
        if( record != null ) {
            return record.isDebug() ;
        }
        return false ;
    }
    
    /**
     * DaoFactoryを取得.
     */
    private DaoFactory getDaoFactory() {
        return ( DaoFactory )GlobalManager.getInstance().get( ServiceDef.DAO_FACTORY ) ;
    }
    
    /**
     * DBが生成されているかチェック.
     */
    private void checkCreate() throws Exception {
        if( isCreate() == false ) {
            throw new IOException( "BaseModelは生成されていません" ) ;
        }
    }
    
    /**
     * コミットチェック.
     */
    private void checkCommit() throws Exception {
        if( record == null ) {
            throw new IOException( "レコードは利用できません" ) ;
        }
        if( record.isUse() == false ) {
            throw new IOException( "レコード処理エラー" ) ;
        }
    }
}

