package org.maachang.seabass.httpd.engine.script;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.maachang.util.FileUtil;
import org.maachang.util.StringUtil;

/**
 * TreeList形式の内容を生成.
 * この内容は、Javascript(SeaBass)のTreeオブジェクト表現を利用する場合のサーバ実行です.
 * 
 * @version 2009/02/17
 * @author masahito suzuki
 * @since SeaBass 1.0.0.
 */
public abstract class ServerTreeList {
    private ServerTreeList() {}
    
    /**
     * 現在のディレクトリ状況.
     */
    private static final Map<String,String> dirs = new ConcurrentHashMap<String,String>() ;
    
    /**
     * 指定ディレクトリ以下の内容をディレクトリリスト形式で取得.
     * @param mode [true]を指定した場合、ディレクトリ＋ファイル形式の内容で取得します.
     * @param base 絶対パス位置を設定します.
     * @param dir 対象のディレクトリ名を設定します.
     * @param head 除外対象のヘッダ名を設定します.
     * @param foot 除外対象のフッダ名を設定します.
     * @return String ディレクトリリストが返されます.
     * @exception Exception 例外.
     */
    public static final String get( boolean mode,String base,String dir,String head,String foot )
        throws Exception {
        return get( mode,base,dir,head,foot,null ) ;
    }
    
    /**
     * 指定ディレクトリ以下の内容をディレクトリリスト形式で取得.
     * @param mode [true]を指定した場合、ディレクトリ＋ファイル形式の内容で取得します.
     * @param base 絶対パス位置を設定します.
     * @param dir 対象のディレクトリ名を設定します.
     * @param head 除外対象のヘッダ名を設定します.
     * @param foot 除外対象のフッダ名を設定します.
     * @param full 除外対象の非表示名を設定します.
     * @return String ディレクトリリストが返されます.
     * @exception Exception 例外.
     */
    public static final String get( boolean mode,String base,String dir,String head,String foot,String full )
        throws Exception {
        if( dir == null ) {
            dir = base ;
        }
        if( !FileUtil.isDirExists( dir ) ) {
            throw new IllegalArgumentException( "指定内容[" + dir + "]はディレクトリではありません" ) ;
        }
        dir = FileUtil.getFullPath( dir ) ;
        base = FileUtil.getFullPath( base ) ;
        StringBuilder buf = new StringBuilder() ;
        buf.append( "?" ).append( FileUtil.getFileName( dir ) ).append( "*0*?" ) ;
        if( !dir.endsWith( "/" ) ) {
            dir += "/" ;
        }
        if( !base.endsWith( "/" ) ) {
            base += "/" ;
        }
        List<String> hArray = null ;
        if( head != null && head.length() > 0 ) {
            hArray = new ArrayList<String>() ;
            if( head.indexOf( "|" ) != -1 ) {
                StringUtil.cutString( hArray,head,"|" ) ;
            }
            else {
                hArray.add( head ) ;
            }
        }
        List<String> fArray = null ;
        if( foot != null && foot.length() > 0 ) {
            fArray = new ArrayList<String>() ;
            if( foot.indexOf( "|" ) != -1 ) {
                StringUtil.cutString( fArray,foot,"|" ) ;
            }
            else {
                fArray.add( foot ) ;
            }
        }
        List<String> fullArray = null ;
        if( full != null && full.length() > 0 ) {
            fullArray = new ArrayList<String>() ;
            if( full.indexOf( "|" ) != -1 ) {
                StringUtil.cutString( fullArray,full,"|" ) ;
            }
            else {
                fullArray.add( full ) ;
            }
        }
        next( mode,buf,base,dir,hArray,fArray,fullArray ) ;
        return buf.toString() ;
    }
    
    private static final boolean next( boolean mode,StringBuilder buf,String base,String dir,List<String> head,List<String> foot,List<String> full )
        throws Exception {
        int len ;
        String[] list = FileUtil.getList( dir ) ;
        if( !dir.endsWith( "/" ) ) {
            dir += "/" ;
        }
        Arrays.sort( list ) ;
        int cnt = 0 ;
        if( list != null && ( len = list.length ) > 0 ) {
            for( int i = 0 ; i < len ; i ++ ) {
                String s = list[ i ] ;
                if( ( head != null && FileUtil.isListWith( true,s,head ) ) || ( foot != null && FileUtil.isListWith( false,s,foot ) ) ||
                    ( full != null && FileUtil.isListEquals( s,full ) ) ) {
                    continue ;
                }
                s = dir + s ;
                if( !FileUtil.isDirExists( s ) ) {
                    continue ;
                }
                else {
                    cnt ++ ;
                    buf.append( "|?" ).append( FileUtil.getFileName( s ) ).append( "*" ).
                        append( FileUtil.getListLength( s,head,foot,full ) ).append( "*?" ) ;
                    if( !s.endsWith( "/" ) ) {
                        s += "/" ;
                    }
                    next( mode,buf,base,s,head,foot,full ) ;
                    buf.append( "|*" ) ;
                    list[ i ] = null ;
                }
            }
            if( mode ) {
                for( int i = 0 ; i < len ; i ++ ) {
                    String s = list[ i ] ;
                    if( s == null || ( head != null && FileUtil.isListWith( true,s,head ) ) || ( foot != null && FileUtil.isListWith( false,s,foot ) ) ||
                        ( full != null && FileUtil.isListEquals( s,full ) ) ) {
                        continue ;
                    }
                    s = dir + s ;
                    if( !FileUtil.isDirExists( s ) ) {
                        cnt ++ ;
                        String linkPath = "/" + s.substring( base.length() ) ;
                        buf.append( "|\\" ).append( FileUtil.getFileName( s ) ).append( "*" ).append( linkPath ).append( "*\\" ) ;
                    }
                    else {
                        continue ;
                    }
                }
            }
        }
        return ( cnt > 0 ) ;
    }
    
}
