package org.maachang.seabass.httpd.engine.script.cache;

import gnu.trove.THashSet;

import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.maachang.BaseMap;
import org.maachang.StringMap;
import org.maachang.atomic.AtomicObject;
import org.maachang.seabass.SeaBassMap;
import org.maachang.seabass.SeaBassScriptEngine;

/**
 * 基底API管理用マネージャ.
 * 
 * @version 2008/08/11
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
public class ApiManager extends BaseMap {
    
    /**
     * コアAPI用バインディングマップ.
     */
    private final AtomicObject<Map<String,Object>> coreApiMap = new AtomicObject<Map<String,Object>>() ;
    
    /**
     * Model用バインディングマップ.
     */
    private final AtomicObject<Map<String,Object>> modelMap = new AtomicObject<Map<String,Object>>() ;
    
    /**
     * ライブラリ用バインディングマップ.
     */
    private final AtomicObject<Map<String,Object>> libMap = new AtomicObject<Map<String,Object>>() ;
    
    /**
     * スクリプトエンジン.
     */
    private final SeaBassScriptEngine engine = new SeaBassScriptEngine() ;
    
    /**
     * シングルトン.
     */
    private static final ApiManager SNGL = new ApiManager() ;
    
    /**
     * オブジェクトを取得します.
     * @return ApiManager オブジェクトが返されます.
     */
    public static final ApiManager getInstance() {
        return SNGL ;
    }
    
    /** コンストラクタ. */
    private ApiManager() {
    }
    
    /**
     * coreApiMap を取得.
     * <BR><BR>
     * @return coreApiMap が返されます.
     */
    public Map<String, Object> getCoreApiMap() {
        return coreApiMap.get() ;
    }
    
    /**
     * coreApiMap を設定.
     * <BR><BR>
     * @param coreApiMap coreApiMap を設定します.
     */
    public void setCoreApiMap(Map<String, Object> coreApiMap) {
        changeNoWrite( coreApiMap ) ;
        this.coreApiMap.set( coreApiMap ) ;
    }
    
    /**
     * libMap を取得.
     * <BR><BR>
     * @return libMap が返されます.
     */
    public Map<String, Object> getLibMap() {
        return libMap.get() ;
    }
    
    /**
     * libMap を設定.
     * <BR><BR>
     * @param libMap libMap を設定します.
     */
    public void setLibMap(Map<String, Object> libMap) {
        changeNoWrite( libMap ) ;
        this.libMap.set( libMap ) ;
    }
    
    /**
     * modelMap を取得.
     * <BR><BR>
     * @return modelMap が返されます.
     */
    public Map<String, Object> getModelMap() {
        return modelMap.get() ;
    }
    
    /**
     * modelMap を設定.
     * <BR><BR>
     * @param modelMap modelMap を設定します.
     */
    public void setModelMap(Map<String, Object> modelMap) {
        changeNoWrite( modelMap ) ;
        this.modelMap.set( modelMap ) ;
    }
    
    /**
     * スクリプトエンジンを取得.
     * @return ScriptEngine スクリプトエンジンが返されます.
     */
    public SeaBassScriptEngine getScriptEngine() {
        return engine ;
    }
    
    /**
     * APIマネージャ内に指定Key内容が存在するかチェック.
     * @param key 対象のキーを設定します.
     * @return boolean [true]の場合、内容が存在します.
     */
    public boolean containsKey(Object key) {
        if( key == null || ( key instanceof String ) == false ||
            ( ( String )key ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        Map<String, Object> coreApi = this.coreApiMap.get() ;
        if( coreApi != null && coreApi.containsKey(key) ) {
            return true ;
        }
        Map<String, Object> lib = this.libMap.get() ;
        if( lib != null && lib.containsKey(key) ) {
            return true ;
        }
        Map<String, Object> model = this.modelMap.get() ;
        if( model != null && model.containsKey(key) ) {
            return true ;
        }
        return false ;
    }
    
    /**
     * APIマネージャ内の指定Key内容を取得.
     * @param key 対象のキーを設定します.
     * @return Object 内容が返されます.
     */
    public Object get(Object key) {
        if( key == null || ( key instanceof String ) == false ||
            ( ( String )key ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        Map<String, Object> coreApi = this.coreApiMap.get() ;
        if( coreApi != null && coreApi.containsKey(key) ) {
            return coreApi.get( key ) ;
        }
        Map<String, Object> lib = this.libMap.get() ;
        if( lib != null && lib.containsKey(key) ) {
            return lib.get( key ) ;
        }
        Map<String, Object> model = this.modelMap.get() ;
        if( model != null && model.containsKey(key) ) {
            return model.get( key ) ;
        }
        return null ;
    }
    
    /**
     * Key一覧を取得.
     * @return Set Key一覧が返されます.
     */
    public Set keySet() {
        Set<Object> ret = new THashSet<Object>() ;
        getAllKey( ret ) ;
        return ret ;
    }
    
    /**
     * Value一覧を取得.
     * @return Set Value一覧が返されます.
     */
    public Collection values() {
        Set<Object> ret = new THashSet<Object>() ;
        getAllValues( ret ) ;
        return ret ;
    }
    
    /**
     * APIマネージャ内のキー一覧を取得.
     * @param set Key一覧をセットするためのSetを設定します.
     */
    public void getAllKey( Set<Object> set ) {
        Map<String, Object> lib = this.libMap.get() ;
        Map<String, Object> model = this.modelMap.get() ;
        Map<String, Object> coreApi = this.coreApiMap.get() ;
        
        if( coreApi != null && coreApi.size() > 0 ) {
            Iterator ita = coreApi.keySet().iterator() ;
            while( ita.hasNext() ) {
                set.add( ( String )ita.next() ) ;
            }
        }
        if( model != null && model.size() > 0 ) {
            Iterator ita = model.keySet().iterator() ;
            while( ita.hasNext() ) {
                set.add( ( String )ita.next() ) ;
            }
        }
        if( lib != null && lib.size() > 0 ) {
            Iterator ita = lib.keySet().iterator() ;
            while( ita.hasNext() ) {
                set.add( ( String )ita.next() ) ;
            }
        }
    }
    
    /**
     * APIマネージャ内のキー一覧を取得.
     * @param set Key一覧をセットするためのSetを設定します.
     */
    public void getAllValues( Set<Object> set ) {
        Map<String, Object> coreApi = this.coreApiMap.get() ;
        if( coreApi != null && coreApi.size() > 0 ) {
            Iterator ita = coreApi.values().iterator() ;
            while( ita.hasNext() ) {
                set.add( ( String )ita.next() ) ;
            }
        }
        Map<String, Object> model = this.modelMap.get() ;
        if( model != null && model.size() > 0 ) {
            Iterator ita = model.values().iterator() ;
            while( ita.hasNext() ) {
                set.add( ( String )ita.next() ) ;
            }
        }
        Map<String, Object> lib = this.libMap.get() ;
        if( lib != null && lib.size() > 0 ) {
            Iterator ita = lib.values().iterator() ;
            while( ita.hasNext() ) {
                set.add( ( String )ita.next() ) ;
            }
        }
    }
    
    /**
     * APIマネージャ内のキー数を取得.
     * @return int キー数が返されます.
     */
    public int size() {
        int ret = 0 ;
        Map<String, Object> coreApi = this.coreApiMap.get() ;
        if( coreApi != null ) {
            ret += coreApi.size() ;
        }
        Map<String, Object> model = this.modelMap.get() ;
        if( model != null ) {
            ret += model.size() ;
        }
        Map<String, Object> lib = this.libMap.get() ;
        if( lib != null ) {
            ret += lib.size() ;
        }
        return ret ;
    }
    
    /**
     * Map内のStringMapを書き込み不可にセット.
     */
    protected void changeNoWrite( Map<String,Object> map ) {
        if( map instanceof SeaBassMap ) {
            Iterator it = ((SeaBassMap)map).keys() ;
            while( it.hasNext() ) {
                Object key = it.next() ;
                Object o = map.get( key ) ;
                if( o instanceof StringMap ) {
                    ((StringMap)o).setNoSetting() ;
                }
            }
        }
    }
    
}

