package org.maachang.seabass.httpd.engine.script.cache;

import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.maachang.util.StringUtil;
import org.maachang.util.thread.LoopThread;

/**
 * Json情報をキャッシュ.
 * 
 * @version 2010/01/15
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
public class CacheJson {
    private static final CacheJson SNGL = new CacheJson() ;
    protected static final long TIMEOUT = 120000L ;
    
    /**
     * オブジェクトを取得.
     * @return CacheJson オブジェクトが返されます.
     */
    public static final CacheJson getInstance() {
        return SNGL ;
    }
    
    /** 格納オブジェクト. **/
    private final Map<String,CacheJsonChild> cache = new ConcurrentHashMap<String,CacheJsonChild>() ;
    
    /** 監視スレッド. **/
    private CacheJsonMonThread thread = null ;
    
    /** コンストラクタ **/
    private CacheJson() {
        CacheJsonMonThread t = null ;
        try {
            t = new CacheJsonMonThread( cache ) ;
        } catch( Exception e ) {
            t.stopThread() ;
            t = null ;
        }
        thread = t ;
    }
    
    /**
     * オブジェクト破棄.
     */
    public void destroy() {
        if( thread != null ) {
            thread.stopThread() ;
            thread = null ;
        }
    }
    
    /**
     * 情報取得.
     * @param key 対象のキーを設定します.
     * @return CacheJsonChild 子要素が返されます.
     */
    public CacheJsonChild get( String key ) {
        return cache.get( key ) ;
    }
    
    /**
     * 情報設定.
     * @param key 対象のキーを設定します.
     * @param child 子要素を設定します.
     */
    public void put( String key,CacheJsonChild child ) {
        cache.put( key,child ) ;
    }
    
    /**
     * 情報削除.
     * @param key 対象のキーを設定します.
     */
    public void remove( String key ) {
        cache.remove( key ) ;
    }
    
    /**
     * バイナリから、１行目のコメントを呼び出してパラメータ名を取得.
     * @param bin 対象のバイナリを設定します.
     * @return String[] ヘッダ内容が返されます.
     * @exception Exception 例外.
     */
    public List<String> jsonHeader( byte[] bin )
        throws Exception {
        if( bin == null || bin.length <= 0 ) {
            return null ;
        }
        int p = 0 ;
        int len = bin.length ;
        for( int i = 0 ; i < len ; i ++ ) {
            if( bin[ i ] == '\r' || bin[ i ] == '\n' ) {
                p = i ;
                break ;
            }
        }
        String h = new String( bin,0,p,"UTF8" ) ;
        h = StringUtil.trim( h ) ;
        if( !h.startsWith( "//" ) ) {
            return null ;
        }
        h = StringUtil.trim( h.substring( 2 ) ) ;
        return StringUtil.cutString( false,false,h,"\t 　|" ) ;
    }
    
    /**
     * 監視スレッド.
     */
    private static final class CacheJsonMonThread extends LoopThread {
        private Map<String,CacheJsonChild> cache = null ;
        CacheJsonMonThread( Map<String,CacheJsonChild> cache )
            throws Exception {
            this.cache = cache ;
            startThread() ;
        }
        protected void clear() {
        }
        protected void toException( Exception e ) {
        }
        protected boolean execution() throws Exception {
            if( isStop() ) {
                return false ;
            }
            if( cache.size() <= 0 ) {
                Thread.sleep( 250 ) ;
                return false ;
            }
            Iterator it = cache.keySet().iterator() ;
            String key ;
            CacheJsonChild ch ;
            while( it.hasNext() ) {
                if( isStop() ) {
                    return false ;
                }
                key = ( String )it.next() ;
                ch = cache.get( key ) ;
                if( ch.getLastAccessTime() + TIMEOUT <= System.currentTimeMillis() ) {
                    cache.remove( key ) ;
                }
                Thread.sleep( 50 ) ;
            }
            return true ;
        }
    }
}
