package org.maachang.seabass.httpd.engine.script.cache;

import java.util.Iterator;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.maachang.manager.GlobalManager;
import org.maachang.seabass.SeaBassContext;
import org.maachang.seabass.SeaBassException;
import org.maachang.seabass.SeaBassStackContext;
import org.maachang.seabass.ServiceDef;
import org.maachang.seabass.httpd.engine.script.SrcScript;
import org.maachang.seabass.httpd.engine.script.js.JsDef;
import org.maachang.util.ObjectArray;

/**
 * キャッシュスクリプト管理.
 * 
 * @version 2007/11/12
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
public class CacheScriptManager {
    
    /**
     * コアAPI用マネージャ.
     */
    private CoreApiManager coreApiManager = null ;
    
    /**
     * ModelAPI用マネージャ.
     */
    private ModelApiManager modelApiManager = null ;
    
    /**
     * ライブラリAPI用マネージャ.
     */
    private LibApiManager libApiManager = null ;
    
    /**
     * 自オブジェクト.
     */
    private static CacheScriptManager thisObject = null ;
    
    /**
     * キャッシュスクリプト初期化処理.
     */
    public static final void init()
        throws Exception {
        thisObject = new CacheScriptManager() ;
        thisObject.coreApiManager = new CoreApiManager() ;
        thisObject.modelApiManager = new ModelApiManager() ;
        thisObject.libApiManager = new LibApiManager() ;
        GlobalManager.getInstance().put( ServiceDef.SCRIPT_CACHE_MANAGER,thisObject ) ;
    }
    
    /**
     * キャッシュスクリプトマネージャを取得.
     * <BR><BR>
     * キャッシュスクリプトマネージャを取得します.
     * <BR>
     * @return CacheScriptManager キャッシュスクリプトマネージャが返されます.
     */
    public static final CacheScriptManager getInstance() {
        return thisObject ;
    }
    
    /**
     * コンストラクタ.
     */
    private CacheScriptManager() throws Exception {
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        this.destroy() ;
    }
    
    /**
     * オブジェクト破棄.
     * <BR><BR>
     * オブジェクトを破棄します.
     */
    protected void destroy() {
        this.coreApiManager = null ;
        this.modelApiManager = null ;
        this.libApiManager = null ;
    }
    
    /**
     * リロード処理.
     * <BR><BR>
     * リロード処理を実施します.
     * <BR>
     * @exception Exception 例外.
     */
    public void reload() throws Exception {
        this.modelApiManager.reload() ;
        this.libApiManager.reload() ;
    }
    
    /**
     * 指定内容をリロード.
     * <BR><BR>
     * 指定された内容だけをリロードします.
     * <BR>
     * @param name リロード対象の条件を設定します.<BR>
     *             [null]の場合、全体をリロードします.
     * @exception Exception 例外.
     */
    public void reload( String name ) throws Exception {
        if( name == null || ( name = name.trim() ).length() <= 0 ) {
            this.reload() ;
        }
        else {
            name = name.toLowerCase() ;
            if( "model".equals( name ) ) {
                this.modelApiManager.reload() ;
            }
            else if( "lib".equals( name ) ) {
                this.libApiManager.reload() ;
            }
            else {
                this.reload() ;
            }
        }
    }
    
    /**
     * スクリプトキャッシュを取得.
     * <BR><BR>
     * スクリプトキャッシュを取得します.
     * <BR>
     * @param name 対象のスクリプト名を設定します.
     * @return SeaBassContext 新しいコンテキストが返されます.
     * @exception Exception 例外.
     */
    public SeaBassContext script( String name ) throws Exception {
        this.modelApiManager.useScript() ;
        this.libApiManager.useScript() ;
        SeaBassContext ret = SeaBassStackContext.getInstance().getContext( ApiManager.getInstance() ) ;
        if( name != null ) {
            ret.setName( name ) ;
        }
        return ret ;
    }
    
    /**
     * スクリプトに対してデフォルトパッケージをセット.
     * @param script 対象のスクリプトを設定します.
     * @return String 付加されたスクリプトが返されます.
     * @exception Exception 例外.
     */
    public String insertPackage( String script )
        throws Exception {
        StringBuilder buf = new StringBuilder() ;
        //JsDef.pushDefaultPackage( buf ) ;
        return buf.append( script ).toString() ;
    }
     
    
    /**
     * スクリプト終了処理.
     * @param context 対象のコンテキストを設定します.
     */
    public void exitScript( SeaBassContext context ) {
        SeaBassStackContext.getInstance().releaseContext( context ) ;
    }
    
    /**
     * 指定パスのソーススクリプトを取得.
     * <BR><BR>
     * 指定パスのソーススクリプトを取得します.
     * <BR>
     * @param path 対象のパス名を設定します.
     * @return SrcScript 対象のソーススクリプトが返されます.
     */
    public SrcScript getSrcScript( String path ) {
        if( path == null || ( path = path.trim() ).length() <= 0 ) {
            return null ;
        }
        //path = path.toLowerCase() ;
        if( this.coreApiManager.isCache( path ) ) {
            return coreApiManager.getSrcScript( path ) ;
        }
        if( this.modelApiManager.isCache( path ) ) {
            modelApiManager.setReloadFlag( true ) ;
            return modelApiManager.getSrcScript() ;
        }
        if( this.libApiManager.isCache( path ) ) {
            libApiManager.setReloadFlag( true ) ;
            return libApiManager.getSrcScript() ;
        }
        return null ;
    }
    
    /**
     * 指定パスの内容が存在するかチェック.
     * <BR><BR>
     * 指定パスの内容が存在するかチェックします.
     * <BR>
     * @param path 対象のパス名を設定します.
     * @return boolean [true]の場合、存在します.
     */
    public boolean isCache( String path ) {
        if( path == null || ( path = path.trim() ).length() <= 0 ) {
            return false ;
        }
        //path = path.toLowerCase() ;
        if( this.coreApiManager.isCache( path ) ) {
            return true ;
        }
        if( this.modelApiManager.isCache( path ) ) {
            return true ;
        }
        if( this.libApiManager.isCache( path ) ) {
            return true ;
        }
        return false ;
    }
    
    /**
     * 指定パスのディレクトリ名を取得.
     * <BR><BR>
     * 指定パスのディレクトリ名を取得します.
     * <BR>
     * @param path 対象のパス名を設定します.
     * @return String 対象のディレクトリが返されます.
     */
    public String getDirectory( String path ) {
        if( path == null || ( path = path.trim() ).length() <= 0 ) {
            return "" ;
        }
        if( this.coreApiManager.isCache( path ) ) {
            return "" ;
        }
        if( this.modelApiManager.isCache( path ) ) {
            return modelApiManager.getDirectory() ;
        }
        if( this.libApiManager.isCache( path ) ) {
            return libApiManager.getDirectory() ;
        }
        return "" ;
    }
    
    /**
     * 読み込み名一覧を取得.
     * <BR><BR>
     * 現在読み込んでいる名前一覧を取得します.
     * <BR>
     * @param context 対象のスクリプトコンテキストを設定します.
     * @return String[] 読み込み名一覧が返されます.
     */
    public String[] getNames( SeaBassContext context ) {
        ObjectArray lst = new ObjectArray() ;
        pushMapKeys( lst,context.getMemory() ) ;
        int len = lst.size() ;
        if( len > 0 ) {
            lst.sort() ;
            String[] ret = new String[ len ] ;
            System.arraycopy( lst.getObjects(),0,ret,0,len ) ;
            lst.clear() ;
            return ret ;
        }
        return null ;
    }
    
    /** エラー表示総行数 **/
    private static final int ERROR_LIST_LENGTH = 15 ;
    
    /** エラー表示オフセット値 **/
    private static final int ERROR_LIST_OFFSET = -5 ; 
    
    /** ライン情報 **/
    private static final String ERROR_LINE = "0000" ;
    
    /**
     * スクリプトパースエラーを表示.
     * @param log ログオブジェクトを設定します.
     * @param pe ParseExceptionを設定します.
     * @param src 対象のソーススクリプトを設定します.
     */
    public static final void viewParseException( Log log,SeaBassException se,SrcScript src ) {
        if( log.isDebugEnabled() ) {
            int line = se.getTopLine() ;
            log.error( "## scriptError(" + se.getTopScriptName() + ") [Line:" + line + "]" ) ;
            if( line > 0 ) {
                line += ERROR_LIST_OFFSET ;
                if( line <= 0 ) {
                    line = 0 ;
                }
                for( int i = 0 ; i < ERROR_LIST_LENGTH ; i ++ ) {
                    String one = src.get( line+i ) ;
                    if( one == null ) {
                        log.error( "## !!maxScriptLine:" + src.lines() ) ;
                        break ;
                    }
                    String ln = String.valueOf( line+i ) ;
                    ln = ERROR_LINE.substring( ln.length() ) + ln ;
                    log.error( "[" + ln + "]:" + one ) ;
                }
            }
        }
    }
    
    private void pushMapKeys( ObjectArray lst,Map map ) {
        if( map != null ) {
            Iterator ita = map.keySet().iterator() ;
            while( ita.hasNext() ) {
                String key = ( String )ita.next() ;
                if( key != null && key.length() > 0 ) {
                    lst.add( key ) ;
                }
            }
        }
    }
}

