package org.maachang.seabass.httpd.engine.script.cache;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.manager.GlobalManager;
import org.maachang.seabass.SeaBassContext;
import org.maachang.seabass.SeaBassException;
import org.maachang.seabass.SeaBassScriptEngine;
import org.maachang.seabass.SeaBassStackContext;
import org.maachang.seabass.ServiceDef;
import org.maachang.seabass.httpd.engine.script.ScriptDef;
import org.maachang.seabass.httpd.engine.script.SrcScript;
import org.maachang.seabass.httpd.engine.script.dao.MasterCache;
import org.maachang.seabass.httpd.engine.script.validate.ServerValidateCache;

/**
 * コアAPIマネージャ.
 * 
 * @version 2008/08/11
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
class CoreApiManager {
    
    /**
     * LOG.
     */
    private static final Log LOG = LogFactory.getLog( CoreApiManager.class ) ;
    
    /**
     * スクリプトエンジン名.
     */
    protected static final String ENGINE_NAME = "js" ;
    
    /**
     * 読み込みデフォルトスクリプト一覧.
     */
    private static final String[] LIST = {
        "org/maachang/seabass/httpd/engine/script/js/default.js",
        "org/maachang/seabass/httpd/engine/script/js/log.js",
        "org/maachang/seabass/httpd/engine/script/js/io.js",
        "org/maachang/seabass/httpd/engine/script/js/file_list.js",
        "org/maachang/seabass/httpd/engine/script/js/utils.js",
        "org/maachang/seabass/httpd/engine/script/js/html.js",
        "org/maachang/seabass/httpd/engine/script/js/cacheJson.js",
        "org/maachang/seabass/httpd/engine/script/js/validate.js",
        "org/maachang/seabass/httpd/engine/script/js/tag.js",
        "org/maachang/seabass/httpd/engine/script/js/validate_tag.js",
        "org/maachang/seabass/httpd/engine/script/js/validate_lite.js",
        "org/maachang/seabass/httpd/engine/script/js/master.js",
        "org/maachang/seabass/httpd/engine/script/js/search.js",
        "org/maachang/seabass/httpd/engine/script/js/input.js",
        "org/maachang/seabass/httpd/engine/script/js/view.js",
        "org/maachang/seabass/httpd/engine/script/js/record.js",
        "org/maachang/seabass/httpd/engine/script/js/base_model.js",
        "org/maachang/seabass/httpd/engine/script/js/dbm.js",
        "org/maachang/seabass/httpd/engine/script/js/http_client.js",
        "org/maachang/seabass/httpd/engine/script/js/mail.js",
        "org/maachang/seabass/httpd/engine/script/js/template.js",
        "org/maachang/seabass/httpd/engine/script/js/report.js",
        "org/maachang/seabass/httpd/engine/script/js/rpc.js",
        "org/maachang/seabass/httpd/engine/script/js/comet.js",
        "org/maachang/seabass/httpd/engine/script/js/mttml.js",
        "org/maachang/seabass/httpd/engine/script/js/startup.js"
    } ;
    
    /**
     * 読み込みデフォルトスクリプト名.
     */
    private static final String[] LIST_NAME = {
        "*base_script",
        "*log",
        "*io",
        "*file_list",
        "*utils",
        "*html",
        "*cacheJSON",
        "*validate",
        "*tag",
        "*validate_tag",
        "*validate_lite",
        "*master",
        "*search",
        "*input",
        "*view",
        "*record",
        "*base_model",
        "*dbm",
        "*http_client",
        "*mail",
        "*template",
        "*report",
        "*rpc",
        "*comet",
        "*mttml",
        "*startup"
    } ;
    
    /**
     * スクリプト名接頭語.
     */
    private static final String HEAD = "*" ;
    
    /**
     * ソーススクリプト.
     */
    private final Map<String,SrcScript> srcScripts = new ConcurrentHashMap<String,SrcScript>() ;
    
    /**
     * コンストラクタ.
     */
    public CoreApiManager() throws Exception {
        try {
            loadScripts() ;
        } catch( Exception e ) {
            LOG.error( "## [coreApi] loadScriptError",e ) ;
            throw e ;
        }
    }
    
    /**
     * 指定パスのソーススクリプトを取得.
     * <BR><BR>
     * 指定パスのソーススクリプトを取得します.
     * <BR>
     * @param path 対象のパス名を設定します.
     * @return SrcScript 対象のソーススクリプトが返されます.
     */
    public SrcScript getSrcScript( String path ) {
        if( isCache( path ) == false ) {
            return null ;
        }
        return srcScripts.get( path ) ;
    }
    
    /**
     * 指定パスがこのキャッシュ内の条件であるかチェック.
     * <BR><BR>
     * 指定パスがこのキャッシュ内の条件であるかチェックします.
     * <BR>
     * @param path チェック対象のパスを設定します.
     * @return boolean [true]の場合はこのキャッシュ内の条件です.
     */
    public boolean isCache( String path ) {
        if( path == null || ( path = path.trim() ).length() <= 0 ||
            path.startsWith( HEAD ) == false ) {
            return false ;
        }
        return true ;
    }
    
    /**
     * スクリプトをロードして実行.
     */
    private void loadScripts() throws Exception {
        SeaBassContext context = null ;
        int len = LIST.length ;
        SrcScript currentSrcScript = null ;
        try {
            // コアAPIを実行して登録する.
            MasterCache masterCache = new MasterCache() ;
            ServerValidateCache validateCache = new ServerValidateCache() ;
            Map<String,Object> shared = new ConcurrentHashMap<String,Object>() ;
            Map<String,Object> lockShared = new ConcurrentHashMap<String,Object>() ;
            Lock lock = new ReentrantLock() ;
            GlobalManager.getInstance().put( ServiceDef.MASTER_DB_CACHE_MANAGER,masterCache ) ;
            GlobalManager.getInstance().put( ServiceDef.VALIDATE_CACHE_MANAGER,validateCache ) ;
            GlobalManager.getInstance().put( ServiceDef.SHARED_MEMORY_MANAGER,shared ) ;
            GlobalManager.getInstance().put( ServiceDef.SCRIPT_SYSTEM_LOCK,lock ) ;
            GlobalManager.getInstance().put( ServiceDef.SCRIPT_LOCK_MANAGER,lockShared ) ;
            context = SeaBassStackContext.getInstance().getContext( null ) ;
            context.put( ScriptDef.MANAGER,GlobalManager.getInstance() ) ;
            context.put( ScriptDef.SCRIPT_CACHE_DB,masterCache ) ;
            context.put( ScriptDef.SCRIPT_SHARED,shared ) ;
            
            // ScrpitEngine実行.
            SeaBassScriptEngine engine = ApiManager.getInstance().getScriptEngine() ;
            this.srcScripts.clear() ;
            for( int i = 0 ; i < len ; i ++ ) {
                String currentScriptName = LIST_NAME[ i ] ;
                if( LOG.isDebugEnabled() ) {
                    LOG.debug( ">read[coreApi] - " + currentScriptName ) ;
                }
                String script = ScriptDef.getScriptByResource( LIST[ i ] ) ;
                script = CacheScriptManager.getInstance().insertPackage( script ) ;
                currentSrcScript = new SrcScript( script ) ;
                this.srcScripts.put( currentScriptName,currentSrcScript ) ;
                context.setName( currentScriptName ) ;
                engine.eval( script,context ) ;
            }
            
            // APIマネージャに実行結果を登録.
            Map<String,Object> coreApi = context.getMemory().getMap() ;
            context.getMemory().recreate() ;
            ApiManager.getInstance().setCoreApiMap( coreApi ) ;
            if( LOG.isDebugEnabled() ) {
                LOG.debug( "** coreApi-cacheを読み込み" ) ;
            }
        } catch( SeaBassException se ) {
            if( currentSrcScript != null ) {
                CacheScriptManager.viewParseException( LOG,se,currentSrcScript ) ;
            }
            throw se ;
        } catch( Exception e ) {
            throw e ;
        } finally {
            SeaBassStackContext.getInstance().releaseContext( context ) ;
        }
    }
}

