package org.maachang.seabass.httpd.engine.script.cron;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.conf.Config;
import org.maachang.conf.ReadIni;
import org.maachang.seabass.SeaBassDef;
import org.maachang.util.FileUtil;
import org.maachang.util.StringUtil;
import org.maachang.util.atomic.AtomicLONG;

/**
 * cron.conf用読み込みファイル.
 *
 * @version 2008/07/01
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
public class CronConfig {
    
    /**
     * LOG.
     */
    private static final Log LOG = LogFactory.getLog( CronConfig.class ) ;
    
    /**
     * 読み込みファイル名.
     */
    private static final String READ_FILE = SeaBassDef.DIRECTORY_CONFIG+"cron.conf" ;
    
    /**
     * チェック間隔.
     */
    private static final long CHECK_TIMING = 5000L ;
    
    /**
     * キャラクタセット.
     */
    private static final String CHARSET = "UTF8" ;
    
    /**
     * 1Cron区切り長.
     */
    private static final int CRON_LINE = 6 ;
    
    /**
     * 読み込み対象データ.
     */
    private Config iniParams = null ;
    
    /**
     * 対象ファイル.
     */
    private String fileName = null ;
    
    /**
     * 最終更新日.
     */
    private long lastUpdate = -1L ;
    
    /**
     * 最終確認時間.
     */
    private long lastCheckTime = -1L ;
    
    /**
     * 世代ID.
     */
    private final AtomicLONG id = new AtomicLONG( 0L ) ;
    
    /**
     * CronId.
     */
    private final AtomicLONG cronId = new AtomicLONG( 0L ) ;
    
    /**
     * Cronスケジュール情報.
     */
    private final ConcurrentLinkedQueue<CronBean> scheduleQueue = new ConcurrentLinkedQueue<CronBean>() ;
    
    /**
     * Cron実行中情報.
     */
    private final ConcurrentHashMap<Long,CronBean> execMap = new ConcurrentHashMap<Long,CronBean>() ;
    
    /**
     * コンストラクタ.
     */
    public CronConfig() {
        
    }
    
    /**
     * オープン.
     * <BR><BR>
     * オープンします.
     * <BR>
     * @exception Exception 例外.
     */
    public void open() throws Exception {
        this.open( READ_FILE ) ;
    }
    
    /**
     * オープン.
     * <BR><BR>
     * オープンします.
     * <BR>
     * @param name 対象のファイル名を設定します.
     * @exception Exception 例外.
     */
    public void open( String name ) throws Exception {
        this.fileName = FileUtil.getFullPath( name ) ;
        try {
            this.updateToReload() ;
        } catch( Exception e ) {
            this.close() ;
            throw e ;
        } finally {
        }
    }
    
    /**
     * クローズ.
     * <BR><BR>
     * クローズ処理.
     */
    public void close() {
        scheduleQueue.clear() ;
        execMap.clear() ;
        iniParams = null ;
        fileName = null ;
        lastUpdate = -1L ;
        lastCheckTime = -1L ;
    }
    
    /**
     * 文字列に変換.
     */
    public String toString() {
        if( iniParams != null ) {
            return iniParams.toString() ;
        }
        return "null" ;
    }
    
    /**
     * Config世代IDを取得.
     * @return long Config世代IDが返されます.
     */
    public long getConfigId() {
        return cronId.get() ;
    }
    
    /**
     * スケジュール対象のCronBeanを取得.
     * @return CronBean 対象のCronBean情報が返されます.
     */
    public CronBean schedulePoll() {
        return scheduleQueue.poll() ;
    }
    
    /**
     * スケジュール対象のCronBeanをスケジュールをセット.
     * @param bean CronBeanをスケジュールに登録します.
     * @return boolean [true]の場合、正常に登録できました.
     */
    public boolean scheduleOffer( CronBean bean ) {
        if( bean != null && bean.getConfigId() == id.get() ) {
            scheduleQueue.offer( bean ) ;
            return true ;
        }
        return false ;
    }
    
    /**
     * 実行するCronBeanが存在するかチェック.
     * @param bean 対象のBeanを設定します.
     * @return boolean [true]の場合、情報は存在します.
     */
    public boolean isExecution( CronBean bean ) {
        if( bean != null && bean.getConfigId() == id.get() ) {
            return execMap.containsKey( bean.getCronId() ) ;
        }
        return false ;
    }
    
    /**
     * 実行されているCronBeanを取得.
     * @param id 対象のCronIdを設定します.
     * @return CronBean 対象のCronBeanが返されます.
     */
    public CronBean getExecution( Long id ) {
        return execMap.get( id ) ;
    }
    
    /**
     * 実行するCronBeanをセット.
     * @param bean 対象のBeanを設定します.
     */
    public void addExecution( CronBean bean ) {
        if( bean != null && bean.getConfigId() == id.get() ) {
            execMap.put( bean.getCronId(),bean ) ;
        }
    }
    
    /**
     * 実行完了したCronBeanを削除.
     * @param bean 対象のBeanを設定します.
     */
    public void removeExecution( CronBean bean ) {
        if( bean != null ) {
            execMap.remove( bean.getCronId() ) ;
        }
    }
    
    /**
     * 実行中のIteratorを取得.
     * @return Iterator Iteratorが返されます.
     */
    public Iterator executions() {
        return execMap.keySet().iterator() ;
    }
    
    /**
     * 更新時には、Reload処理を行います.
     */
    protected boolean updateToReload() {
        boolean ret = false ;
        if( lastCheckTime + CHECK_TIMING <= System.currentTimeMillis() ) {
            try {
                long tm = FileUtil.getLastTime( fileName ) ;
                if( lastUpdate == -1 || lastUpdate != tm ) {
                    LOG.info( "* * * Cronコンフィグ読み込み開始" ) ;
                    this.reload() ;
                    this.createConfigByCronBeans() ;
                    LOG.info( "* * * Cronコンフィグ読み込み正常終了 " + scheduleQueue.size() + "件" ) ;
                    ret = true ;
                }
            } catch( Exception e ) {
                LOG.error( "* * * Cronコンフィグ読み込み失敗",e ) ;
            }
            lastCheckTime = System.currentTimeMillis() ;
        }
        return ret ;
    }
    
    /**
     * 再読み込み.
     */
    private final void reload()
        throws Exception {
        BufferedReader buf = null ;
        if( FileUtil.isFileExists( fileName ) == true ) {
            try {
                buf = new BufferedReader(
                    new InputStreamReader(
                        new FileInputStream( fileName ),CHARSET ) ) ;
                Config iniParams = new Config() ;
                ReadIni.analisys( iniParams,buf ) ;
                buf.close() ;
                buf = null ;
                long last = FileUtil.getLastTime( fileName ) ;
                // 読み込みデータを設定.
                this.iniParams = iniParams ;
                this.lastUpdate = last ;
            } catch( Exception e ) {
                throw e ;
            } finally {
                if( buf != null ) {
                    try {
                        buf.close() ;
                    } catch( Exception e ) {
                    }
                }
            }
        }
    }
    
    /**
     * Cron実行内容を生成.
     */
    private final void createConfigByCronBeans() throws Exception {
        scheduleQueue.clear() ;
        if( iniParams == null || iniParams.isSection( "cron-tab" ) == false ) {
            return ;
        }
        long thisId = id.inc() ;
        int len = iniParams.size( "cron-tab","cron" ) ;
        for( int i = 0 ; i < len ; i ++ ) {
            String s = iniParams.get( "cron-tab","cron",i ) ;
            //List<String> one = StringUtil.getCsv( false,s,"," ) ;
            List<String> one = StringUtil.getCsv( false,s," " ) ;
            if( one == null ) {
                LOG.warn( "Cron内容[no" + (i+1) + "]の解析に失敗しました" ) ;
                continue ;
            }
            int lenJ = one.size() ;
            if( lenJ != CRON_LINE ) {
                LOG.warn( "Cron内容[no" + (i+1) + " data:" + s + "]のパラメータ条件は不正です("+lenJ+"/"+CRON_LINE+")" ) ;
                continue ;
            }
            CronBean bean = new CronBean( thisId,cronId.inc() ) ;
            try {
                for( int j = 0 ; j < lenJ ; j ++ ) {
                    String ss = one.get( j ) ;
                    if( ss == null || ( ss = ss.trim() ).length() <= 0 || "*".equals( ss ) ) {
                        continue ;
                    }
                    if( ss.startsWith( "(" ) && ss.endsWith( ")" ) ) {
                        String tm = ss.substring( 1,ss.length() -1 ) ;
                        tm = tm.trim() ;
                        int cd = Integer.parseInt( tm ) ;
                        if( cd <= 0 ) {
                            cd = -1 ;
                        }
                        switch( j ) {
                            case 0 :
                                bean.setMinuteTm( cd ) ;
                                break ;
                            case 1 :
                                bean.setHourTm( cd ) ;
                                break ;
                            case 2 :
                                bean.setDateTm( cd ) ;
                                break ;
                            case 3 :
                            case 4 :
                            case 5 :
                                bean = null ;
                                LOG.warn( "Cron内容[no" + (i+1) + " data:" + s + "]の月,曜日,スクリプト内容は条件指定ができません" ) ;
                                break ;
                        }
                    }
                    else {
                        switch( j ) {
                            case 0 :
                                bean.setMinute( Integer.parseInt( ss ) ) ;
                                break ;
                            case 1 :
                                bean.setHour( Integer.parseInt( ss ) ) ;
                                break ;
                            case 2 :
                                bean.setDate( Integer.parseInt( ss ) ) ;
                                break ;
                            case 3 :
                                bean.setMonth( Integer.parseInt( ss ) ) ;
                                break ;
                            case 4 :
                                bean.setWeek( Integer.parseInt( ss ) ) ;
                                break ;
                            case 5 :
                                bean.setScript( ss ) ;
                                break ;
                        }
                    }
                    if( bean == null ) {
                        break ;
                    }
                }
            } catch( Exception e ) {
                LOG.warn( "Cron内容[no" + (i+1) + " data:" + s + "]の解析中にエラーが発生しました",e ) ;
                bean = null ;
            }
            if( bean != null ) {
                if( bean.getScript() == null ) {
                    LOG.warn( "Cron内容[no" + (i+1) + " data:" + s + "]に対して、実行スクリプトが設定されていません" ) ;
                    bean = null ;
                    continue ;
                }
                else if( bean.getMinute() <= -1 && bean.getMinuteTm() <= -1 &&
                    bean.getHour() <= -1 && bean.getHourTm() <= -1 &&
                    bean.getDate() <= -1 && bean.getDateTm() <= -1 &&
                    bean.getMonth() <= -1 && bean.getWeek() <= -1 ) {
                    LOG.warn( "Cron内容[no" + (i+1) + " data:" + s + "]に対して、実行可能条件が設定されていません" ) ;
                    bean = null ;
                    continue ;
                }
                scheduleQueue.offer( bean ) ;
            }
        }
    }
}

