package org.maachang.seabass.httpd.engine.script.dao;

import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.maachang.util.thread.LoopThread;

/**
 * マスターキャッシュオブジェクト.
 * 
 * @version 2008/07/14
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
public class MasterCache {
    
    /**
     * デフォルトキャッシュ破棄時間.
     */
    private static final long DEF_DESTROY_TIME = 300000L ;
    
    /**
     * 最小キャッシュ破棄時間.
     */
    private static final long MIN_DESTROY_TIME = 30000L ;
    
    /**
     * 最大キャッシュ破棄時間.
     */
    private static final long MAX_DESTROY_TIME = 2678400000L ;
    
    /**
     * キャッシュ格納領域.
     */
    private final Map<String,Object> cache = new ConcurrentHashMap<String,Object>() ;
    
    /**
     * キャッシュ監視オブジェクト.
     */
    private MasterCacheMonitor mon = null ;
    
    /**
     * コンストラクタ.
     * @exception Exception 例外.
     */
    public MasterCache() throws Exception {
        this( DEF_DESTROY_TIME ) ;
    }
    
    /**
     * コンストラクタ.
     * @param time キャッシュ破棄時間を設定します.
     * @exception Exception 例外.
     */
    public MasterCache( long time ) throws Exception {
        if( time <= -1 ) {
            time = MAX_DESTROY_TIME ;
        }
        else if( time <= MIN_DESTROY_TIME ) {
            time = MIN_DESTROY_TIME ;
        }
        else if( time >= MAX_DESTROY_TIME ) {
            time = MAX_DESTROY_TIME ;
        }
        this.mon = new MasterCacheMonitor( cache,time ) ;
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        this.destroy() ;
    }
    
    /**
     * オブジェクト破棄.
     */
    public void destroy() {
        if( mon != null ) {
            mon.destroy() ;
        }
        if( cache != null ) {
            cache.clear() ;
        }
        mon = null ;
    }
    
    /**
     * マスターキャッシュクリア.
     */
    public void clear() {
        cache.clear() ;
    }
    
    /**
     * キャッシュ監視を行わない情報を追加.
     * @param key 対象のキー名を設定します.
     * @param value 対象の要素を設定します.
     */
    public void put( String key,Object value ) {
        if( key == null || ( key = key.trim() ).length() <= 0 || value == null ) {
            return ;
        }
        cache.put( key,value ) ;
    }
    
    /**
     * キャッシュ監視を行う情報を追加.
     * @param key 対象のキー名を設定します.
     * @param value 対象の要素を設定します.
     */
    public void putByCache( String key,Object value ) {
        if( key == null || ( key = key.trim() ).length() <= 0 ) {
            return ;
        }
        cache.put( key,new MasterCacheChild( value ) ) ;
    }
    
    /**
     * キャッシュ内容をクリア.
     * @param key 対象のキー名を設定します.
     */
    public void clearCache( String key ) {
        if( key == null || ( key = key.trim() ).length() <= 0 ) {
            return ;
        }
        Object o = cache.get( key ) ;
        if( o instanceof MasterCacheChild ) {
            ( ( MasterCacheChild )o ).clearValue() ;
        }
    }
    
    /**
     * 情報を削除.
     * @param key 対象のキー名を設定します.
     */
    public void remove( String key ) {
        if( key == null || ( key = key.trim() ).length() <= 0 ) {
            return ;
        }
        cache.remove( key ) ;
    }
    
    /**
     * 情報を取得.
     * @param key 対象のキー名を設定します.
     */
    public Object get( String key ) {
        if( key == null || ( key = key.trim() ).length() <= 0 ) {
            return null ;
        }
        return cache.get( key ) ;
    }
    
    /**
     * 情報が存在するかチェック.
     * @param key 対象のキー名を設定します.
     * @return boolean [true]の場合、キー名の内容が存在します.
     */
    public boolean containsKey( String key ) {
        return cache.containsKey( key ) ;
    }
    
    /**
     * マスターキャッシュサイズを取得.
     * @return int マスターキャッシュサイズが返されます.
     */
    public int size() {
        return cache.size() ;
    }
    
    /**
     * マスターキャッシュ一覧を取得.
     * @return Object[] マスターキャッシュ一覧が返されます.
     */
    public Object[] getKeys() {
        return cache.keySet().toArray() ;
    }
}

/**
 * キャッシュ監視スレッド.
 */
class MasterCacheMonitor extends LoopThread {
    private long time = -1L ;
    private Map<String,Object> cache = null ;
    protected MasterCacheMonitor( Map<String,Object> cache,long time )
        throws Exception {
        this.time = time ;
        this.cache = cache ;
        this.startThread() ;
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        this.destroy() ;
    }
    
    /**
     * オブジェクト破棄.
     */
    public void destroy() {
        super.stopThread() ;
    }
    
    /**
     * オブジェクトクリア.
     */
    protected void clear() {
        cache = null ;
        time = -1 ;
    }
    
    /**
     * 実行処理.
     * @exception Exception 例外.
     */
    protected boolean execution() throws Exception {
        if( cache.size() <= 0 ) {
            return false ;
        }
        Iterator<String> it = cache.keySet().iterator() ;
        while( it.hasNext() ) {
            if( super.isStop() == true ) {
                break ;
            }
            String key = it.next() ;
            Object o = cache.get( key ) ;
            if( o == null || ( o instanceof MasterCacheChild ) == false ) {
                Thread.sleep( 30L ) ;
                continue ;
            }
            MasterCacheChild ch = ( MasterCacheChild )o ;
            if( ch.getLastAccessTime() != -1 &&
                ch.getLastAccessTime() + time <= System.currentTimeMillis() ) {
                ch.clearValue() ;
            }
            Thread.sleep( 30L ) ;
        }
        return true ;
    }
}
