package org.maachang.seabass.httpd.engine.script.scripts;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.maachang.seabass.SeaBassDef;
import org.maachang.seabass.httpd.HttpdErrorDef;
import org.maachang.seabass.httpd.HttpdMode;
import org.maachang.seabass.httpd.HttpdStateException;
import org.maachang.seabass.httpd.engine.script.Script;
import org.maachang.seabass.httpd.engine.script.ScriptDef;
import org.maachang.util.FileUtil;

/**
 * Application用スクリプトファクトリ.
 * 
 * @version 2007/08/24
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
public class ApplicationScriptFactory {
    
    /**
     * カレントディレクトリ.
     */
    private String currentDirectory = null ;
    
    /**
     * アプリケーションスクリプト格納用.
     */
    private final Map<String,Script> apps = new ConcurrentHashMap<String,Script>() ;
    
    /**
     * コンストラクタ.
     */
    private ApplicationScriptFactory() {
        
    }
    
    /**
     * アプリケーションスクリプトを全てクリア.
     */
    public void clear() {
        apps.clear() ;
    }
    
    /**
     * コンストラクタ.
     * <BR><BR>
     * 条件を指定してオブジェクトを生成します.
     * <BR>
     * @param directory 対象のカレントディレクトリを設定します.
     * @exception Exception 例外.
     */
    public ApplicationScriptFactory( String currentDirectory )
        throws Exception {
        currentDirectory = ScriptDef.trimCurrentDirectory( currentDirectory ) ;
        this.currentDirectory = currentDirectory ;
    }
    
    /**
     * アプリケーションスクリプトを取得.
     * <BR><BR>
     * アプリケーションスクリプトを取得します.
     * <BR>
     * @param name 対象のアプリケーションスクリプト名を設定します.
     * @return Script アプリケーションスクリプトが返されます.
     * @exception Exception 例外.
     */
    public Script getApplication( String name )
        throws Exception {
        if( name == null || ( name = name.trim() ).length() <= 0 ) {
            throw new HttpdStateException( HttpdErrorDef.HTTP11_500,
                "対象アプリケーション名は不正です" ) ;
        }
        String key = ScriptManager.cutScriptPlus( name ) ;
        Script ret = null ;
        // コンパイルキャッシュを行う場合.
        if( !HttpdMode.noNompileCache ) {
            ret = apps.get( key ) ;
            if( ret == null ) {
                if( ( ret = ScriptManager.getScript( name,currentDirectory ) ) != null ) {
                    apps.put( key,ret ) ;
                }
            }
            if( ret == null ) {
                throw new HttpdStateException( HttpdErrorDef.HTTP11_500,
                    "対象アプリケーション名["+name+"]は存在しません" ) ;
            }
        }
        // コンパイルキャッシュを行わない場合.
        else {
            if( ( ret = ScriptManager.getScript( name,currentDirectory ) ) == null ) {
                throw new HttpdStateException( HttpdErrorDef.HTTP11_500,
                    "対象アプリケーション名["+name+"]は存在しません" ) ;
            }
        }
        return ret ;
    }
    
    /**
     * 指定アプリケーションスクリプトが存在するかチェック.
     * <BR><BR>
     * 指定されたアプリケーションスクリプトが存在するかチェックします.
     * <BR>
     * @param name 対象名を設定します.
     * @return boolean [true]の場合、存在します.
     */
    public boolean isApplication( String name ) {
        if( name == null || ( name = name.trim() ).length() <= 0 ) {
            return false ;
        }
        if( name.startsWith( "/" ) && !name.startsWith( currentDirectory ) ) {
            name = currentDirectory + SeaBassDef.DIRECTORY_APPLICATION + name.substring( 1 ) ;
        }
        if ( FileUtil.isFileExists( name ) ) {
            return true ;
        }
        name = ScriptManager.cutScriptPlus( name ) ;
        return ( FileUtil.isFileExists( name + ".ms" ) || FileUtil.isFileExists( name + ".mhtml" ) ) ;
    }
    
    /**
     * カレントディレクトリを取得.
     * <BR><BR>
     * 設定されているカレントディレクトリを取得します.
     * <BR>
     * @return String カレントディレクトリが返されます.
     */
    public String getCurrentDirectory() {
        return currentDirectory ;
    }
}

