package org.maachang.seabass.httpd.engine.script.scripts;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.maachang.util.FileUtil;
import org.maachang.util.StringUtil;

/**
 * スクリプトInclude処理.
 * 
 * @version 2009/08/04
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
public abstract class IncludeScript {
    private IncludeScript() {}
    
    private static final String INCLUDE = "#include" ;
    private static final int INCLUDE_LENGTH = INCLUDE.length() ;
    
    private static final String DEFINE = "#define" ;
    private static final int DEFINE_LENGTH = DEFINE.length() ;
    
    /**
     * Include/Defineを行い、スクリプトファイルを取得.
     * @param name 対象のファイル名を設定します.
     * @return String スクリプト情報が返されます
     * @exception Exception 例外.
     */
    public static final String readScript( String name )
        throws Exception {
        return readScript( name,null ) ;
    }
    
    /**
     * Include/Defineを行い、スクリプトファイルを取得.
     * @param name 対象のファイル名を設定します.
     * @param charset 対象のCharsetを設定します.
     * @return String スクリプト情報が返されます
     * @exception Exception 例外.
     */
    public static final String readScript( String name,String charset )
        throws Exception {
        if( name == null || ( name = name.trim() ).length() <= 0 ) {
            return null ;
        }
        if( charset == null || ( charset = charset.trim() ).length() <= 0 ) {
            charset = "UTF8" ;
        }
        name = FileUtil.getFullPath( name ) ;
        String script = FileUtil.getFileByString( name,charset ) ;
        script = include( name,script,charset ) ;
        return define( script ) ;
    }
    
    /**
     * スクリプトInclude.
     * @param path 対象のパスを設定します.
     * @param script 対象のスクリプト情報を設定します.
     * @param charset 対象のCharsetを設定します.
     * @return String Includeされたスクリプト情報が返されます.
     * @exception Exception 例外.
     */
    public static final String include( String path,String script )
        throws Exception {
        return include( path,script,"UTF8" ) ;
    }
    
    /**
     * スクリプトInclude.
     * @param path 対象のパスを設定します.
     * @param script 対象のスクリプト情報を設定します.
     * @param charset 対象のCharsetを設定します.
     * @return String Includeされたスクリプト情報が返されます.
     * @exception Exception 例外.
     */
    public static final String include( String path,String script,String charset )
        throws Exception {
        if( script == null || script.length() <= 0 ) {
            return "" ;
        }
        return include( new HashSet<String>(),0,path,script,charset ) ;
    }
    
    /** スクリプトInclude実行処理. **/
    private static final String include( Set<String> includeFiles,int n,String path,String script,String charset )
        throws Exception {
        if( script == null || script.length() <= 0 ) {
            return "" ;
        }
        path = FileUtil.getFullPath( path ) ;
        if( includeFiles.contains( path ) ) {
            return "" ;
        }
        includeFiles.add( path ) ;
        int p,b ;
        if( ( p = StringUtil.indexToNotCote( script,INCLUDE,0 ) ) == -1 ) {
            return script ;
        }
        int len = script.length() ;
        b = 0 ;
        path = FileUtil.getDirName( path ) ;
        if( !( path.endsWith( "/" ) || path.endsWith( "\\" ) ) ) {
            path += "/" ;
        }
        StringBuilder buf = new StringBuilder() ;
        int end ;
        int start ;
        char c ;
        String file ;
        while( true ) {
            start = -1 ;
            buf.append( script.substring( b,p ) ) ;
            end = endPosToInclude( script,len,p ) ;
            if( end == -1 ) {
                b = p ;
                p = StringUtil.indexToNotCote( script,INCLUDE,b+1 ) ;
                if( p == -1 ) {
                    buf.append( script.substring( b ) ) ;
                    break ;
                }
                continue ;
            }
            c = script.charAt( end ) ;
            for( int i = p+INCLUDE_LENGTH ; i < end+1 ; i ++ ) {
                if( c == script.charAt( i ) ) {
                    start = i ;
                    break ;
                }
            }
            file = cutEnter( StringUtil.trim( script.substring( start+1,end ) ) ) ;
            end ++ ;
            if( !( file.startsWith( "/" ) || file.startsWith( "\\" ) ) ) {
                file = path + file ;
            }
            else {
                file = file.substring( 1 ) ;
            }
            file = FileUtil.getFullPath( file ) ;
            buf.append( IncludeScript.include( includeFiles,n+1,file,FileUtil.getFileByString( file,charset ),charset ) ) ;
            b = end ;
            p = StringUtil.indexToNotCote( script,INCLUDE,b ) ;
            if( p == -1 ) {
                buf.append( script.substring( b ) ) ;
                break ;
            }
            if( n == 0 ) {
                includeFiles.clear() ;
            }
        }
        return buf.toString() ;
    }
    
    /** Includeの終端を取得. **/
    private static final int endPosToInclude( String script,int len,int index ) {
        char c,b ;
        char cote = 0 ;
        b = 0 ;
        for( int i = index + INCLUDE_LENGTH ; i < len ; i ++ ) {
            c = script.charAt( i ) ;
            if( cote != 0 ) {
                if( b != '\\' && c == cote ) {
                    return i ;
                }
            }
            else if( c == '\'' || c == '\"' ) {
                cote = c ;
            }
            else if( c == '\r' || c == '\n' ) {
                return -1 ;
            }
            if( b == '\\' && b == c ) {
                b = 0 ;
            }
            else {
                b = c ;
            }
        }
        return -1 ;
    }
    
    /** コーテーション情報から、改行などの情報を除外. **/
    private static final String cutEnter( String string ) {
        int len = string.length() ;
        char c ;
        StringBuilder buf = new StringBuilder() ;
        for( int i = 0 ; i < len ; i ++ ) {
            c = string.charAt( i ) ;
            if( c == '\r' || c == '\n' ) {
                continue ;
            }
            buf.append( c ) ;
        }
        return buf.toString() ;
    }
    
    /**
     * define定義を展開.
     * @param script 対象のスクリプトを設定します.
     * @return String 変換されたスクリプトが返されます.
     * @exception Exception 例外.
     */
    public static final String define( String script )
        throws Exception {
        if( script == null || script.length() <= 0 ) {
            return "" ;
        }
        Map<String,String> defineMap = new HashMap<String,String>() ;
        int p,b ;
        if( ( p = StringUtil.indexToNotCote( script,DEFINE,0 ) ) == -1 ) {
            return script ;
        }
        int len = script.length() ;
        StringBuilder buf = new StringBuilder() ;
        int end ;
        b = 0 ;
        // 初めに#define定義内容を取得して、元のscriptから#define定義内容を除外させる.
        while( true ) {
            buf.append( script.substring( b,p ) ) ;
            end = endPosToDefine( script,len,p ) ;
            if( end == -1 ) {
                b = p ;
                p = StringUtil.indexToNotCote( script,DEFINE,b+1 ) ;
                if( p == -1 ) {
                    buf.append( script.substring( b ) ) ;
                    break ;
                }
                continue ;
            }
            String value = convertDefineInfo( script,p+1,end ) ;
            cutToMap( defineMap,value ) ;
            value = null ;
            b = end + 1 ;
            p = StringUtil.indexToNotCote( script,DEFINE,b ) ;
            if( p == -1 ) {
                buf.append( script.substring( b ) ) ;
                break ;
            }
        }
        script = buf.toString() ;
        // 次に#define定義内容を、スクリプトに割り当てる.
        Iterator it = defineMap.keySet().iterator() ;
        List<_SortDefine> lst = new ArrayList<_SortDefine>() ;
        while( it.hasNext() ) {
            String s = ( String )it.next() ;
            if( s != null ) {
                lst.add( new _SortDefine( s,defineMap.get( s ) ) ) ;
            }
        }
        it = null ;
        defineMap = null ;
        len = lst.size() ;
        if( len > 0 ) {
            Collections.sort( lst ) ;
            _SortDefine d ;
            for( int i = 0 ; i < len ; i ++ ) {
                d = lst.get( i ) ;
                script = StringUtil.changeString( script,d.getKey(),d.getValue() ) ;
            }
        }
        return script ;
    }
    
    /** Defineの終端を取得. **/
    private static final int endPosToDefine( String script,int len,int index ) {
        char c,b,bb ;
        char cote = 0 ;
        b = bb = 0 ;
        boolean nextFlag = false ;
        for( int i = index + DEFINE_LENGTH ; i < len ; i ++ ) {
            c = script.charAt( i ) ;
            if( cote != 0 ) {
                if( b != '\\' && c == cote ) {
                    cote = 0 ;
                }
            }
            else if( c == '\'' || c == '\"' ) {
                cote = c ;
            }
            else if( bb != '\\' && c == '\\' ) {
                nextFlag = true ;
            }
            else if( c == '\r' || c == '\n' ) {
                if( nextFlag ) {
                    if( c == '\r' && i+1 < len && script.charAt( i+1 ) == '\n' ) {
                        i ++ ;
                    }
                    nextFlag = false ;
                    b = bb = c ;
                    continue ;
                }
                if( c == '\r' && i+1 < len && script.charAt( i+1 ) == '\n' ) {
                    return i + 1 ;
                }
                return i ;
            }
            if( b == '\\' && b == c ) {
                if( cote == 0 ) {
                    nextFlag = false ;
                }
                b = 0 ;
                bb = c ;
            }
            else {
                b = bb = c ;
            }
        }
        return -1 ;
    }
    
    /** Defineデータの￥マーク情報を除外. **/
    private static final String convertDefineInfo( String script,int start,int end )
        throws Exception {
        String value = StringUtil.trim( script.substring( start+DEFINE_LENGTH,end ) ) ;
        if( StringUtil.indexToNotCote( value,"\\",0 ) == -1 ) {
            return value ;
        }
        int len = value.length() ;
        StringBuilder buf = new StringBuilder() ;
        boolean firstFlag = true ;
        char cote = 0 ;
        char c,b,bb ;
        char x ;
        boolean mode ;
        int cnt ;
        b = bb = 0 ;
        for( int i = 0 ; i < len ; i ++ ) {
            c = value.charAt( i ) ;
            if( cote != 0 ) {
                if( b != '\\' && c == cote ) {
                    cote = 0 ;
                }
            }
            else if( c == '\'' || c == '\"' ) {
                cote = c ;
            }
            else if( c == '\\' ) {
                if( bb != '\\' && c == '\\' && i+1 < len && value.charAt( i+1 ) != '\\' ) {
                    mode = false ;
                    cnt = i+1 ;
                    for( int j = i+1 ; j < len ; j ++ ) {
                        cnt ++ ;
                        x = value.charAt( j ) ;
                        if( x == '\r' || x == '\n' ) {
                            i = j-1 ;
                            mode = true ;
                            break ;
                        }
                        else if( x == ' ' || x == '\t' ) {
                            continue ;
                        }
                        break ;
                    }
                    if( mode || cnt == len ) {
                        continue ;
                    }
                }
                else if( i+1 == len && c == '\\' ) {
                    break ;
                }
            }
            else if( c == '\r' || c == '\n' ) {
                firstFlag = true ;
                continue ;
            }
            if( b == '\\' && b == c ) {
                b = 0 ;
                bb = c ;
            }
            else {
                b = bb = c ;
            }
            if( firstFlag ) {
                if( c == ' ' || c == '　' || c == '\t' ) {
                    continue ;
                }
                firstFlag = false ;
            }
            buf.append( c ) ;
        }
        return buf.toString() ;
    }
    
    /** #define定義内容を分解してMapにセット **/
    private static final void cutToMap( Map<String,String> map,String value ) {
        int len = value.length() ;
        char c ;
        int end = 0 ;
        for( int i = 0 ; i < len ; i ++ ) {
            c = value.charAt( i ) ;
            if( c == ' ' || c == '　' || c == '\t' ) {
                end = i ;
                break ;
            }
        }
        map.put( StringUtil.trim( value.substring( 0,end ) ),
            StringUtil.trim( value.substring( end ) ) ) ;
    }
    
    /**
     * ソート用変換データ.
     */
    private static class _SortDefine implements Comparable {
        private String key ;
        private String value ;
        private int keyLen ;
        private _SortDefine() {
        }
        public _SortDefine( String key,String value ) {
            this.key = key ;
            this.value = value ;
            this.keyLen = key.length() ;
        }
        public int compareTo( Object o ) {
            if( o == null || ( o instanceof _SortDefine ) == false ) {
                return -1 ;
            }
            return (( _SortDefine )o).keyLen - keyLen ;
        }
        public String getKey() {
            return key ;
        }
        public String getValue() {
            return value ;
        }
    }
}

