package org.maachang.seabass.httpd.engine.script.scripts;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.atomic.AtomicObject;
import org.maachang.seabass.SeaBassDef;
import org.maachang.seabass.httpd.HttpdErrorDef;
import org.maachang.seabass.httpd.HttpdStateException;
import org.maachang.seabass.httpd.engine.script.ExecutionScript;
import org.maachang.seabass.httpd.engine.script.Script;
import org.maachang.seabass.httpd.engine.script.ScriptDef;
import org.maachang.seabass.httpd.engine.script.SrcScript;
import org.maachang.seabass.httpd.engine.script.doc.SeaBassDoc;
import org.maachang.seabass.httpd.engine.script.doc.SeaBassDocAnalysis;
import org.maachang.util.FileUtil;

/**
 * SeaBassドキュメントスクリプト.
 * 
 * @version 2009/01/19
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
public class SeaBassDocScript implements Script {
    
    /**
     * LOG.
     */
    private static final Log LOG = LogFactory.getLog( SeaBassDocScript.class ) ;
    
    /**
     * 対象パス.
     */
    protected String path = null ;
    
    /**
     * ファイル名.
     */
    protected String fileName = null ;
    
    /**
     * カレントディレクトリ.
     */
    protected String currentDirectory = null ;
    
    /**
     * SeaBassDocオブジェクト.
     */
    private final AtomicObject<SeaBassDoc> seabassDoc = new AtomicObject<SeaBassDoc>() ;
    
    /**
     * MHTMLオブジェクト.
     */
    private final AtomicObject<MhtmlScript> mhtml = new AtomicObject<MhtmlScript>() ;
    
    /**
     * コンストラクタ.
     * <BR><BR>
     * 条件を指定してオブジェクトを生成します.
     * <BR>
     * @param path 対象のパスを設定します.
     * @param currentDirectory カレントディレクトリを設定します.
     * @exception Exception 例外.
     */
    public SeaBassDocScript( String path,String currentDirectory )
        throws Exception {
        if( path == null || ( path = path.trim() ).length() <= 0 ) {
            throw new HttpdStateException( HttpdErrorDef.HTTP11_500,
                "スクリプトの指定パスは不正です" ) ;
        }
        if( path.endsWith( ScriptManager.SDOC_PLUS ) == false ) {
            path += ScriptManager.SDOC_PLUS ;
        }
        currentDirectory = ScriptDef.trimCurrentDirectory( currentDirectory ) ;
        //currentDirectory = ScriptDef.trimCurrentDirectory( currentDirectory ) ;
        String fileName = new StringBuilder().append( currentDirectory ).
            append( SeaBassDef.DIRECTORY_APPLICATION ).
            append( ( ( path.startsWith("/" ) ) ? path.substring( 1,path.length() ) : path ) ).
            toString() ;
        if( FileUtil.isFileExists( fileName ) == false ) {
            throw new HttpdStateException( HttpdErrorDef.HTTP11_500,
                "SeaBassドキュメントファイル[" + path + "]は存在しません" ) ;
        }
        this.path = path.substring( 0,path.length()-ScriptManager.SCRIPT_PLUS.length() ) ;
        this.fileName = fileName ;
        this.currentDirectory = currentDirectory ;
        this.readScript() ;
    }
    
    /**
     * 情報クリア.
     */
    protected void clear() {
        path = null ;
        fileName = null ;
        currentDirectory = null ;
        seabassDoc.set( null ) ;
        mhtml.set( null ) ;
    }
    
    /**
     * 基本スクリプトを取得.
     * <BR><BR>
     * 基本スクリプトを取得します.
     * <BR>
     * @return ExecutionScript 基本スクリプトが返されます.
     * @exception Exception 例外.
     */
    public ExecutionScript getScript() throws Exception {
        this.readScript() ;
        MhtmlScript mh = mhtml.get() ;
        if( mh == null ) {
            return null ;
        }
        // デフォルトパラメータとして、SeaBassドキュメントを設置.
        mh.getScript().putDefault( "doc",seabassDoc.get() ) ;
        return mh.getScript() ;
    }
    
    /**
     * スクリプトパスを取得.
     * <BR><BR>
     * スクリプトパスを取得します.
     * <BR>
     * @return String スクリプトパスが返されます.
     */
    public String getPath() {
        return path ;
    }
    
    /**
     * スクリプトファイル名を取得.
     * <BR><BR>
     * スクリプトファイル名を取得します.
     * <BR>
     * @return String スクリプトファイル名が返されます.
     */
    public String getScriptName() {
        return fileName ;
    }
    
    /**
     * スクリプトタイプを取得.
     * <BR><BR>
     * スクリプトタイプを取得します.
     * <BR>
     * @return int スクリプトタイプが返されます.
     */
    public int getScriptType() {
        return ScriptManager.SCRIPT_TYPE_BY_SDOC ;
    }
    
    /**
     * スクリプト内容を取得.
     * <BR><BR>
     * スクリプト内容を取得します.
     * <BR>
     * @return SrcScript スクリプト内容が返されます.
     */
    public SrcScript getSrcScript() {
        try {
            this.readScript() ;
        } catch( Exception e ) {}
        MhtmlScript mh = mhtml.get() ;
        if( mh == null ) {
            return null ;
        }
        return mh.getSrcScript() ;
    }
    
    /**
     * スクリプトが利用可能かチェック.
     * <BR><BR>
     * スクリプトが利用可能かチェックします.
     * <BR>
     * @return boolean [true]の場合、利用可能です.
     */
    public boolean isScript() {
        return ( seabassDoc.get() != null && mhtml.get() != null ) ;
    }
    
    /**
     * スクリプト内容を読み込む.
     */
    protected void readScript()
        throws Exception {
        SeaBassDoc doc = seabassDoc.get() ;
        if( doc != null ) {
            if( doc.isUpdate() ) {
                SeaBassDoc befDoc = this.seabassDoc.get() ;
                MhtmlScript befMhtml = this.mhtml.get() ;
                try {
                    load() ;
                } catch( Exception e ) {
                    LOG.error( "SeaBassドキュメント解析に失敗しました",e ) ;
                    // エラーの場合は、前の内容を利用する.
                    this.seabassDoc.set( befDoc ) ;
                    this.mhtml.set( befMhtml ) ;
                }
            }
        }
        else {
            try {
                load() ;
            } catch( HttpdStateException hs ) {
                throw hs ;
            } catch( Exception e ) {
                throw new HttpdStateException( e,HttpdErrorDef.HTTP11_500,
                    "SeaBassドキュメントの初期化に失敗しました" ) ;
            }
        }
    }
    
    /**
     * スクリプトローディング.
     */
    protected void load() throws Exception {
        MhtmlScript mh = null ;
        SeaBassDoc doc = SeaBassDocAnalysis.analysis( fileName,"UTF8" ) ;
        if( mhtml.get() == null || mhtml.get().isReload() ) {
            mh = new MhtmlScript( doc,currentDirectory ) ;
        }
        this.seabassDoc.set( doc ) ;
        if( mh != null ) {
            this.mhtml.set( mh ) ;
        }
    }
}
