package org.maachang.seabass.httpd.engine.script.scripts;

import org.maachang.bdb.Bdb;
import org.maachang.bdb.BdbKeyValue;
import org.maachang.seabass.dbm.BdbFactory;
import org.maachang.seabass.httpd.HttpdErrorDef;
import org.maachang.seabass.httpd.HttpdStateException;
import org.maachang.seabass.httpd.engine.script.ExecutionScript;
import org.maachang.seabass.httpd.engine.script.Script;
import org.maachang.seabass.httpd.engine.script.SrcScript;
import org.maachang.seabass.httpd.engine.script.js.JsDef;
import org.maachang.seabass.httpd.engine.script.taglib.AnalysisTagLib;
import org.maachang.util.ConvertParam;
import org.maachang.util.atomic.AtomicINT;
import org.maachang.util.atomic.AtomicLONG;
import org.maachang.util.atomic.AtomicOBJECT;

/**
 * テンプレート用スクリプト情報.
 * 
 * @version 2008/07/13
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
public class TemplateScript implements Script {
    
    /**
     * BDBテンプレート格納名.
     */
    private static final String BDB_NAME = "seabass@template" ;
    
    /**
     * DBM名前ヘッダ.
     */
    protected static final String DBM_HEADER = "@mc@template:" ;
    
    /**
     * DBM更新ヘッダ.
     */
    protected static final String DBM_UPDATE_HEADER = "@mc@template@update:" ;
    
    /**
     * 対象パス.
     */
    protected String path = null ;
    
    /**
     * カレントディレクトリ.
     */
    protected String currentDir = null ;
    
    /**
     * DBMアクセス名.
     */
    protected byte[] baseName = null ;
    protected byte[] updateName = null ;
    
    /**
     * ファイル更新時間.
     */
    protected final AtomicLONG updateTime = new AtomicLONG( -1L ) ;
    
    /**
     * スクリプト内容.
     */
    private final AtomicOBJECT<SrcScript> srcScript = new AtomicOBJECT<SrcScript>() ;
    
    /**
     * 実行スクリプトオブジェクト.
     */
    protected final AtomicOBJECT<ExecutionScriptImpl> executionScript = new AtomicOBJECT<ExecutionScriptImpl>() ;
    
    /**
     * デフォルトパッケージ更新ID.
     */
    private final AtomicINT packageId = new AtomicINT( -1 ) ;
    
    /**
     * コンストラクタ.
     */
    private TemplateScript() {
    }
    
    /**
     * コンストラクタ.
     * <BR><BR>
     * 条件を指定してオブジェクトを生成します.
     * <BR>
     * @param path 対象のパスを設定します.
     * @exception Exception 例外.
     */
    public TemplateScript( String path )
        throws Exception {
        if( path == null || ( path = path.trim() ).length() <= 0 ) {
            throw new HttpdStateException( HttpdErrorDef.HTTP11_500,
                "mhtmlの指定パスは不正です" ) ;
        }
        byte[] baseName = getBaseKey( path ) ;
        Bdb dbm = BdbFactory.get( BDB_NAME ) ;
        BdbKeyValue kv = BdbFactory.getBdbKeyValue() ;
        if( dbm.get( kv.setKey( baseName ) ) == false ) {
            kv.clear() ;
            throw new HttpdStateException( HttpdErrorDef.HTTP11_500,
                "対象templateファイル[" + path + "]は存在しません" ) ;
        }
        kv.clear() ;
        this.path = path ;
        this.baseName = baseName ;
        this.updateName = getUpdateKey( path ) ;
        this.currentDir = "./" ;
    }
    
    /**
     * 情報クリア.
     */
    protected void clear() {
        updateTime.set( -1L ) ;
        srcScript.set( null ) ;
        executionScript.set( null ) ;
        packageId.set( -1 ) ;
    }
    
    /**
     * 基本スクリプトを取得.
     * <BR><BR>
     * 基本スクリプトを取得します.
     * <BR>
     * @return ExecutionScript 基本スクリプトが返されます.
     * @exception Exception 例外.
     */
    public ExecutionScript getScript() throws Exception {
        Bdb dbm = BdbFactory.get( BDB_NAME ) ;
        BdbKeyValue kv = BdbFactory.getBdbKeyValue() ;
        try {
            if( dbm.get( kv.setKey( baseName ) ) == false ) {
                kv.clear() ;
                clear() ;
                throw new HttpdStateException( HttpdErrorDef.HTTP11_500,
                    "対象mhtmlファイル[" + path + "]は存在しません" ) ;
            }
            else if( executionScript.get() == null || updateCheck() == true ||
                this.packageId.get() != JsDef.getDefaultPackageId() ) {
                readScript() ;
            }
            return executionScript.get() ;
        } finally {
            kv.clear() ;
        }
    }
    
    /**
     * スクリプトパスを取得.
     * <BR><BR>
     * スクリプトパスを取得します.
     * <BR>
     * @return String スクリプトパスが返されます.
     */
    public String getPath() {
        return this.path ;
    }
    
    /**
     * スクリプトファイル名を取得.
     * <BR><BR>
     * スクリプトファイル名を取得します.
     * <BR>
     * @return String スクリプトファイル名が返されます.
     */
    public String getScriptName() {
        return this.path ;
    }
    
    /**
     * スクリプトタイプを取得.
     * <BR><BR>
     * スクリプトタイプを取得します.
     * <BR>
     * @return int スクリプトタイプが返されます.
     */
    public int getScriptType() {
        return ScriptManager.SCRIPT_TYPE_BY_HTML ;
    }
    
    /**
     * スクリプト内容を取得.
     * <BR><BR>
     * スクリプト内容を取得します.
     * <BR>
     * @return SrcScript スクリプト内容が返されます.
     */
    public SrcScript getSrcScript() {
        SrcScript ret = srcScript.get() ;
        if( ret == null ) {
            try {
                readScript() ;
                ret = srcScript.get() ;
            } catch( Exception e ) {
                ret = null ;
            }
        }
        return ret ;
    }
    
    /**
     * スクリプトが利用可能かチェック.
     * <BR><BR>
     * スクリプトが利用可能かチェックします.
     * <BR>
     * @return boolean [true]の場合、利用可能です.
     */
    public boolean isScript() {
        return ( executionScript.get() == null ) ? false : true ;
    }
    
    /**
     * スクリプト内容を読み込む.
     */
    protected void readScript()
        throws Exception {
        Bdb dbm = BdbFactory.get( BDB_NAME ) ;
        BdbKeyValue kv = BdbFactory.getBdbKeyValue() ;
        if( dbm.get( kv.setKey( baseName ) ) == false ) {
            clear() ;
            return ;
        }
        String script = new String( kv.getValue(),CHARSET ) ;
        kv.clear() ;
        script = ReadScriptPlus.directInclude( currentDir,script ) ;
        script = AnalysisTagLib.analysis( script ) ;
        script = TemplateScriptAnalysis.analysis( script ) ;
        
        // 出力内容は、まとめて返信するように変更.
        script = new StringBuilder().
            append( "var ___$_$_$rs$buf$_$ = createStrBuf() ;\n" ).
            append( "httpPrint = function( str ) { " ).
                append( "if( !useString(str) ) {return}; " ).
                append( "___$_$_$rs$buf$_$.ad( str ); } ;\n" ).
            append( "context.remove( \"_$validate_script\" );\n" ).
            append( "exitMHTMLCache(); \n" ).
            append( script ).append( "\n" ).
            append( "httpPrintln( ___$_$_$rs$buf$_$.ts() ) ;\n ___$_$_$rs$buf$_$.clear();\n___$_$_$rs$buf$_$=null;\n" ).
            //append( "var r = ___$_$_$rs$buf$_$.ts() ;\n ___$_$_$rs$buf$_$.clear();\n___$_$_$rs$buf$_$=null;\n" ).
            //append( "r = r.getBytes( 'UTF8' ) ;\n responseToBinary( r ) ;\n" ).
            toString() ;
        
        // 出力内容はまとめて返さない.
        //script = new StringBuilder().
        //    append( "context.remove( \"_$validate_script\" );\n" ).
        //    append( "exitMHTMLCache(); \n" ).
        //    append( script ).toString() ;
        
        StringBuilder buf = new StringBuilder() ;
        //int pid = JsDef.pushDefaultPackage( buf ) ;
        int pid = JsDef.getPackageId() ;
        ReadScriptPlus.addMHTML( buf,script,path ) ;
        script = buf.toString() ;
        buf = null ;
        
        script = ReadScriptPlus.convert( false,script ) ;
        this.srcScript.set( new SrcScript( script ) ) ;
        this.executionScript.set( new ExecutionScriptImpl(
            getScriptType(),this.getPath(),script ) ) ;
        script = null ;
        this.updateTime.set( getUpdate() ) ;
        this.packageId.set( pid ) ;
    }
    
    /**
     * 更新チェック.
     */
    private boolean updateCheck() throws Exception {
        long tm = getUpdate() ;
        if( tm != updateTime.get() ) {
            return true ;
        }
        return false ;
    }
    
    /**
     * 更新情報を取得.
     */
    private long getUpdate() throws Exception {
        Bdb dbm = BdbFactory.get( BDB_NAME ) ;
        BdbKeyValue kv = BdbFactory.getBdbKeyValue() ;
        if( dbm.get( kv.setKey( updateName ) ) == false || kv.getValue() == null ||
            kv.getValue().length < 8 ) {
            kv.clear() ;
            throw new IllegalArgumentException( "更新確認データは不正です" ) ;
        }
        long ret = ConvertParam.convertLong( 0,kv.getValue() ) ;
        kv.clear() ;
        return ret ;
    }
    
    /**
     * 指定パス上のテンプレート内容が存在するかチェック.
     * @param path 対象のパスを設定します.
     * @return boolean [true]の場合、存在します.
     * @exception Exception 例外.
     */
    public static final boolean isTemplate( String path ) throws Exception {
        if( path == null || ( path = path.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        Bdb dbm = BdbFactory.get( BDB_NAME ) ;
        BdbKeyValue kv = BdbFactory.getBdbKeyValue() ;
        boolean ret = dbm.get( kv.setKey( getBaseKey( path ) ) ) ;
        kv.clear() ;
        return ret ;
    }
    
    /**
     * テンプレート更新時間を取得.
     * @param path 対象のパスを設定します.
     * @return long 更新時間が返されます.
     * @exception Exception 例外.
     */
    public static final long getTime( String path ) throws Exception {
        if( path == null || ( path = path.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        Bdb dbm = BdbFactory.get( BDB_NAME ) ;
        BdbKeyValue kv = BdbFactory.getBdbKeyValue() ;
        if( dbm.get( kv.setKey( getUpdateKey( path ) ) ) == false ||
            kv.getValue() == null || kv.getValue().length < 8 ) {
            kv.clear() ;
            throw new IllegalArgumentException( "更新確認データは不正です" ) ;
        }
        long ret = ConvertParam.convertLong( 0,kv.getValue() ) ;
        kv.clear() ;
        return ret ;
    }
    
    /**
     * 基本キー名変換.
     * @param key 対象のキー名を設定します.
     * @return byte[] 変換されたバイナリが返されます.
     * @exception Exception 例外.
     */
    public static final byte[] getBaseKey( String key ) throws Exception {
        return new StringBuilder().append( DBM_HEADER ).
            append( key ).toString().getBytes( CHARSET ) ;
    }
    
    /**
     * 更新キー名変換.
     * @param key 対象のキー名を設定します.
     * @return byte[] 変換されたバイナリが返されます.
     * @exception Exception 例外.
     */
    public static final byte[] getUpdateKey( String key ) throws Exception {
        return new StringBuilder().append( DBM_UPDATE_HEADER ).
            append( key ).toString().getBytes( CHARSET ) ;
    }
}
