package org.maachang.seabass.httpd.engine.script.taglib;

import gnu.trove.THashSet;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.util.Set;

import org.maachang.conf.Config;
import org.maachang.conf.ReadIni;
import org.maachang.util.Digest;
import org.maachang.util.DirUtil;
import org.maachang.util.FileUtil;
import org.maachang.util.atomic.AtomicLONG;
import org.maachang.util.atomic.AtomicOBJECT;

/**
 * タグライブラリ定義管理.
 * 
 * @version 2008/07/19
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
public class DefineTagLib {
    
    /**
     * デフォルト読み込みタグライブラリ.
     */
    private static final String[] DEF_TAGLIB = {
        "org/maachang/seabass/httpd/engine/script/js/tag.tlib"
    } ;
    
    /**
     * 拡張タグライブラリ読み込みディレクトリ.
     */
    private static final String CUSTOM_TAGLIB_DIR = "./taglib" ;
    
    /**
     * 拡張タグ拡張子.
     */
    private static final String CUSTOM_TAGLIB_PLUS = ".tlib" ;
    
    /**
     * 再読み込み確認時間.
     */
    private static final long CHECK_TIME = 30000L ;
    
    /**
     * イベント条件群.
     */
    private static final String[] EVENTS = {
        "onclick",
        "ondblclick",
        "onkeydown",
        "onkeypress",
        "onkeyup",
        "onkeyup",
        "onmousedown",
        "onmouseup",
        "onmouseover",
        "onmouseout",
        "onmousemove",
        "onload",
        "onunload",
        "onfocus",
        "onblur",
        "onsubmit",
        "onreset",
        "onchange",
        "onresize",
        "onmove",
        "ondragdrop",
        "onabort",
        "onerror",
        "onselect"
    } ;
    
    /**
     * 更新確認コード.
     */
    private String sha1 = null ;
    
    /**
     * 前回確認時間.
     */
    private final AtomicLONG beforeTiem = new AtomicLONG( -1L ) ;
    
    /**
     * デフォルト読み込みタグライブラリ用コンフィグファイル.
     */
    private final AtomicOBJECT<Config> defConfig = new AtomicOBJECT<Config>() ;
    
    /**
     * データ管理.
     */
    private final AtomicOBJECT<Config> custom = new AtomicOBJECT<Config>() ;
    
    /**
     * イベントチェック用.
     */
    private final AtomicOBJECT<Set<String>> event = new AtomicOBJECT<Set<String>>() ;
    
    /**
     * シングルトン.
     */
    private static final DefineTagLib SNGL = new DefineTagLib() ;
    
    /**
     * オブジェクトを取得.
     * @return DefineTagLib オブジェクトが返されます.
     */
    public static final DefineTagLib getInstance() {
        return SNGL ;
    }
    
    /**
     * コンストラクタ.
     */
    private DefineTagLib() {
    }
    
    /**
     * 初期化処理.
     */
    private void init() throws Exception {
        if( defConfig.get() == null ) {
            this.sha1 = null ;
            this.beforeTiem.set( -1L ) ;
            this.defConfig.set( readDefault() ) ;
            this.custom.set( new Config() ) ;
            Set<String> h = new THashSet<String>() ;
            for( int i = 0 ; i < EVENTS.length ; i ++ ) {
                h.add( EVENTS[ i ] ) ;
            }
            event.set( h ) ;
        }
    }
    
    /**
     * 指定タグライブラリ名の内容を含むコンフィグ情報を取得.
     * @param name 対象のタグライブラリ名を設定します.
     * @return Config コンフィグ情報が返されます.
     * @exception Exception 例外.
     */
    public Config getConfig( String name ) throws Exception {
        this.init() ;
        if( defConfig.get().isSection( name ) ) {
            return defConfig.get() ;
        }
        else {
            if( CHECK_TIME + beforeTiem.get() <= System.currentTimeMillis() ) {
                beforeTiem.set( System.currentTimeMillis() ) ;
                String chk = getSha1() ;
                if( chk == null ) {
                    custom.get().clear() ;
                }
                else if( chk != sha1 ) {
                    sha1 = chk ;
                    readCustom( custom.get() ) ;
                }
            }
            if( custom.get().isSection( name ) ) {
                return custom.get() ;
            }
        }
        return null ;
    }
    
    /**
     * イベント要素チェック.
     * @param key 対象のキーを設定します.
     * @return boolean [true]の場合、イベント情報です.
     */
    public boolean isEvent( String key ) throws Exception {
        this.init() ;
        return event.get().contains( key ) ;
    }
    
    /**
     * デフォルトタグライブラリを読み込む.
     */
    private static final Config readDefault() throws Exception {
        BufferedReader br = null ;
        try {
            Config ret = new Config() ;
            int len = DEF_TAGLIB.length ;
            for( int i = 0 ; i < len ; i ++ ) {
                br = new BufferedReader( new InputStreamReader(
                        Thread.currentThread().getContextClassLoader().getResourceAsStream( DEF_TAGLIB[ i ] ),"UTF8" ) ) ;
                ReadIni.analisys( ret,br ) ;
                br.close() ;
                br = null ;
            }
            return ret ;
        } finally {
            if( br != null ) {
                try {
                    br.close() ;
                } catch( Exception e ) {
                }
            }
        }
    }
    
    /**
     * 拡張タグライブラリを読み込む.
     */
    private static final void readCustom( Config custom ) throws Exception {
        if( FileUtil.isDirExists( CUSTOM_TAGLIB_DIR ) == false ) {
            custom.clear() ;
            return ;
        }
        String[] list = DirUtil.getDirByUseFileName( CUSTOM_TAGLIB_DIR,CUSTOM_TAGLIB_PLUS ) ;
        if( list == null || list.length <= 0 ) {
            return ;
        }
        custom.clear() ;
        int len = list.length ;
        for( int i = 0 ; i < len ; i ++ ) {
            String file = CUSTOM_TAGLIB_DIR + list[ i ] ;
            BufferedReader br = null ;
            try {
                br = new BufferedReader( new InputStreamReader( new FileInputStream( file ) ) ) ;
                ReadIni.analisys( custom,br ) ;
                br.close() ;
                br = null ;
            } catch( Exception e ) {
                custom.clear() ;
                throw e ;
            } finally {
                if( br != null ) {
                    try {
                        br.close() ;
                    } catch( Exception e ) {
                    }
                }
            }
        }
    }
    
    /**
     * SHA1を生成.
     */
    private static final String getSha1() throws Exception {
        if( FileUtil.isDirExists( CUSTOM_TAGLIB_DIR ) == false ) {
            return null ;
        }
        String[] list = DirUtil.getDirByUseFileName( CUSTOM_TAGLIB_DIR,CUSTOM_TAGLIB_PLUS ) ;
        if( list == null || list.length <= 0 ) {
            return null ;
        }
        StringBuilder buf = new StringBuilder() ;
        int len = list.length ;
        for( int i = 0 ; i < len ; i ++ ) {
            if( i != 0 ) {
                buf.append( ":" ) ;
            }
            buf.append( list[ i ] ).append( "@" ).append( FileUtil.getLastTime( CUSTOM_TAGLIB_DIR + list[ i ] ) ) ;
        }
        byte[] bin = buf.toString().getBytes( "UTF8" ) ;
        buf = null ;
        return Digest.convert( "SHA1",bin ) ;
    }
    
}

