package org.maachang.seabass.net ;

import java.io.IOException;
import java.io.OutputStream;

/**
 * Chunked用OutputStream.
 * 
 * @version 2007/08/26
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
class ChunkedOutputStream extends OutputStream {
    private static final int DEFAULT_BUF = 1024 ;
    private static final byte[] ENTER = new byte[] {(byte)'\r', (byte)'\n'};
    private static final byte[] CLOSE_CHUNKED = new byte[] {(byte) '0',(byte)'\r', (byte)'\n'};
    private final byte[] cache = new byte[DEFAULT_BUF];
    private OutputStream stream = null;
    private int cachePosition = 0;
    private boolean wroteLastChunk = false;
    private boolean isCloseFlag = false ;
    
    public ChunkedOutputStream() {
        
    }
    
    public ChunkedOutputStream(OutputStream stream) throws IOException {
        create( stream ) ;
    }
    
    public OutputStream create(OutputStream stream) throws IOException {
        clear() ;
        this.stream = stream;
        return this ;
    }
    
    public void clear() {
        stream = null;
        cachePosition = 0;
        wroteLastChunk = false;
        isCloseFlag = false ;
    }
    
    public boolean isChunked() {
        return ( stream != null ) ;
    }
    
    protected void flushCache() throws IOException {
        if (cachePosition > 0) {
            chunedHead( cachePosition ) ;
            stream.write(cache, 0, cachePosition);
            stream.write(ENTER, 0, ENTER.length);
            stream.flush() ;
            cachePosition = 0;
        }
    }
    
    public void finish() throws IOException {
        flushCache();
        if (!wroteLastChunk) {
            stream.write(CLOSE_CHUNKED, 0, CLOSE_CHUNKED.length);
            stream.write(ENTER, 0, ENTER.length);
            stream.flush() ;
            wroteLastChunk = true;
        }
    }
    
    public void write(int b) throws IOException {
        cache[cachePosition] = (byte) b;
        cachePosition++;
        if (cachePosition >= DEFAULT_BUF) {
            flushCache();
        }
    }
    
    public void write(byte b[]) throws IOException {
        this.write(b, 0, b.length);
    }
    
    public void write(byte src[], int off, int len) throws IOException {
        /*
        if (len >= DEFAULT_BUF - cachePosition) {
            flushCacheWithAppend(src, off, len);
        } else {
            System.arraycopy(src, off, cache, cachePosition, len);
            cachePosition += len;
        }
        */
        int mLen = off + len ;
        for( int i = off ; i < mLen ; i ++ ) {
            write( src[i] ) ;
        }
    }
    
    protected void flushCacheWithAppend(byte bufferToAppend[], int off, int len) throws IOException {
        if( cachePosition > 0 ) {
            chunedHead( cachePosition + len ) ;
            stream.write(cache, 0, cachePosition);
            stream.write(bufferToAppend, off, len);
            stream.write(ENTER, 0, ENTER.length);
            stream.flush() ;
            cachePosition = 0;
        }
        else {
            chunedHead( len ) ;
            stream.write(bufferToAppend, off, len);
            stream.write(ENTER, 0, ENTER.length);
            stream.flush() ;
        }
    }
    
    public void flush() throws IOException {
        flushCache() ;
    }
    
    public void close() throws IOException {
        IOException io = null ;
        if( !isCloseFlag ) {
            try {
                flushCache() ;
            } catch( IOException e ) {
                io = e ;
            }
            try {
                finish();
            } catch( IOException e ) {
                io = e ;
            }
            try {
                stream.close() ;
            } catch( IOException e ) {
                io = e ;
            }
            if( io != null ) {
                throw io ;
            }
            isCloseFlag = true ;
        }
    }
    
    private void chunedHead( int value ) throws IOException {
        long m = 0x00000000f0000000L ;
        long p = 28L ;
        boolean f = false ;
        while( true ) {
            int x = (int)((value & m)>>p) ;
            if( f || x != 0 ) {
                stream.write( HEX_LIST[x] ) ;
                f=true;
            }
            if( p == 0 ) break ;
            p -= 4L ;
            m = m >> 4L ;
        }
        stream.write( ENTER ) ;
    }
    
    private static final byte[] HEX_LIST = {
        (byte)'0',(byte)'1',(byte)'2',(byte)'3',(byte)'4',(byte)'5',(byte)'6',(byte)'7',
        (byte)'8',(byte)'9',(byte)'A',(byte)'B',(byte)'C',(byte)'D',(byte)'E',(byte)'F' } ;
}
