package org.maachang.seabass.net ;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Map;

import org.maachang.bdb.BdbKeyValue;
import org.maachang.manager.GlobalManager;
import org.maachang.seabass.ServiceDef;
import org.maachang.seabass.httpd.HttpdHeaders;
import org.maachang.seabass.httpd.HttpdParams;
import org.maachang.seabass.httpd.HttpdRequest;
import org.maachang.seabass.httpd.engine.HttpdCookie;
import org.maachang.seabass.httpd.engine.HttpdDef;
import org.maachang.seabass.httpd.engine.session.HttpdSession;
import org.maachang.seabass.httpd.engine.session.HttpdSessionManager;
import org.maachang.seabass.net.nio.ConnectionInfo;
import org.maachang.util.BaseRandom;
import org.maachang.util.Xor128Random;


/**
 * HTTPD基本リクエスト実装.
 *
 * @version 2007/08/20
 * @author  masahito suzuki
 * @since   SeaBass 1.00
 */
public class HttpdRequestImpl implements HttpdRequest {
    
    /**
     * セッションランダムオブジェクト.
     */
    private final BaseRandom random = new Xor128Random() ;
    
    /**
     * BdbKeyValueオブジェクト.
     */
    private final BdbKeyValue keyValue = new BdbKeyValue() ;
    
    /**
     * 処理タイプ.(GET/POST).
     */
    private String method = null ;
    
    /**
     * HTTPバージョン.
     */
    private String httpVersion = null ;
    
    /**
     * クライアントアクセス時のURL.
     */
    private String clientAccessURL = null ;
    
    /**
     * URL相対パス.
     */
    private String urlPath = null ;
    
    /**
     * リモートアドレス.
     */
    private String remoteAddress = null ;
    
    /**
     * リモートポート番号.
     */
    private int remotePort = -1 ;
    
    /**
     * パラメータ.
     */
    private final HttpdParamsImpl query = new HttpdParamsImpl() ;
    
    /**
     * ヘッダ.
     */
    private final HttpdHeadersImpl header = new HttpdHeadersImpl() ;
    
    /**
     * セッション.
     */
    private HttpdSession session = null ;
    
    /**
     * コネクション情報.
     */
    private ConnectionInfo conn = null ;
    
    /**
     * GZIP圧縮許可フラグ.
     */
    private boolean gzipFlag = false ;
    
    /**
     * クッキー.
     */
    private Map<String,String> cookie = null ;
    private boolean readCookieFlag = false ;
    
    /**
     * Cookieフラグ.
     */
    private boolean cookieFlag = false ;
    
    /**
     * SSLフラグ.
     */
    private boolean sslFlag = false ;
    
    /**
     * ディレクトリフラグ.
     */
    private boolean dirFlag = false ;
    
    /**
     * フルパスターゲット名.
     */
    private String fullPath = null ;
    
    /**
     * Seabass32暗号化フラグ.
     */
    private boolean seabass32 = false ;
    
    /**
     * Seabass32暗号化に対するオプション.
     */
    private String[] seabass32Opt = new String[ 2 ] ;
    
    /**
     * コンストラクタ.
     */
    public HttpdRequestImpl() {
    }
    
    /**
     * 情報生成.
     */
    public void create() {
        clear() ;
    }
    
    /**
     * 情報クリア.
     */
    public void clear() {
        keyValue.clear() ;
        query.clear() ;
        header.clear() ;
        method = null ;
        httpVersion = null ;
        urlPath = null ;
        clientAccessURL = null ;
        remoteAddress = null ;
        remotePort = -1 ;
        session = null ;
        conn = null ;
        gzipFlag = false ;
        cookie = null ;
        cookieFlag = false ;
        readCookieFlag = false ;
        sslFlag = false ;
        dirFlag = false ;
        fullPath = null ;
        seabass32 = false ;
        seabass32Opt[0] = null ;
        seabass32Opt[1] = null ;
    }
    
    /**
     * オブジェクトを複製.
     * @return HttpdRequest 複製されたオブジェクトが返されます.
     * @exception Exception 例外.
     */
    public HttpdRequest copy() throws Exception {
        if( conn == null ) {
            throw new IOException( "オブジェクト構成は不正です" ) ;
        }
        ConnectionInfo c = conn.copy() ;
        HttpdRequestImpl ret = new HttpdRequestImpl() ;
        query.copy( ret.query ) ;
        header.copy( ret.header ) ;
        ret.method = method ;
        ret.httpVersion = httpVersion ;
        ret.urlPath = urlPath ;
        ret.clientAccessURL = clientAccessURL ;
        ret.remoteAddress = remoteAddress ;
        ret.remotePort = remotePort ;
        ret.session = session ;
        ret.conn = c ;
        ret.gzipFlag = gzipFlag ;
        ret.cookie = cookie ;
        ret.cookieFlag = cookieFlag ;
        ret.readCookieFlag = readCookieFlag ;
        ret.sslFlag = sslFlag ;
        ret.dirFlag = dirFlag ;
        ret.fullPath = fullPath ;
        ret.seabass32 = seabass32 ;
        ret.keyValue.clear() ;
        return ret ;
    }
    
    /**
     * HTTPメソッドを設定.
     * <BR><BR>
     * HTTPメソッドを設定します.
     * <BR>
     * @param method HTTPメソッドを設定します.
     */
    public void setMethod( String method ) {
        this.method = method ;
    }
    
    /**
     * HTTPメソッドを取得.
     * <BR><BR>
     * HTTPメソッドを取得します.
     * <BR>
     * @return String HTTPメソッドが返されます.
     */
    public String getMethod() {
        return method ;
    }
    
    /**
     * HTTPバージョンを設定.
     * <BR><BR>
     * HTTPバージョンを設定します.
     * <BR>
     * @param httpVersion HTTPバージョンが返されます.<BR>
     *                    [HTTP/0.9]の場合は、HTTPバージョン0.9です.<BR>
     *                    [HTTP/1.0]の場合は、HTTPバージョン1.0です.<BR>
     *                    [HTTP/1.1]の場合は、HTTPバージョン1.1です.
     */
    public void setHttpVersion( String httpVersion ) {
        this.httpVersion = httpVersion ;
    }
    
    /**
     * HTTPバージョンを取得.
     * <BR><BR>
     * HTTPバージョンを取得します.
     * <BR>
     * @return String HTTPバージョンが返されます.
     *                [HTTP/0.9]の場合は、HTTPバージョン0.9です.<BR>
     *                [HTTP/1.0]の場合は、HTTPバージョン1.0です.<BR>
     *                [HTTP/1.1]の場合は、HTTPバージョン1.1です.
     */
    public String getHttpVersion() {
        return httpVersion ;
    }
    
    /**
     * HTTPバージョン番号を取得.
     * <BR><BR>
     * HTTPバージョン番号を取得します.
     * <BR>
     * @return String HTTPバージョン番号が返されます.
     */
    public String getVersion() {
        return httpVersion.substring( "HTTP/".length() ) ;
    }
    
    /**
     * クライアントがアクセスした生URLを取得.
     * @return String URL相対パスが返されます.
     */
    public String getClientAccessURL() {
        return clientAccessURL ;
    }
    
    /**
     * URL相対パスを設定.
     * <BR><BR>
     * URL相対パスを設定します.
     * <BR>
     * @param urlPath URL相対パスを設定します.
     */
    public void setUrlPath( String urlPath ) {
        if( this.urlPath == null ) {
            clientAccessURL = urlPath ;
        }
        this.urlPath = urlPath ;
    }
    
    /**
     * URL相対パスを取得.
     * <BR><BR>
     * URL相対パスを取得します.
     * <BR>
     * @return String URL相対パスが返されます.
     */
    public String getUrlPath() {
        return urlPath ;
    }
    
    /**
     * リモートアドレス／ポートを設定.
     * @param addr 対象のアドレスを設定します.
     * @param port 対象のポート番号を設定します.
     */
    public void setRemote( String addr,int port ) {
        this.remoteAddress = addr ;
        this.remotePort = port ;
    }
    
    /**
     * リモートアドレスを取得.
     * <BR><BR>
     * リモートアドレスが返されます.
     * <BR>
     * @return String リモートアドレスが返されます.
     */
    public String getRemoteAddress() {
        return remoteAddress ;
    }
    
    /**
     * リモートポート番号を取得.
     * <BR><BR>
     * リモートポート番号が返されます.
     * <BR>
     * @return int リモートポート番号が返されます.
     */
    public int getRemotePort() {
        return remotePort ;
    }
    
    /**
     * クエリオブジェクトを取得.
     * <BR><BR>
     * クエリオブジェクトを取得します.
     * <BR>
     * @return HttpdParams クエリオブジェクトが返されます.
     */
    public HttpdParams getQuery() {
        return query ;
    }
    
    /**
     * ヘッダオブジェクトを取得.
     * <BR><BR>
     * ヘッダオブジェクトを取得します.
     * <BR>
     * @return HttpdHeaders ヘッダオブジェクトが返されます.
     */
    public HttpdHeaders getHeader() {
        return header ;
    }
    
    /**
     * クッキーを取得.
     * @return Map<String,String> クッキー情報が返されます.
     * @exception Exception 例外.
     */
    public Map<String,String> getCookie() throws Exception {
        if( !this.readCookieFlag ) {
            this.readCookieFlag = true ;
            String s = HttpdCookie.getCookie( this ) ;
            if( s != null ) {
                this.cookie = HttpdCookie.analysisCookie( s ) ;
            }
        }
        return this.cookie ;
    }
    
    /**
     * このリクエストに対するセッションオブジェクトを取得.
     * <BR><BR>
     * このリクエストに対するセッションオブジェクトが返されます.
     * <BR>
     * @return HttpdSession セッション情報が返されます.
     * @exception Exceptino 例外.
     */
    public HttpdSession getSession()
        throws Exception {
        if( session == null ) {
            // セッション付与を行うファイルタイプの場合.
            if( isSessionId() ) {
                HttpdSessionManager sessionManager = ( HttpdSessionManager )GlobalManager.getValue(
                    ServiceDef.MANAGER_BY_SESSION ) ;
                String id = null ;
                // CookieのセッションIDで処理.
                Map<String,String> ana = getCookie() ;
                if( ana != null ) {
                    id = ana.get( HttpdSession.SESSION_NAME ) ;
                    if( id != null && id.length() == HttpdSessionManager.SESSION_ID_LENGTH ) {
                        session = sessionManager.getSession( random,keyValue,id ) ;
                    }
                    cookieFlag = true ;
                }
                // パラメータのセッションIDで処理.
                if( session == null ) {
                    id = query.getParam( HttpdDef.SESSION_PARAM ) ;
                    if( id != null &&
                        id.length() == HttpdSessionManager.SESSION_ID_LENGTH &&
                        sessionManager.getSession( random,keyValue,id ) != null ) {
                        session = sessionManager.getSession( random,keyValue,id ) ;
                    }
                    else {
                        session = sessionManager.getNewSession( random,keyValue ) ;
                    }
                    // クエリに新しく割り当てられたセッションパラメータを付加.
                    query.setParam( HttpdDef.SESSION_PARAM,session.getSessionId() ) ;
                    cookieFlag = false ;
                }
            }
            // セッション付与を行わないタイプの場合.
            else {
                //cookieFlag = ( HttpdCookie.getCookie( this ) != null ) ;
                cookieFlag = false ;
            }
            //LOG.debug( "session:" + session.toString() ) ;
        }
        return session ;
    }
    
    /**
     * コネクション情報を設定.
     * <BR><BR>
     * コネクション情報を設定します.
     * <BR>
     * @param conn 対象のコネクション情報を設定します.
     */
    public void setConnectionInfo( ConnectionInfo conn ) {
        if( conn.getElement().getSslElement() != null ) {
            sslFlag = true ;
        }
        else {
            sslFlag = false ;
        }
        this.conn = conn ;
    }
    
    /**
     * コネクション情報を取得.
     * <BR><BR>
     * コネクション情報を取得します.
     * <BR>
     * @param ConnectionInfo 対象のコネクション情報が返されます.
     */
    public ConnectionInfo getConnectionInfo() {
        return this.conn ;
    }
    
    /**
     * KeepAliveタイムアウトを設定.
     * <BR><BR>
     * KeepAliveタイムアウトを設定します.
     * <BR>
     * @param timeout 対象のKeepAliveタイムアウト値を設定します.
     */
    public void setKeepAliveTimeout( int timeout ) {
        if( timeout > 0L ) {
            this.conn.setTimeout( timeout ) ;
        }
    }
    
    /**
     * KeepAliveタイムアウトを取得.
     * <BR><BR>
     * KeepAliveタイムアウトを取得します.
     * <BR>
     * @return long KeepAliveタイムアウト値が返されます.
     */
    public long getKeepAliveTimeout() {
        if( conn == null ) {
            return -1L ;
        }
        return conn.getTimeout() ;
    }
    
    /**
     * KeepAliveカウントを取得.
     * <BR><BR>
     * KeepAliveカウントを取得します.
     * <BR>
     * @return int KeepAliveカウント値が返されます.
     */
    public int getKeepAliveCount() {
        if( conn == null ) {
            return -1 ;
        }
        return this.conn.getCount() ;
    }
    
    /**
     * ランダムオブジェクトを取得.
     * @return BaseRandom ランダムオブジェクトが返されます.
     */
    public BaseRandom getRandom() {
        return random ;
    }
    
    /**
     * BdbKeyValueオブジェクトを取得.
     * @return BdbKeyValue BdbKeyValueオブジェクトが返されます.
     */
    public BdbKeyValue getBdbKeyValue() {
        return keyValue ;
    }
    
    /**
     * GZIP圧縮を許可するか設定.
     * <BR><BR>
     * SeaBassがGZIP圧縮を許可している場合[true]を設定します.
     * <BR>
     * @param flag [true]の場合、GZIP圧縮を許可します.
     */
    public void setGzip( boolean flag ) {
        this.gzipFlag = flag ;
    }
    
    /**
     * イントラネットアドレスでアクセスされたかチェック.
     * @return boolean [true]の場合、イントラネットアドレスでアクセスされています.
     */
    public boolean isIntranetAddress() {
        return this.conn.isIntranetAddress() ;
    }
    
    /**
     * GZIP圧縮を許可しているかチェック.
     * <BR><BR>
     * SeaBassがGZIP圧縮を許可しているかチェックします.
     * <BR>
     * @return boolean [true]の場合、許可されています.
     */
    public boolean isGzip() {
        return gzipFlag ;
    }
    
    /**
     * Cookieが存在するかチェック.
     * <BR><BR>
     * Cookieが存在するかチェックします.
     * <BR>
     * @return boolean [true]の場合、Cookieが存在します.
     */
    public boolean isCookie() {
        return cookieFlag ;
    }
    
    /**
     * 対象プロトコルがSSLであるかチェック.
     * <BR><BR>
     * 対象プロトコルがSSLであるかチェックします.
     * <BR>
     * @return boolean [true]の場合、プロトコルはSSLです.
     */
    public boolean isSsl() {
        return sslFlag ;
    }
    
    /**
     * このリクエストが有効かチェック.
     * <BR><BR>
     * このリクエストが有効かチェックします.
     * <BR>
     * @return [true]が返された場合、有効です.
     */
    public boolean isUse() {
        return ( conn == null || conn.isUse() == false ) ? false : true ;
    }
    
    /**
     * IPアドレス、ポート番号を文字列として取得.
     * <BR><BR>
     * IPアドレス、ポート番号を文字列として取得します.
     * <BR>
     * @return String 文字列が返されます.
     */
    public String toIpPort() {
        return new StringBuilder().append( conn.getInetAddress().getHostAddress() ).
            append( "/" ).append( conn.getPort() ).toString() ;
    }
    
    /**
     * フルパスを設定.
     * @param path 対象のフルパスを設定します.
     */
    public void setFullPath( String path ) {
        this.fullPath = path ;
    }
    
    /**
     * フルパスを取得.
     * @return String フルパスが返されます.
     */
    public String getFullPath() {
        return fullPath ;
    }
    
    /**
     * Seabass32暗号フラグを設定.
     * @param flg [true]を設定した場合、Seabass32暗号化フラグがONになります.
     * @param host 対象のホスト名を設定します.
     * @param sessionId 対象のセッションIDを設定します.
     */
    public void setSeabass32( boolean flg,String host,String sessionId ) {
        this.seabass32 = flg ;
        if( seabass32 ) {
            seabass32Opt[ 0 ] = host ;
            seabass32Opt[ 1 ] = sessionId ;
        }
    }
    
    /**
     * Seabass32暗号フラグを取得.
     * @return boolean [true]の場合、Seabass暗号化フラグはONです.
     */
    public boolean isSeabass32() {
        return seabass32 ;
    }
    
    /**
     * Seabass32暗号化オプションを取得.
     * @return String[] Seabass32暗号化オプションが返されます.
     */
    public String[] getSeabass32Opt() {
        return seabass32Opt ;
    }
    
    /**
     * SessionId付与許可拡張子.
     */
    private static final String[] SESSION_LIST = {
        ".ms",".mhtml",".js" } ;
    
    /**
     * SessionID付与許可拡張子かチェック.
     * @return boolean [true]の場合、セッションIDを付与します.
     */
    protected boolean isSessionId() {
        String path = this.urlPath ;
        // 拡張ヘッダ：非セッション付与処理が指定されている場合.
        if( "true".equals( header.getHeader( HttpdDef.NOT_SESSION,0 ) ) ) {
            return false ;
        }
        int x = path.lastIndexOf( "/" ) ;
        if( x >= 0 ) {
            path = path.substring( x ) ;
        }
        // 拡張子なしの場合、セッションIDを付与する.
        if( path.lastIndexOf( "." ) <= -1 ) {
            return true ;
        }
        int len = SESSION_LIST.length ;
        path = path.trim().toLowerCase() ;
        for( int i = 0 ; i < len ; i ++ ) {
            if( path.endsWith( SESSION_LIST[ i ] ) ) {
                return true ;
            }
        }
        return false ;
    }
    
    /**
     * アクセスURLがディレクトリの場合設定.
     */
    public void toDirectory() {
        dirFlag = true ;
    }
    
    /**
     * アクセスURLがディレクトリかチェック.
     * @return boolean [true]の場合、ディレクトリです.
     */
    public boolean isDirectory() {
        return dirFlag ;
    }
}
