package org.maachang.seabass.net.nio ;

import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;

import org.maachang.seabass.net.NetInfo;
import org.maachang.seabass.net.NioNotConnectException;

/**
 * １つのコネクションオブジェクト.
 * 
 * @version 2008/05/28
 * @author  masahito suzuki
 * @since   SeaBass 1.0.0
 */
class ConnectionInfoImpl implements ConnectionInfo {
    
    private KeepAliveBean keepAlive = null ;
    private NioElement element = null ;
    private InetAddress address = null ;
    private int port = -1 ;
    private boolean cloesFlag = false ;
    private boolean cometFlag = false ;
    private boolean localFlag = false ;
    private boolean isIntranetFlag = false ;
    private final NioInputStream inputStream = new NioInputStream() ;
    private final NioOutputStream outputStream = new NioOutputStream() ;
    
    /**
     * コンストラクタ.
     */
    public ConnectionInfoImpl() {
        
    }
    
    protected void finalize() throws Exception {
        this.destroy() ;
    }
    
    /**
     * オブジェクト破棄.
     */
    public void destroy() {
        if( element != null ) {
            element.destroy() ;
        }
        inputStream.close() ;
        outputStream.clear() ;
        keepAlive = null ;
        element = null ;
        address = null ;
        port = -1 ;
        cloesFlag = false ;
        cometFlag = false ;
        localFlag = false ;
        isIntranetFlag = false ;
    }
    
    /**
     * 生成処理.
     * @param element 対象のNIO要素を設定します.
     * @return ConnectionInfo このオブジェクトが返されます.
     * @exception Exception 例外.
     */
    public ConnectionInfo create( NioElement element )
        throws Exception {
        this.clear() ;
        if( element == null ) {
            throw new IllegalArgumentException( "通信要素がnullで渡されました" ) ;
        }
        else if( element.isUse() == false ) {
            return null ;
        }
        KeepAliveBean keepAlive = ( KeepAliveBean )element.getObject() ;
        if( keepAlive == null ) {
            keepAlive = KeepAliveFactory.getInstance().getBean() ;
            element.setObject( keepAlive ) ;
        }
        InetAddress address = element.getChannel().socket().getInetAddress() ;
        String netAddr = address.getHostAddress() ;
        this.element = element ;
        this.keepAlive = keepAlive ;
        this.address = address ;
        this.port = element.getChannel().socket().getPort() ;
        this.localFlag = NetInfo.THIS_IP_LIST.contains( netAddr ) ;
        this.isIntranetFlag = ( netAddr.startsWith( "10." ) ||
            netAddr.startsWith( "172.16." ) ||
            netAddr.startsWith( "192.168." ) ||
            netAddr.equals( "127.0.0.1" ) ) ;
        return this ;
    }
    
    /**
     * オブジェクトクリア.
     */
    public void clear() {
        inputStream.close() ;
        if( cometFlag ) {
            outputStream.cometClear() ;
        }
        else {
            outputStream.clear() ;
        }
        keepAlive = null ;
        element = null ;
        address = null ;
        port = -1 ;
        cloesFlag = false ;
        cometFlag = false ;
        localFlag = false ;
        isIntranetFlag = false ;
    }
    
    /**
     * オブジェクト複製.
     * @return ConnectionInfo 複製されたオブジェクトが返されます.
     * @exception Exception 例外.
     */
    public ConnectionInfo copy() throws Exception {
        if( isUse() == false ) {
            throw new NioNotConnectException( "オブジェクトは既に破棄されています" ) ;
        }
        ConnectionInfoImpl ret = new ConnectionInfoImpl() ;
        // comet用として、InputStreamを停止.
        inputStream.close() ;
        ret.keepAlive = keepAlive ;
        ret.element = element ;
        ret.address = address ;
        ret.port = port ;
        ret.cloesFlag = cloesFlag ;
        ret.cometFlag = cometFlag ;
        ret.localFlag = localFlag ;
        ret.isIntranetFlag = isIntranetFlag ;
        return ret ;
    }
    
    /**
     * Nio要素を取得.
     * @return NioElement Nio要素が返されます.
     */
    public NioElement getElement() {
        if( !isUseElement() ) {
            throw new NioNotConnectException( "オブジェクトは既に破棄されています" ) ;
        }
        return element ;
    }
    
    /**
     * KeepAliveタイムアウト値を設定.
     * @param time KeepAliveタイムアウト値を設定します.
     */
    public void setTimeout( int time ) {
        if( !isUseElement() ) {
            throw new NioNotConnectException( "オブジェクトは既に破棄されています" ) ;
        }
        keepAlive.setTime( time ) ;
    }
    
    /**
     * KeepAliveタイムアウト値を取得.
     * @return long KeepAliveタイムアウト値が返されます.
     */
    public long getTimeout() {
        if( !isUseElement() ) {
            throw new NioNotConnectException( "オブジェクトは既に破棄されています" ) ;
        }
        return keepAlive.getTime() ;
    }
    
    /**
     * KeepAliveカウント値を取得.
     * @return int KeepAliveカウント値が返されます.
     */
    public int getCount() {
        if( !isUseElement() ) {
            throw new NioNotConnectException( "オブジェクトは既に破棄されています" ) ;
        }
        return keepAlive.getCount() ;
    }
    
    /**
     * OutputStreamを取得.
     * @return OutputStream OutputStreamが返されます.
     */
    public OutputStream getOutputStream()
        throws Exception {
        if( !isUseElement() ) {
            throw new NioNotConnectException( "オブジェクトは既に破棄されています" ) ;
        }
        return outputStream.create( element ) ;
    }
    
    /**
     * InputStreamを取得.
     * @return InputStreamが返されます.
     */
    public InputStream getInputStream()
        throws Exception {
        if( !isUseElement() ) {
            throw new NioNotConnectException( "オブジェクトは既に破棄されています" ) ;
        }
        return inputStream.create( element ) ;
    }
    
    /**
     * InputStreamを取得.
     * @param timeout タイムアウト値を設定します.
     * @return InputStreamが返されます.
     */
    public InputStream getInputStream( long timeout )
        throws Exception {
        if( !isUseElement() ) {
            throw new NioNotConnectException( "オブジェクトは既に破棄されています" ) ;
        }
        return inputStream.create( element,timeout ) ;
    }
    
    /**
     * 接続先アドレスを取得.
     * @return InetAddress 接続先アドレスが返されます.
     */
    public InetAddress getInetAddress() {
        if( !isUseElement() ) {
            throw new NioNotConnectException( "オブジェクトは既に破棄されています" ) ;
        }
        return address ;
    }
    
    /**
     * 接続先ポートを取得.
     * @return int 接続先ポート番号が返されます.
     */
    public int getPort() {
        if( !isUseElement() ) {
            throw new NioNotConnectException( "オブジェクトは既に破棄されています" ) ;
        }
        return port ;
    }
    
    /**
     * コネクションクローズフラグを設定.
     * @param mode [true]の場合、コネクションクローズを行います.
     */
    public void setCloseFlag( boolean mode ) {
        if( !isUseElement() ) {
            throw new NioNotConnectException( "オブジェクトは既に破棄されています" ) ;
        }
        cloesFlag = mode ;
        element.setCloseFlag( mode ) ;
    }
    
    /**
     * KeepAliveが有効な場合のコネクション再利用チェック.
     * @return boolean [true]の場合、再利用可能です.
     */
    public boolean recyclingConnection() {
        if( !isUseElement() ) {
            throw new NioNotConnectException( "オブジェクトは既に破棄されています" ) ;
        }
        int cnt = keepAlive.getCount() - 1 ;
        if( cnt <= 0 ) {
            //KeepAliveFactory.getInstance().getBean( keepAlive ) ;
            //return true ;
            element.setCloseFlag( true ) ;
            return false ;
        }
        keepAlive.setCount( cnt ) ;
        element.setExecutionFlag( false ) ;
        element.setKeepAliveTime( keepAlive.getTime() ) ;
        return true ;
    }
    
    /**
     * ローカルアドレスでアクセスされたかチェック.
     * @return boolean [true]の場合、ローカルアドレスでアクセスされています.
     */
    public boolean isLocalAddress() {
        return localFlag ;
    }
    
    /**
     * イントラネットアドレスでアクセスされたかチェック.
     * @return boolean [true]の場合、イントラネットアドレスでアクセスされています.
     */
    public boolean isIntranetAddress() {
        return isIntranetFlag ;
    }
    
    /**
     * コネクションクローズフラグを取得.
     * @return boolean [true]の場合、コネクションはクローズします.
     */
    public boolean isCloseFlag() {
        return cloesFlag ;
    }
    
    /**
     * オブジェクト有効チェック.
     * @return boolean [true]の場合、有効です.
     */
    public boolean isUse() {
        boolean ret = ( element != null && element.getChannel() != null ) ;
        if( ret == false && element != null ) {
            destroy() ;
        }
        return ret ;
    }
    
    /**
     * 通信要素有効チェック.
     * @return boolean [true]の場合、有効です.
     */
    public boolean isUseElement() {
        return (element != null) ;
    }
    
    /**
     * コメット実行設定.
     */
    public void cometOn() {
        cometFlag = true ;
    }
    
    /**
     * コメット実行フラグ.
     */
    public boolean isComet() {
        return cometFlag ;
    }
}
