package org.maachang.seabass.net.nio ;

import org.maachang.util.atomic.AtomicINT;
import org.maachang.util.atomic.AtomicLONG;


/**
 * KeepAlive情報を管理.
 * 
 * @version 2008/05/28
 * @author  masahito suzuki
 * @since   SeaBass 1.0.0
 */
public class KeepAliveFactory {
    
    /** クローズ条件破棄時間. **/
    protected static final int CLOSE_TIME = 1 ;
    
    /**
     * デフォルト通信タイムアウト値.
     */
    private static final long DEF_TIMEOUT = 5000L ;
    
    /**
     * デフォルトSSL通信タイムアウト値.
     */
    private static final long DEF_SSL_TIMEOUT = 5000L ;
    
    /**
     * 追加タイムアウト値.
     */
    private static final long ADD_TIME = 2500L ;
    
    /**
     * デフォルトKeepAlive時間.
     */
    private static final int DEF_TIME = 3 ;
    
    /**
     * デフォルトKeepAliveカウント.
     */
    private static final int DEF_COUNT = 100 ;
    
    /**
     * タイムアウト時間.
     */
    private final AtomicLONG timeout = new AtomicLONG( -1L ) ;
    
    /**
     * SSLタイムアウト時間.
     */
    private final AtomicLONG sslTimeout = new AtomicLONG( -1L ) ;
    
    /**
     * KeepAlive時間.
     */
    private final AtomicINT time = new AtomicINT( -1 ) ;
    
    /**
     * KeepAlive回数.
     */
    private final AtomicINT count = new AtomicINT( -1 ) ;
    
    /**
     * シングルトン.
     */
    private static final KeepAliveFactory SNGL = new KeepAliveFactory() ;
    
    /**
     * オブジェクトを取得.
     * @return KeepAliveFactory オブジェクトが返されます.
     */
    public static final KeepAliveFactory getInstance() {
        return SNGL ;
    }
    
    /**
     * コンストラクタ.
     */
    private KeepAliveFactory() {
        timeout.set( DEF_TIMEOUT ) ;
        sslTimeout.set( DEF_SSL_TIMEOUT ) ;
        time.set( DEF_TIME ) ;
        count.set( DEF_COUNT ) ;
    }
    
    /**
     * KeepAlive値を設定.
     * @param time 対象の時間を設定します.
     * @param count 対象のカウントを設定します.
     */
    public void set( int time,int count ) {
        this.time.set( time ) ;
        this.count.set( count ) ;
    }
    
    /**
     * 通信タイムアウト値を設定.
     * @param timeout 通信タイムアウト値を設定します.
     */
    public void setTimeout( long timeout ) {
        this.timeout.set( timeout ) ;
    }
    
    /**
     * SSL通信タイムアウト値を設定.
     * @param timeout SSL通信タイムアウト値を設定します.
     */
    public void setSslTimeout( long timeout ) {
        this.sslTimeout.set( timeout ) ;
    }
    
    /**
     * 新しいKeepAlive値を取得.
     * @return KeepAliveBean 新しいKeepAliveBeanが返されます.
     */
    public KeepAliveBean getBean() {
        return new KeepAliveBean( time.get(),count.get() ) ;
    }
    
    /**
     * 新しいKeepAlive値を取得.
     * @param out 既存のKeepAliveBeanを設定します.
     */
    public void getBean( KeepAliveBean out ) {
        out.setTime( time.get() ) ;
        out.setCount( count.get() ) ;
    }
    
}
