package org.maachang.seabass.net.nio ;

import java.io.IOException;
import java.io.OutputStream;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.maachang.manager.GlobalManager;
import org.maachang.seabass.ServiceDef;
import org.maachang.seabass.conf.MimeConfig;
import org.maachang.seabass.httpd.HttpdErrorDef;
import org.maachang.seabass.httpd.HttpdHeaders;
import org.maachang.seabass.httpd.HttpdVersionDef;
import org.maachang.seabass.httpd.engine.HttpdDef;
import org.maachang.seabass.httpd.engine.HttpdTimestamp;
import org.maachang.seabass.httpd.engine.script.scripts.ScriptManager;
import org.maachang.util.FileUtil;
import org.maachang.util.StringUtil;

/**
 * HTTPD出力用.
 * 
 * @version 2007/08/19
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
public class NetWriteHttpd {
    
    /**
     * キャラクタセット.
     */
    private static final String CHARSET = "UTF8" ;
    
    /**
     * 改行情報.
     */
    private static final String ENTER = "\r\n" ;
    
    /**
     * URL.
     */
    private String url = null ;
    
    /**
     * データ出力用.
     */
    private OutputStream outputStream = null ;
    
    /**
     * ヘッダキャッシュ.
     */
    private final OutputHeader cacheHeader = new OutputHeader() ;
    
    /**
     * コンテンツ長.
     */
    private String contentLength = null ;
    
    /**
     * GZIPモード.
     */
    private boolean gzipFlag = false ;
    
    /**
     * GZIP名.
     */
    private String gzipName = null ;
    
    /**
     * not-cacheフラグ.
     */
    private boolean notCacheFlag = false ;
    
    /**
     * クローズフラグ.
     */
    private boolean closeFlag = true ;
    
    /**
     * KeepAliveFlag.
     */
    private boolean keepAliveFlag = false ;
    
    /**
     * KeepAlive情報.
     */
    private String[] keepAliveInfo = null ;
    
    /**
     * chunkedFlag.
     */
    private boolean chunkedFlag = false ;
    
    /**
     * ContentLengthFlag.
     */
    private boolean contentLengthFlag = false ;
    
    /**
     * Connectionなしフラグ.
     */
    private boolean noConnection = false ;
    
    /**
     * HTTPバージョン.
     */
    private String version = null ;
    
    /**
     * HTTPバージョン(Double).
     */
    private double versionDouble = 0.0d ;
    
    /**
     * ステータス.
     */
    private int state = 200 ;
    
    /**
     * 生成フラグ.
     */
    private boolean useFlag = false ;
    
    /**
     * Output生成フラグ.
     */
    private boolean newOutputFlag = false ;
    
    /**
     * コンストラクタ.
     */
    public NetWriteHttpd() {
        
    }
    
    /**
     * 情報生成.
     * <BR><BR>
     * 条件を指定してオブジェクトを設定します.
     * <BR>
     * @param connectionInfo コネクション情報を設定します.
     * @param url 対象のURLを設定します.
     * @param state 対象のステータスを設定します.
     * @param version HTTPバージョンを設定します.
     * @param versionDouble HTTPバージョン(double)を設定します.
     * @param noConnection コネクションを付加する場合は、falseを設定します.
     * @exception Exception 例外.
     */
    public void create( ConnectionInfo connectinoInfo,String url,int state,String version,double versionDouble,
        boolean noConnection )
        throws Exception {
        if( connectinoInfo == null || connectinoInfo.isUse() == false ) {
            throw new IOException( "コネクションオブジェクトは無効です" ) ;
        }
        if( url == null || ( url = url.trim() ).length() <= 0 ) {
            throw new IOException( "URLは不正です" ) ;
        }
        this.close() ;
        if( version == null || ( version = version.trim() ).length() <= 0 ) {
            version = HttpdDef.VERSION_11 ;
        }
        this.version = version ;
        this.versionDouble = versionDouble ;
        this.noConnection = noConnection ;
        this.state = state ;
        this.outputStream = connectinoInfo.getOutputStream() ;
        this.url = url ;
        this.useFlag = true ;
    }
    
    /**
     * オブジェクトクリア.
     */
    public void clear() {
        url = null ;
        outputStream = null ;
        cacheHeader.clear() ;
        contentLength = null ;
        gzipFlag = false ;
        gzipName = null ;
        notCacheFlag = false ;
        closeFlag = true ;
        keepAliveFlag = false ;
        keepAliveInfo = null ;
        noConnection = false ;
        chunkedFlag = false ;
        contentLengthFlag = false ;
        version = null ;
        versionDouble = 0.0d ;
        state = 200 ;
        useFlag = false ;
        newOutputFlag = false ;
    }
    
    /**
     * オブジェクトクローズ.
     * <BR><BR>
     * オブジェクトをクローズします.
     */
    public void close() {
        if( outputStream != null ) {
            try {
                outputStream.close() ;
            } catch( Exception e ) {
            }
        }
        clear() ;
    }
    
    /**
     * オブジェクト更新.
     */
    public void finish() throws IOException {
        if( outputStream != null ) {
            try {
                outputStream.flush() ;
            } catch( Exception e ) {
            }
        }
    }
    
    /**
     * 基本ヘッダを生成.
     * <BR><BR>
     * 基本ヘッダを生成します.
     */
    public void baseHeader() {
        outputHttp( version,state ) ;
    }
    
    /**
     * ヘッダを追加.
     * <BR><BR>
     * ヘッダを追加します.
     * <BR>
     * @param key 対象のキーを設定します.
     * @param value 対象のValueを設定します.
     */
    public void pushHeader( String key,String value ) {
        this.outputHeader( key,value ) ;
    }
    
    /**
     * ヘッダ群を追加.
     * <BR><BR>
     * ヘッダ群を追加します.
     * <BR>
     * @param header 対象のヘッダ群を設定します.
     */
    public void pushHeaders( HttpdHeaders header ) {
        if( header == null || header.size() <= 0 ) {
            return ;
        }
        int len ;
        String key ;
        Iterator it = header.iterator() ;
        while( it.hasNext() ) {
            key = ( String )it.next() ;
            len = header.size( key ) ;
            for( int i = 0 ; i < len ; i ++ ) {
                this.outputHeader( key,header.getHeader( key,i ) ) ;
            }
        }
    }
    
    /**
     * キャッシュをOFFに設定.
     * <BR><BR>
     * HTTPキャッシュをOFFに設定します.
     */
    public void cacheOff() {
        notCacheFlag = true ;
    }
    
    /**
     * クローズヘッダをONに設定.
     * <BR><BR>
     * HTTPクローズヘッダをONに設定します.
     */
    public void closeOn() {
        closeFlag = true ;
    }
    
    /**
     * クローズヘッダをOFFに設定.
     * <BR><BR>
     * HTTPクローズヘッダをOFFに設定します.
     */
    public void closeOff() {
        closeFlag = false ;
    }
    
    /**
     * クローズヘッダ条件を取得.
     * <BR><BR>
     * HTTPクローズヘッダ条件を取得します.
     * <BR>
     * @return boolean [true]の場合、ONになります.
     */
    public boolean isClose() {
        return closeFlag ;
    }
    
    /**
     * チャングフラグを取得.
     * <BR><BR>
     * チャングフラグを取得します.
     * <BR>
     * @return boolean [true]の場合は、チャングは有効です.
     */
    public boolean isChunked() {
        return chunkedFlag ;
    }
    
    /**
     * コンテンツ長がヘッダに設定されているかを取得.
     * @return boolean [true]の場合、コンテンツフラグが設定されています.
     */
    public boolean isContentLengthHeader() {
        return contentLengthFlag ;
    }
    
    /**
     * GZIPモードを取得.
     * @return boolean [true]の場合、GZIP圧縮が必要です.
     */
    public boolean isGZIP() {
        return gzipFlag ;
    }
    
    /**
     * GZIP名を設定.
     * @param name GZIP名を設定します.
     */
    public void setGZIPName( String name ) {
        this.gzipName = name ;
    }
    
    /**
     * KeepAlive情報を設定.
     * @param timeout タイムアウト値を設定します.
     * @param max 最大値を設定します.
     */
    public void setKeepAlive( String timeout,String max ) {
        keepAliveInfo = new String[]{ timeout,max } ;
        keepAliveFlag = true ;
    }
    
    /**
     * OutputStreamを取得.
     * <BR><BR>
     * OutputStreamを取得します.
     * <BR>
     * @param seabass32 Seabass32暗号フラグを設定します.
     * @return OutputStream OutputStreamが返されます.
     * @exception Exception 例外.
     */
    public OutputStream getOutputStream( boolean seabass32 )
        throws Exception {
        if( this.outputStream == null ) {
            throw new IOException( "コネクションオブジェクトは生成されていません" ) ;
        }
        OutputStream ret = null ;
        if( !cacheHeader.isClear() ) {
            writeHeader( seabass32 ) ;
            ret = this.outputStream ;
        }
        else {
            ret = this.outputStream ;
        }
        //if( !newOutputFlag ) {
        //    NioElement em = ((NioOutputStream)this.outputStream).getNioElement() ;
            // コネクションクローズで、SSLアクセスでない場合、
            // ソケット入力をシャットダウンする.
            //if( closeFlag && !em.isSSL() ) {
            //    em.shutdownInput() ;
            //}
        //    newOutputFlag = true ;
        //}
        return ret ;
    }
    
    /**
     * ヘッダ情報が出力可能かチェック.
     * <BR><BR>
     * ヘッダ情報が出力可能かチェックします.
     * <BR>
     * @return boolean ヘッダが出力可能かチェックします.
     */
    public boolean isWriteHeader() {
        return !cacheHeader.isClear() ;
    }
    
    /**
     * 指定ヘッダが存在するかチェック.
     * <BR><BR>
     * 指定ヘッダが存在するかチェックします.
     * <BR>
     * @param key 対象のキーを設定します.
     * @return boolean [true]の場合存在します.
     */
    public boolean isHeader( String key ) {
        return cacheHeader.isHeader( key ) ;
    }
    
    /**
     * コンテンツ長が設定されているかチェック.
     * <BR><BR>
     * コンテンツ長が設定されているかチェックします.
     * <BR>
     * @return boolean [true]の場合、設定されています.
     */
    public boolean isContentLength() {
        return ( contentLength != null ) ? true : false ;
    }
    
    /**
     * オブジェクトが生成されているかチェック.
     * @return boolean [true]の場合、生成されています.
     */
    public boolean useObject() {
        return useFlag ;
    }
    
    /**
     * outputStreamのflushが必要がチェック.
     * @return boolean [true]の場合、flush処理が必要です.
     */
    public boolean isFlush() {
        return this.outputStream != null &&
            ((NioOutputStream)this.outputStream).isData() ;
    }
    
    /**
     * HTTP情報を出力.
     */
    private void outputHttp( String version,int state ) {
        if( !cacheHeader.isClear() ) {
            cacheHeader.clear() ;
        }
        cacheHeader.put( "HTTP/" ).put( version ).
            put( " " ).put( String.valueOf( state ) ).put( " " ).
            put( HttpdErrorDef.convertErrorMessage( state ) ).put( " " ).
            put( ENTER ) ;
        // デフォルトヘッダを出力.
        cacheHeader.add( "Server",HttpdVersionDef.getPrintServerName() ) ;
        cacheHeader.add( "Date",HttpdTimestamp.getNowTimestamp() ) ;
        // コネクションなしフラグがOFF.
        if( !noConnection ) {
            // HTTPバージョンが1.1以外の場合は、KeepAliveは無効とする.
            if( versionDouble <= 1.0d ) {
                closeFlag = true ;
                chunkedFlag = false ;
            }
        }
    }
    
    /**
     * キャッシュOFFの設定.
     */
    private void notCache() {
        if( !cacheHeader.isHeader( "Expires" ) && !cacheHeader.isHeader( "Pragma" ) ) {
            cacheHeader.add( "Expires","Thu, 01 Jan 1970 01:00:00 GMT" ) ;
            //cacheHeader.add( "Expires",HttpdTimestamp.getTimestamp( 0L ) );
            //cacheHeader.add( "Expires","-1" );
            cacheHeader.add( "Pragma","no-cache" ) ;
            if( isHeader( HttpdDef.VALUE_LAST_UPDATE ) == false ) {
                cacheHeader.add( HttpdDef.VALUE_LAST_UPDATE,HttpdTimestamp.getNowTimestamp() ) ;
            }
            cacheHeader.add( "Cache-Control","private" ) ;
            cacheHeader.add( "Cache-Control","no-cache" ) ;
            cacheHeader.add( "Cache-Control","no-store" ) ;
            cacheHeader.add( "Cache-Control","max-age=0" ) ;
        }
    }
    
    /** ヘッダ追加条件. **/
    private static final String CONTENT_LENGTH = "content-length" ;
    private static final Set<String> noSetHeaderNames ;
    static {
        noSetHeaderNames = new HashSet<String>() ;
        noSetHeaderNames.add( "connection" ) ;
        noSetHeaderNames.add( "transfer-encoding" ) ;
        noSetHeaderNames.add( "content-encoding" ) ;
        noSetHeaderNames.add( "keep-alive" ) ;
    }
    
    /**
     * ヘッダ情報を出力.
     */
    private void outputHeader( String key,String value ) {
        if( key == null || ( key = key.trim() ).length() <= 0 ||
            value == null || ( value = value.trim() ).length() <= 0 ) {
            return ;
        }
        String k = key.toLowerCase() ;
        // 外部から設定できないヘッダ条件の場合は、設定しない.
        if( noSetHeaderNames.contains( k ) ) {
            return ;
        }
        // コンテンツ長を示すヘッダ定義の場合は、その内容を保管.
        else if( k.equals( CONTENT_LENGTH ) ) {
            if( value != null && ( value = value.trim() ).length() > 0 && StringUtil.isNumber( value ) ) {
                if( "0".equals( value ) ) {
                    // content-length :0の場合は、クローズモードをONにセット.
                    closeFlag = true ;
                }
                contentLength = value ;
            }
        }
        else {
            cacheHeader.add( key,value ) ;
        }
    }
    
    /**
     * ヘッダ情報を出力.
     */
    private void writeHeader( boolean seabass32 )
        throws Exception {
        if( !cacheHeader.isClear() ) {
            try {
                
                // closeモードがONの場合は、KeepAliveは行わない.
                if( closeFlag ) {
                    keepAliveFlag = false ;
                }
                
                // mimeTypeが設定されていない場合.
                if( !isContentType( cacheHeader ) ) {
                    if( state >= 400 ) {
                        cacheHeader.add( HttpdDef.VALUE_CONTENT_TYPE,HttpdDef.ERROR_PAGE_MIME_TYPE ) ;
                    }
                    else {
                        noSetByContentType( cacheHeader,url ) ;
                    }
                }
                
                boolean version11Flag = versionDouble >= 1.1d ;
                // HTTPバージョンが1.1の場合.
                if( version11Flag ) {
                    // チャンクは有効.
                    chunkedFlag = true ;
                }
                else {
                    // チャンク／KeepAliveは無効.
                    keepAliveFlag = false ;
                    chunkedFlag = false ;
                    closeFlag = true ;
                }
                
                // KeepAliveが有効な場合.
                if( keepAliveFlag && keepAliveInfo != null ) {
                    cacheHeader.add( "Connection","Keep-Alive" ) ;
                    // KeepAliveヘッダをセット.
                    if( keepAliveInfo[ 0 ] != null ) {
                        cacheHeader.add( "Keep-Alive",keepAliveInfo[ 0 ] ) ;
                    }
                    if( keepAliveInfo[ 1 ] != null ) {
                        cacheHeader.add( "Keep-Alive",keepAliveInfo[ 1 ] ) ;
                    }
                }
                // KeepAliveが無効な場合.
                else {
                    cacheHeader.add( "Connection","close" ) ;
                    keepAliveFlag = false ;
                    closeFlag = true ;
                }
                
                // コンテンツ長が設定されている場合.
                if( contentLength != null && !cacheHeader.isHeader( "Content-Length" ) ) {
                    cacheHeader.add( "Content-Length",contentLength ) ;
                    contentLength = null ;
                    gzipName = null ;
                    chunkedFlag = false ;
                }
                // チャングが有効な場合.
                else if( chunkedFlag && !cacheHeader.isHeader( "Transfer-Encoding" ) ) {
                    cacheHeader.add( "Transfer-Encoding","chunked" ) ;
                    // GZIP圧縮条件が存在する場合.
                    if( gzipName != null ) {
                        cacheHeader.add( "Content-Encoding",gzipName ) ;
                        gzipFlag = true ;
                    }
                }
                
                // 非キャッシュ設定の場合.
                if( notCacheFlag ) {
                    notCache() ;
                }
                // チャンクヘッダが付加されている場合.
                chunkedFlag = cacheHeader.isHeader( "Transfer-Encoding" ) ;
                // コンテンツ長が付加されている場合.
                contentLengthFlag = cacheHeader.isHeader( "Content-Length" ) ;
                
                // ヘッダ終点を設定.
                String hd = cacheHeader.get() ;
//System.out.println( "################## httpVersion:" + versionDouble ) ;
//System.out.println( "################## gzip:" + gzipFlag ) ;
//System.out.println( "################## chunkedFlag:" + chunkedFlag ) ;
//System.out.println( "################## header:" + hd ) ;
                byte[] writeHd = hd.getBytes( CHARSET ) ;
                hd = null ;
                this.outputStream.write( writeHd,0,writeHd.length ) ;
                this.outputStream.flush() ;
                writeHd = null ;
            } catch( Exception e ) {
                if( this.outputStream != null ) {
                    try {
                        finish() ;
                    } catch( Exception ee ) {
                    }
                    try {
                        this.outputStream.close() ;
                    } catch( Exception ee ) {
                    }
                    this.outputStream = null ;
                }
                throw e ;
            }
        }
    }
    
    /*
     * コンテンツタイプが設定されていない場合.
     */
    private static final void noSetByContentType( OutputHeader buf,String url ) {
        url = FileUtil.getFileName( url ) ;
        if( url.indexOf( "." ) == -1 && url.endsWith( ScriptManager.SCRIPT_PLUS ) == true ) {
            buf.add( HttpdDef.VALUE_CONTENT_TYPE,HttpdDef.AJAX_MIME_TYPE ) ;
        }
        else if( url.endsWith( ScriptManager.SCRIPT_HTML_PLUS ) == true ) {
            buf.add( HttpdDef.VALUE_CONTENT_TYPE,HttpdDef.MHTML_MIME_TYPE ) ;
        }
        else {
            MimeConfig conf = ( MimeConfig )GlobalManager.getValue(
                ServiceDef.MANAGER_BY_MIME_TYPE ) ;
            String mime = conf.getNameByMimeType( url ) ;
            if( mime.indexOf( "xml" ) != -1 || mime.indexOf( "html" ) != -1 ) {
                buf.add( HttpdDef.VALUE_CONTENT_TYPE,HttpdDef.MIME_CHARSET_UTF8 ) ;
            }
            else {
                buf.add( HttpdDef.VALUE_CONTENT_TYPE,mime ) ;
            }
        }
    }
    
    /**
     * コンテンツタイプが設定されていないかチェック.
     */
    private static final boolean isContentType( OutputHeader buf ) {
        return buf.isHeader( HttpdDef.VALUE_CONTENT_TYPE ) ;
    }
}
