package org.maachang.seabass.net.nio ;

import java.io.IOException;
import java.net.Socket;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.seabass.net.NioException;
import org.maachang.seabass.net.ssl.SslElement;
import org.maachang.seabass.net.ssl.SslOption;
import org.maachang.util.atomic.AtomicBOOL;
import org.maachang.util.atomic.AtomicINT;
import org.maachang.util.atomic.AtomicLONG;
import org.maachang.util.atomic.AtomicOBJECT;

/**
 * Nio受信データ要素.
 * 
 * @version 2008/05/28
 * @author  masahito suzuki
 * @since   SeaBass 1.0.0
 */
public class NioElement implements Comparable<NioElement> {
    
    /**
     * ログ.
     */
    private static final Log LOG = LogFactory.getLog( NioElement.class ) ;
    
    /**
     * SelectionKey.
     */
    private final AtomicOBJECT<SelectionKey> key = new AtomicOBJECT<SelectionKey>() ;
    
    /**
     * SocketChannel.
     */
    private final AtomicOBJECT<SocketChannel> channel = new AtomicOBJECT<SocketChannel>() ;
    
    /**
     * 書き込み可能を示すバッファ.
     */
    private final AtomicOBJECT<IsWriteQueue> isWriteQueue = new AtomicOBJECT<IsWriteQueue>() ;
    
    /**
     * 最新時間.
     */
    private final AtomicLONG time = new AtomicLONG( -1L ) ;
    
    /**
     * KeepAliveタイム.
     */
    private final AtomicLONG keepAliveTime = new AtomicLONG( -1L ) ;
    
    /**
     * 処理中フラグ.
     */
    private final AtomicBOOL executionFlag = new AtomicBOOL( false ) ;
    
    /**
     * クローズフラグ.
     */
    private final AtomicBOOL closeFlag = new AtomicBOOL( false ) ; ;
    
    /**
     * 付加オブジェクト.
     */
    private final AtomicOBJECT<Object> object = new AtomicOBJECT<Object>( null ) ;
    
    /**
     * SSL要素.
     */
    private final AtomicOBJECT<SslElement> sslElement = new AtomicOBJECT<SslElement>( null ) ;
    
    /**
     * SSLオプション.
     */
    private final AtomicOBJECT<SslOption> sslOption = new AtomicOBJECT<SslOption>( null ) ;
    
    /**
     * 受信バッファ.
     */
    private final ReceiveBuffer recvBuffer = new ReceiveBuffer() ;
    
    /**
     * 送信バッファ.
     */
    private final SendBuffer sndBuffer = new SendBuffer() ; ;
    
    /**
     * KeepAliveフラグ.
     */
    private final AtomicBOOL keepAliveFlag = new AtomicBOOL( false ) ;
    
    /**
     * stackフラグ.
     */
    protected final AtomicBOOL stackFlag = new AtomicBOOL( false ) ;
    
    /**
     * シーケンス番号フラグ.
     */
    private final AtomicINT seq = new AtomicINT(-1) ;
    
    /**
     * opt_writeフラグ.
     */
    private final AtomicBOOL optWrite = new AtomicBOOL( false ) ;
    
    /**
     * コンストラクタ.
     */
    protected NioElement() {
    }
    
    /**
     * コンストラクタ.
     * @param ch 対象のチャネル情報を設定します.
     * @exception Exception 例外.
     */
    protected NioElement( SocketChannel ch ) throws Exception {
        this.create( ch ) ;
    }
    
    /**
     * コンストラクタ.
     * @param ch 対象のチャネル情報を設定します.
     * @param queue 書き込み通知キューを設定します.
     * @exception Exception 例外.
     */
    protected NioElement( SocketChannel ch,IsWriteQueue queue ) throws Exception {
        this.create( ch,queue ) ;
    }
    
    /**
     * オブジェクト利用中に変更.
     */
    protected void restart() {
        this.stackFlag.set( false ) ;
    }
    
    /**
     * オブジェクト生成.
     * @param ch 対象のチャネル情報を設定します.
     * @exception Exception 例外.
     */
    protected void create( SocketChannel ch ) throws Exception {
        this.create( ch,null ) ;
        this.seq.set( -1 ) ;
    }
    
    /**
     * オブジェクト生成.
     * @param ch 対象のチャネル情報を設定します.
     * @param queue 書き込み通知キューを設定します.
     * @exception Exception 例外.
     */
    protected void create( SocketChannel ch,IsWriteQueue queue ) throws Exception {
        this.stackFlag.set( false ) ;
        if( ch == null) {
            throw new NioException( "ソケットチャネルは不正です" ) ;
        }
        this.channel.set( ch ) ;
        this.isWriteQueue.set( queue ) ;
        this.time.set( System.currentTimeMillis() ) ;
        this.executionFlag.set( false ) ;
        this.closeFlag.set( false ) ;
        this.object.set( null ) ;
        this.sndBuffer.create() ;
        this.recvBuffer.create() ;
        this.optWrite.set( false ) ;
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        destroy() ;
    }
    
    /**
     * オブジェクト破棄.
     */
    public void destroy() {
        SocketChannel ch = channel.get() ;
        if( ch != null ) {
            channel.set( null ) ;
            if( sslElement.get() != null ) {
                sslElement.get().destroy() ;
            }
            clearOpMode() ;// キークリア.
            NioTcpUtil.closeChannel( ch ) ;
            //NioTcpUtil.shutdownChannel( ch ) ;
            isWriteQueue.set( null ) ;
            time.set( -1L ) ;
            keepAliveTime.set( -1L ) ;
            executionFlag.set( false ) ;
            closeFlag.set( false ) ;
            object.set( null ) ;
            sslElement.set( null ) ;
            sslOption.set( null ) ;
            keepAliveFlag.set( false ) ;
            optWrite.set( false ) ;
            sndBuffer.clear() ;
            recvBuffer.reset( true ) ;
            int s = seq.get() ;
            seq.set( -1 ) ;
            if( !stackFlag.get() ) {
//LOG.info( "### ID(" + seq.get() + ")通信要素を破棄(" + (channel.get() != null) + ")",new Exception( "trace" ) ) ;
                if( LOG.isDebugEnabled() ) {
                    LOG.debug( "## NioElement.destroy()->" + s ) ;
                }
                stackFlag.set( true ) ;
                if( NioElementPoolingFactory.isPush() ) {
                    NioElementPoolingFactory.push( this ) ;
                }
            }
        }
    }
    
    /**
     * 書き込み開始キューを設定.
     * @param queue 書き込み開始キューを設定します.
     */
    protected void setIsWriteQueue( IsWriteQueue queue ) {
        isWriteQueue.set( queue ) ;
    }
    
    /**
     * SSL条件を追加.
     * @param sslOption 対象のSSLオプションを設定します.
     * @param sslElement 対象のSSL要素を追加します.
     */
    protected void setSsl( SslOption op ) {
        if( op != null ) {
            clearSsl() ;
            sslOption.set( op ) ;
        }
    }
    
    /**
     * SSL要素を取得.
     * @return SslOption SSL要素が返されます.
     */
    public SslOption getSslOption() {
        return sslOption.get() ;
    }
    
    /**
     * 新しいSSLオプションを生成.
     * @exception Exception 例外.
     */
    public void createSsl() throws Exception {
        SslOption op ;
        if( ( op = sslOption.get() ) != null && sslElement.get() == null ) {
            sslElement.set( new SslElement( this,op.getSSLEngine() ) ) ;
        }
    }
    
    /**
     * SSL要素をクリア.
     */
    public void clearSsl() {
        SslElement em ;
        if( ( em = sslElement.get() ) != null ) {
            sslElement.set( null ) ;
            em.destroy() ;
        }
    }
    
    /**
     * SSL要素を取得.
     * @return SslElement SSL要素が返されます.
     */
    public SslElement getSslElement() {
        return sslElement.get() ;
    }
    
    /**
     * SSLが有効かチェック.
     * @return boolean [true]の場合、SSLは有効です.
     */
    public boolean isSSL() {
        return sslElement.get() != null ;
    }
    
    /**
     * SelectionKeyを設定.
     * @param k SelectionKeyを設定します.
     */
    public void setKey( SelectionKey k ) {
        optWrite.set( false ) ;
        key.set( k ) ;
    }
    
    /**
     * SelectionKeyを取得.
     * @return SelectionKey SelectionKeyが返されます.
     */
    public SelectionKey getKey() {
        return key.get() ;
    }
    
    /**
     * SocketChannelを取得.
     * @return SocketChannel SocketChannelが返されます.
     */
    public SocketChannel getChannel() {
        return channel.get() ;
    }
    
    /**
     * 受信バッファを取得.
     * @return ReceiveBuffer 受信バッファを取得します.
     */
    public ReceiveBuffer getReceiveBuffer() {
        return recvBuffer ;
    }
    
    /**
     * 受信バッファが存在するかチェック.
     * @return boolean [true]の場合、存在します.
     */
    public boolean isReceiveBuffer() {
        return recvBuffer.isData() ;
    }
    
    /**
     * 書き込み開始/再開を通知.
     */
    public void restartWrite() {
        SelectionKey skey ;
        if( !optWrite.get() && ( skey = key.get() ) != null ) {
            isWriteQueue.get().append( this ) ;
            skey.selector().wakeup() ;
        }
    }
    
    /**
     * wakeup.
     */
    public void wakeup() {
        SelectionKey skey ;
        if( ( skey = key.get() ) != null ) {
            skey.selector().wakeup() ;
        }
    }
    
    /**
     * SelectionKeyモード登録.
     * @param interestOption 追加モードを設定します.
     * @return boolean [true]の場合、設定成功です.
     * @exception IOException I/O例外.
     */
    public boolean opMode( int interestOption ) throws IOException {
        boolean ret = false ;
        SelectionKey sKey = key.get() ;
        if( sKey != null ) {
            if( ( interestOption | SelectionKey.OP_WRITE ) != 0 ) {
                if( !optWrite.get() ) {
                    sKey.interestOps( sKey.interestOps() | SelectionKey.OP_WRITE ) ;
                    optWrite.set( true ) ;
                    ret = true ;
                }
                else if( !sndBuffer.isRead() ) {
                    sKey.interestOps( sKey.interestOps() & (~SelectionKey.OP_WRITE ) ) ;
                    optWrite.set( false ) ;
                    ret = true ;
                }
            }
            else if( ( interestOption | SelectionKey.OP_READ ) != 0 ) {
                    sKey.interestOps( ( sKey.interestOps() & (~SelectionKey.OP_WRITE ) ) | SelectionKey.OP_READ ) ;
                    optWrite.set( false ) ;
                    ret = true ;
            }
            else {
                sKey.interestOps( sKey.interestOps() & (~SelectionKey.OP_WRITE) ) ;
                optWrite.set( false ) ;
                ret = true ;
            }
        }
        return ret ;
    }
    
    /**
     * SelectionKeyモードクリア処理.
     */
    protected void clearOpMode() {
        optWrite.set( false ) ;
        SelectionKey sKey = key.get() ;
        key.set( null ) ;
        if( sKey != null ) {
            sKey.attach( null ) ;
            try {
                sKey.interestOps( 0 ) ;
            } catch( Exception e ) {
            }
            try {
                sKey.cancel() ;
            } catch( Exception e ) {
            }
        }
    }
    
    /**
     * OP_WRITE状態かチェック.
     * @return boolean [true]の場合、OP_WRITEがONになっています.
     */
    public boolean isOpWrite() {
        return optWrite.get() ;
    }
    
    /**
     * OpWriteモードを設定.
     * @param mode OpWriteモードを設定します.
     */
    public void setOpWriteFlag( boolean mode ) {
        optWrite.set( mode ) ;
    }
    
    /**
     * 書き込みバッファオブジェクトを取得.
     * @return SendBuffer 書き込みバッファオブジェクトが返されます.
     */
    public SendBuffer getWriteBuffer() {
        return sndBuffer ;
    }
    
    /**
     * 書き込みバッファオブジェクトが存在するかチェック.
     * @return boolean [true]の場合、存在します.
     */
    public boolean isWriteBuffer() {
        return sndBuffer.isRead() ;
    }
    
    /**
     * 書き込みバッファの初期化.
     */
    public void recreateWriteBuffer() {
        sndBuffer.clear() ;
        sndBuffer.create() ;
    }
    
    /**
     * クローズフラグをセット.
     * @param mode [true]の場合、クローズとなります.
     */
    public void setCloseFlag( boolean mode ) {
        closeFlag.set( mode ) ;
    }
    
    /**
     * クローズフラグを取得.
     * @return boolean [true]の場合、クローズです.
     */
    public boolean isCloseFlag() {
        return closeFlag.get() ;
    }
    
    /**
     * 更新処理.
     */
    public void update() {
        time.set( System.currentTimeMillis() ) ;
    }
    
    /**
     * 最新時間を取得.
     * @return long 最新時間が返されます.
     */
    public long getTime() {
        return time.get() ;
    }
    
    /**
     * 処理中フラグを設定.
     * @param flg [true]の場合、処理中フラグをONにします.
     */
    protected void setExecutionFlag( boolean flg ) {
        executionFlag.set( flg ) ;
        if( flg == false ) {
            update() ;
        }
    }
    
    /**
     * 処理中フラグを取得.
     * @return boolean [true]の場合、処理中です.
     */
    protected boolean isExecutionFlag() {
        return executionFlag.get() ;
    }
    
    /**
     * KeepAlive時間を設定.
     * @param time KeepAlive時間を設定します.
     */
    public void setKeepAliveTime( int time ) {
        keepAliveTime.set( (long)time * 1000L ) ;
    }
    
    /**
     * タイムアウト値を設定.
     * @param time タイムアウト値を設定します.
     */
    public void setTimeout( long time ) {
        keepAliveTime.set( time ) ;
    }
    
    /**
     * KeepAlive時間を取得.
     * @return long KeepAlive時間が返されます.
     */
    public long getKeepAliveTime() {
        return keepAliveTime.get() ;
    }
    
    /**
     * KeepAlive監視フラグを設定.
     * @param mode [true]の場合、監視中です.
     */
    public void setKeepAliveMonitor( boolean mode ) {
        keepAliveFlag.set( mode ) ;
    }
    
    /**
     * KeepAlive監視フラグを取得.
     * @return boolean [true]の場合、監視中です.
     */
    public boolean isKeepAliveMonitor() {
        return keepAliveFlag.get() ;
    }
    
    /**
     * 付加オブジェクトを設定.
     * @param o 付加オブジェクトを設定します.
     */
    public void setObject( Object o ) {
        object.set( o ) ;
    }
    
    /**
     * 付加オブジェクトを取得.
     * @return Object 付加オブジェクトが返されます.
     */
    public Object getObject() {
        return object.get() ;
    }
    
    /**
     * シーケンスIDを取得.
     * @return int シーケンスIDが返されます.
     */
    public int getSequenceId() {
        return seq.get() ;
    }
    
    /**
     * シーケンスIDを付加.
     * @param seq 対象のシーケンスIDを設定します.
     */
    public void setSequenceId( int seq ) {
        this.seq.set( seq ) ;
    }
    
    /**
     * オブジェクトが同じかチェックします.
     * @param o 比較元を設定します.
     * @return boolean [true]の場合、同じです.
     */
    public boolean equals( NioElement o ) {
        return ( o == this ) ;
    }
    
    /**
     * オブジェクト比較.
     * @param o 比較元を設定します.
     * @return int 比較結果が返されます.
     */
    public int compareTo(NioElement o) {
        return o.hashCode() - this.hashCode() ;
    }
    
    /**
     * IPアドレスとポート番号を取得.
     * @return String IPアドレスとポート番号が返されます.
     */
    public String getIpPort() {
        if( channel.get() == null ) {
            return "UNKNOWN:????" ;
        }
        Socket s = channel.get().socket() ;
        return new StringBuilder().append( s.getInetAddress().getHostAddress() ).
            append( ":" ).append( s.getPort() ).toString() ;
    }
    
    /**
     * オブジェクトが有効かチェック.
     * @return boolean [true]の場合、有効です.
     */
    public boolean isUse() {
        return ( channel.get() != null ) ;
    }
}
