package org.maachang.seabass.net.nio ;

import org.maachang.util.atomic.AtomicIntSeq;


/**
 * KeepAlive情報を管理.
 * 
 * @version 2008/05/28
 * @author  masahito suzuki
 * @since   SeaBass 1.0.0
 */
public class NioPoolManager {
    
    /**
     * 実行スレッド空き待ちリトライ数.
     */
    protected static final int POOL_SPACE_RETRY = 3 ;
    
    /**
     * Poolスレッド群.
     */
    private NioPoolThread[] pools = null ;
    
    /**
     * 管理Poolスレッド数.
     */
    private int maxPoolLength = 0 ;
    
    /**
     * スレッド実行項番管理.
     */
    private AtomicIntSeq execId = null ;
    
    /**
     * コンストラクタ.
     */
    private NioPoolManager() {}
    
    /**
     * コンストラクタ.
     * @param pools 対象のPoolスレッド群を設定します.
     */
    public NioPoolManager( NioPoolThread[] pools ) {
        this.pools = pools ;
        this.maxPoolLength = pools.length ;
        this.execId = new AtomicIntSeq( maxPoolLength ) ;
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        destroy() ;
    }
    
    /**
     * オブジェクト破棄.
     */
    public void destroy() {
        if( pools != null ) {
            for( int i = 0 ; i < maxPoolLength ; i ++ ) {
                if( pools[ i ] != null ) {
                    try {
                        pools[ i ].destroy() ;
                    } catch( Exception e ) {
                    }
                    pools[ i ] = null ;
                }
            }
            pools = null ;
        }
        maxPoolLength = 0 ;
        execId = null ;
    }
    
    /**
     * 実行処理.
     * @param element 対象の要素を設定します.
     * @exception Exception 例外.
     */
    public void execution( NioElement element )
        throws Exception {
        // シーケンス番号にあわせて処理.
        pools[ execId.get() ].startExecution( element ) ;
    }
    
    /**
     * 有効プーリングスレッド数を取得.
     * @return int 有効プーリングスレッド数が返されます.
     */
    public int getUseLength() {
        if( pools == null ) {
            return 0 ;
        }
        int ret = 0 ;
        int len = maxPoolLength ;
        for( int i = 0 ; i < len ; i ++ ) {
            if( pools[ i ] != null && pools[ i ].isStop() == false ) {
                ret ++ ;
            }
        }
        return ret ;
    }
    
    /**
     * プールスレッド数を取得.
     * @return int プールスレッド数が返されます.
     */
    public int size() {
        return maxPoolLength ;
    }
    
    /**
     * 指定位置のプーリングスレッドを取得.
     * @param no 対象の項番を設定します.
     * @return NioPoolThread プーリングスレッドが返されます.
     */
    public NioPoolThread getThread( int no ) {
        if( pools == null ) {
            return null ;
        }
        if( no < 0 || no >= pools.length ) {
            return null ;
        }
        return pools[ no ] ;
    }
}
