package org.maachang.seabass.net.nio ;

import java.io.IOException;
import java.io.InputStream;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.seabass.net.HttpReceiveCallback;
import org.maachang.seabass.net.HttpdRequestImpl;
import org.maachang.seabass.net.HttpdResponseImpl;
import org.maachang.util.atomic.AtomicBOOL;
import org.maachang.util.atomic.AtomicINT;
import org.maachang.util.atomic.AtomicIntSeq;
import org.maachang.util.atomic.AtomicLONG;
import org.maachang.util.thread.LoopThread;

/**
 * 処理用プーリングスレッド.
 * 
 * @version 2008/05/28
 * @author  masahito suzuki
 * @since   SeaBass 1.0.0
 */
public class NioPoolThread extends LoopThread {
    
    /**
     * ログ.
     */
    private static final Log LOG = LogFactory.getLog( NioPoolThread.class ) ;
    
    /**
     * 受信タイムアウト値.
     */
    private static final long TIMEOUT = 5000L ;
    
    /**
     * 送信タイムアウトを設定.
     */
    private static final long SEND_TIMEOUT = 15000L ;
    
    /**
     * 一時的に利用するバイナリ長.
     */
    private static final int TEMP_BINARY = NioInputStream.BUFFER_LENGTH ;
    
    /**
     * コールバックオブジェクト.
     */
    private HttpReceiveCallback call = null ;
    
    /**
     * キューオブジェクト.
     */
    private ReceiveLinkQueue queue = null ;
    
    /**
     * シーケンスID発行オブジェクト.
     */
    private AtomicIntSeq sequence = null ;
    
    /**
     * Active数管理.
     */
    private AtomicINT actionLength = null ;
    
    /**
     * 最終プログラム実行時間.
     */
    private final AtomicLONG lastAccessTime = new AtomicLONG( -1L ) ;
    
    /**
     * スレッド実行フラグ.
     */
    private final AtomicBOOL executionThread = new AtomicBOOL( false ) ;
    
    /**
     * リクエストバッファ.
     */
    private final RequestBuffer requestBuffer = new RequestBuffer() ;
    
    /**
     * コネクション情報.
     */
    private final ConnectionInfoImpl connectionInfo = new ConnectionInfoImpl() ;
    
    /**
     * リクエスト情報.
     */
    private final HttpdRequestImpl request = new HttpdRequestImpl() ;
    
    /**
     * レスポンス情報.
     */
    private final HttpdResponseImpl response = new HttpdResponseImpl() ;
    
    /**
     * 一時的に利用するバイナリ.
     */
    private final byte[] tmpBinary = new byte[ TEMP_BINARY ] ;
    
    /**
     * コンストラクタ.
     */
    private NioPoolThread() {
        
    }
    
    /**
     * コンストラクタ.
     * @param call コールバックオブジェクトを設定します.
     * @param actionLength アクティブ実行数管理長オブジェクトを設定します.
     * @param sequence シーケンスID割り当てオブジェクトを設定します.
     * @param threadId 対象のスレッドIDを設定します.
     * @exception Exception 例外.
     */
    public NioPoolThread( HttpReceiveCallback call,
        AtomicINT actionLength,AtomicIntSeq sequence,int threadId )
        throws Exception {
        if( call == null || actionLength == null || sequence == null ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        this.call = call ;
        this.queue = new ReceiveLinkQueue() ;
        this.actionLength = actionLength ;
        this.sequence = sequence ;
        super.threadId.set( threadId ) ;
        super.setPriority( PriorityDef.POOL_THREAD ) ;
        this.startThread() ;
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        this.destroy() ;
    }
    
    /**
     * オブジェクト破棄.
     */
    public void destroy() {
        super.stopThread() ;
    }
    
    /**
     * オブジェクトクリア.
     */
    protected void clear() {
        queue = null ;
        call = null ;
    }
    
    /**
     * スレッドが処理実行中かチェック.
     * @return boolean [true]の場合、実行中です.
     */
    public boolean isExecutionThread() {
        return executionThread.get() ;
    }
    
    /**
     * 実行処理待ち数を取得.
     * @return int 実行処理待ち数が返されます.
     */
    public int getQueueLength() {
        return queue.size() ;
    }
    
    /**
     * 最終実行時間を取得.
     * @return long 最終実行時間が返されます.
     */
    public long getLastAccessTime() {
        return this.lastAccessTime.get() ;
    }
    
    /**
     * オブジェクト実行命令.
     * @param em 対象のNioElementオブジェクトを設定します.
     * @exception Exception 例外.
     */
    public void startExecution( NioElement em )
        throws Exception {
        queue.append( em ) ;
    }
    
    /**
     * 実行処理.
     * @exception Exception 例外.
     */
    protected boolean execution() throws Exception {
        long time = 0L ;
        NioElement element = null ;
        while( true ) {
            // 通信キューが取得できた場合.
            if( ( element = queue.getQueue() ) != null ) {
                // 実行開始.
                if( LOG.isDebugEnabled() ) {
                    LOG.debug( "## (start)NioPoolThread->execution(" + threadId + ") - IP:[" + element.getIpPort() + "]" ) ;
                    time = System.currentTimeMillis() ;
                }
                executionThread.set( true ) ;
                this.lastAccessTime.set( System.currentTimeMillis() ) ;
                actionLength.inc() ;
                try {
                    executionServer( element ) ;
                } catch( Exception e ) {
                    element.destroy() ;
                    throw e ;
                } catch( Error err ) {
                    element.destroy() ;
                    throw err ;
                } finally {
                    if( LOG.isDebugEnabled() ) {
                        time = System.currentTimeMillis() - time ;
                        LOG.debug( "## (end)NioPoolThread->execution(" + threadId + ") - " + time + "msec." ) ;
                    }
                    // 実行終了.
                    executionThread.set( false ) ;
                    actionLength.dec() ;
                }
            }
        }
    }
    
    /**
     * 例外ハンドラ.
     */
    protected void toException( Throwable e ) {
        //LOG.warn( "error",e ) ;
    }
    
    /**
     * １つのコネクションを処理.
     */
    private final void executionServer( NioElement element ) throws Exception {
        String url = null ;
        long time = -1L ;
        int id = -1 ;
        HttpReceiveCallback execReq = this.call ;
        ConnectionInfoImpl connInfo = this.connectionInfo ;
        HttpdRequestImpl req = this.request ;
        HttpdResponseImpl resp = this.response ;
        RequestBuffer reqBuf = this.requestBuffer ;
        try {
            connInfo.clear() ;
            if( LOG.isDebugEnabled() ) {
                time = System.currentTimeMillis() ;
            }
            url = null ;
            id = sequence.get() ;
            element.setSequenceId( id ) ;
            if( connInfo.create( element ) == null ) {
                // コネクション切断.
                if( element != null ) {
                    element.destroy() ;
                    element = null ;
                }
                return ;
            }
            resp.clear() ;
            req.clear() ;
            // HTTPヘッダ＋Body読み込み.
            createHttpdRequest( req,connInfo,reqBuf,id ) ;
            // URLを取得.
            if( LOG.isDebugEnabled() ) {
                url = req.getUrlPath() ;
            }
            //if( LOG.isDebugEnabled() ) {
            //    url = req.getUrlPath() ;
            //    LOG.info( new StringBuilder().append( " ## ***** start " ).
            //        append( "receive(" ).append( id ).append( ") - [" ).
            //        append( connInfo.getInetAddress().getHostAddress() ).
            //        append( ":" ).append( connInfo.getPort() ).
            //        append( "] - threadId:(" ).append( threadId ).
            //        append( ") url:" + url ).toString() ) ;
            //}
            // 実行処理.
            execReq.execution( tmpBinary,connInfo,id,req,resp ) ;
            
            // コメットフラグがOFFの場合の処理.
            // コメットフラグがONの場合は、何もしない.
            if( !connInfo.isComet() ) {
                resp.flushAndClose() ;
                if( connInfo.isUse() ) {
                    if( !connInfo.isCloseFlag() ) {
                        connInfo.recyclingConnection() ;
                    }
                }
                else {
                    if( element != null ) {
                        element.destroy() ;
                        element = null ;
                    }
                    connInfo.clear() ;
                    return ;
                }
            }
            // コメット処理の場合は、KeepAliveは処理しない.
            else {
                element.setKeepAliveTime(-1) ;
            }
            // コネクション終了条件の場合.
            if( element.isCloseFlag() ) {
                connInfo.clear() ;
                element.setKeepAliveTime(0) ;
            }
            // 終了ログ.
            if( LOG.isDebugEnabled() ) {
                LOG.debug( new StringBuilder().append( " ## @@@@@ end " ).
                    append( "## id[" ).append( id ).
                    append( "]-threadId:(" ).append( threadId ).append( ") url:" ).append( url ).
                    append( " " ).append( (System.currentTimeMillis() - time) ).append( "msec" ).toString() ) ;
            }
        } catch( Exception e ) {
            if( LOG.isDebugEnabled() ) {
                StringBuilder buf = new StringBuilder() ;
                buf.append( " ## $$$$$ error " ).
                    append( "## id[" ).append( id ).
                    append( "]-threadId:(" ).append( threadId ) ;
                if( url != null ) {
                    buf.append( ") url:" ).append( url ).append( " " ) ;
                }
                else {
                    buf.append( ") " ) ;
                }
                buf.append( ( System.currentTimeMillis() - time ) ).append( "msec - message:" ).
                    append( e.getMessage() ) ;
                LOG.debug( buf.toString(),e ) ;
            }
            throw e ;
        } finally {
            connInfo.clear() ;
            resp.clear() ;
            req.clear() ;
        }
    }
    
    /**
     * HttpdRequesを生成.
     */
    private static final boolean createHttpdRequest( HttpdRequestImpl request,ConnectionInfo info,RequestBuffer buf,int seqId )
        throws Exception {
        if( info == null || info.isUse() == false ) {
            throw new IOException( "コネクション情報は利用できません" ) ;
        }
        InputStream in = null ;
        try {
            in = info.getInputStream( TIMEOUT ) ;
            CreateHttpdRequest.receiveHttpRequest( request,info,seqId,in,buf ) ;
            in.close() ;
            in = null ;
            // 送信タイムアウト値を設定.
            info.getElement().setTimeout( SEND_TIMEOUT ) ;
            return true ;
        } catch( Exception e ) {
            throw e ;
        } finally {
            if( in != null ) {
                in.close() ;
            }
        }
    }
}
