package org.maachang.seabass.net.nio ;

import java.nio.ByteBuffer;
import java.nio.channels.CancelledKeyException;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.util.Iterator;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.seabass.net.NetUtil;
import org.maachang.seabass.net.ssl.SslOption;
import org.maachang.util.thread.LoopThread;

/**
 * NIO-SSLシングルサーバ処理.
 * 
 * @version 2008/05/28
 * @author  masahito suzuki
 * @since   SeaBass 1.0.0
 */
class NioSingleSslServer extends LoopThread {
    
    /**
     * ログ.
     */
    private static final Log LOG = LogFactory.getLog( NioSingleSslServer.class ) ;
    
    /**
     * Selectorタイムアウト.
     */
    private static final int TIMEOUT = NioSslUtil.TIMEOUT ;
    
    /**
     * サーバソケットチャネル.
     */
    private ServerSocketChannel server = null ;
    
    /**
     * セレクタ.
     */
    private Selector selector = null ;
    
    /**
     * 書き込み可能キュー.
     */
    private IsWriteQueue isWriteQueue = null ;
    
    /**
     * 実行Poolスレッド管理.
     */
    private NioPoolManager man = null ;
    
    /**
     * SSLOption.
     */
    private SslOption option = null ;
    
    /**
     * 送信バッファ.
     */
    private final ByteBuffer sendBuffer = NetUtil.byteBuffer( NioOutputStream.BUFFER_LENGTH ) ;
    
    /**
     * コンストラクタ.
     */
    private NioSingleSslServer() {
        
    }
    
    /**
     * コンストラクタ.
     * @param channel 対象のサーバソケットを設定します.
     * @param option 対象のSSLオプションを設定します.
     * @param man 対象のプーリングマネージャをを設定します.
     * @exception Exception 例外.
     */
    public NioSingleSslServer( ServerSocketChannel channel,SslOption option,NioPoolManager man )
        throws Exception {
        if( channel == null || man == null || option == null ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        this.server = channel ;
        this.selector = Selector.open() ;
        this.option = option ;
        this.server.register( this.selector,SelectionKey.OP_ACCEPT ) ;
        this.isWriteQueue = new IsWriteQueue() ;
        this.man = man ;
        super.setPriority( PriorityDef.SSL_THREAD ) ;
        this.startThread() ;
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        this.destroy() ;
    }
    
    /**
     * オブジェクト破棄.
     */
    public void destroy() {
        super.stopThread() ;
    }
    
    /**
     * オブジェクトクリア.
     */
    protected void clear() {
        if( this.selector != null ) {
            try {
                this.selector.close() ;
            } catch( Exception e ) {
            }
        }
        if( this.server != null ) {
            try {
                this.server.close() ;
            } catch( Exception e ) {
            }
        }
        this.selector = null ;
        this.server = null ;
        this.isWriteQueue = null ;
        this.man = null ;
        this.option = null ;
    }
    
    /**
     * 実行処理.
     * @exception Exception 例外.
     */
    protected boolean execution() throws Exception {
        int counter = 0 ;
        Iterator<SelectionKey> itr ;
        SelectionKey key = null ;
        NioElement em = null ;
        while( true ) {
            NioTcpUtil.isWrite( isWriteQueue ) ;
            if( ( counter = selector.selectNow() ) <= 0 ) {
                Thread.yield();
                counter = selector.select( TIMEOUT ) ;
            }
            if( counter > 0 ) {
                Thread.yield();
                itr = selector.selectedKeys().iterator() ;
                while(itr.hasNext()) {
                    try {
                        key = itr.next() ;
                        itr.remove() ;
                        // keyが存在する場合.
                        if( key != null ) {
                            // use-socket.
                            if( !key.isValid() ) {
                                em = ( NioElement )key.attachment() ;
                                NioTcpUtil.destroyConnect( key,em ) ;
                            }
                            // accept.
                            else if( !NioSslUtil.toAcceptBySingle( selector,key,isWriteQueue,option ) &&
                                ( em = ( NioElement )key.attachment() ) != null ) {
                                // write.
                                if( key.isWritable() ) {
                                    NioSslUtil.isWrite( selector,em,key,sendBuffer ) ;
                                }
                                // read.
                                if( key.isReadable() ) {
                                    NioSslUtil.isRead( selector,em,man,key ) ;
                                }
                            }
                            // エラー及びタイムアウト検知.
                            if( !key.isValid() || em == null ||
                                em.getTime() + NioTcpUtil.DESTROY_TIMEOUT < System.currentTimeMillis() ) {
                                LOG.warn( "### 通信エラーを検知" ) ;
                                if( em != null ) {
                                    NioTcpUtil.destroyConnect( key,em ) ;
                                }
                                else if( key != null ) {
                                    NioTcpUtil.destroyKey( key ) ;
                                }
                            }
                        }
                    } catch( ClosedChannelException ce ) {
                        if( em != null ) {
                            NioTcpUtil.destroyConnect( key,em ) ;
                        }
                        else if( key != null ) {
                            NioTcpUtil.destroyKey( key ) ;
                        }
                    } catch( CancelledKeyException cee ) {
                        if( em != null ) {
                            NioTcpUtil.destroyConnect( key,em ) ;
                        }
                        else if( key != null ) {
                            NioTcpUtil.destroyKey( key ) ;
                        }
                    } catch( Exception e ) {
                        if( LOG.isDebugEnabled() ) {
                            LOG.debug( "SSL通信エラー:",e ) ;
                        }
                        if( em != null ) {
                            NioTcpUtil.destroyConnect( key,em ) ;
                        }
                        else if( key != null ) {
                            NioTcpUtil.destroyKey( key ) ;
                        }
                    }
                    key = null ;
                    em = null ;
                }
            }
        }
    }
}
