package org.maachang.seabass.net.nio ;

import java.nio.ByteBuffer;
import java.nio.channels.CancelledKeyException;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.util.Iterator;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.seabass.net.NetUtil;
import org.maachang.util.atomic.AtomicIntSeq;
import org.maachang.util.thread.LoopThread;


/**
 * Nio-SSL用受信サーバ群管理.
 * 
 * @version 2008/05/28
 * @author  masahito suzuki
 * @since   SeaBass 1.0.0
 */
class NioSslReceiveServerArray {
    
    /**
     * デフォルトサイズ.
     */
    private static final int MIN_SIZE = 1 ;
    
    /**
     * NioReceiveServer.
     */
    private NioSslReceiveServer[] recvs = null ;
    
    /**
     * シーケンスID.
     */
    private AtomicIntSeq seq = null ;
    
    /**
     * コンストラクタ.
     */
    private NioSslReceiveServerArray() {
        
    }
    
    /**
     * コンストラクタ.
     * @param man 対象のプーリングマネージャをを設定します.
     * @param size 受信スレッド数を設定します.
     * @exception Exception 例外.
     */
    public NioSslReceiveServerArray( NioPoolManager man,int size )
        throws Exception {
        if( man == null ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        if( size <= MIN_SIZE ) {
            size = MIN_SIZE ;
        }
        NioSslReceiveServer[] rv = new NioSslReceiveServer[ size ] ;
        for( int i = 0 ; i < size ; i ++ ) {
            rv[ i ] = new NioSslReceiveServer( man ) ;
        }
        this.recvs = rv ;
        this.seq = new AtomicIntSeq( size ) ;
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        destroy() ;
    }
    
    /**
     * オブジェクト破棄.
     */
    public void destroy() {
        int len = recvs.length ;
        for( int i = 0 ; i < len ; i ++ ) {
            recvs[ i ].destroy() ;
        }
    }
    
    /**
     * Acceptされた要素を受信用にセット.
     * @param element 対象のElementを設定します.
     * @exception Exception 例外.
     */
    public void setElement( NioElement element )
        throws Exception {
        if( element == null ) {
            return ;
        }
        recvs[ seq.get() ].setElement( element ) ;
    }
}

/**
 * 1つの受信ソケット.
 */
class NioSslReceiveServer extends LoopThread {
    
    /**
     * ログ.
     */
    private static final Log LOG = LogFactory.getLog( NioSslReceiveServer.class ) ;
    
    /**
     * Selectorタイムアウト.
     */
    private static final int TIMEOUT = NioSslUtil.TIMEOUT ;
    
    /**
     * Selector.
     */
    private Selector selector = null ;
    
    /**
     * 読み込み可能キュー.
     */
    private IsReadQueue isRecvQueue = null ;
    
    /**
     * 書き込み可能キュー.
     */
    private IsWriteQueue isWriteQueue = null ;
    
    /**
     * 実行Poolスレッド管理.
     */
    private NioPoolManager man = null ;
    
    /**
     * 送信バッファ.
     */
    private final ByteBuffer sendBuffer = NetUtil.byteBuffer( NioOutputStream.BUFFER_LENGTH ) ;
    
    /**
     * コンストラクタ.
     */
    private NioSslReceiveServer() {
        
    }
    
    /**
     * コンストラクタ.
     * @param man 対象のプーリングマネージャをを設定します.
     * @exception Exception 例外.
     */
    public NioSslReceiveServer( NioPoolManager man )
        throws Exception {
        this.selector = Selector.open() ;
        this.isRecvQueue = new IsReadQueue() ;
        this.isWriteQueue = new IsWriteQueue() ;
        this.man = man ;
        super.setPriority( PriorityDef.SSL_THREAD ) ;
        this.startThread() ;
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        this.destroy() ;
    }
    
    /**
     * オブジェクト破棄.
     */
    public void destroy() {
        super.stopThread() ;
    }
    
    /**
     * AcceptされたNioElementを受信用にセット.
     * @param element 対象のElementを設定します.
     * @exception Exception 例外.
     */
    public void setElement( NioElement element )
        throws Exception {
        element.setIsWriteQueue( isWriteQueue ) ;
        isRecvQueue.append( element ) ;
        selector.wakeup() ;
    }
    
    /**
     * オブジェクトクリア.
     */
    protected void clear() {
        if( this.selector != null ) {
            try {
                this.selector.close() ;
            } catch(Exception e) {}
        }
        this.selector = null ;
    }
    
    /**
     * 実行処理.
     * @exception Exception 例外.
     */
    protected boolean execution() throws Exception {
        int counter = 0 ;
        Iterator<SelectionKey> itr ;
        SelectionKey key = null ;
        NioElement em = null ;
        while( true ) {
            NioTcpUtil.isSendOrAccept( selector,isRecvQueue,isWriteQueue ) ;
            if( ( counter = selector.selectNow() ) <= 0 ) {
                //Thread.yield();
                counter = selector.select( TIMEOUT ) ;
            }
            if( counter > 0 ) {
                //Thread.yield();
                itr = selector.selectedKeys().iterator() ;
                while(itr.hasNext()) {
                    NioTcpUtil.isAccept( selector,isRecvQueue ) ;
                    try {
                        key = itr.next() ;
                        itr.remove() ;
                        // keyが存在する場合.
                        if( key != null ) {
                            if( ( em = ( NioElement )key.attachment() ) != null && key.isValid() ) {
                                // write.
                                if( key.isWritable() ) {
                                    NioSslUtil.isWrite( selector,em,key,sendBuffer ) ;
                                }
                                // read.
                                if( key.isReadable() ) {
                                    NioSslUtil.isRead( selector,em,man,key ) ;
                                }
                            }
                            // エラー及びタイムアウト検知.
                            if( !key.isValid() || em == null ||
                                em.getTime() + NioTcpUtil.DESTROY_TIMEOUT < System.currentTimeMillis() ) {
                                LOG.warn( "### 通信エラーを検知" ) ;
                                if( em != null ) {
                                    NioTcpUtil.destroyConnect( key,em ) ;
                                }
                                else if( key != null ) {
                                    NioTcpUtil.destroyKey( key ) ;
                                }
                            }
                        }
                    } catch( ClosedChannelException ce ) {
                        if( em != null ) {
                            NioTcpUtil.destroyConnect( key,em ) ;
                        }
                        else if( key != null ) {
                            NioTcpUtil.destroyKey( key ) ;
                        }
                    } catch( CancelledKeyException cee ) {
                        if( em != null ) {
                            NioTcpUtil.destroyConnect( key,em ) ;
                        }
                        else if( key != null ) {
                            NioTcpUtil.destroyKey( key ) ;
                        }
                    } catch( Exception e ) {
                        if( LOG.isDebugEnabled() ) {
                            LOG.debug( "SSL通信エラー:",e ) ;
                        }
                        if( em != null ) {
                            NioTcpUtil.destroyConnect( key,em ) ;
                        }
                        else if( key != null ) {
                            NioTcpUtil.destroyKey( key ) ;
                        }
                    }
                    key = null ;
                    em = null ;
                }
            }
        }
    }
}

