package org.maachang.seabass.net.nio ;

import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.seabass.net.NioException;
import org.maachang.seabass.net.ssl.SslElement;
import org.maachang.seabass.net.ssl.SslOption;

/**
 * NIO-SSLサーバユーティリティ.
 * 
 * @version 2008/11/15
 * @author  masahito suzuki
 * @since   SeaBass 1.0.0
 */

public abstract class NioSslUtil {
    
    /**
     * Selectorタイムアウト値.
     */
    protected static final int TIMEOUT = 50 ;
    
    /**
     * ログ.
     */
    private static final Log LOG = LogFactory.getLog( NioSslUtil.class ) ;
    
    /**
     * accept処理.
     * <p>単一スレッド用</p>
     */
    protected static final boolean toAcceptBySingle( final Selector selector,final SelectionKey key,final IsWriteQueue queue,final SslOption option )
        throws Exception {
        // accept.
        if( key.isValid() && key.isAcceptable() ) {
            SocketChannel sc = null ;
            try {
                // 新しいAcceptをセット.
                ServerSocketChannel ssc = ( ServerSocketChannel )key.channel() ;
                sc = ssc.accept() ;
                if( sc != null ) {
                    NioTcpUtil.setOption( true,sc ) ;
                    NioElement em = NioElementPoolingFactory.get( sc,queue ) ;
                    //NioElement em = new NioElement( sc,queue ) ;
                    em.setSsl( option ) ;
                    em.createSsl() ;
                    SelectionKey regKey = sc.register( selector,SelectionKey.OP_READ,em ) ;
                    if( regKey != null ) {
                        em.setKey( regKey ) ;
                        em.update() ;
                    }
                    else {
                        NioTcpUtil.destroyElement( em ) ;
                    }
                }
            } catch( Exception e ) {
                //LOG.warn( "## accept失敗",e ) ;
                if( sc != null ) {
                    try { sc.close() ; } catch( Exception ee ) {}
                }
            }
            return true ;
        }
        return false ;
    }
    
    /**
     * accept処理.
     * <p>複数スレッド用</p>
     */
    protected static boolean toAcceptByPlural( final SelectionKey key,final NioSslReceiveServerArray recvServer,final SslOption option )
        throws Exception {
        // accept.
        if( key.isValid() && key.isAcceptable() ) {
            SocketChannel sc = null ;
            try {
                // 新しいAcceptをセット.
                ServerSocketChannel ssc = ( ServerSocketChannel )key.channel() ;
                sc = ssc.accept() ;
                if( sc != null ) {
                    NioTcpUtil.setOption( true,sc ) ;
                    NioElement em = NioElementPoolingFactory.get( sc ) ;
                    //NioElement em = new NioElement( sc ) ;
                    em.setSsl( option ) ;
                    em.createSsl() ;
                    em.update() ;
                    recvServer.setElement( em ) ;
                }
            } catch( Exception e ) {
                //LOG.warn( "## accept失敗",e ) ;
                if( sc != null ) {
                    try { sc.close() ; } catch( Exception ee ) {}
                }
            }
            return true ;
        }
        return false ;
    }
    
    /**
     * 受信処理.
     */
    protected static void isRead( final Selector selector,final NioElement em,final NioPoolManager man,final SelectionKey key )
        throws Exception {
        SslElement ssl = em.getSslElement() ;
        boolean handShake = false ;
        boolean result = true ;
        // ハンドシェイク処理.
        if( !ssl.isHandshake() ) {
            result = ssl.doHandshake( key,true,false,null,em.getWriteBuffer() ) ;
            handShake = true ;
        }
        // ハンドシェイク処理が完了している場合.
        if( result ) {
            /////////////////////////////////////
            // ハンドシェイクが終了している場合.
            /////////////////////////////////////
            // 受信情報が存在しない場合.
            if( ssl.read( handShake ) == -1 ) {
                NioTcpUtil.destroyElement( em ) ;
                return ;
            }
            // データ送信がはじめての場合.
            else if( !em.isExecutionFlag() ) {
                // 通信タイムアウトを設定.
                em.update() ;
                // 通信開始を設定.
                em.setKeepAliveTime(-1) ;
                // 接続条件をPoolキューに通知.
                em.setExecutionFlag( true ) ;
                // Pooling実行スレッドに登録.
                man.execution( em ) ;
            }
        }
    }
    
    /**
     * 送信処理.
     */
    protected static void isWrite( final Selector selector,final NioElement em,final SelectionKey key,final ByteBuffer buf ) throws Exception {
        // write.
        SslElement ssl = em.getSslElement() ;
        // 書き込みバッファを取得.
        SendBuffer wBuf = em.getWriteBuffer() ;
        if( !ssl.isHandshake() ) {
            ssl.doHandshake( key,false,true,buf,wBuf ) ;
        }
        // ハンドシェイク処理.
        else {
            /////////////////////////////////////
            // ハンドシェイクが終了している場合.
            /////////////////////////////////////
            int mode = send( em,buf,wBuf ) ;
            if( mode != RESULT_SEND_LAST && !wBuf.isRead() ) {
                mode = RESULT_SEND_NONE ;
            }
            // 送信データが存在しない場合.
            if( mode == RESULT_SEND_NONE || mode == RESULT_SEND_LAST ) {
                // 書き込みの終端を検知.
                if( mode == RESULT_SEND_LAST ) {
                    // 読み込み条件をセット.
                    em.opMode( SelectionKey.OP_READ ) ;
                    // 送信バッファを再構成.
                    em.recreateWriteBuffer() ;
                }
                else {
                    // 書き込み内容が存在しない場合.
                    // 読み込み条件をセット.
                    em.opMode( SelectionKey.OP_READ ) ;
                }
            }
        }
    }
    
    // ******* 送信結果情報 ********.
    private static final int RESULT_SEND_NONE = 0 ;
    private static final int RESULT_SEND_SUCCESS = 1 ;
    private static final int RESULT_SEND_LAST = 2 ;
    
    /**
     * データ送信処理.
     */
    protected static final int send( final NioElement emt,ByteBuffer buf,final SendBuffer wBuf ) throws Exception {
        int res = wBuf.read( emt.getChannel(),buf ) ;
        if( res <= 0 ) {
            if( res <= -1 ) {
                NioTcpUtil.destroyElement( emt ) ;
                throw new NioException( "通信処理は既に停止しています" ) ;
            }
            return RESULT_SEND_NONE ;
        }
        emt.update() ;
        // 最終データを取得.
        if( res == Integer.MAX_VALUE ) {
            return RESULT_SEND_LAST ;
        }
        return RESULT_SEND_SUCCESS ;
    }
}
