package org.maachang.seabass.net.nio ;

import java.io.IOException;
import java.net.Socket;
import java.nio.ByteBuffer;
import java.nio.channels.Channel;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;

import org.maachang.seabass.net.NioException;

/**
 * NIO-TCP/IPサーバユーティリティ.
 * 
 * @version 2008/11/15
 * @author  masahito suzuki
 * @since   SeaBass 1.0.0
 */
abstract class NioTcpUtil {
    private NioTcpUtil(){}
    
    /** 基本通信条件. **/
    //private static final int RCVBUF = 25188 ;
    //private static final int SENDBUF = 43800 ;
    private static final int RCVBUF = 65535 ;
    //private static final int SENDBUF = 65535 ;
    //private static final int RCVBUF = 0x00100000 ;
    //private static final int SENDBUF = 0x00100000 ;
    private static final int SENDBUF = 0x00080000 ;
    private static final int LINGER = 0 ;
    
    /**
     * Selectorタイムアウト値.
     */
    protected static final int TIMEOUT = 50 ;
    
    /**
     * 破棄タイムアウト値.
     */
    protected static final long DESTROY_TIMEOUT = 5000L ;
    
    /**
     * accept処理.
     * <p>単一スレッド用</p>
     */
    protected static final boolean toAcceptBySingle( final Selector selector,final SelectionKey key,
        final IsWriteQueue queue )
        throws Exception {
        if( key.isValid() && key.isAcceptable() ) {
            SocketChannel sc = null ;
            try {
                // 新しいAcceptをセット.
                ServerSocketChannel ssc = ( ServerSocketChannel )key.channel() ;
                sc = ssc.accept() ;
                if( sc != null ) {
                    setOption( false,sc ) ;
                    NioElement em = NioElementPoolingFactory.get( sc,queue ) ;
                    //NioElement em = new NioElement( sc,queue ) ;
                    SelectionKey regKey = sc.register( selector,SelectionKey.OP_READ,em ) ;
                    if( regKey != null ) {
                        em.setKey( regKey ) ;
                        em.opMode( SelectionKey.OP_READ ) ;
                    }
                    else {
                        destroyElement( em ) ;
                    }
                }
                return true ;
            } catch( Exception e ) {
                if( sc != null ) {
                    try { sc.close() ; } catch( Exception ee ) {}
                }
            }
        }
        return false ;
    }
    
    /**
     * accept処理.
     * <p>複数スレッド用</p>
     */
    protected static final boolean toAcceptByPlural( final SelectionKey key,final NioReceiveServerArray recvServer )
        throws Exception {
        if( key.isValid() && key.isAcceptable() ) {
            SocketChannel sc = null ;
            try {
                // 新しいAcceptをセット.
                ServerSocketChannel ssc = ( ServerSocketChannel )key.channel() ;
                if( ( sc = ssc.accept() ) != null ) {
                    setOption( false,sc ) ;
                    NioElement em = NioElementPoolingFactory.get( sc ) ;
                    //NioElement em = new NioElement( sc ) ;
                    recvServer.setElement( em ) ;
                }
                return true ;
            } catch( Exception e ) {
                if( sc != null ) {
                    try { sc.close() ; } catch( Exception ee ) {}
                }
            }
        }
        return false ;
    }
    
    /**
     * 受信処理.
     */
    protected static final boolean executionRead( final Selector selector,final SelectionKey key,
        final NioPoolManager man,final ByteBuffer buffer )
        throws Exception {
        // read.
        NioElement em ;
        if( ( em = getNioElement( key ) ) != null ) {
            try {
                // 差分データ読み込み処理.
                if( !recv( em,buffer ) ) {
                    // データが-1バイトの場合は、クライアント
                    // コネクション切断なので、要素を破棄する.
                    destroyConnect( key,em ) ;
                    return false ;
                }
                // 最初のデータ送信.
                else if( !em.isExecutionFlag() ) {
                    // 通信開始を設定.
                    em.setKeepAliveTime(-1) ;
                    // 接続条件をPoolキューに通知.
                    em.setExecutionFlag( true ) ;
                    // Pooling実行スレッドに登録.
                    man.execution( em ) ;
                    em.update() ;
                }
                return true ;
            } catch( IOException e ) {
                destroyElement( em ) ;
            }
            return false ;
        }
        else {
            throw new NioException( "通信処理は既に停止しています" ) ;
        }
    }
    
    /**
     * 受信処理.
     */
    protected static final boolean executionRead( final Selector selector,final SelectionKey key,final NioElement em,
        final NioPoolManager man,final ByteBuffer buffer )
        throws Exception {
        // 差分データ読み込み処理.
        try {
            if( !recv( em,buffer ) ) {
                // データが-1バイトの場合は、クライアント
                // コネクション切断なので、要素を破棄する.
                destroyConnect( key,em ) ;
                return false ;
            }
            // 最初のデータ送信.
            else if( !em.isExecutionFlag() ) {
                // 通信開始を設定.
                em.setKeepAliveTime(-1) ;
                // 接続条件をPoolキューに通知.
                em.setExecutionFlag( true ) ;
                // Pooling実行スレッドに登録.
                man.execution( em ) ;
                em.update() ;
            }
            return true ;
        } catch( IOException e ) {
            destroyElement( em ) ;
        }
        return false ;
    }
    
    /**
     * 送信処理.
     */
    protected static final void executionWrite( final Selector selector,final SelectionKey key,final ByteBuffer buf )
        throws Exception {
        // write.
        NioElement em ;
        if( ( em = getNioElement( key ) ) != null ) {
            SendBuffer wBuf = em.getWriteBuffer() ;
            int mode = send( em,buf,wBuf ) ;
            if( mode != RESULT_SEND_LAST && !wBuf.isRead() ) {
                mode = RESULT_SEND_NONE ;
            }
            // 送信データが存在しない場合.
            if( mode == RESULT_SEND_NONE || mode == RESULT_SEND_LAST ) {
                // 書き込みの終端を検知.
                if( mode == RESULT_SEND_LAST ) {
                    // 読み込み条件をセット.
                    em.opMode( SelectionKey.OP_READ ) ;
                    // 送信バッファを再構成.
                    em.recreateWriteBuffer() ;
                    // クローズモードの場合.
                    if( em.isCloseFlag() ) {
                        destroyConnect( key,em ) ;
                    }
                }
                // 書込み終端でない場合.
                else {
                    // 書き込み内容が存在しない場合.
                    // 読み込み条件をセット.
                    em.opMode( SelectionKey.OP_READ ) ;
                }
            }
        }
        else {
            throw new NioException( "通信処理は既に停止しています" ) ;
        }
    }
    
    /**
     * 送信処理.
     */
    protected static final void executionWrite( final Selector selector,final SelectionKey key,final NioElement em,final ByteBuffer buf )
        throws Exception {
        SendBuffer wBuf = em.getWriteBuffer() ;
        int mode = send( em,buf,wBuf ) ;
        if( mode != RESULT_SEND_LAST && !wBuf.isRead() ) {
            mode = RESULT_SEND_NONE ;
        }
        // 送信データが存在しない場合.
        if( mode == RESULT_SEND_NONE || mode == RESULT_SEND_LAST ) {
            // 書き込みの終端を検知.
            if( mode == RESULT_SEND_LAST ) {
                // 読み込み条件をセット.
                em.opMode( SelectionKey.OP_READ ) ;
                // 送信バッファを再構成.
                em.recreateWriteBuffer() ;
                // クローズモードの場合.
                if( em.isCloseFlag() ) {
                    destroyConnect( key,em ) ;
                }
            }
            // 書込み終端でない場合.
            else {
                // 書き込み内容が存在しない場合.
                // 読み込み条件をセット.
                em.opMode( SelectionKey.OP_READ ) ;
            }
        }
    }
    
    /**
     * Acceptソケット設定.
     */
    protected static final void setOption( final boolean sslMode,final SocketChannel channel )
        throws Exception {
        channel.configureBlocking(false) ;
        Socket soc = channel.socket() ;
        soc.setReuseAddress( true ) ;
        //soc.setSoLinger( true,LINGER ) ;
        soc.setSoLinger( false,LINGER ) ;
        soc.setSendBufferSize( SENDBUF ) ;
        soc.setReceiveBufferSize( RCVBUF ) ;
        soc.setKeepAlive( false ) ;
        soc.setTcpNoDelay( true ) ;
        //soc.setTcpNoDelay( false ) ;
        //soc.setOOBInline( true ) ;
        soc.setOOBInline( false ) ;
        soc.setPerformancePreferences( 1,2,0 ) ;
        soc.setTrafficClass( 0x00000004 | 0x00000008 | 0x00000010 ) ;
    }
    
    /**
     * Accept,書き込み処理待ち.
     */
    protected static final boolean isSendOrAccept( final Selector selector,final IsReadQueue read,
        final IsWriteQueue write ) throws Exception {
        boolean ret = false ;
        NioElement element = null ;
        // 送信データ存在確認.
        while( ( element = write.getQueue() ) != null ) {
            ret = true ;
            if( element.isWriteBuffer() ) {
                element.opMode( SelectionKey.OP_WRITE ) ;
            }
            else {
                element.opMode( SelectionKey.OP_READ ) ;
            }
        }
        // Accept存在確認.
        while( ( element = read.getQueue() ) != null ) {
            ret = true ;
            SelectionKey key = element.getChannel().register( selector,SelectionKey.OP_READ,element ) ;
            if( key != null ) {
                element.setKey( key ) ;
                element.opMode( SelectionKey.OP_READ ) ;
            }
            else {
                destroyElement( element ) ;
            }
        }
        return ret ;
    }
    
    /**
     * Accept処理待ち.
     */
    protected static final boolean isAccept( final Selector selector,final IsReadQueue read ) throws Exception {
        boolean ret = false ;
        NioElement element ;
        // Accept存在確認.
        while( ( element = read.getQueue() ) != null ) {
            ret = true ;
            SelectionKey key = element.getChannel().register( selector,SelectionKey.OP_READ,element ) ;
            if( key != null ) {
                element.setKey( key ) ;
                element.opMode( SelectionKey.OP_READ ) ;
            }
            else {
                destroyElement( element ) ;
            }
        }
        return ret ;
    }
    
    /**
     * 書き込み処理待ち.
     */
    protected static final boolean isWrite( final IsWriteQueue queue ) throws Exception {
        boolean ret = false ;
        NioElement element = null ;
        // 送信データ存在確認.
        while( ( element = queue.getQueue() ) != null ) {
            ret = true ;
            if( element.isWriteBuffer() ) {
                element.opMode( SelectionKey.OP_WRITE ) ;
            }
            else {
                element.opMode( SelectionKey.OP_READ ) ;
            }
        }
        return ret ;
    }
    
    /**
     * 通信破棄.
     */
    protected static final void destroyConnect( final SelectionKey key,final NioElement em ) {
        Channel ch = em.getChannel() ;
        shutdownChannel( ch ) ;
        closeChannel( ch  ) ;
        destroyKey( key ) ;
        destroyElement( em ) ;
    }
    
    /**
     * Elementを破棄する.
     */
    protected static final void destroyElement( final NioElement em ) {
        if( em != null ) {
            em.destroy() ;
        }
    }
    
    /**
     * Keyが無効な場合の処理.
     */
    protected static final void destroyKey( final SelectionKey key ) {
        if( key != null ) {
            NioElement em = ( NioElement )key.attachment() ;
            key.attach( null ) ;
            if( em != null ) {
                em.destroy() ;
            }
            closeChannel( key.channel() ) ;
            try {
                key.interestOps( 0 ) ;
            } catch( Exception e ) {
            }
            try {
                key.cancel() ;
            } catch( Exception e ) {
            }
        }
    }
    
    /**
     * チャネルクローズ.
     */
    protected static final void closeChannel( final Channel ch ) {
        if( ch != null ) {
            if( ch instanceof SocketChannel ) {
                shutdownChannel( ch ) ;
                try {
                    ((SocketChannel)ch).socket().close() ;
                } catch( Exception e ) {
                }
            }
            try {
                ch.close() ;
            } catch( Exception e ) {
            }
        }
    }
    
    /**
     * チャネルシャットダウン.
     */
    protected static final void shutdownChannel( final Channel ch ) {
        if( ch != null && ch instanceof SocketChannel ) {
            Socket s = (( SocketChannel )ch).socket() ;
            if( s != null ) {
                try {
                    s.shutdownInput() ;
                } catch( Exception e ) {
                }
                try {
                    s.shutdownOutput() ;
                } catch( Exception e ) {
                }
            }
            s = null ;
        }
    }
    
    /**
     * 要素取得.
     */
    protected static final NioElement getNioElement( final SelectionKey key ) throws Exception {
        NioElement ret = null ;
        // 要素取得に失敗および、取得要素が不正.
        if( !key.isValid() || ( ret = ( NioElement )key.attachment() ) == null ) {
            destroyKey( key ) ;
            ret = null ;
        }
        return ret ;
    }
    
    // ******* 送信結果情報 ********.
    private static final int RESULT_SEND_NONE = 0 ;
    private static final int RESULT_SEND_SUCCESS = 1 ;
    private static final int RESULT_SEND_LAST = 2 ;
    
    /**
     * データ送信.
     */
    protected static final int send( final NioElement emt,ByteBuffer buf,final SendBuffer wBuf ) throws Exception {
        int res = wBuf.read( emt.getChannel(),buf ) ;
        if( res <= 0 ) {
            if( res <= -1 ) {
                destroyElement( emt ) ;
                throw new NioException( "通信処理は既に停止しています" ) ;
            }
            return RESULT_SEND_NONE ;
        }
        emt.update() ;
        // 最終データを取得.
        if( res == Integer.MAX_VALUE ) {
            return RESULT_SEND_LAST ;
        }
        return RESULT_SEND_SUCCESS ;
    }
    
    /**
     * データ受信処理.
     */
    protected static final boolean recv( final NioElement emt,final ByteBuffer buffer ) throws Exception {
        buffer.clear() ;
        int len = emt.getChannel().read( buffer ) ;
        if( len <= 0 ) {
            if( len == 0 ) {
                emt.update() ;
                return true ;
            }
            return false ;
        }
        emt.update() ;
        buffer.flip() ;
        emt.getReceiveBuffer().put( buffer ) ;
        return true ;
    }
    
    /**
     * SelectionKey内を表示.
     */
    protected static final String keyByString( SelectionKey key ) {
        StringBuilder buf = new StringBuilder() ;
        buf.append( "key(" ).append( key.isValid() ).append( ") " ) ;
        if( key.isAcceptable() ) {
            buf.append( "[ACCEPT] " ) ;
        }
        if( key.isConnectable() ) {
            buf.append( "[CONNECT] " ) ;
        }
        if( key.isReadable() ) {
            buf.append( "[READ] " ) ;
        }
        if( key.isWritable() ) {
            buf.append( "[WRITE]" ) ;
        }
        return buf.toString() ;
    }
    
}
