package org.maachang.seabass.net.nio ;

import java.io.IOException;
import java.net.SocketTimeoutException;
import java.nio.ByteBuffer;
import java.util.concurrent.ConcurrentLinkedQueue;

import org.maachang.seabass.net.NioException;
import org.maachang.util.atomic.AtomicBOOL;
import org.maachang.util.atomic.Blocking;

/**
 * 受信バッファ.
 * 
 * @version 2008/11/27
 * @author  masahito suzuki
 * @since   SeaBass 1.0.0
 */
public class ReceiveBuffer {
    private static final long WAIT = 50L ;
    private final ConcurrentLinkedQueue<byte[]> queue = new ConcurrentLinkedQueue<byte[]>() ;
    private final Blocking blocking = new Blocking() ;
    private final AtomicBOOL useConnectFlag = new AtomicBOOL( false ) ;
    
    /**
     * コンストラクタ.
     */
    public ReceiveBuffer() {
    }
    
    protected void finalize() throws Exception {
        this.destroy() ;
    }
    
    /**
     * オブジェクト生成.
     */
    protected void create() {
        useConnectFlag.set( true ) ;
    }
    
    /**
     * オブジェクト破棄.
     */
    protected void destroy() {
        reset( true ) ;
    }
    
    /**
     * 内容リセット.
     */
    public void reset() {
        reset( false ) ;
    }
    
    /**
     * 内容リセット.
     * @param mode [true]の場合、コネクション切断を通知します.
     */
    public void reset( boolean mode ) {
        if( mode ) {
            useConnectFlag.set( false ) ;
        }
        queue.clear() ;
        blocking.freeAll() ;// 全解除.
    }
    
    /**
     * データセット.
     * @param buf ByteBufferオブジェクトを設定します.
     * @exception Exception 例外.
     */
    public void put( ByteBuffer buf ) throws Exception {
        if( !useConnectFlag.get() ) {
            throw new NioException( "オブジェクトは既にクローズしています" ) ;
        }
        if( buf.position() > 0 ) {
            buf.flip() ;
        }
        int len = buf.limit() - buf.position() ;
        if( len > 0 ) {
            byte[] b = new byte[ len ] ;
            buf.get( b ) ;
            queue.offer( b ) ;
            blocking.freeAll() ;// 全解除.
        }
        buf.clear() ;
    }
    
    /**
     * 受信情報を読み込む.
     * @param timeout 対象のタイムアウト値を設定します.
     * @return byte[] 受信したバイナリ情報が返されます.
     * @exception IOException 例外.
     */
    public byte[] get( long timeout )
        throws IOException {
        if( !useConnectFlag.get() ) {
            throw new NioException( "オブジェクトは既にクローズしています" ) ;
        }
        byte[] ret = queue.poll() ;
        if( ret != null ) {
            return ret ;
        }
        if( timeout <= 0L ) {
            while( true ) {
                try {
                    if( blocking.stop( WAIT ) ) {
                        if( !useConnectFlag.get() ) {
                            throw new NioException( "オブジェクトは既にクローズしています" ) ;
                        }
                    }
                } catch( InterruptedException ie ) {
                    throw new IOException( "割り込み例外(InterruptedException)が発生しました" ) ;
                }
                if( ( ret = queue.poll() ) != null ) {
                    return ret ;
                }
            }
        }
        else {
            long cnt = 0L ;
            while( true ) {
                try {
                    if( blocking.stop( WAIT ) ) {
                        cnt += WAIT ;
                        if( cnt >= timeout ) {
                            throw new SocketTimeoutException( "受信タイムアウトを検知:" + timeout ) ;
                        }
                        if( !useConnectFlag.get() ) {
                            throw new NioException( "オブジェクトは既にクローズしています" ) ;
                        }
                    }
                } catch( InterruptedException ie ) {
                    throw new IOException( "割り込み例外(InterruptedException)が発生しました" ) ;
                }
                if( ( ret = queue.poll() ) != null ) {
                    return ret ;
                }
            }
        }
    }
    
    /**
     * 読み込み可能データが存在するかチェック.
     * @return boolean [true]の場合、読み込み可能データが存在します.
     */
    public boolean isData() {
        return queue.size() > 0 ;
    }
    
    /**
     * オブジェクトが有効かチェック.
     * @return boolean [true]の場合、有効です.
     */
    public boolean isUse() {
        return useConnectFlag.get() ;
    }
}

