package org.maachang.seabass.net.nio ;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;
import java.util.concurrent.ConcurrentLinkedQueue;

import org.maachang.seabass.net.NioException;
import org.maachang.util.atomic.AtomicBOOL;
import org.maachang.util.atomic.Blocking;

/**
 * 送信バッファ.
 * 
 * @version 2008/11/27
 * @author  masahito suzuki
 * @since   SeaBass 1.0.0
 */
public class SendBuffer {
    private static final int MAX_QUEUE = 256 ;
    private static final int RESTART_QUEUE_LENGTH = (int)( (double)MAX_QUEUE * 0.65d ) ;
    private static final int WAIT_TIME = 50 ;
    private static final int BUFFER = NioOutputStream.BUFFER_LENGTH ;
    
    protected static final byte[] LAST_BYTE = new byte[0] ;
    private final ConcurrentLinkedQueue<byte[]> queue = new ConcurrentLinkedQueue<byte[]>() ;
    private final Blocking blocking = new Blocking() ;
    private final AtomicBOOL useConnectFlag = new AtomicBOOL( false ) ;
    private final AtomicBOOL lastFlag = new AtomicBOOL( false ) ;
    private final AtomicBOOL waitFlag = new AtomicBOOL( false ) ;
    
    /**
     * コンストラクタ.
     */
    public SendBuffer() {
        useConnectFlag.set( true ) ;
    }
    
    protected void finalize() throws Exception {
        this.destroy() ;
    }
    
    /**
     * オブジェクト破棄.
     */
    protected void destroy() {
        useConnectFlag.set( false ) ;
        queue.clear() ;
        lastFlag.set( false ) ;
        blocking.freeAll() ;// 全解除.
    }
    
    /**
     * オブジェクト生成.
     */
    public void create() {
        queue.clear() ;
        lastFlag.set( false ) ;
        useConnectFlag.set( true ) ;
        blocking.freeAll() ;// 全解除.
    }
    
    /**
     * 内容クリア
     */
    public void clear() {
        useConnectFlag.set( false ) ;
        queue.clear() ;
        lastFlag.set( false ) ;
        blocking.freeAll() ;// 全解除.
    }
    
    /**
     * バッファ情報をソケットチャネルに出力.
     * @param buf 書き込み対象のバッファオブジェクトを設定します.
     * @param channel 対象のソケットチャネルを設定します.
     * @return int 送信長が返されます.
     *             [Integer.MAX_VALUE]の場合は、終端を取得しました.
     * @exception Exception 例外.
     */
    public int read( SocketChannel channel,ByteBuffer buf )
        throws Exception {
        byte[] b ;
        if( !useConnectFlag.get() ) {
            throw new NioException( "オブジェクトは既にクローズしています" ) ;
        }
        int ret = 0 ;
        if( ( b = queue.poll() ) != null && b.length > 0 ) {
            buf.clear() ;
            buf.put( b ) ;
            buf.flip() ;
            ret = channel.write( buf ) ;
            if( waitFlag.get() && RESTART_QUEUE_LENGTH > queue.size() ) {
                blocking.freeAll() ;// 全解除.
            }
            return ret ;
        }
        // 終端データを取得した場合.
        if( b == LAST_BYTE ) {
            ret = Integer.MAX_VALUE ;
        }
        return ret ;
    }
    
    /** 既定値を越す情報数が存在する場合は、待機. **/
    private void waitTo() throws IOException {
        if( queue.size() > MAX_QUEUE ) {
            waitFlag.set( true ) ;
            try {
                while( true ) {
                    blocking.waitTo( WAIT_TIME ) ;
                    if( queue.size() > RESTART_QUEUE_LENGTH ) {
                        if( !useConnectFlag.get() || lastFlag.get() ) {
                            throw new NioException( "オブジェクトは既にクローズしています" ) ;
                        }
                    }
                    else {
                        break ;
                    }
                }
            } finally {
                waitFlag.set( false ) ;
            }
        }
    }
    
    /**
     * データセット.
     * @param buf ByteBufferオブジェクトを設定します.
     * @return boolean [true]の場合、書き込まれました.
     * @exception IOException I/O例外.
     */
    public boolean write( ByteBuffer buf ) throws IOException {
        if( !useConnectFlag.get() || lastFlag.get() ) {
            throw new NioException( "オブジェクトは既にクローズしています" ) ;
        }
        waitTo() ;
        if( buf.position() > 0 ) {
            buf.flip() ;
        }
        int len = buf.limit() - buf.position() ;
        if( len > BUFFER ) {
            byte[] bin ;
            int bLen ;
            while( true ) {
                bLen = len ;
                if( bLen > BUFFER ) {
                    bLen = BUFFER ;
                }
                bin = new byte[ bLen ] ;
                buf.get( bin ) ;
                queue.offer( bin ) ;
                len -= bLen ;
                if( bLen != BUFFER ) {
                    return true ;
                }
            }
        }
        if( len <= 0 ) {
            return false ;
        }
        byte[] bin = new byte[ len ] ;
        buf.get( bin ) ;
        queue.offer( bin ) ;
        return true ;
    }
    
    /**
     * 読み込み可能データが存在するかチェック.
     * @return boolean [true]の場合、読み込み可能データが存在します.
     */
    public boolean isRead() {
        return queue.size() > 0 ;
    }
    
    /**
     * オブジェクトが有効かチェック.
     * @return boolean [true]の場合、有効です.
     */
    public boolean isUse() {
        return useConnectFlag.get() ;
    }
    
    /**
     * 書き込み終了設定.
     */
    public void last() {
        if( !lastFlag.get() ) {
            lastFlag.set( true ) ;
            queue.offer( LAST_BYTE ) ;// lastデータを出力.
        }
    }
    
    /**
     * 書き込み終了フラグがセットされている.
     * @return boolean [true]の場合、書き込み終了は行われています.
     */
    public boolean lastFlag() {
        return lastFlag.get() ;
    }
}

