package org.maachang.seabass.net.ssl ;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;

import javax.net.ssl.SSLEngine;
import javax.net.ssl.SSLEngineResult;
import javax.net.ssl.SSLException;
import javax.net.ssl.SSLEngineResult.HandshakeStatus;
import javax.net.ssl.SSLEngineResult.Status;

import org.maachang.seabass.net.NioException;
import org.maachang.seabass.net.nio.NioElement;
import org.maachang.seabass.net.nio.SendBuffer;
import org.maachang.util.atomic.AtomicBOOL;

/**
 * SSLソケット要素.
 *  
 * @version 2008/06/01
 * @author  masahito suzuki
 * @since   SeaBass 1.0.0
 */
public class SslElement {
    private NioElement element = null ;
    private SSLEngine sslEngine = null ;
    
    protected ByteBuffer requestBB;
    
    private int appBBSize;
    private int netBBSize;
    
    private ByteBuffer inNetBB = null ;
    private ByteBuffer outNetBB = null ;
    
    private static ByteBuffer hsBB = ByteBuffer.allocate(0);
    
    private HandshakeStatus initialHSStatus;
    private boolean initialHSComplete = false ;
    
    private final AtomicBOOL shutdownFlag = new AtomicBOOL( false ) ;
    private final AtomicBOOL openFlag = new AtomicBOOL( false ) ;
    
    /**
     * コンストラクタ.
     */
    private SslElement() {
        
    }
    
    /**
     * コンストラクタ.
     * @param element 通信要素を設定します.
     * @param engine SSLEngineを設定します.
     * @exception Exception 例外.
     */
    public SslElement( final NioElement element,final SSLEngine engine ) throws Exception {
        if( element == null || !element.isUse() || engine == null ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        this.init( engine ) ;
        String[] supportCiphers = engine.getSupportedCipherSuites() ;
        if( supportCiphers != null ) {
            engine.setEnabledCipherSuites( supportCiphers ) ;
        }
        this.element = element ;
        this.sslEngine = engine ;
        this.openFlag.set( true ) ;
        this.shutdownFlag.set( false ) ;
    }
    
    protected void finalize() throws Exception {
        destroy() ;
    }
    
    /** SSL要素初期化処理. **/
    protected void init( final SSLEngine engine ) throws Exception {
        engine.beginHandshake() ;
        initialHSStatus = HandshakeStatus.NEED_UNWRAP;
        netBBSize = engine.getSession().getPacketBufferSize();
        inNetBB = ByteBuffer.allocate(netBBSize);
        outNetBB = ByteBuffer.allocate(netBBSize);
        outNetBB.position(0);
        outNetBB.limit(0);
        appBBSize = engine.getSession().getApplicationBufferSize();
        requestBB = ByteBuffer.allocate(appBBSize);
    }
    
    /**
     * オブジェクト破棄.
     */
    public void destroy() {
        if( openFlag.get() ) {
            openFlag.set( false ) ;
        }
        inNetBB = null ;
        outNetBB = null ;
        initialHSStatus = null ;
    }
    
    /** 読み込み処理. **/
    private int nioRead( final ByteBuffer buf ) throws Exception {
        element.update() ;
        int ret ;
        try {
            ret = element.getChannel().read( buf ) ;
        } catch( Exception e ) {
            ret = -1 ;
        }
        return ret ;
    }
    
    /**
     * ハンドシェイク再開処理.
     * @exception Exception 例外.
     */
    public void beginHandshake() throws Exception {
        initialHSStatus = HandshakeStatus.NEED_UNWRAP;
        inNetBB.clear() ;
        outNetBB.clear() ;
        requestBB.clear() ;
        sslEngine.beginHandshake() ;
        shutdownFlag.set( false ) ;
    }
    
    /** データ書込み処理. **/
    private int tryFlush( final ByteBuffer tmp,final SendBuffer wBuf,final ByteBuffer buf ) throws Exception {
        if( wBuf != null ) {
            // 書き込み対象情報(buf)が存在する場合.
            if( buf.hasRemaining() ) {
                // 書き込みバッファに現在内容を追加書き込み.
                wBuf.write( buf ) ;
            }
            // 書き込みバッファが存在する場合.
            int result = wBuf.read( element.getChannel(),tmp ) ;
            if( result > 0 ) {
                element.update() ;
                if ( wBuf.isRead() ) {
                    element.opMode( SelectionKey.OP_WRITE ) ;
                    return 1 ;
                }
                return 0 ;
            }
            // コネクションが切断されていることを検知.
            else if( result == -1 ) {
                throw new IOException( "コネクションは切断されています" ) ;
            }
            // 書き込み可能情報が存在しない場合はWRITE処理を止める.
            element.opMode( SelectionKey.OP_READ ) ;
            return 2 ;
        }
        throw new IOException( "コネクションは切断されています" ) ;
    }
    
    /** 書き込みバッファに書込み処理. **/
    private boolean tryFlushByWBuf( final ByteBuffer buf ) throws IOException {
        SendBuffer wBuf = element.getWriteBuffer() ;
        if( wBuf != null ) {
            if( buf.hasRemaining() ) {
                // 書き込みバッファに情報を出力.
                wBuf.write( buf ) ;
                element.opMode( SelectionKey.OP_WRITE ) ;
                element.update() ;
                return true ;
            }
            else {
                buf.clear() ;
                return false ;
            }
        }
        throw new IOException( "コネクションは切断されています" ) ;
    }
    
    /** リクエストバッファサイズリサイズ **/
    protected void resizeRequestBB() {
        resizeRequestBB(appBBSize);
    }
    
    /** リクエストバッファサイズリサイズ **/
    protected void resizeRequestBB(final int remaining) {
        if (requestBB.remaining() < remaining) {
            ByteBuffer bb = ByteBuffer.allocate(requestBB.capacity() * 2);
            requestBB.flip();
            bb.put(requestBB);
            requestBB = bb;
        }
    }
    
    /** レスポンスバッファサイズリサイズ **/
    private void resizeResponseBB() {
        ByteBuffer bb = ByteBuffer.allocate(netBBSize);
        inNetBB.flip();
        bb.put(inNetBB);
        inNetBB = bb;
    }
    
    /**
     * ハンドシェイク処理.
     * @param sk SelectionKeyを設定します.
     * @param readMode 読み込み処理が可能な場合は、処理します.
     * @param writeMode 書込み処理が可能な場合は、処理します.
     * @param writeTmpBuffer 書込みテンポラリバッファを設定します.
     * @param wBuf 書込みバッファを設定します.
     * @return boolean [true]の場合、ハンドシェイク処理は終了しています.
     * @exception Exception 例外.
     */
    public boolean doHandshake(final SelectionKey sk,final boolean readMode,final boolean writeMode,
        final ByteBuffer writeTmpBuffer,final SendBuffer wBuf ) throws Exception {
        if( !openFlag.get() ) {
            throw new IOException( "オブジェクトは既にクローズしています" ) ;
        }
        if( initialHSComplete ) {
            return true ;
        }
        SSLEngineResult result;
        // 書き込み処理の場合.
        if( writeMode ) {
            int c = tryFlush(writeTmpBuffer,wBuf,outNetBB) ;
            switch( c ) {
                // 要求書き込み内容が全て書き込まれた場合.
                case 0 :
                    if( initialHSStatus == HandshakeStatus.FINISHED ) {
                        initialHSComplete = true ;
                        setOpMode( SelectionKey.OP_READ ) ;
                    }
                    return initialHSComplete ;
                case 1 : 
                    return initialHSComplete ;
            }
        }
        switch (initialHSStatus) {
            case NEED_UNWRAP:
                // 読み込み処理.
                if( readMode ) {
                    if ( nioRead(inNetBB) == -1) {
                        element.destroy() ;
                        return initialHSComplete ;
                    }
                }
                else {
                    return initialHSComplete ;
                }
// need loop.
L_NEED_ID:
                while (initialHSStatus == HandshakeStatus.NEED_UNWRAP) {
                    resizeRequestBB();
                    inNetBB.flip();
                    result = sslEngine.unwrap(inNetBB, requestBB);
                    inNetBB.compact();
                    initialHSStatus = result.getHandshakeStatus();
                    switch (result.getStatus()) {
                        case OK:
                            switch (initialHSStatus) {
                                case NOT_HANDSHAKING:
                                    throw new IOException("Not handshaking during initial handshake");
                                case NEED_TASK:
                                    initialHSStatus = doTasks();
                                    break;
                                case FINISHED:
                                    initialHSComplete = true ;
                                    break L_NEED_ID;
                            }
                            break;
                        case BUFFER_UNDERFLOW:
                            netBBSize = sslEngine.getSession().getPacketBufferSize();
                            if (netBBSize > inNetBB.capacity()) {
                                resizeResponseBB();
                            }
                            setOpMode( SelectionKey.OP_READ ) ;
                            break L_NEED_ID;
                        case BUFFER_OVERFLOW:
                            appBBSize = sslEngine.getSession().getApplicationBufferSize();
                            break;
                        default: //CLOSED:
                            throw new IOException("Received" + result.getStatus() +
                            "during initial handshaking");
                    }
                }
                if (initialHSStatus != HandshakeStatus.NEED_WRAP) {
                    break;
                }
            case NEED_WRAP:
                outNetBB.clear();
                result = sslEngine.wrap(hsBB, outNetBB);
                outNetBB.flip();
                initialHSStatus = result.getHandshakeStatus();
                switch (result.getStatus()) {
                    case OK:
                        if (initialHSStatus == HandshakeStatus.NEED_TASK) {
                            initialHSStatus = doTasks();
                        }
                        if (sk != null) {
                            setOpMode( SelectionKey.OP_WRITE ) ;
                        }
                        break;
                    default: // BUFFER_OVERFLOW/BUFFER_UNDERFLOW/CLOSED:
                    throw new IOException("Received" + result.getStatus() +
                        "during initial handshaking");
                }
                break;
            default: // NOT_HANDSHAKING/NEED_TASK/FINISHED
                throw new IOException("Invalid Handshaking State" + initialHSStatus);
        }
        if( initialHSComplete ) {
            setOpMode( SelectionKey.OP_READ ) ;
        }
        return initialHSComplete ;
    }
    
    /** SSLタスク処理 **/
    private SSLEngineResult.HandshakeStatus doTasks() {
        Runnable runnable;
        while ((runnable = sslEngine.getDelegatedTask()) != null) {
            runnable.run();
        }
        return sslEngine.getHandshakeStatus();
    }
    
    /**
     * データ読み込み処理.
     * @param mode [true]の場合、情報を読み込まず処理します.
     * @return int データ読み込み長が返されます.
     * @exception Exception 例外.
     */
    public int read( boolean mode ) throws Exception {
        if( !openFlag.get() || !initialHSComplete ) {
            if( !openFlag.get() ) {
                return -1 ;
            }
            if ( !initialHSComplete ) {
                throw new IllegalStateException( "read処理はハンドシェイク実行中で呼び出せません" );
            }
        }
        SSLEngineResult result;
        if( !mode ) {
            requestBB.clear() ;
            if (nioRead(inNetBB) == -1) {
                return -1;
            }
        }
        if( !requestBB.hasRemaining() ) {
            return 0 ;
        }
        int readLen = 0 ;
        do {
            resizeRequestBB();
            inNetBB.flip();
            result = sslEngine.unwrap(inNetBB, requestBB);
            inNetBB.compact();
            if ( result.getStatus()==Status.OK || result.getStatus()==Status.BUFFER_UNDERFLOW ) {
                readLen += result.bytesProduced();
                if (result.getHandshakeStatus() == HandshakeStatus.NEED_TASK) {
                    doTasks();
                }
                if ( result.getStatus() == Status.BUFFER_UNDERFLOW ) {
                    break;
                }
            }else if ( result.getStatus()==Status.BUFFER_OVERFLOW && readLen>0 ) {
                break;
            }else if( result.getStatus() == Status.CLOSED ) {
                return -1 ;
            }else {
                throw new IOException("sslEngine error during data read: " + result.getStatus());
            }
        } while ( (inNetBB.position() != 0));
        
        
        int ret = requestBB.position() ;
        if( ret > 0 ) {
            element.getReceiveBuffer().put( requestBB ) ;
            element.update() ;
            requestBB.clear() ;
        }
        return ret ;
    }
    
    /**
     * データ書込み.
     * @param src 書込み対象の情報を設定します.
     * @return int 書き込まれた情報長が返されます.
     * @exception Exception 例外.
     */
    public int write( final ByteBuffer src) throws IOException {
        if( !openFlag.get() || !initialHSComplete ) {
            if( !openFlag.get() ) {
                throw new IOException( "オブジェクトは既にクローズしています" ) ;
            }
            if ( !initialHSComplete ) {
                throw new IllegalStateException( "write処理はハンドシェイク実行中で呼び出せません" );
            }
        }
        int retValue = 0;
        if (outNetBB.hasRemaining() && !tryFlushByWBuf(outNetBB)) {
            return retValue;
        }
        outNetBB.clear();
        SSLEngineResult result = sslEngine.wrap(src, outNetBB);
        retValue = result.bytesConsumed();
        outNetBB.flip();
        switch (result.getStatus()) {
            case OK:
                if (result.getHandshakeStatus() == HandshakeStatus.NEED_TASK) {
                    doTasks();
                }
                break;
            default:
                throw new IOException("sslEngine error during data write: " +
                result.getStatus());
        }
        if (outNetBB.hasRemaining()) {
            tryFlushByWBuf(outNetBB);
        }
        return retValue;
    }
    
    /**
     * シャットダウン処理.
     * @return boolean [true]の場合既にシャットダウンしています.
     * @exception Exception 例外.
     */
    public boolean shutdown() throws Exception {
        if( !openFlag.get() ) {
            throw new IOException( "オブジェクトは既にクローズしています" ) ;
        }
        if (shutdownFlag.get()) {
            return true ;
        }
        shutdownFlag.set(true);
        sslEngine.closeOutbound();
        // 出力バッファが存在する場合は処理しない.
        if (tryFlushByWBuf(outNetBB)) {
            return false;
        }
        outNetBB.clear();
        SSLEngineResult result = sslEngine.wrap(hsBB, outNetBB);
        if (result.getStatus() != Status.CLOSED) {
            throw new SSLException("Improper close state:" + result.getStatus());
        }
        outNetBB.flip();
        if (outNetBB.hasRemaining()) {
            tryFlushByWBuf(outNetBB);
        }
        return (!outNetBB.hasRemaining() && (result.getHandshakeStatus() != HandshakeStatus.NEED_WRAP));
    }
    
    /**
     * ハンドシェイク中かチェック.
     * @return boolean [false]の場合、ハンドシェイク中です.
     * @exception Exception 例外.
     */
    public boolean isHandshake() throws Exception {
        if( !openFlag.get() ) {
            throw new IOException( "オブジェクトは既にクローズしています" ) ;
        }
        return initialHSComplete ;
    }
    
    /**
     * シャットダウン完了チェック.
     * @return boolean [true]の場合、シャットダウンは完了しています.
     */
    public boolean isShutdown() {
        return shutdownFlag.get() ;
    }
    
    /**
     * OP_MODE設定.
     */
    private void setOpMode( int mode ) {
        SelectionKey key = element.getKey() ;
        if( key == null ) {
            return ;
        }
        if( ( mode | SelectionKey.OP_WRITE ) != 0 ) {
            key.interestOps( key.interestOps() | SelectionKey.OP_WRITE ) ;
            element.setOpWriteFlag( true ) ;
        }
        else if( ( mode | SelectionKey.OP_READ ) != 0 ) {
            key.interestOps( ( key.interestOps() & (~SelectionKey.OP_WRITE) ) | SelectionKey.OP_READ ) ;
            element.setOpWriteFlag( false ) ;
        }
        else {
            key.interestOps( key.interestOps() & (~SelectionKey.OP_WRITE) ) ;
            element.setOpWriteFlag( false ) ;
        }
    }
}

