package org.maachang.seabass.service.client;

import java.net.URLEncoder;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.maachang.manager.GlobalManager;
import org.maachang.seabass.ServiceDef;
import org.maachang.seabass.conf.Seabass32Key;
import org.maachang.seabass.httpd.HttpdRequest;
import org.maachang.seabass.httpd.HttpdVersionDef;
import org.maachang.seabass.httpd.engine.HttpdDef;
import org.maachang.seabass.httpd.engine.Seabass32;
import org.maachang.seabass.session.engine.PersistenceDefine;
import org.maachang.util.BaseRandom;
import org.maachang.util.Digest;
import org.maachang.util.RandomId;

/**
 * HTTPサービス.
 * 
 * @version 2007/10/18
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
public class HttpService {
    
    /** デフォルトユーザエージェント. **/
    private static final String DEFAULT_USER_AGENT = "seabass("+HttpdVersionDef.getName()+
        ")-" + HttpdVersionDef.getVersion() + "_" + HttpdVersionDef.getDate() ;
    
    /**
     * HTTPプロトコルが設定されていない場合、セット.
     */
    private static final String trimUrl( String url ) {
        if( url == null || ( url = url.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "URLが設定されていません" ) ;
        }
        String s = url.trim() ;
        if( s.startsWith( "http://" ) || s.startsWith( "https://" ) ) {
            return url ;
        }
        return "http://" + url ;
    }
    
    /**
     * 指定URLに対してPOSTでアクセス.
     * <BR><BR>
     * 指定URLに対してPOSTでアクセスします.
     * <BR>
     * @param url 対象のURLを設定します.
     * @param param 対象のパラメータを設定します.
     * @param header 対象のヘッダを設定します.
     * @param userAgent 対象のユーザエージェントを設定します.
     * @return HttpResult 結果情報が返されます.
     * @exception Exception 例外.
     */
    public static final HttpResult post( String url,Map<String,Object> param,HttpHeader header,String userAgent )
        throws Exception {
        url = trimUrl( url ) ;
        if( userAgent == null || ( userAgent = userAgent.trim() ).length() <= 0 ) {
            userAgent = DEFAULT_USER_AGENT ;
        }
        return HttpClientAccess.connection( true,url,null,header,param,userAgent ) ;
    }
    
    /**
     * 指定URLに対してGETでアクセス.
     * <BR><BR>
     * 指定URLに対してGETでアクセスします.
     * <BR>
     * @param url 対象のURLを設定します.
     * @param param 対象のパラメータを設定します.
     * @param header 対象のヘッダを設定します.
     * @param userAgent 対象のユーザエージェントを設定します.
     * @return HttpResult 結果情報が返されます.
     * @exception Exception 例外.
     */
    public static final HttpResult get( String url,Map<String,Object> param,HttpHeader header,String userAgent )
        throws Exception {
        url = trimUrl( url ) ;
        if( userAgent == null || ( userAgent = userAgent.trim() ).length() <= 0 ) {
            userAgent = DEFAULT_USER_AGENT ;
        }
        return HttpClientAccess.connection( false,url,null,header,param,userAgent ) ;
    }
    
    /**
     * 指定URLに対してBodyに情報を設定してアクセス.
     * <BR><BR>
     * 指定URLに対してBodyに情報を設定してアクセスします.
     * <BR>
     * @param url 対象のURLを設定します.
     * @param body 対象のBodyを設定します.
     * @param param 対象のパラメータを設定します.
     * @param header 対象のヘッダを設定します.
     * @param contentType 対象のコンテンツタイプを設定します.
     * @param userAgent 対象のユーザエージェントを設定します.
     * @return HttpResult 結果情報が返されます.
     * @exception Exception 例外.
     */
    public static final HttpResult body( String url,byte[] body,Map<String,Object> param,HttpHeader header,String contentType,String userAgent )
        throws Exception {
        url = trimUrl( url ) ;
        if( userAgent == null || ( userAgent = userAgent.trim() ).length() <= 0 ) {
            userAgent = DEFAULT_USER_AGENT ;
        }
        return HttpClientAccess.connectionByInputBody( url,null,body,contentType,header,param,userAgent ) ;
    }
    
    /** Seabass32用RND64コード. **/
    private static final String RND64 = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz+-" ;
    
    /**
     * Seabass32アクセス.
     * <BR><BR>
     * 指定URLに対して、Seabass32暗号化でアクセスします.
     * <BR>
     * @param req リクエストオブジェクトを設定します.
     * @param sessionId セッションIDを設定します.
     * @param url 対象のURLを設定します.
     * @param param 対象のパラメータを設定します.
     * @param header 対象のヘッダを設定します.
     * @param userAgent 対象のユーザエージェントを設定します.
     * @return HttpResult 結果情報が返されます.
     * @exception Exception 例外.
     */
    public static final HttpResult seabass32( HttpdRequest req,
        String url,Map<String,Object> param,HttpHeader header,String userAgent )
        throws Exception {
        url = trimUrl( url ) ;
        return seabass32( req.getRandom(),req.getSession().getSessionId(),url,param,header,userAgent,null ) ;
    }
    
    /**
     * Seabass32アクセス.
     * <BR><BR>
     * 指定URLに対して、Seabass32暗号化でアクセスします.
     * <BR>
     * @param req リクエストオブジェクトを設定します.
     * @param sessionId セッションIDを設定します.
     * @param url 対象のURLを設定します.
     * @param param 対象のパラメータを設定します.
     * @param header 対象のヘッダを設定します.
     * @param userAgent 対象のユーザエージェントを設定します.
     * @param oKey オリジナル暗号キーを設定します.
     * @return HttpResult 結果情報が返されます.
     * @exception Exception 例外.
     */
    public static final HttpResult seabass32( HttpdRequest req,
        String url,Map<String,Object> param,HttpHeader header,String userAgent,String oKey )
        throws Exception {
        url = trimUrl( url ) ;
        return seabass32( req.getRandom(),req.getSession().getSessionId(),url,param,header,userAgent,oKey ) ;
    }
    
    /**
     * Seabass32アクセス.
     * <BR><BR>
     * 指定URLに対して、Seabass32暗号化でアクセスします.
     * <BR>
     * @param rand ランダムオブジェクトを設定します.
     * @param sessionId セッションIDを設定します.
     * @param url 対象のURLを設定します.
     * @param param 対象のパラメータを設定します.
     * @param header 対象のヘッダを設定します.
     * @param userAgent 対象のユーザエージェントを設定します.
     * @return HttpResult 結果情報が返されます.
     * @exception Exception 例外.
     */
    public static final HttpResult seabass32( BaseRandom rand,String sessionId,
        String url,Map<String,Object> param,HttpHeader header,String userAgent )
        throws Exception {
        url = trimUrl( url ) ;
        return seabass32( rand,sessionId,url,param,header,userAgent,null ) ;
    }
    
    /**
     * Seabass32アクセス.
     * <BR><BR>
     * 指定URLに対して、Seabass32暗号化でアクセスします.
     * <BR>
     * @param rand ランダムオブジェクトを設定します.
     * @param sessionId セッションIDを設定します.
     * @param url 対象のURLを設定します.
     * @param param 対象のパラメータを設定します.
     * @param header 対象のヘッダを設定します.
     * @param userAgent 対象のユーザエージェントを設定します.
     * @param oKey オリジナル暗号キーを設定します.
     * @return HttpResult 結果情報が返されます.
     * @exception Exception 例外.
     */
    public static final HttpResult seabass32( BaseRandom rand,String sessionId,
        String url,Map<String,Object> param,HttpHeader header,String userAgent,String oKey )
        throws Exception {
        if( rand == null || url == null || ( url = url.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        url = trimUrl( url ) ;
        if( userAgent == null || ( userAgent = userAgent.trim() ).length() <= 0 ) {
            userAgent = DEFAULT_USER_AGENT ;
        }
        if( sessionId == null || sessionId.length() <= 0 ) {
            sessionId = RandomId.get( PersistenceDefine.SESSION_KEY_LENGTH,rand ) ;
        }
        Seabass32Key seabass32Key = ( Seabass32Key )GlobalManager.getInstance().get( ServiceDef.MANAGER_BY_SEABASS32 ) ;
        String key = seabass32Key.getSeabass32() ;
        boolean rand32 = seabass32Key.isRandom() ;
        // ランダムキーの場合は、keyはoKeyを利用.
        if( rand32 ) {
            boolean defokey = false ;
            if( oKey == null || oKey.length() <= 0 ) {
                if( (oKey = seabass32Key.getDefaultKey()) == null ) {
                    oKey = seabass32Key.createDefaultKey() ;
                }
                defokey = true ;
            }
            int h = oKey.length() / 2 ;
            key = Digest.convert( "MD5",( oKey.substring( h ) + oKey.substring( 0,h ) ).getBytes( "UTF8" ) ) ;
            if( defokey ) {
                oKey = "_" ;
            }
        }
        String s = RND64 ;
        StringBuilder buf = new StringBuilder() ;
        buf.append( sessionId ) ;
        for( int i = 0 ; i < 32 ; i ++ ) {
            buf.append( s.charAt( rand.nextInt() & 0x0000003f ) ) ;
        }
        String sid = Digest.convert( "MD5",buf.toString().getBytes( "UTF8" ) ) ;
        buf = null ;
        s = null ;
        // URL部分のみ取得.
        String burl = url ;
        int p = burl.indexOf( "?" ) ;
        if( p >= 0 ) {
            burl = burl.substring( 0,p ) ;
        }
        int pLen = 0 ;
        p = burl.indexOf( "http://" ) ;
        pLen = 7 ;
        if( p <= -1 ) {
            p = burl.indexOf( "https://" ) ;
            pLen = 8 ;
        }
        if( p >= 0 ) {
            p = burl.indexOf( "/",p+pLen ) ;
            if( p >= 0 ) {
                burl = burl.substring( p ) ;
            }
        }
        String secKey = Digest.convert( "MD5",burl.getBytes( "UTF8" ) ) + sid ;
        burl = null ;
        if( param != null && param.size() > 0 ) {
            buf = new StringBuilder() ;
            Iterator it = param.keySet().iterator() ;
            boolean nextFlag = false ;
            Object o ;
            while( it.hasNext() ) {
                s = ( String )it.next() ;
                if( nextFlag ) {
                    buf.append( "&" ) ;
                }
                o = param.get( s ) ;
                if( o instanceof String ) {
                    buf.append( s ).append( "=" ).append( URLEncoder.encode( (String)o,"UTF8" ) ) ;
                }
                else {
                    buf.append( s ).append( "=" ).append( URLEncoder.encode( o.toString(),"UTF8" ) ) ;
                }
                nextFlag = true ;
            }
            o = null ;
            it = null ;
            String sendParams = buf.toString() ;
            buf = null ;
            sendParams = Seabass32.encToString( sendParams,key,secKey ) ;
            param.clear() ;
            param.put( HttpdDef.SEABASS32_PARAM_NAME,sendParams ) ;
            param.put( HttpdDef.SEABASS32_SID,sid ) ;
            // ランダムキーの場合は、セッティング用キーを付加.
            if( rand32 ) {
                param.put( HttpdDef.SEABASS32_KEY,oKey ) ;
            }
        }
        else {
            if( param == null ) {
                param = new HashMap<String,Object>() ;
            }
            param.put( HttpdDef.SEABASS32_PARAM_NAME,"" ) ;
            param.put( HttpdDef.SEABASS32_SID,sid ) ;
            // ランダムキーの場合は、セッティング用キーを付加.
            if( rand32 ) {
                param.put( HttpdDef.SEABASS32_KEY,oKey ) ;
            }
        }
        HttpResult ret = HttpClientAccess.connection( true,url,null,header,param,userAgent ) ;
        ret.setBinary( Seabass32.dec( new String( ret.getBinary(),"UTF8" ),key,secKey ) ) ;
        return ret ;
    }
}
