package org.maachang.seabass.service.mail;

import java.io.IOException;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.util.atomic.AtomicOBJECT;
import org.maachang.util.thread.LoopThread;

/**
 * Queueに格納してメール送信.
 * 
 * @version 2009/08/10
 * @author masahito suzuki
 * @since SeaBass 1.0.0
 */
public class SendMailQueue {
    
    /** メール送信用スレッド. **/
    private static class LoopSendMailThread extends LoopThread {
        private static final Log LOG = LogFactory.getLog( LoopSendMailThread.class ) ;
        private static final int WAIT = 500 ;
        BlockingQueue<SendMail> mailQueue = null ;
        private LoopSendMailThread() {}
        protected LoopSendMailThread( BlockingQueue<SendMail> queue ) throws Exception {
            this.mailQueue = queue ;
            this.startThread() ;
        }
        /**
         * デストラクタ.
         */
        protected void finalize() throws Exception {
            this.destroy() ;
        }
        
        /**
         * オブジェクト破棄.
         */
        public void destroy() {
            super.stopThread() ;
        }
        
        /**
         * 実行処理.
         * @return boolean [false]の場合、一定期間待機します.
         * @exception Exception 例外.
         */
        protected boolean execution() throws Exception {
            SendMail mail ;
            if( mailQueue.size() > 0 ) {
                while( true ) {
                    Thread.sleep( WAIT ) ;
                    if( ( mail = mailQueue.poll() ) != null ) {
                        try {
                            SendMailService.sendMail( mail ) ;
                        } catch( Exception e ) {
                            LOG.error( "## sendMailError",e ) ;
                        }
                    }
                }
            }
            else {
                Thread.sleep( WAIT ) ;
            }
            return false ;
        }
        
        /** 情報クリア. **/
        protected void clear() {
            mailQueue = null ;
        }
    }
    
    /**
     * メール送信用データ管理キュー.
     */
    private final AtomicOBJECT<BlockingQueue<SendMail>> queue = new AtomicOBJECT<BlockingQueue<SendMail>>() ;
    
    /** メール送信用スレッド. **/
    private final AtomicOBJECT<LoopSendMailThread> thread = new AtomicOBJECT<LoopSendMailThread>() ;
    
    /**
     * コンストラクタ.
     * @exception Exception 例外.
     */
    public SendMailQueue() throws Exception {
        this( -1 ) ;
    }
    
    /**
     * コンストラクタ.
     * @param max キュー格納最大数を設定します.
     * @exception Exception 例外.
     */
    public SendMailQueue( int max ) throws Exception {
        if( max <= 0 ) {
            queue.set( new LinkedBlockingQueue<SendMail>() ) ;
        }
        else {
            queue.set( new LinkedBlockingQueue<SendMail>( max ) ) ;
        }
        thread.set( new LoopSendMailThread( queue.get() ) ) ;
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        destroy() ;
    }
    
    /**
     * オブジェクト破棄.
     */
    public void destroy() {
        LoopSendMailThread th = thread.get() ;
        thread.set( null ) ;
        if( th != null ) {
            th.destroy() ;
        }
        queue.set( null ) ;
    }
    
    /**
     * メール送信処理.
     * @param mailBean 送信情報を設定します.
     * @exception Exception 例外.
     */
    public void sendMail( SendMail mailBean ) throws Exception {
        BlockingQueue<SendMail> q = queue.get() ;
        if( q == null ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( mailBean == null ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        q.put( mailBean ) ;
    }
    
    /**
     * キューに格納されているメール情報数を取得.
     * @return int キューに格納されているサイズが返されます.
     * @exception Exception 例外.
     */
    public int size() throws Exception {
        BlockingQueue<SendMail> q = queue.get() ;
        if( q == null ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        return q.size() ;
    }
    
    /**
     * オブジェクトが有効かチェック.
     * @return boolean [true]の場合、有効です.
     */
    public boolean isUse() {
        return queue.get() != null ;
    }
    
    
    
    /**
     * オブジェクトを取得.
     * @return SendMailQueue オブジェクトが返されます.
     * @exception Exception 例外.
     */
    public static final SendMailQueue getInstance() throws Exception {
        if( SNGL == null ) {
            throw new IOException( "送信メールキューは利用できません" ) ;
        }
        return SNGL ;
    }
    
    /**
     * シングルトン初期化処理.
     * @exception Exception 例外.
     */
    public static final void initSingle() throws Exception {
        SNGL = new SendMailQueue() ;
    }
    
    /**
     * シングルトン初期化処理.
     * @param max キュー格納最大数を設定します.
     * @exception Exception 例外.
     */
    public static final void initSingle( int max ) throws Exception {
        SNGL = new SendMailQueue( max ) ;
    }
    
    /**
     * シングルトンオブジェクト破棄.
     */
    public static final void destroySingle() {
        if( SNGL != null ) {
            SNGL.destroy() ;
            SNGL = null ;
        }
    }
    
    /** シングルトン **/
    private static SendMailQueue SNGL = null ;
}

