package org.maachang.seabass.session.engine ;

import java.io.IOException;

import org.maachang.bdb.Bdb;
import org.maachang.bdb.BdbKeyValue;
import org.maachang.bdb.BdbManager;
import org.maachang.util.BaseRandom;

/**
 * 永続化セッション管理オブジェクト.
 * 
 * @version 2008/05/25
 * @author masahito suzuki
 * @since PersistenceSession 1.00
 */
class PersistenceEngineImpl implements PersistenceEngine {
    
    /**
     * 生成待機カウント.
     */
    private static final int WAIT_LOOP = 25 ;
    
    /**
     * 生成待機時間.
     */
    private static final int WAIT_TIME = 100 ;
    
    /**
     * セッションキャッシュ管理情報.
     */
    private SessionCache sessionCache = null ;
    
    /**
     * セッションID生成管理情報.
     */
    private CreateSession createSession = null ;
    
    /**
     * コンストラクタ.
     */
    private PersistenceEngineImpl() {
        
    }
    
    /**
     * コンストラクタ.
     * @param manager Bdbマネージャを設定します.
     * @param cacheTimeout キャッシュタイムアウト値を設定します.
     * @param sessionTimeout セッションタイムアウト値を設定します.
     * @exception Exception 例外.
     */
    public PersistenceEngineImpl( BdbManager manager,int cacheTimeout,int sessionTimeout )
        throws Exception {
        Bdb bdb = manager.get( "_sessions_" ) ;
        SessionCache sessionCache = new SessionCache( bdb,cacheTimeout,sessionTimeout ) ;
        if( sessionCache.isUse() == false ) {
            boolean flg = false ;
            for( int i = 0 ; i < WAIT_LOOP ; i ++ ) {
                if( sessionCache.isUse() == true ) {
                    flg = true ;
                    break ;
                }
                try { Thread.sleep( WAIT_TIME ) ; } catch( Exception e ) {}
            }
            if( flg == false ) {
                throw new IOException( "生成に失敗しました" ) ;
            }
        }
        CreateSession createSession = new CreateSession( sessionCache ) ;
        this.sessionCache = sessionCache ;
        this.createSession = createSession ;
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        destroy() ;
    }
    
    /**
     * オブジェクトを破棄します.
     */
    public void destroy() {
        if( createSession != null ) {
            createSession.destroy() ;
        }
        createSession = null ;
        if( sessionCache != null ) {
            sessionCache.destroy() ;
        }
        sessionCache = null ;
    }
    
    /**
     * 新しいセッションキーを生成.
     * @param keyValue スレッドで管理しているKeyValueオブジェクトを設定します.
     * @param baseRandom ランダムオブジェクトを設定します.
     * @return String 新しいSessionKeyが返されます.
     * @exception Exception 例外.
     */
    public String createSessionId( BdbKeyValue keyValue,BaseRandom baseRandom )
        throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは無効な状態です" ) ;
        }
        return createSession.createSessionKey( keyValue,baseRandom ) ;
    }
    
    /**
     * セッションに１つの要素を設定.
     * @param keyValue スレッドで管理しているKeyValueオブジェクトを設定します.
     * @param sessionId 対象のセッションIDを設定します.
     * @param key 対象のKeyを設定します.
     * @param value 対象の要素を設定します.
     * @exception Exception 例外.
     */
    public void put( BdbKeyValue keyValue,String sessionId,String key,byte[] value )
        throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは無効な状態です" ) ;
        }
        sessionCache.put( keyValue,sessionId,key,value ) ;
    }
    
    /**
     * セッションを削除.
     * @param keyValue スレッドで管理しているKeyValueオブジェクトを設定します.
     * @param sessionId 対象のセッションIDを設定します.
     * @exception Exception 例外.
     */
    public void remove( BdbKeyValue keyValue,String sessionId ) throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは無効な状態です" ) ;
        }
        sessionCache.remove( keyValue,sessionId ) ;
    }
    
    /**
     * セッションから１つの要素を削除.
     * @param keyValue スレッドで管理しているKeyValueオブジェクトを設定します.
     * @param sessionId 対象のセッションIDを設定します.
     * @param key 対象のKeyを設定します.
     * @exception Exception 例外.
     */
    public void remove( BdbKeyValue keyValue,String sessionId,String key )
        throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは無効な状態です" ) ;
        }
        sessionCache.remove( keyValue,sessionId,key ) ;
    }
    
    /**
     * セッションから１つの要素を取得.
     * @param keyValue スレッドで管理しているKeyValueオブジェクトを設定します.
     * @param sessionId 対象のセッションIDを設定します.
     * @param key 対象のKeyを設定します.
     * @return byte[] １つの要素が返されます.
     * @exception Exception 例外.
     */
    public byte[] get( BdbKeyValue keyValue,String sessionId,String key )
        throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは無効な状態です" ) ;
        }
        return sessionCache.get( keyValue,sessionId,key ) ;
    }
    
    /**
     * 指定セッション更新時間を取得.
     * @param keyValue スレッドで管理しているKeyValueオブジェクトを設定します.
     * @param sessionId 対象のセッションIDを設定します.
     * @return long 対象の更新時間が返されます.
     * @exception Exception 例外.
     */
    public long getLastUpdateTime( BdbKeyValue keyValue,String sessionId ) throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは無効な状態です" ) ;
        }
        return sessionCache.getLastUpdateTime( keyValue,sessionId ) ;
    }
    
    /**
     * 指定セッション情報が存在するかチェック.
     * @param keyValue スレッドで管理しているKeyValueオブジェクトを設定します.
     * @param sessionId 対象のセッションIDを設定します.
     * @return boolean [true]の場合、セッションIDは存在します.
     */
    public boolean containsKey( BdbKeyValue keyValue,String sessionId ) {
        if( isUse() == false ) {
            return false ;
        }
        boolean ret = false ;
        if( ( ret = sessionCache.containsKey( keyValue,sessionId ) ) == false ) {
            return createSession.containsKey( sessionId ) ;
        }
        return ret ;
    }
    
    /**
     * Bdbオブジェクトを取得.
     * @return Bdb Bdbオブジェクトが返されます.
     */
    public Bdb getBdb() {
        return sessionCache.io() ;
    }
    
    /**
     * セッション削除時間を取得.
     * @return long セッション削除時間が返されます.
     */
    public long getDeleteSessionTime() {
        if( isUse() == false ) {
            return -1L ;
        }
        return sessionCache.getDeleteSessionTime() ;
    }
    
    /**
     * このオブジェクトが有効かチェック.
     * @return boolean [true]が返された場合、オブジェクトは有効です.
     */
    public boolean isUse() {
        return ( sessionCache == null || createSession == null || createSession.isUse() == false ) ?
            false : true ;
    }
}

