package org.maachang.seabass.session.engine ;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.maachang.bdb.Bdb;
import org.maachang.bdb.BdbKeyValue;
import org.maachang.util.ConvertParam;
import org.maachang.util.thread.LoopThread;

/**
 * セッションキャッシュ管理.
 *  
 * @version 2008/05/25
 * @author  masahito suzuki
 * @since  PersistenceSession 1.00
 */
class SessionCache {
    
    /**
     * ログオブジェクト
     */
    protected static final Log LOG = LogFactory.getLog( SessionCache.class ) ;
    
    /**
     * 基本キー名.
     */
    protected static final String BASE_SESSION_KEY = "@session@base:" ;
    
    /**
     * データ格納要素名.
     */
    protected static final String VALUE_SESSION_KEY = "@session@value:" ;
    
    /**
     * キャッシュ削除時間 : デフォルト値.
     * 5分.
     */
    private static final long DEFAULT_DELETE_CACHE = 300000L ;
    
    /**
     * キャッシュ削除時間 : 最小値.
     * 1分.
     */
    private static final long MIN_DELETE_CACHE = 60000L ;
    
    /**
     * キャッシュ削除時間 : 最大値.
     * 30分.
     */
    private static final long MAX_DELETE_CACHE = 1800000L ;
    
    /**
     * セッション削除時間 : デフォルト値.
     * 1日.
     */
    private static final long DEFAULT_DELETE_SESSION = 86400000L ;
    
    /**
     * セッション削除時間 : 最小値.
     * 30分.
     */
    private static final long MIN_DELETE_SESSION = 1800000L ;
    
    /**
     * セッション削除時間 : 最大値.
     * 1週間.
     */
    private static final long MAX_DELETE_SESSION = 604800000L ;
    
    /**
     * セッション管理用Bdb.
     */
    private Bdb io = null ;
    
    /**
     * キャッシュデータ格納オブジェクト.
     */
    private Map<String,CacheValue> cache = null ;
    
    /**
     * キャッシュタイムアウト.
     */
    private long cacheTimeout = -1L ;
    
    /**
     * セッションタイムアウト.
     */
    private long sessionTimeout = -1L ;
    
    /**
     * キャッシュ監視スレッド.
     */
    private CacheSessionMonThread cacheThread = null ;
    
    /**
     * セッション監視スレッド.
     */
    private BdbSessionMonThread sessionThread = null ;
    
    /**
     * コンストラクタ.
     */
    private SessionCache() {
        
    }
    
    /**
     * コンストラクタ.
     * @param io セッション管理用Bdbを設定します.
     * @param cacheTimeout キャッシュタイムアウト値を設定します.
     * @param sessionTimeout セッションタイムアウト値を設定します.
     * @exception Exception 例外.
     */
    public SessionCache( Bdb io,long cacheTimeout,long sessionTimeout )
        throws Exception {
        if( cacheTimeout <= -1 ) {
            cacheTimeout = DEFAULT_DELETE_CACHE ;
        }
        else if( cacheTimeout <= MIN_DELETE_CACHE ) {
            cacheTimeout = MIN_DELETE_CACHE ;
        }
        else if( cacheTimeout >= MAX_DELETE_CACHE ) {
            cacheTimeout = MAX_DELETE_CACHE ;
        }
        if( sessionTimeout <= -1 ) {
            sessionTimeout = DEFAULT_DELETE_SESSION ;
        }
        else if( sessionTimeout <= MIN_DELETE_SESSION ) {
            sessionTimeout = MIN_DELETE_SESSION ;
        }
        else if( sessionTimeout >= MAX_DELETE_SESSION ) {
            sessionTimeout = MAX_DELETE_SESSION ;
        }
        this.io = io ;
        this.cacheTimeout = cacheTimeout ;
        this.sessionTimeout = sessionTimeout ;
        this.cache = new ConcurrentHashMap<String,CacheValue>() ;
        this.cacheThread = new CacheSessionMonThread( this.cache,this.cacheTimeout ) ;
        this.sessionThread = new BdbSessionMonThread( this.io,this.cache,this.sessionTimeout ) ;
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        destroy() ;
    }
    
    /**
     * オブジェクトを破棄.
     */
    public void destroy() {
        if( sessionThread != null ) {
            sessionThread.destroy() ;
        }
        sessionThread = null ;
        if( cacheThread != null ) {
            cacheThread.stopThread() ;
        }
        cacheThread = null ;
        io = null ;
        if( cache != null ) {
            cache.clear() ;
        }
        cache = null ;
        cacheTimeout = -1L ;
        sessionTimeout = -1L ;
    }
    
    /**
     * セッションに１つの要素を設定.
     * @param keyValue スレッドで管理しているKeyValueオブジェクトを設定します.
     * @param sessionId 対象のセッションIDを設定します.
     * @param key 対象のKeyを設定します.
     * @param value 対象の要素を設定します.
     * @exception Exception 例外.
     */
    public void put( BdbKeyValue keyValue,String sessionId,String key,byte[] value )
        throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( sessionId == null ||
            ( sessionId = sessionId.trim() ).length() != PersistenceDefine.SESSION_KEY_LENGTH ||
            key == null || ( key = key.trim() ).length() <= 0 ||
            value == null || value.length <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        sessionId = SessionCache.toSessionId( sessionId ) ;
        // セッションIDでロック.
        String skey = convertSessionKey( sessionId ) ;
        String vKey = convertSessionValueKey( sessionId,key ) ;
        byte[] skeyBin = skey.getBytes( "UTF8" ) ;
        // 新しいセッションIDで登録する.
        keyValue.clear() ;
        if( io.get( keyValue.setKey( skeyBin ) ) == false ) {
            SessionChild ch = new SessionChild() ;
            ch.getChild().add( key.getBytes( "UTF8" ) ) ;
            byte[] b = ch.toBinary() ;
            ch = null ;
            io.put( keyValue.create( skeyBin,b ) ) ;
            io.put( keyValue.create( vKey.getBytes( "UTF8" ),value ) ) ;
        }
        // 既にセッションIDが存在する.
        else {
            // 指定セッションIDのキャッシュは削除.
            cache.remove( sessionId ) ;
            byte[] b = keyValue.getValue() ;
            SessionChild ch = new SessionChild( b ) ;
            b = null ;
            
            // 指定要素が新しい要素の場合.
            if( ch.containsKey( key ) == false ) {
                ch.update() ;
                ch.getChild().add( key.getBytes( "UTF8" ) ) ;
                b = ch.toBinary() ;
                io.put( keyValue.create( skeyBin,b ) ) ;
                b = null ;
            }
            // 新しい要素は既に存在する.
            else {
                // セッション時間の更新.
                setSessionTime( keyValue,io,sessionId ) ;
            }
            // 要素をセットする.
            io.put( keyValue.create( vKey.getBytes( "UTF8" ),value ) ) ;
        }
    }
    
    /**
     * セッションを削除.
     * @param keyValue スレッドで管理しているKeyValueオブジェクトを設定します.
     * @param sessionId 対象のセッションIDを設定します.
     * @exception Exception 例外.
     */
    public void remove( BdbKeyValue keyValue,String sessionId ) throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( sessionId == null ||
            ( sessionId = sessionId.trim() ).length() != PersistenceDefine.SESSION_KEY_LENGTH ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        sessionId = SessionCache.toSessionId( sessionId ) ;
        // セッションIDでロック.
        removeSession( keyValue,io,cache,sessionId ) ;
    }
    
    /**
     * セッションから１つの要素を削除.
     * @param keyValue スレッドで管理しているKeyValueオブジェクトを設定します.
     * @param sessionId 対象のセッションIDを設定します.
     * @param key 対象のKeyを設定します.
     * @exception Exception 例外.
     */
    public void remove( BdbKeyValue keyValue,String sessionId,String key )
        throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( sessionId == null ||
            ( sessionId = sessionId.trim() ).length() != PersistenceDefine.SESSION_KEY_LENGTH ||
            key == null || ( key = key.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        sessionId = SessionCache.toSessionId( sessionId ) ;
        // セッションIDでロック.
        cache.remove( sessionId ) ;
        setSessionTime( keyValue,io,sessionId ) ;
        String vKey = convertSessionValueKey( sessionId,key ) ;
        io.remove( keyValue.setKey( vKey.getBytes( "UTF8" ) ) ) ;
    }
    
    /**
     * セッションから１つの要素を取得.
     * @param keyValue スレッドで管理しているKeyValueオブジェクトを設定します.
     * @param sessionId 対象のセッションIDを設定します.
     * @param key 対象のKeyを設定します.
     * @return byte[] １つの要素が返されます.
     * @exception Exception 例外.
     */
    public byte[] get( BdbKeyValue keyValue,String sessionId,String key )
        throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( sessionId == null ||
            ( sessionId = sessionId.trim() ).length() != PersistenceDefine.SESSION_KEY_LENGTH ||
            key == null || ( key = key.trim() ).length() <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        sessionId = SessionCache.toSessionId( sessionId ) ;
        byte[] ret = null ;
        boolean cacheSessionFlag = cache.containsKey( sessionId ) ;
        // セッション名に対してキャッシュ情報が存在する場合.
        if( cacheSessionFlag == true ) {
            CacheValue cacheVal = cache.get( sessionId ) ;
            // キャッシュから、Keyの位置を取得.
            int no = cacheVal.search( key ) ;
            if( no >= 0 ) {
                // キャッシュにKeyが存在する場合.
                setSessionTime( keyValue,io,sessionId ) ;
                ret = cacheVal.get( no ) ;
            }
        }
        // キャッシュから情報を取得できない場合.
        if( ret == null ) {
            // Bdbから情報を取得.
            String vKey = convertSessionValueKey( sessionId,key ) ;
            if( io.get( keyValue.setKey( vKey.getBytes( "UTF8" ) ) ) ) {
                // BdbにKey情報が存在する場合は、時間更新して、情報を取得し
                // その内容をCache化する.
                ret = keyValue.getValue() ;
                setSessionTime( keyValue,io,sessionId ) ;
                if( ret != null ) {
                    // キャッシュにSessionID情報が存在しない場合は
                    // 生成して、格納する.
                    CacheValue cacheVal = null ;
                    if( cacheSessionFlag == false ) {
                        cacheVal = new CacheValue() ;
                        cache.put( sessionId,cacheVal ) ;
                    }
                    else {
                        cacheVal = cache.get( sessionId ) ;
                    }
                    cacheVal.put( key,ret ) ;
                }
            }
        }
        return ret ;
    }
    
    /**
     * 指定セッション情報が存在するかチェック.
     * @param keyValue スレッドで管理しているKeyValueオブジェクトを設定します.
     * @param sessionId 対象のセッションIDを設定します.
     * @return boolean [true]の場合、セッションIDは存在します.
     */
    public boolean containsKey( BdbKeyValue keyValue,String sessionId ) {
        if( isUse() == false ) {
            return false ;
        }
        if( sessionId == null ||
            ( sessionId = sessionId.trim() ).length() != PersistenceDefine.SESSION_KEY_LENGTH ) {
            return false ;
        }
        sessionId = SessionCache.toSessionId( sessionId ) ;
        try {
            return io.containsKey( keyValue.setKey( convertSessionKey( sessionId ).getBytes( "UTF8" ) ) ) ;
        } catch( Exception e ) {
        }
        return false ;
    }
    
    /**
     * 指定セッション更新時間を取得.
     * @param keyValue スレッドで管理しているKeyValueオブジェクトを設定します.
     * @param sessionId 対象のセッションIDを設定します.
     * @return long 対象の更新時間が返されます.
     * @exception Exception 例外.
     */
    public long getLastUpdateTime( BdbKeyValue keyValue,String sessionId ) throws Exception {
        if( isUse() == false ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( sessionId == null ||
            ( sessionId = sessionId.trim() ).length() != PersistenceDefine.SESSION_KEY_LENGTH ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        long ret = -1L ;
        sessionId = SessionCache.toSessionId( sessionId ) ;
        ret = getSessionTime( keyValue,io,sessionId ) ;
        return ret ;
    }
    
    /**
     * Bdbオブジェクトを取得.
     * @return Bdb Bdbオブジェクトが返されます.
     */
    public Bdb io() {
        return ( isUse() ) ? io : null ;
    }
    
    /**
     * セッション削除時間を取得.
     * @return int セッション削除時間が返されます.
     */
    public long getDeleteSessionTime() {
        return sessionTimeout ;
    }
    
    /**
     * このオブジェクトが有効かチェック.
     * @return boolean [true]が返された場合、オブジェクトは有効です.
     */
    public boolean isUse() {
        return ( io == null || io.isClose() ||
            cache == null || 
            cacheThread == null || cacheThread.isStop() == true ||
            sessionThread == null || sessionThread.isStop() == true ) ?
            false : true ;
    }
    
    /**
     * 指定Keyに対するセッションタイマーを取得.
     */
    protected static final long getSessionTime( BdbKeyValue keyValue,Bdb io,String key ) throws Exception {
        if( key == null || ( key = key.trim() ).length() <= 0 ) {
            return -1L ;
        }
        if( key.startsWith( BASE_SESSION_KEY ) == false ) {
            key = convertSessionKey( key ) ;
        }
        if( io.get( keyValue.setKey( key.getBytes( "UTF8" ) ),0,8 ) ) {
            byte[] val = keyValue.getValue() ;
            if( val == null || val.length == 8 ) {
                return ConvertParam.convertLong( 0,val ) ;
            }
        }
        return -1L ;
    }
    
    /**
     * 指定Keyに対するセッションタイマーを更新.
     */
    protected static final void setSessionTime( BdbKeyValue keyValue,Bdb io,String key ) throws Exception {
        if( key == null || ( key = key.trim() ).length() <= 0 ) {
            return ;
        }
        if( key.startsWith( BASE_SESSION_KEY ) == false ) {
            key = convertSessionKey( key ) ;
        }
        if( io.containsKey( keyValue.setKey( key.getBytes( "UTF8" ) ) ) ) {
            io.put( keyValue.setValue( ConvertParam.convertLong( System.currentTimeMillis() ) ),0,8 ) ;
        }
    }
    
    /**
     * 指定Keyを削除.
     */
    protected static final void removeSession( BdbKeyValue keyValue,Bdb io,Map<String,CacheValue> cache,String key )
        throws Exception {
        if( key == null || ( key = key.trim() ).length() <= 0 ) {
            return ;
        }
        if( key.startsWith( BASE_SESSION_KEY ) ) {
            key = toSessionId( key ) ;
        }
        if( cache.containsKey( key ) ) {
            cache.remove( key ) ;
        }
        String baseKey = convertSessionKey( key ) ;
        byte[] baseKeyBin = baseKey.getBytes( "UTF8" ) ;
        if( io.get( keyValue.setKey( baseKeyBin ) ) ) {
            SessionChild ch = new SessionChild( keyValue.getValue() ) ;
            keyValue.setValue( null ) ;
            ArrayList<byte[]> lst = ch.getChild() ;
            int len = lst.size() ;
            String vKey ;
            for( int i = 0 ; i < len ; i ++ ) {
                vKey = convertSessionValueKey( key,new String( lst.get( i ),"UTF8" ) ) ;
                io.remove( keyValue.create( vKey.getBytes( "UTF8" ),null ) ) ;
            }
            io.remove( keyValue.setKey( baseKeyBin ) ) ;
        }
    }

    
    /**
     * セッションキーに変換.
     */
    protected static final String convertSessionKey( String name ) {
        return BASE_SESSION_KEY+name ;
    }
    
    /**
     * MDBM格納のセッションKeyから、セッションIDを取得.
     */
    protected static final String toSessionId( String name ) {
        if( name.startsWith( BASE_SESSION_KEY ) ) {
            return name.substring( BASE_SESSION_KEY.length() ) ;
        }
        return name ;
    }
    
    /**
     * セッション要素キーに変換.
     */
    protected static final String convertSessionValueKey( String sessionId,String key ) {
        return new StringBuilder().
            append( VALUE_SESSION_KEY ).
            append( sessionId ).append( "." ).
            append( key ).toString() ;
    }
    
    /**
     * キャッシュ監視スレッド.
     */
    private static class CacheSessionMonThread extends LoopThread {
        private long timeout = -1L ;
        private Map<String,CacheValue> map = null ;
        
        private CacheSessionMonThread() {
            
        }
        
        public CacheSessionMonThread( Map<String,CacheValue> map,long timeout )
            throws Exception {
            this.map = map ;
            this.timeout = timeout ;
            startThread() ;
        }
        
        protected void clear() {
            this.map = null ;
        }
        
        protected boolean execution() throws Exception {
            if( map.size() <= 0 ) {
                Thread.sleep( 1000 ) ;
            }
            else {
                for( Iterator<String> keys = map.keySet().iterator() ; keys.hasNext() ; ) {
                    Thread.sleep( 250 ) ;
                    String key = keys.next() ;
                    CacheValue val = map.get( key ) ;
                    if( val == null || val.size() <= 0 ||
                        val.getLastTime() + timeout <= System.currentTimeMillis() ) {
                        if( LOG.isDebugEnabled() ) {
                            LOG.debug( "## キャッシュセッション[" + key + "]を削除しました" ) ;
                        }
                        keys.remove() ;
                    }
                }
                Thread.sleep( 250 ) ;
            }
            return false ;
        }
        
    }
    
    /**
     * Session監視スレッド.
     */
    private static class BdbSessionMonThread extends LoopThread {
        private byte[] baseKeyHeader = null ;
        private long timeout = -1L ;
        private Bdb io = null ;
        private Map<String,CacheValue> cache = null ;
        private final BdbKeyValue keyValue = new BdbKeyValue() ;
        
        private BdbSessionMonThread() {
        }
        
        public BdbSessionMonThread( Bdb io,Map<String,CacheValue> cache,long timeout )
            throws Exception {
            this.io = io ;
            this.cache = cache ;
            this.timeout = timeout ;
            this.baseKeyHeader = BASE_SESSION_KEY.getBytes( "UTF8" ) ;
            startThread() ;
        }
        
        public void destroy() {
            stopThread() ;
        }
        
        protected void clear() {
            this.io = null ;
            this.cache = null ;
        }
        
        protected void toException( Exception e ) {
            LOG.warn( "## sessionMonitor-error",e ) ;
        }
        
        private final boolean toHead( byte[] value,byte[] chk ) {
            int len = chk.length ;
            if( len >= value.length ) {
                return false ;
            }
            for( int i = 0 ; i < len ; i ++ ) {
                if( chk[ i ] != value[ i ] ) {
                    return false ;
                }
            }
            return true ;
        }
        
        protected boolean execution() throws Exception {
            byte[] key ;
            String sessionId ;
            long time ;
            Thread.sleep( WAIT ) ;
            if( io.size() > 0 ) {
                Iterator<byte[]> keys = io.keys( keyValue ) ;
                while( keys.hasNext() ) {
                    Thread.sleep( WAIT ) ;
                    if( isStop() ) {
                        break ;
                    }
                    key = keys.next() ;
                    if( key != null && key.length > 0 && !io.isClose() && toHead( key,baseKeyHeader ) ) {
                        // 文字のセッションIDを取得.
                        sessionId = SessionCache.toSessionId( new String( key,"UTF8" ) ) ;
                        // セッションタイムアウトチェック.
                        if( io.get( keyValue,0,8 ) ) {
                            // 永続化セッションタイムアウトをチェック.
                            time = ConvertParam.convertLong( 0,keyValue.getValue() ) ;
                            if( time <= -1 || time + timeout <= System.currentTimeMillis() ) {
                                SessionCache.removeSession( keyValue,io,cache,sessionId ) ;
                                LOG.info( "## 永続化セッション["+ sessionId + "]を削除しました" ) ;
                            }
                        }
                    }
                }
            }
            return true ;
        }
        private static final long WAIT = 30L ;
    }
}

/**
 * SessionChild.
 */
class SessionChild {
    private long lastMod = -1L ;
    private ArrayList<byte[]> child = null ;
    
    public SessionChild() {
        this.lastMod = System.currentTimeMillis() ;
        this.child = new ArrayList<byte[]>() ;
    }
    
    public SessionChild( byte[] bin ) throws Exception {
        int p = 0 ;
        long time = ConvertParam.convertLong( p,bin ) ;
        p += 8 ;
        int len = ConvertParam.convertInt( p,bin ) ;
        p += 4 ;
        ArrayList<byte[]> lst = new ArrayList<byte[]>() ;
        for( int i = 0 ; i < len ; i ++ ) {
            int strLen = ConvertParam.convertInt( p,bin ) ;
            p += 4 ;
            if( strLen <= 0 ) {
                continue ;
            }
            byte[] b = new byte[ strLen ] ;
            System.arraycopy( bin,p,b,0,strLen ) ;
            p += strLen ;
            lst.add( b ) ;
        }
        this.lastMod = time ;
        this.child = lst ;
    }
    
    protected void finalize() {
        destroy() ;
    }
    
    public void destroy() {
        lastMod = -1L ;
        child = null ;
    }
    
    public long lastMod() {
        return lastMod ;
    }
    
    public void update() {
        lastMod = System.currentTimeMillis() ;
    }
    
    public boolean containsKey( String key ) {
        try {
            return containsKey( key.getBytes( "UTF8" ) ) ;
        } catch( Exception e ) {
        }
        return false ;
    }
    
    public boolean containsKey( byte[] key ) {
        return ( search( key ) >= 0 ) ;
    }
    
    public int search( String key ) {
        try {
            return search( key.getBytes( "UTF8" ) ) ;
        } catch( Exception e ) {
        }
        return -1 ;
    }
    
    public int search( byte[] key ) {
        int klen = key.length ;
        int len = child.size() ;
        for( int i = 0 ; i < len ; i ++ ) {
            byte[] x = child.get( i ) ;
            if( x.length == klen ) {
                boolean flg = true ;
                for( int j = 0 ; j < klen ; j ++ ) {
                    if( x[ j ] != key[ j ] ) {
                        flg = false ;
                        break ;
                    }
                }
                if( flg == true ) {
                    return i ;
                }
            }
        }
        return -1 ;
    }
    
    public ArrayList<byte[]> getChild() {
        return child ;
    }
    
    public byte[] toBinary() throws Exception {
        int len = child.size() ;
        int max = 8 + 4 ;
        for( int i = 0 ; i < len ; i ++ ) {
            byte[] b = child.get( i ) ;
            max += 4 + b.length ;
        }
        byte[] ret = new byte[ max ] ;
        int p = 0 ; 
        ConvertParam.convertLong( ret,p,lastMod ) ;
        p += 8 ;
        ConvertParam.convertInt( ret,p,len ) ;
        p += 4 ;
        for( int i = 0 ; i < len ; i ++ ) {
            byte[] b = child.get( i ) ;
            int strLen = b.length ;
            ConvertParam.convertInt( ret,p,strLen ) ;
            p += 4 ;
            System.arraycopy( b,0,ret,p,strLen ) ;
            p += strLen ;
        }
        return ret ;
    }
    
}

