package org.maachang.bdb ;

import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.locks.Lock;

/**
 * [Transaction]Berkeley DB wrapper実装.
 *  
 * @version 2010/09/27
 * @author  masahito suzuki
 * @since   bdb 1.0.5
 */
public class TransactionBdb {
    private static final int TYPE_PUT = 0 ;     // トランザクションPut.
    private static final int TYPE_REMOVE = 1 ;  // トランザクションRemove.
    
    private Bdb bdb = null ;
    private final Map<KeyBinary,TransElement> trans = new HashMap<KeyBinary,TransElement>() ;
    private BdbKeyValue keyValue = null ;
    private final KeyBinary keyBinary = new KeyBinary() ;
    
    private int _transLength = 0 ;
    
    /**
     * コンストラクタ.
     */
    public TransactionBdb() {
        
    }
    
    /**
     * コンストラクタ.
     * @param bdb bdbオブジェクトを設定します.
     * @exception Exception 例外.
     */
    public TransactionBdb( Bdb bdb ) throws Exception {
        open( bdb ) ;
    }
    
    /**
     * コンストラクタ.
     * @param keyValue BdbKeyValueオブジェクトを設定します.
     * @param bdb bdbオブジェクトを設定します.
     * @exception Exception 例外.
     */
    public TransactionBdb( BdbKeyValue keyValue,Bdb bdb ) throws Exception {
        open( keyValue,bdb ) ;
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        close() ;
    }
    
    /**
     * オブジェクト生成.
     * @param bdb Bdbオブジェクトを設定します.
     * @exception Exception 例外.
     */
    public void open( Bdb bdb ) throws Exception {
        if( bdb == null ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        if( !bdb.isClose() ) {
            close() ;
        }
        this.close() ;
        this.bdb = bdb ;
        this.keyValue = BdbManager.getBdbKeyValue() ;
    }
    
    /**
     * オブジェクト生成.
     * @param keyValue BdbKeyValueオブジェクトを設定します.
     * @param bdb Bdbオブジェクトを設定します.
     * @exception Exception 例外.
     */
    public void open( BdbKeyValue keyValue,Bdb bdb ) throws Exception {
        if( bdb == null ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        if( !bdb.isClose() ) {
            close() ;
        }
        if( keyValue == null ) {
            keyValue = BdbManager.getBdbKeyValue() ;
        }
        this.close() ;
        this.bdb = bdb ;
        this.keyValue = keyValue ;
    }
    
    /**
     * オブジェクトクローズ.
     * <p>クローズした場合は、ロールバックされます</p>
     */
    public void close() {
        trans.clear() ;
        bdb = null ;
        _transLength = 0 ;
    }
    
    /**
     * オブジェクトが有効かチェック.
     * @return boolean [true]の場合、有効です.
     */
    public boolean isClose() {
        return bdb == null || bdb.isClose() ;
    }
    
    /**
     * コミット処理.
     * @exception Exception 例外.
     */
    public void commit() throws Exception {
        if( isClose() ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        try {
            Iterator it = trans.keySet().iterator() ;
            while( it.hasNext() ) {
                TransElement em =  trans.get( ( KeyBinary )it.next() ) ;
                if( em != null ) {
                    switch( em.getType() ) {
                        case TYPE_PUT :
                            bdb.put( keyValue.create( em.getKey(),em.getValue() ) ) ;
                            keyValue.clear() ;
                            break ;
                        case TYPE_REMOVE :
                            bdb.remove( keyValue.create( em.getKey(),null ) ) ;
                            keyValue.clear() ;
                            break ;
                    }
                }
            }
        } finally {
            trans.clear() ;
            _transLength = 0 ;
        }
    }
    
    /**
     * ロールバック処理.
     * @exception Exception 例外.
     */
    public void rollback() throws Exception {
        if( isClose() ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        trans.clear() ;
        _transLength = 0 ;
    }
    
    /**
     * 情報をセット.
     * @param key 対象のキーを設定します.
     * @param value 対象の要素を設定します.
     * @exception Exception 例外.
     */
    public void put( byte[] key,byte[] value )
        throws Exception {
        if( isClose() ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( key == null || key.length <= 0 ||
            value == null || value.length <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        key = copyKey( key ) ;
        KeyBinary kb = new KeyBinary( key ) ;
        TransElement em = trans.get( kb ) ;
        boolean useSrc = bdb.containsKey( keyValue.create( key,null ) ) ;
        if( ( em == null && !useSrc ) || ( em != null && em.type == TYPE_REMOVE ) ) {
            _transLength = _transLength + 1 ;
        }
        if( em == null ) {
            trans.put( kb,new TransElement( TYPE_PUT,key,value ) ) ;
        }
        else {
            em.create( TYPE_PUT,key,value ) ;
        }
    }
    
    /**
     * 情報を削除.
     * @param key 対象のキーを設定します.
     * @return byte[] 対象の上方が返されます.
     * @exception Exception 例外.
     */
    public byte[] remove( byte[] key ) throws Exception {
        if( isClose() ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( key == null || key.length <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        byte[] ret = null ;
        boolean useSrc = bdb.containsKey( keyValue.create( key,null ) ) ;
        keyValue.clear() ;
        // トランザクション中のデータを取得.
        key = copyKey( key ) ;
        KeyBinary kb = new KeyBinary( key ) ;
        TransElement em = trans.get( kb ) ;
        if( em != null ) {
            switch( em.type ) {
                // REMOVE条件が既に存在する場合は、何もしない.
                case TYPE_REMOVE :
                    return null ;
                // PUT条件の場合は、そのBinaryを取得.
                case TYPE_PUT :
                    ret = em.getValue() ;
                    if( useSrc ) {
                        // 元条件が存在する場合は、トランザクションデータに削除条件をセット.
                        em.create( TYPE_REMOVE,key,null ) ;
                        _transLength = _transLength - 1 ;
                    }
                    else {
                        // 元条件が存在しない場合は、トランザクション中のデータを物理削除.
                        trans.remove( kb ) ;
                        _transLength = _transLength - 1 ;
                    }
                    return ret ;
            }
        }
        else if( useSrc ) {
            // 元条件が存在する場合は、トランザクションデータに削除条件をセット.
            trans.put( kb,new TransElement( TYPE_REMOVE,key,null ) ) ;
            _transLength = _transLength - 1 ;
            // 現在の元情報を取得.
            if( bdb.get( keyValue.create( key,null ) ) ) {
                ret = keyValue.getValue() ;
            }
            keyValue.clear() ;
        }
        return ret ;
    }
    
    /**
     * 情報を取得.
     * @param key 対象のキーを設定します.
     * @return byte[] 対象の要素が返されます.
     * @exception Exception 例外.
     */
    public byte[] get( byte[] key ) throws Exception {
        if( isClose() ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( key == null || key.length <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        byte[] ret = null ;
        TransElement em = trans.get( keyBinary.create( key ) ) ;
        if( em != null ) {
            if( em.type == TYPE_REMOVE ) {
                return null ;
            }
            ret = em.getValue() ;
        }
        else {
            keyValue.create( key,null ) ;
            if( bdb.get( keyValue ) ) {
                ret = keyValue.getValue() ;
                keyValue.clear() ;
            }
        }
        return ret ;
    }
    
    /**
     * 情報存在チェック.
     * @param key 対象のキーを設定します.
     * @return boolean [true]の場合、情報は存在します.
     * @exception Exception 例外.
     */
    public boolean containsKey(byte[] key) throws Exception {
        if( isClose() ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( key == null || key.length <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        TransElement em = trans.get( keyBinary.create( key ) ) ;
        if( em != null ) {
            return ( em.type != TYPE_REMOVE ) ;
        }
        boolean ret = bdb.containsKey( keyValue.create( key,null ) ) ;
        keyValue.clear() ;
        return ret ;
    }
    
    /**
     * サイズを取得.
     * @return long 現在格納されているサイズが返されます.
     * @exception Exception 例外.
     */
    public long size() throws Exception {
        if( isClose() ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        return bdb.size() + _transLength ;
    }
    
    /**
     * オープン名を取得.
     * @return String オープン名が返されます.
     * @exception Exception 例外.
     */
    public String getName() throws Exception {
        if( isClose() ) {
            throw new IOException( "オブジェクトは既にクローズしています" ) ;
        }
        return bdb.getName() ;
    }
    
    /**
     * 読み込み用同期オブジェクトを取得.
     * @return Lock 読み込み用同期オブジェクトが返されます.
     * @exception Exception 例外.
     */
    public Lock getRSync() throws Exception {
        if( isClose() ) {
            throw new IOException( "オブジェクトは既にクローズしています" ) ;
        }
        return bdb.getRSync() ;
    }
    
    /**
     * 書き込み用同期オブジェクトを取得.
     * @return Lock 書き込み用同期オブジェクトが返されます.
     * @exception Exception 例外.
     */
    public Lock getWSync() throws Exception {
        if( isClose() ) {
            throw new IOException( "オブジェクトは既にクローズしています" ) ;
        }
        return bdb.getWSync() ;
    }
    
    /** キーコピー **/
    private static final byte[] copyKey( byte[] k ) {
        int len = k.length ;
        byte[] ret = new byte[ len ] ;
        System.arraycopy( k,0,ret,0,len ) ;
        return ret ;
    }
}

/**
 * １つのトランザクション命令.
 */
class TransElement {
    
    int type ;
    private byte[] key ;
    private byte[] value ;
    
    public TransElement( int type,byte[] key,byte[] value ) {
        this.type = type ;
        this.key = key ;
        this.value = value ;
    }
    
    public TransElement create( int type,byte[] key,byte[] value ) {
        this.type = type ;
        this.key = key ;
        this.value = value ;
        return this ;
    }
    
    public int getType() {
        return type ;
    }
    
    public byte[] getKey() {
        return key ;
    }
    
    public byte[] getValue() {
        return value ;
    }
}

class KeyBinary {
    private static final long FNV_64_CODE = 0xCBF29CE484222325L;
    private int length ;
    private byte[] key ;
    private int hash ;
    public KeyBinary() {
        
    }
    
    public KeyBinary( byte[] key ) {
        create( key ) ;
    }
    
    public KeyBinary create( byte[] key ) {
        int len = key.length ;
        long h = FNV_64_CODE;
        for (int i = 0; i < len; i++) {
            h ^= key[i];
            h += (h << 1L) + (h << 4L) + (h << 5L) + (h << 7L)
                    + (h << 8L) + (h << 40L);
        }
        hash = (int)(h&0x00000000ffffffffL);
        length = len ;
        this.key = key ;
        return this ;
    }
    
    public byte[] getKey() {
        return key ;
    }
    
    public int hashCode() {
        return hash ;
    }
    
    public boolean equals( Object o ) {
        if( ( o instanceof KeyBinary ) == false ) {
            return false ;
        }
        KeyBinary k = (KeyBinary)o ;
        if( k.hash != hash && k.length != length ) {
            return false ;
        }
        byte[] b = k.key ;
        for( int i = 0 ; i < length ; i ++ ) {
            if( key[ i ] != b[ i ] ) {
                return false ;
            }
        }
        return true ;
    }
}
