package org.maachang.bdb ;

import java.io.IOException;
import java.util.Arrays;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.concurrent.locks.Lock;

/**
 * [Transaction]Berkeley DB Queue実装.
 *  
 * @version 2010/09/27
 * @author  masahito suzuki
 * @since   bdb 1.0.5
 */
public class TransactionQueue {
    private TransactionBdb bdb = null ;
    private long max = -1L ;
    private double warning = 0.0f ;
    private long warnLen ;
    private boolean transactionFlag = false ;
    private String name = null ;
    private Lock wsync = null ;
    protected final byte[] tmpBin = new byte[ 8 ] ;
    
    // _$firstPos$_.
    private static final byte[] FIRST_NAME = new byte[]{
        95,36,102,105,114,115,116,80,111,115,36,95 } ;
    // _$lastPos$_
    private static final byte[] LAST_NAME = new byte[]{
        95,36,108,97,115,116,80,111,115,36,95 } ;
    
    /**
     * コンストラクタ.
     * @param keyValue BdbKeyValueオブジェクトを設定します.
     * @param bdb 対象のBDBオブジェクトを設定します.
     * @exception Exception 例外.
     */
    public TransactionQueue( BdbKeyValue keyValue,Bdb bdb )
        throws Exception {
        this( keyValue,bdb,BQueueImpl.DEFAULT_MAX,BQueueImpl.DEFAULT_WARNING ) ;
    }
    
    /**
     * コンストラクタ.
     * @param keyValue BdbKeyValueオブジェクトを設定します.
     * @param bdb 対象のBDBオブジェクトを設定します.
     * @param max キュー格納最大数を設定します.
     * @param warning 警告値を設定します.<br>たとえば、75％の場合は、0.75dで設定します.
     * @exception Exception 例外.
     */
    public TransactionQueue( BdbKeyValue keyValue,Bdb bdb,long max,double warning )
        throws Exception {
        if( bdb == null || bdb.isClose() ||
            max < 128 || warning < 0.25d ) {
            if( bdb == null || bdb.isClose() ) {
                throw new IllegalArgumentException( "不正なbdbマネージャが設定されています" ) ;
            }
            if( max < 128L ) {
                throw new IllegalArgumentException( "最大格納数は、128以上を設定してください" ) ;
            }
            if( warning < 0.25d ) {
                throw new IllegalArgumentException( "警告値は、25%以上を設定してください" ) ;
            }
        }
        this.max = max ;
        this.warning = warning ;
        this.warnLen = (long)( (double)max * warning ) ;
        this.name = bdb.getName() ;
        this.wsync = bdb.getWSync() ;
        this.bdb = new TransactionBdb( keyValue,bdb ) ;
        this.wsync.lock() ;
        Arrays.fill( tmpBin,(byte)0 ) ;
        // データ初期化.
        try {
            int n = 0 ;
            if( !this.bdb.containsKey( FIRST_NAME ) ) {
                this.bdb.put( FIRST_NAME,tmpBin ) ;
                n ++ ;
            }
            if( !this.bdb.containsKey( LAST_NAME ) ) {
                this.bdb.put( LAST_NAME,tmpBin ) ;
                n ++ ;
            }
            if( n > 0 ) {
                this.bdb.commit() ;
            }
        } finally {
            this.wsync.unlock() ;
        }
    }
    
    /**
     * デストラクタ.
     */
    protected void finalize() throws Exception {
        close() ;
    }
    
    /**
     * オブジェクトクローズ.
     */
    public void close() {
        if( bdb != null ) {
            bdb.close() ;
        }
        bdb = null ;
        max = -1L ;
        warning = 0.0f ;
        warnLen = 0 ;
        name = null ;
        if( wsync != null && transactionFlag ) {
            wsync.unlock() ;
            transactionFlag = false ;
        }
        wsync = null ;
    }
    
    /**
     * オブジェクトが有効かチェック.
     * @return boolean [true]の場合、有効です.
     */
    public boolean isClose() {
        return bdb == null || bdb.isClose() ;
    }
    
    /**
     * トランザクション開始処理.
     * @exception Exception 例外.
     */
    public void startTransaction() throws Exception {
        if( isClose() ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( !transactionFlag ) {
            wsync.lock() ;
            transactionFlag = true ;
        }
    }
    
    /**
     * コミット処理.
     * @exception Exception 例外.
     */
    public void commit() throws Exception {
        if( isClose() ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( transactionFlag ) {
            bdb.commit() ;
            transactionFlag = false ;
            wsync.unlock() ;
        }
    }
    
    /**
     * ロールバック処理.
     * @exception Exception 例外.
     */
    public void rollback() throws Exception {
        if( isClose() ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        if( transactionFlag ) {
            bdb.rollback() ;
            transactionFlag = false ;
            wsync.unlock() ;
        }
    }
    
    /**
     * 情報追加.
     * @param value 対象の要素を設定します.
     * @exception Exception 例外.
     */
    public void offer( byte[] value )
        throws Exception {
        if( value == null || value.length <= 0 ) {
            throw new IllegalArgumentException( "引数は不正です" ) ;
        }
        startTransaction() ; // トランザクション開始でない場合は、開始処理.
        if( this.length() >= max ) {
            throw new IOException( bdb.getName() + ":キューは満杯です" ) ;
        }
        bdb.put( _getSequence(),value ) ;
    }
    
    /**
     * 情報取得.
     * <p>この処理を実行した場合、取得された内容は削除されます.</p>
     * @return byte[] 対象の要素が返されます.
     * @exception Exception 例外.
     */
    public byte[] poll() throws Exception {
        startTransaction() ; // トランザクション開始でない場合は、開始処理.
        long first = _getFirstPos() ;
        long last = _getLastPos() ;
        byte[] ret ;
        while( true ) {
            if( last < first ) {
                return null ;
            }
            ConvertParam.convertLong( tmpBin,0,first ) ;
            if( (ret = bdb.remove( tmpBin )) != null ) {
                _putFirstPos( first ) ;
                return ret ;
            }
            first ++ ;
        }
    }
    
    /**
     * 情報取得.
     * <p>この処理を実行した場合、取得された内容は削除されません.</p>
     * @return byte[] 対象の要素が返されます.
     * @exception Exception 例外.
     */
    public byte[] peek() throws Exception {
        startTransaction() ; // トランザクション開始でない場合は、開始処理.
        long first = _getFirstPos() ;
        long last = _getLastPos() ;
        byte[] ret ;
        while( true ) {
            if( last < first ) {
                return null ;
            }
            ConvertParam.convertLong( tmpBin,0,first ) ;
            if( (ret = bdb.get( tmpBin )) != null ) {
                return ret ;
            }
            first ++ ;
        }
    }
    
    /**
     * Iteratorを取得.
     * @return Iterator<byte[]> Iteratorを取得します.
     * @exception Exception 例外.
     */
    public Iterator<byte[]> iterator() throws Exception {
        startTransaction() ; // トランザクション開始でない場合は、開始処理.
        return new TransactionQueueIterator( this,bdb,_getFirstPos() ) ;
    }
    
    /**
     * 格納情報数を取得.
     * @return long 格納情報数が返されます.
     * @exception Exception 例外.
     */
    public long length() throws Exception {
        if( isClose() ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        return bdb.size() - 2 ;
    }
    
    /**
     * 最大格納数を取得.
     * @return long 最大格納数が返されます.
     * @exception Exception 例外.
     */
    public long getMaxLength() throws Exception {
        if( isClose() ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        return max ;
    }
    
    /**
     * 警告値を取得.
     * @return double 警告値が返されます.
     * @exception Exception 例外.
     */
    public double getWarning() throws Exception {
        if( isClose() ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        return warning ;
    }
    
    /**
     * キュー名を取得.
     * @exception Exception 例外.
     */
    public String getName() throws Exception {
        if( isClose() ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        return name ;
    }
    
    /**
     * 現在の格納数が警告値を越しているかチェック.
     * @return boolean [true]の場合、警告値を越しています.
     * @exception Exception 例外.
     */
    public boolean isWarning() throws Exception {
        if( isClose() ) {
            throw new IOException( "オブジェクトは既に破棄されています" ) ;
        }
        return ( warnLen <= ( this.length() ) ) ;
    }
    
    /** シーケンスIDを取得. **/
    private byte[] _getSequence() throws Exception {
        byte[] b = bdb.get( LAST_NAME ) ;
        long ret = ConvertParam.convertLong( 0,b ) ;
        if( ret >= Long.MAX_VALUE ) {
            throw new IOException( "Long(64bit)のシーケンスIDが枯渇しました" ) ;
        }
        ret += 1 ;
        byte[] bb = ConvertParam.convertLong( ret ) ;
        bdb.put( LAST_NAME,bb ) ;
        return bb ;
    }
    
    /** 開始位置を取得. **/
    protected long _getFirstPos() throws Exception {
        return ConvertParam.convertLong( 0,bdb.get( FIRST_NAME ) ) ;
    }
    
    /** 終了位置を取得. **/
    protected long _getLastPos() throws Exception {
        return ConvertParam.convertLong( 0,bdb.get( LAST_NAME ) ) ;
    }
    
    /** 開始位置を保存. **/
    protected void _putFirstPos( long pos ) throws Exception {
        ConvertParam.convertLong( tmpBin,0,pos ) ;
        bdb.put( FIRST_NAME,tmpBin ) ;
    }
}

/** Iterator. **/
class TransactionQueueIterator implements Iterator<byte[]> {
    private TransactionQueue queue ;
    private TransactionBdb bdb ;
    private long pos ;
    private byte[] target = null ;
    private boolean endFlag = false ;
    TransactionQueueIterator( TransactionQueue queue,TransactionBdb bdb,long firstPos ) {
        this.queue = queue ;
        this.bdb = bdb ;
        this.pos = firstPos ;
    }
    protected void finalize() throws Exception {
        close() ;
    }
    public void close() {
        bdb = null ;
        pos = -1L ;
        endFlag = true ;
        target = null ;
    }
    public void remove() {
        throw new UnsupportedOperationException() ;
    }
    
    public boolean hasNext() {
        if( bdb.isClose() ) {
            throw new RuntimeException( "オブジェクトは既に破棄されています" ) ;
        }
        if( endFlag ) {
            return false ;
        }
        if( target != null ) {
            return true ;
        }
        _get() ;
        if( endFlag ) {
            return false ;
        }
        return true ;
    }
    public byte[] next() {
        if( bdb.isClose() ) {
            throw new RuntimeException( "オブジェクトは既に破棄されています" ) ;
        }
        if( target != null ) {
            byte[] ret = target ;
            target = null ;
            return ret ;
        }
        _get() ;
        if( endFlag ) {
            throw new NoSuchElementException() ;
        }
        if( target != null ) {
            byte[] ret = target ;
            target = null ;
            return ret ;
        }
        throw new NoSuchElementException() ;
    }
    private void _get() {
        if( endFlag ) {
            return ;
        }
        long p = pos ;
        try {
            byte[] tmpBin = queue.tmpBin ;
            byte[] ret ;
            long last = queue._getLastPos() ;
            while( true ) {
                if( last < p ) {
                    endFlag = true ;
                    ret = null ;
                    return ;
                }
                ConvertParam.convertLong( tmpBin,0,p ) ;
                if( ( ret = bdb.get( tmpBin ) ) != null ) {
                    target = ret ;
                    pos = p ;
                    return ;
                }
                p ++ ;
            }
            
        } catch( Exception e ) {
            throw new RuntimeException( e ) ;
        }
    }
}
