package org.maachang.atomic ;

import java.io.Serializable;
import java.util.concurrent.atomic.AtomicStampedReference;

/**
 * AtomicCountObject.
 * 
 * @version 2008/12/01
 * @author masahito suzuki
 * @since MaachangAtomic 1.00
 */
public class AtomicCountObject<T> implements Serializable {
    private static final long serialVersionUID = -89137212990217595L;
    private final AtomicStampedReference<T> ato = new AtomicStampedReference<T>( null,0 ) ;
    
    /**
     * コンストラクタ.
     */
    public AtomicCountObject() {
        this( null,0 ) ;
    }
    
    /**
     * コンストラクタ.
     * @param f 初期値を設定します.
     */
    public AtomicCountObject( T t ) {
        this( t,0 ) ;
    }
    
    /**
     * コンストラクタ.
     * @param f 初期値を設定します.
     * @param count 対象のカウント値を設定します.
     */
    public AtomicCountObject( T t,int count ) {
        set( t,count ) ;
    }
    
    /**
     * コンストラクタ.
     * @param count 対象のカウント値を設定します.
     */
    public AtomicCountObject( int count ) {
        setCount( count ) ;
    }
    
    /**
     * Object値を取得.
     * @return T Object値が返されます.
     */
    public T get() {
        int[] c = new int[ 1 ] ;
        return ato.get( c ) ;
    }
    
    /**
     * カウント値を取得.
     * @return int カウント値が返されます.
     */
    public int getCount() {
        int[] c = new int[ 1 ] ;
        ato.get( c ) ;
        return c[ 0 ] ;
    }
    
    /**
     * Objectを設定.
     * @param t 対象オブジェクトを設定します.
     */
    public void set( T t ) {
        int[] c = new int[ 1 ] ;
        T old ;
        do {
            old = ato.get( c );
        } while( !ato.compareAndSet( old,t,c[0],c[0] ) ) ;
    }
    
    /**
     * Objectを設定.
     * @param t 対象オブジェクトを設定します.
     * @param count 対象のカウント値を設定します.
     */
    public void set( T t,int count ) {
        int[] c = new int[ 1 ] ;
        T old ;
        do {
            old = ato.get( c );
        } while( !ato.compareAndSet( old,t,c[0],count ) ) ;
    }
    
    /**
     * カウント値を設定.
     * @param count 対象のカウント値を設定します.
     */
    public void setCount( int count ) {
        int[] c = new int[ 1 ] ;
        T old ;
        do {
            old = ato.get( c );
        } while( !ato.compareAndSet( old,old,c[0],count ) ) ;
    }
    
    /**
     * カウント値を１インクリメント.
     * @return int １インクリメントしたカウント値が返されます.
     */
    public int inc() {
        int[] c = new int[ 1 ] ;
        T old ;
        do {
            old = ato.get( c );
        } while( !ato.compareAndSet( old,old,c[0],c[0]+1 ) ) ;
        return c[0]+1;
    }
    
    /**
     * カウント値を１デクリメント.
     * @return int １デクリメントしたカウント値が返されます.
     */
    public int dec() {
        int[] c = new int[ 1 ] ;
        T old ;
        do {
            old = ato.get( c );
        } while( !ato.compareAndSet( old,old,c[0],c[0]-1 ) ) ;
        return c[0]-1;
    }
    
    /**
     * 文字列変換.
     * @return String 文字列が返されます.
     */
    public String toString() {
        return String.valueOf( get() ) ;
    }
    
    protected AtomicStampedReference getAtomicStampedReference() {
        return ato ;
    }
}
