package org.maachang.atomic ;

import java.io.*;
import java.util.concurrent.atomic.*;

/**
 * Atomic拡張可能配列.
 * 
 * @version 2008/01/16
 * @author masahito suzuki
 * @since MaachangAtomic 1.00
 */
public class AtomicMoveArray<T> implements Serializable {
    private static final long serialVersionUID = -70907164927496291L;
    private final AtomicObject<AtomicReferenceArray<T>> moveArray = new AtomicObject<AtomicReferenceArray<T>>() ;
    
    /**
     * コンストラクタ.
     */
    private AtomicMoveArray() {
        moveArray.set( null ) ;
    }
    
    /**
     * コンストラクタ.
     * @param size 対象のサイズを設定します.
     */
    public AtomicMoveArray( int size ) {
        if( size <= 0 ) {
            throw new IllegalArgumentException( "配列数が不正です" ) ;
        }
        moveArray.set( new AtomicReferenceArray<T>( size ) ) ;
    }
    
    /**
     * 配列をクリア.
     */
    public void clear() {
        moveArray.set( null ) ;
    }
    
    /**
     * Object値を取得.
     * @param no 対象の項番を設定します.
     * @return T Object値が返されます.
     */
    public T get( int no ) {
        if( no < 0 || no >= size() ) {
            throw new IllegalArgumentException( "指定範囲は不正です" ) ;
        }
        AtomicReferenceArray<T> ary = moveArray.get() ;
        if( ary == null ) {
            throw new IllegalArgumentException( "配列情報は存在しません" ) ;
        }
        return ary.get( no ) ;
    }
    
    /**
     * Objectを設定.
     * @param t 対象オブジェクトを設定します.
     * @param no 対象の項番を設定します.
     */
    public void set( T t,int no ) {
        if( no < 0 || no >= size() ) {
            throw new IllegalArgumentException( "指定範囲は不正です" ) ;
        }
        AtomicReferenceArray<T> ary = moveArray.get() ;
        if( ary == null ) {
            throw new IllegalArgumentException( "配列情報は存在しません" ) ;
        }
        setting( ary,t,no ) ;
    }
    
    /**
     * サイズを取得.
     * @return int サイズが返されます.
     */
    public int size() {
        AtomicReferenceArray<T> ary = moveArray.get() ;
        if( ary == null ) {
            return 0 ;
        }
        return ary.length() ;
    }
    
    /**
     * 配列サイズを指定サイズに増減.
     * @param length 新しい配列サイズを設定します.
     */
    public void newSize( int length ) {
        if( length <= 0 ) {
            throw new IllegalArgumentException( "配列数が不正です" ) ;
        }
        if( moveArray.get() == null ) {
            moveArray.set( new AtomicReferenceArray<T>( length ) ) ;
            return ;
        }
        else if( size() == length ) {
            return ;
        }
        AtomicReferenceArray<T> src = null ;
        AtomicReferenceArray<T> dest = null ;
        AtomicReference<AtomicReferenceArray<T>> o = moveArray.getAtomicReference() ;
        do {
            src = moveArray.get() ;
            if( src == null ) {
                throw new IllegalArgumentException( "配列情報は存在しません" ) ;
            }
            int srcLen = src.length() ;
            dest = new AtomicReferenceArray<T>( length ) ;
            int len = ( srcLen >= length ) ? length : srcLen ;
            copyArrays( src,0,dest,0,len ) ;
        } while( !o.compareAndSet( src,dest ) ) ;
    }
    
    /**
     * 配列コピー.
     * @param src コピー元を設定します.
     * @param srcPos コピー元オフセット値を設定します.
     * @param dest コピー先を設定します.
     * @param destPos コピー先オフセット値を設定します.
     * @param len コピー長を設定します.
     */
    public static final void arraycopy( AtomicMoveArray src,int srcPos,AtomicMoveArray dest,int destPos,int len ) {
        for( int i = 0 ; i < len ; i ++ ) {
            dest.set( src.get( srcPos+i ),destPos+i ) ;
        }
    }
    
    private static final void copyArrays( AtomicReferenceArray src,int sp,AtomicReferenceArray dest,int dp,int len ) {
        for( int i = 0 ; i < len ; i ++ ) {
            setting( dest,src.get( sp+i ),dp+i ) ;
        }
    }
    
    private static final void setting( AtomicReferenceArray ary,Object t,int no ) {
        while( !ary.compareAndSet( no,ary.get( no ),t ) ) {}
    }
}
