package org.maachang.seabass ;

import java.util.*;
import pnuts.lang.*;
import org.maachang.*;

/**
 * SeaBassMap.
 * 
 * @version 2008/01/17
 * @author masahito suzuki
 * @since Seabass 1.00
 */
public class SeaBassMap extends BaseMap implements Map {
    private static final int INITIAL_CAPACITY = 8;
    private static final float LOAD_FACTOR = 0.75f;
    
    private transient KeyValue[] table = null ;
    private int count = 0 ;
    private int threshold = 0 ;
    
    public SeaBassMap() {
        this( INITIAL_CAPACITY ) ;
    }
    
    public SeaBassMap( int n ) {
        this.threshold = (int) (n * LOAD_FACTOR);
        this.table = new KeyValue[n];
    }
    
    public void clear() {
        KeyValue tab[] = table;
        for (int i = 0; i < tab.length; i++) {
            tab[i] = null;
        }
        count = 0;
    }
    
    public Object get(Object interned) {
        int hash = interned.hashCode() & 0x7fffffff;
        int i = hash & (table.length - 1);
        KeyValue b = table[i];
        while (true) {
            if (b == null) {
                return b;
            }
            if (interned == b.name || ( interned != null && interned.equals( b.name )) ) {
                return b.value;
            }
            b = b.chain;
        }
    }
    
    public NamedValue lookup(Object interned) {
        int hash = interned.hashCode() & 0x7fffffff;
        KeyValue[] tab = this.table;
        int i = hash & (tab.length - 1);
        KeyValue b = tab[i];
        while (b != null) {
            if (interned == b.name || ( interned != null && interned.equals( b.name )) ) {
                break ;
            }
            b = b.chain;
        }
        if (b != null) {
            return b;
        }
        return null;
    }
    
    public boolean containsKey(Object interned) {
        return lookup( interned ) != null ;
    }
    
    public Object put(Object interned, Object value) {
        int hash = interned.hashCode() & 0x7fffffff;
        int i = hash & (table.length - 1);
        for (KeyValue b = table[i]; b != null; b = b.chain) {
            if (interned == b.name || ( interned != null && interned.equals( b.name )) ) {
                Object ret = b.get() ;
                b.set(value);
                return ret ;
            }
        }
        addElement(hash, (String)interned, value, i);
        return null ;
    }
    
    public KeyValue removeElement(Object interned) {
        int hash = interned.hashCode() & 0x7fffffff;
        int i = hash & (table.length - 1);
        KeyValue prev = table[i];
        KeyValue b = prev;
        while (b != null) {
            KeyValue next = b.chain;
            if (interned == b.name || ( interned != null && interned.equals( b.name )) ) {
                count--;
                if (prev == b) {
                    table[i] = next;
                } else {
                    prev.chain = next;
                }
                return b;
            }
            prev = b;
            b = next;
        }
        return b;
    }
    
    public Object remove( Object interned) {
        KeyValue ret = removeElement( interned ) ;
        if( ret != null ) {
            return ret.value ;
        }
        return null ;
    }
    
    public int size() {
        return count;
    }
    
    public boolean isEmpty() {
        return count <= 0 ;
    }
    
    public void putAll(Map toMerge) {
        if( toMerge == null ) {
            return ;
        }
        Iterator it = null ;
        if( toMerge instanceof SeaBassMap ) {
            it = ((SeaBassMap)toMerge).bindings() ;
            while( it.hasNext() ) {
                KeyValue km = (KeyValue)it.next() ;
                if( km != null ) {
                    put( km.getName(),km.get() ) ;
                }
            }
        }
        else {
            it = toMerge.keySet().iterator() ;
            while( it.hasNext() ) {
                Object key = it.next() ;
                if( key != null ) {
                    put( key,toMerge.get( key ) ) ;
                }
            }
        }
    }
    
    public Set keySet() {
        return new SeaBassMapSet( this ) ;
    }
    
    public Iterator bindings() {
        return new SeaBassMemoryIterator(0);
    }
    
    public Iterator keys() {
        return new SeaBassMemoryIterator(1);
    }
    
    public Collection values() {
        Set<Object> ret = new HashSet<Object>() ;
        Iterator it = new SeaBassMemoryIterator(2) ;
        while( it.hasNext() ) {
            ret.add( it.next() ) ;
        }
        return ret ;
    }
    
    public String toString() {
        boolean flg = false ;
        StringBuilder buf = new StringBuilder( "{" ) ;
        Iterator it = bindings() ;
        while( it.hasNext() ) {
            KeyValue kv = (KeyValue)it.next() ;
            if( kv.value == this ) {
                continue ;
            }
            if( flg ) {
                buf.append( "," ) ;
            }
            buf.append( kv.name ).append( ":" ).append( kv.value ) ;
            flg = true ;
        }
        return buf.append( "}" ).toString() ;
    }
    
    static class SeaBassMapSet implements Set<Object> {
        private SeaBassMap map = null ;
        private SeaBassMapSet() {
        }
        protected SeaBassMapSet( SeaBassMap map ) {
            this.map = map ;
        }
        public int size() {
            return map.size() ;
        }
        public boolean isEmpty() {
            return map.isEmpty() ;
        }
        public boolean contains(Object o) {
            return map.containsKey( o ) ;
        }
        public Iterator<Object> iterator() {
            return map.keys() ;
        }
        public Object[] toArray() {
            ArrayList<Object> ret = new ArrayList<Object>() ;
            Iterator it = map.keys() ;
            while( it.hasNext() ) {
                ret.add( it.next() ) ;
            }
            if( ret.size() <= 0 ) {
                return new Object[0] ;
            }
            return ret.toArray() ;
        }
        public <Object> Object[] toArray(Object[] a) {
            return null ;
        }
        public boolean add(Object e) {
            return false ;
        }
        public boolean remove(Object o) {
            return false ;
        }
        public boolean containsAll(Collection<?> c) {
            return false ;
        }
        public boolean addAll(Collection<? extends Object> c) {
            return false ;
        }
        public boolean retainAll(Collection<?> c) {
            return false ;
        }
        public boolean removeAll(Collection<?> c) {
            return false ;
        }
        public void clear() {
            map = null ;
        }
        public boolean equals(Object o) {
            return this == o ;
        }
        public int hashCode() {
            return map.hashCode() ;
        }
    }
    
    private class SeaBassMemoryIterator implements Iterator {
        KeyValue bind = null;
        int index = table.length;
        int kind;
        protected SeaBassMemoryIterator(int kind) {
            this.kind = kind;
        }
        public boolean hasNext() {
            while (bind == null && index > 0) {
                bind = table[--index];
            }
            return bind != null;
        }
        public Object next() {
            while (bind == null && index > 0) {
                bind = table[--index];
            }
            if (bind != null) {
                KeyValue b = bind;
                bind = b.chain;
                if (kind == 0) {
                    return b;
                } else if (kind == 1) {
                    return b.name;
                } else {
                    return b.value;
                }
            }
            throw new NoSuchElementException();
        }
        public void remove() {
            throw new UnsupportedOperationException() ;
        }
    }
    
    private void addElement(int hash, String interned, Object value, int index) {
        table[index] = new KeyValue(hash, interned, value, table[index]);
        if (count++ >= threshold) {
            ensureCapacity(table.length * 2);
        }
    }
    
    private void ensureCapacity(int newCapacity) {
        KeyValue[] oldTable = table;
        int oldCapacity = oldTable.length;
        if (oldCapacity == 1 << 30) {
            threshold = Integer.MAX_VALUE;
            return;
        }
        KeyValue[] newTable = new KeyValue[newCapacity];
        KeyValue[] src = table;
        for (int j = 0; j < src.length; j++) {
            KeyValue b = src[j];
            if (b != null) {
                src[j] = null;
                do {
                    KeyValue next = b.chain;
                    int i = b.hash & (newCapacity - 1);
                    b.chain = newTable[i];
                    newTable[i] = b;
                    b = next;
                } while (b != null);
            }
        }
        table = newTable;
        threshold = (int) (newCapacity * LOAD_FACTOR);
    }
    
    protected static class KeyValue implements pnuts.lang.NamedValue {
        String name = null ;
        Object value = null ;
        int hash = -1 ;
        KeyValue chain = null ;
        public KeyValue(int h, Object nm, Object v, KeyValue n) {
            value = v;
            chain = n;
            name = (String)nm;
            hash = h;
        }
        public Object get() {
            return value;
        }
        public void set(Object v) {
            value = v;
        }
        public String getName() {
            return name;
        }
    }
}
