package org.pnuts.util;

import java.util.* ;
import java.lang.reflect.* ;
import java.util.concurrent.* ;

import pnuts.lang.* ;
import gnu.trove.*;
import org.maachang.* ;
import org.maachang.seabass.mod.object.* ;
import org.maachang.reflect.* ;

/**
 * java情報検索(クラス呼び出し/メソッド呼び出し/フィールド呼び出し).
 * <p>詳しくは、org.maachang.reflect系のJavadocを参照してください。</p>
 * 
 * @version 2008/12/11
 * @author masahito suzuki
 * @since SeaBassScriptEngine 1.00
 */
public abstract class SearchJavaInfo {
    private SearchJavaInfo(){}
    
    /**
     * 指定パッケージ名のクラス情報を取得.
     * @param cl 対象のクラスローダを設定します.
     * @param name 対象のクラス名を設定します.
     * @return Class 対象のクラス情報が返されます.
     * @exception ClassNotFoundException クラスロード例外.
     */
    public static final Class getClass( String name )
        throws ClassNotFoundException {
        if( name == null || name.length() <= 0 ) {
            throw new PnutsException( "クラス名が指定されていません" ) ;
        }
        return FastReflect.getClass( name ) ;
    }
    
    /**
     * 指定パッケージ名のクラス情報を取得.
     * @param cl 対象のクラスローダを設定します.
     * @param name 対象のクラス名を設定します.
     * @return Class 対象のクラス情報が返されます.
     * @exception ClassNotFoundException クラスロード例外.
     */
    public static final Class getClass( ClassLoader cl,String name )
        throws ClassNotFoundException {
        if( name == null || name.length() <= 0 ) {
            throw new PnutsException( "クラス名が指定されていません" ) ;
        }
        return FastReflect.getClass( cl,name ) ;
    }
    
    /**
     * コンストラクタ実行.
     * @param c 対象のクラスオブジェクトを設定します.
     * @param args 対象の引数パラメータを設定します.
     * @param types 対象の引数パラメータタイプを設定します.
     * @return Object 生成されたオブジェクトが返されます.
     * @exception Exception 例外.
     */
    public static final Object newInstance( Class c, Object args[],Class types[] )
        throws Exception {
        if( c == null ) {
            throw new PnutsException( "生成対象のクラスが設定されていません" ) ;
        }
        Object ret = null ;
        try {
            ret = FastReflect.newInstanceTo( (ClassLoader)null,c.getName(),args,types ) ;
        } catch( InvocationTargetException it ) {
            // リフレクションラッパー例外が返された場合.
            invocationTargetException( it,null ) ;
        }
        return ret ;
    }
    
    /**
     * フィールド呼び出し.
     * @param target 実行対象のターゲットオブジェクトを設定します.
     * @param name 実行対象のメソッド名を設定します.
     * @return Object フィールド情報が返されます.
     * @exception Exception 例外.
     */
    public static final Object getField( Object target,String name )
        throws Exception {
        if( target == null ) {
            throw new PnutsException( "ターゲットオブジェクトが設定されていません" ) ;
        }
        if (target instanceof Map){
            return ((Map)target).get(name) ;
        }
        Class c = null ;
        if( target instanceof Class ) {
            c = (Class)target ;
            target = null ;
        }
        else {
            c = target.getClass() ;
        }
        return FastReflect.getField( c,target,name ) ;
    }
    
    /**
     * フィールド呼び出し.
     * @param c 実行対象のクラスを設定します.
     * @param target 実行対象のターゲットオブジェクトを設定します.
     * @param name 実行対象のメソッド名を設定します.
     * @return Object フィールド情報が返されます.
     * @exception Exception 例外.
     */
    public static final Object getField( Class c,Object target,String name )
        throws Exception {
        if( c == null && target == null ) {
            throw new PnutsException( "実行引数は不正です" ) ;
        }
        if( c == null ) {
            if( target instanceof Class ) {
                c = (Class)target ;
                target = null ;
            }
            else {
                c = target.getClass() ;
            }
        }
        else if (target instanceof Map){
            return ((Map)target).get(name) ;
        }
        return FastReflect.getField( c,target,name ) ;
    }
    
    /**
     * フィールド設定.
     * @param target 実行対象のターゲットオブジェクトを設定します.
     * @param name 実行対象のメソッド名を設定します.
     * @param o 設定対象のオブジェクトを設定します.
     * @exception Exception 例外.
     */
    public static final void setField( Object target,String name,Object o )
        throws Exception {
        if( target == null ) {
            throw new PnutsException( "ターゲットオブジェクトが設定されていません" ) ;
        }
        if (target instanceof Map){
            ((Map)target).put(name,o) ;
            return ;
        }
        Class c = null ;
        if( target instanceof Class ) {
            c = (Class)target ;
            target = null ;
        }
        else {
            c = target.getClass() ;
        }
        FastReflect.setField( target.getClass(),target,name,o ) ;
    }
    
    /**
     * フィールド設定.
     * @param c 実行対象のクラスを設定します.
     * @param target 実行対象のターゲットオブジェクトを設定します.
     * @param name 実行対象のメソッド名を設定します.
     * @param o 設定対象のオブジェクトを設定します.
     * @return Object フィールド情報が返されます.
     * @exception Exception 例外.
     */
    public static final void setField( Class c,Object target,String name,Object o )
        throws Exception {
        if( c == null && target == null ) {
            throw new PnutsException( "実行引数は不正です" ) ;
        }
        if( c == null || CLAZZ_NAME.equals( c.getName() ) ) {
            if( target instanceof Class ) {
                c = (Class)target ;
                target = null ;
            }
            else {
                c = target.getClass() ;
            }
        }
        else if (target instanceof Map){
            ((Map)target).put(name,o) ;
            return ;
        }
        FastReflect.setField( c,target,name,o ) ;
    }
    
    /**
     * メソッド呼び出し.
     * @param context 対象のコンテキストを設定します.
     * @param target 実行対象のターゲットオブジェクトを設定します.
     * @param name 実行対象のメソッド名を設定します.
     * @param args 実行対象の引数を設定します.
     * @return Object 実行結果が返されます.
     * @exception Exception 例外.
     */
    public static final Object callMethod( Context context,Object target,String name,Object[] args )
        throws Exception {
        if (target == null ) {
            throw new PnutsException( "targetがnullで定義されています" ) ;
        }
        Object ret = null ;
        try {
            if (target instanceof Map && ((Map)target).containsKey(name) ) {
                Object o = ((Map)target).get(name) ;
                if (o != null && o instanceof PnutsFunction) {
                    return ((PnutsFunction)o).call(args, context);
                }
            }
            Class c = null ;
            if( target instanceof Class ) {
                c = (Class)target ;
                target = null ;
            }
            else {
                c = target.getClass() ;
            }
            ret = FastReflect.invoke( c,target,name,args ) ;
        } catch( InvocationTargetException it ) {
            // リフレクションラッパー例外が返された場合.
            invocationTargetException( it,context ) ;
        }
        return ret ;
    }
    
    /**
     * メソッド呼び出し.
     * @param context 対象のコンテキストを設定します.
     * @param target 実行対象のターゲットオブジェクトを設定します.
     * @param name 実行対象のメソッド名を設定します.
     * @param args 実行対象の引数を設定します.
     * @param types 実行対象の引数タイプを設定します.
     * @return Object 実行結果が返されます.
     * @exception Exception 例外.
     */
    public static final Object callMethod( Context context,Object target,String name,Object[] args,Class[] types )
        throws Exception {
        if (target == null ) {
            throw new PnutsException( "targetがnullで定義されています" ) ;
        }
        Object ret = null ;
        try {
            if (target instanceof Map){
                if( ((Map)target).containsKey(name) ) {
                    Object o = ((Map)target).get(name) ;
                    if (o != null && o instanceof PnutsFunction) {
                        return ((PnutsFunction)o).call(args, context);
                    }
                }
            }
            Class c = null ;
            if( target instanceof Class ) {
                c = (Class)target ;
                target = null ;
            }
            else {
                c = target.getClass() ;
            }
            ret = FastReflect.invokeTo( (ClassLoader)null,c,target,name,args,types ) ;
        } catch( InvocationTargetException it ) {
            // リフレクションラッパー例外が返された場合.
            invocationTargetException( it,context ) ;
        }
        return ret ;
    }
    
    /**
     * メソッド呼び出し.
     * @param context 対象のコンテキストを設定します.
     * @param c 実行対象のクラスを設定します.
     * @param target 実行対象のターゲットオブジェクトを設定します.
     * @param name 実行対象のメソッド名を設定します.
     * @param args 実行対象の引数を設定します.
     * @param types 実行対象の引数タイプを設定します.
     * @return Object 実行結果が返されます.
     * @exception Exception 例外.
     */
    public static final Object callMethod( Context context,Class c,Object target,String name,Object[] args,Class[] types )
        throws Exception {
        if( c == null && target == null ) {
            throw new PnutsException( "実行引数は不正です" ) ;
        }
        Object ret = null ;
        try {
            if (target instanceof Map){
                if( ((Map)target).containsKey(name) ) {
                    Object o = ((Map)target).get(name) ;
                    if (o != null && o instanceof PnutsFunction) {
                        return ((PnutsFunction)o).call(args, context);
                    }
                }
            }
            if( c == null || CLAZZ_NAME.equals( c.getName() ) ) {
                if( target instanceof Class ) {
                    c = (Class)target ;
                    target = null ;
                }
                else {
                    c = target.getClass() ;
                }
            }
            ret = FastReflect.invokeTo( (ClassLoader)null,c,target,name,args,types ) ;
        } catch( InvocationTargetException it ) {
            // リフレクションラッパー例外が返された場合.
            invocationTargetException( it,context ) ;
        }
        return ret ;
    }
    
    /** リフレクションラッパー例外が存在する場合の処理. **/
    private static final void invocationTargetException( InvocationTargetException it,pnuts.lang.Context context )
        throws Exception {
        if( context == null ) {
            context = Context.getCurrentContext() ;
        }
        Throwable t = it.getCause() ;
        if( t == null ) {
            throw it ;
        }
        if( t instanceof Error ) {
            if( t instanceof pnuts.lang.PnutsException ) {
                throw (pnuts.lang.PnutsException)t ;
            }
            throw new pnuts.lang.PnutsException( t,context ) ;
            //throw (Error)t ;
        }
        else if( t instanceof Exception ) {
            if( t instanceof pnuts.lang.PnutsException ) {
                throw (pnuts.lang.PnutsException)t ;
            }
            throw new pnuts.lang.PnutsException( t,context ) ;
            //throw (Exception)t ;
        }
        throw it ;
    }
    
    private static final String CLAZZ_NAME = "java.lang.Class" ;
}
