/*
 * PnutsException.java
 *
 * Copyright (c) 1997-2007 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "LICENSE.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Enumeration;
import java.util.Vector;

/**
 * This is a wrapper class for Exception to be thrown.
 * 
 * @author Toyokazu Tomatsu
 */
public class PnutsException extends RuntimeException {

    static final long serialVersionUID = -5567045200989654942L;

    /**
     * @serial
     */
    protected int line;

    /**
     * @serial
     */
    protected int column;

    /**
     * @serial
     */
    protected Object file;

    /**
     * @serial
     */
    protected Throwable throwable;

    /**
     * @serial
     */
    protected String contextName;
    
    /**
     * errorScriptName.
     */
    protected String errorScriptName = "unknown" ;


    protected transient Vector trace;

    protected transient Object operation;

    /**
     * Constructor
     */
    public PnutsException() {
        this("");
    }

    /**
     * Constructor
     * 
     * @param msg
     *            the error message
     */
    public PnutsException(String msg) {
        this( new PnutsDetailException( msg ),null ) ;
    }

    /**
     * Constructor
     * 
     * @param msg
     *            the error message
     * @param context
     *            the context in which the error occurs
     */
    public PnutsException(String msg, Context context) {
        this( new PnutsDetailException( msg ),context ) ;
    }

    /**
     * This constructor creates a PnutsException using i18n resources in
     * pnuts.properties.
     */
    public PnutsException(String key, Object param[], Context context) {
        this("pnuts.lang.pnuts", key, param, context);
    }

    /**
     * This constructor creates a PnutsException using i18n resources in
     * pnuts.properties.
     */
    public PnutsException(String bundleName, String key, Object param[],
            Context context) {
        this(new PnutsDetailException( Runtime.getMessage(bundleName, key, param) ), context);
    }

    public PnutsException(Throwable t) {
        this( t,null ) ;
    }

    /**
     * Constructor
     * 
     * @param t
     *            a Throwable
     * @param context
     *            the context
     */
    public PnutsException(Throwable t, Context context) {
        super( t ) ;
        if( context == null ) {
            context = Context.getCurrentContext() ;
        }
        this.contextName = context.getName();
        if (t instanceof PnutsException) {
            PnutsException p = (PnutsException) t;
            this.throwable = p.throwable;
            this.trace = p.trace;
            this.file = p.file;
            this.line = p.line;
            this.column = p.column;
            this.errorScriptName = p.errorScriptName ;
        } else if (t instanceof ParseException) {
            this.throwable = t;
            this.line = ((ParseException) t).getErrorLine();
            this.column = ((ParseException) t).getErrorColumn();
            this.file = context.getScriptSource();
            this.errorScriptName = context.getErrorScriptName() ;
        } else {
            this.throwable = t;
            this.line = context.beginLine;
            this.column = context.beginColumn;
            this.file = context.getScriptSource();
            this.errorScriptName = context.getErrorScriptName() ;
        }
        addPnutsErrorLineInfo( context ) ;// 新しいエラー行情報を保存.
    }

    /**
     * Constructor
     * 
     * @deprecated
     * @param t
     *            a Throwable
     * @param operation
     *            a Method or a Constructor
     * @param context
     *            the context
     */
    public PnutsException(Throwable t, Object operation, Context context) {
        super( t ) ;
        if( context == null ) {
            context = Context.getCurrentContext() ;
        }
        if (t instanceof PnutsException) {
            PnutsException p = (PnutsException) t;
            this.throwable = p.throwable;
            this.trace = p.trace;
            this.file = p.file;
            this.line = p.line;
            this.column = p.column;
            this.errorScriptName = p.errorScriptName ;
        } else if (t instanceof ParseException) {
            this.throwable = t;
            this.line = ((ParseException) t).getErrorLine();
            this.column = ((ParseException) t).getErrorColumn();
            this.file = context.getScriptSource();
            this.errorScriptName = context.getErrorScriptName() ;
        } else {
            this.throwable = t;
            this.line = context.beginLine;
            this.column = context.beginColumn;
            this.file = context.getScriptSource();
            this.errorScriptName = context.getErrorScriptName() ;
        }
        this.contextName = context.getName();
        this.operation = operation;
        addPnutsErrorLineInfo( context ) ;// 新しいエラー行情報を保存.
    }
    
    /**
     * エラー条件をセット.
     */
    private void addPnutsErrorLineInfo( pnuts.lang.Context context ) {
        if( context == null ) {
            context = Context.getCurrentContext() ;
        }
        PnutsErrorLineInfo e = (PnutsErrorLineInfo)context.getCurrentPackage().get( "_$pnuts_error_info" ) ;
        if( e == null ) {
            e = new PnutsErrorLineInfo() ;
            context.getCurrentPackage().set( "_$pnuts_error_info",e ) ;
        }
        e.add( file,contextName,errorScriptName,line,column ) ;
    }
    
    /**
     * エラー条件をセット.
     */
    protected static final void addErrorLineInfo( pnuts.lang.Context context ) {
        if( context == null ) {
            context = Context.getCurrentContext() ;
        }
        PnutsErrorLineInfo e = (PnutsErrorLineInfo)context.getCurrentPackage().get( "_$pnuts_error_info" ) ;
        if( e == null ) {
            e = new PnutsErrorLineInfo() ;
            context.getCurrentPackage().set( "_$pnuts_error_info",e ) ;
        }
        e.add( context.getScriptSource(),context.getName(),context.getErrorScriptName(),context.beginLine,context.beginColumn ) ;
    }
    
    /**
     * エラー条件を取得.
     */
    public static final PnutsErrorLineInfo getPnutsErrorLineInfo( pnuts.lang.Context context ) {
        if( context == null ) {
            context = Context.getCurrentContext() ;
        }
        return (PnutsErrorLineInfo)context.getCurrentPackage().get( "_$pnuts_error_info" ) ;
    }
    
    /**
     * エラー条件をクリア.
     */
    public static final void clearPnutsErrorLineInfo( pnuts.lang.Context context ) {
        PnutsErrorLineInfo info = getPnutsErrorLineInfo( context ) ;
        if( info != null ) {
            info.clear() ;
        }
    }

    public String getMessage(){
        try {
            if (throwable != null && throwable != this) {
                return throwable.getMessage();
            } else {
                return super.getMessage();
            }
        } catch( Exception e ) {
            return "unknown" ;
        }
    }

    /**
     * Returns the root cause
     */
    public Throwable getThrowable() {
        return throwable;
    }
    
    public String getErrorScriptName() {
        return errorScriptName ;
    }
    
    protected void setErrorScriptName( String name ) {
        this.errorScriptName = name ;
    }

    String position() {
        StringBuilder buf = new StringBuilder();
        //if (file != null || contextName != null || errorScriptName != null) {
        if( contextName != null || errorScriptName != null) {
            buf.append('[');
            if (contextName != null) {
                buf.append(contextName).append(", ");
            }
            //if (file != null) {
                //buf.append(file);
            //}
            if( errorScriptName != null ) {
                buf.append( errorScriptName ).append( ", ") ;
            }
            if (line > 0) {
                buf.append("line: " + line);
                if (column > 0){
                    buf.append(" column: " + column);
                }
            }
            buf.append(']');
        }
        if (buf.length() > 0) {
            buf.append(":\n");
        }
        return buf.toString();
    }

    /**
     * Returns an enumeration of PnutsException.TraceInfo objects
     */
    public Enumeration getBackTrace() {
        if (trace != null){
            return trace.elements();
        } else {
            return null;
        }
    }

    void printBackTrace(PrintWriter pw, String indent) {
        for (Enumeration e = trace.elements(); e.hasMoreElements();) {
            pw.print(indent);
            Object elem = e.nextElement();
            if (elem == this){
                pw.println(getClass() + "@" + System.identityHashCode(this));
            } else {
                pw.println(elem);
            }
        }
    }

    String trace() {
        StringWriter sw = new StringWriter();
        printBackTrace(new PrintWriter(sw), "    ");
        return sw.toString();
    }

    public void printStackTrace(PrintWriter writer) {
        writer.write( toString() ) ;
        if (throwable != null && throwable != this) {
            throwable.printStackTrace(writer);
        } else {
            super.printStackTrace(writer);
        }
    }

    public void printStackTrace(PrintStream ps) {
        ps.println( toString() ) ;
        if (throwable != null && throwable != this) {
            throwable.printStackTrace(ps);
        } else {
            super.printStackTrace(ps);
        }
    }
    
    public void setContext( Context context ) {
        this.line = context.beginLine;
        this.column = context.beginColumn;
        this.contextName = context.getName();
        this.file = context.getScriptSource();
        this.errorScriptName = context.getErrorScriptName() ;
    }

    /**
     * Returns the line number where the error occured.
     */
    public int getLine() {
        return line;
    }
    
    public void setLine( int line ) {
        this.line = line ;
    }

    /**
     * Returns the column number where the error occured.
     * -1: unknown 
     */
    public int getColumn() {
        return column;
    }
    
    public void setColumn( int column ) {
        this.column = column ;
    }

    /**
     * Returns the script source (usually a URL) where the error occured.
     */
    public Object getScriptSource() {
        return file;
    }

    void backtrace(TraceInfo traceInfo) {
        if (trace == null) {
            trace = new Vector();
        }
        trace.addElement(traceInfo);
    }

    public String toString() {
        String s = position();
        if (trace != null) {
            s += trace();
        }
        if (operation != null) {
            s += operation.toString() + "->";
        }
        return s + throwable.getClass().getName() + " : "
                + throwable.getMessage();
    }

    /**
     * A node of a call-chain, which represents a position of a certain function
     * call
     */
    public static class TraceInfo {
        Object target;

        Object frame;

        Object[] arguments;

        Object scriptSource;
        
        String scriptName;

        int line;

        int column;

        protected TraceInfo() {
        }

        /**
         * Constructor
         *
         * @param frame the function's name or the class of the constructor
         * @param args the arguments
         * @param scriptSource the object from which the script was read
         * @param scriptName
         * @param line the line
         * @param column the column
         */
        public TraceInfo(Object frame, Object[] args, Object scriptSource,
                String scriptName,int line, int column) {
            this(null, frame, args, scriptSource, scriptName, line, column);
        }

        /**
         * Constructor
         *
         * @param target the target object
         * @param methodName the method name
         * @param args the arguments
         * @param scriptSource the object from which the script was read
         * @param scriptName
         * @param line the line
         * @param column the column
         */
        public TraceInfo(Object target, Object methodName, Object[] args, Object scriptSource,
                String scriptName,int line, int column) {
            this.target = target;
            this.frame = methodName;
            this.arguments = args;
            this.scriptSource = scriptSource;
            this.scriptName = scriptName ;
            this.line = line;
            this.column = column;
        }

        /**
         * Gets the source of the script where the function call was taken
         * place. It is usually a URL object, though the script source could be
         * any object.
         */
        public Object getScriptSource() {
            return scriptSource;
        }
        
        public String getScriptName() {
            return scriptName ;
        }

        /**
         * The actual arguments of the function call
         */
        public Object[] getArguments() {
            return (Object[])arguments.clone();
        }

        /**
         * The line number of the place where the function call was taken place.
         */
        public int getLine() {
            return line;
        }

        /**
         * The column number of the place where the function call was taken place.
         */
        public int getColumn(){
            return column;
        }

        /**
         * Gets the callee that throws an exception.
         *
         * @return either of a function name, method name, or Class object.
         */
        public Object getFrame() {
            return frame;
        }

        /**
         * Gets the target object of the method call that causes an exception
         *
         * @return the target object of the method call that causes an exception
         */
        public Object getTargetObject(){
            return target;
        }

        public String toString() {
            StringBuilder sbuf = new StringBuilder();
            if (target != null){
                try {
                    sbuf.append(Pnuts.format(target));
                } catch (Throwable t){
                    sbuf.append("?");
                }
                sbuf.append(".");
            }
            sbuf.append(String.valueOf(frame));
            try {
                String args = Runtime.format(arguments, 64);
                sbuf.append("(" + args.substring(1, args.length() - 1) + ")");
            } catch (Throwable tt) {
                sbuf.append("(?)");
            }
            if( scriptName != null ) {
                sbuf.append(" [");
                sbuf.append(scriptName);
                sbuf.append(':');
                sbuf.append(" line: " + line);
                if (column >= 0){
                    sbuf.append(" column: " + column);
                }
                sbuf.append(']');
            }
            else if (scriptSource != null) {
                sbuf.append(" [");
                sbuf.append(scriptSource);
                sbuf.append(':');
                sbuf.append(" line: " + line);
                if (column >= 0){
                    sbuf.append(" column: " + column);
                }
                sbuf.append(']');
            }
            return sbuf.toString();
        }
    }
}
