/*
 * SymbolTable.java
 * 
 * Copyright (c) 1997-2005 Sun Microsystems, Inc. All Rights Reserved.
 * 
 * See the file "LICENSE.txt" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Enumeration;
import java.util.NoSuchElementException;

import org.maachang.atomic.* ;
import org.pnuts.util.* ;

class SymbolTable implements Cloneable, Serializable {

    private static final int INITIAL_CAPACITY = 8;
    private static final float LOAD_FACTOR = 0.75f;
    
    public final PnutsLock lock = new PnutsLock() ;

    SymbolTable parent;
    transient AtomicObject<AtomicArray<Binding>> table;
    protected void setTable( AtomicArray<Binding> t ) {
        this.table.set( t ) ;
    }
    protected AtomicArray<Binding> getTable() {
        return table.get();
    }
    private AtomicInt count = null ;
    protected void setCount( int c ) {
        count.set( c ) ;
    }
    protected int getCount() {
        return count.get() ;
    }
    private AtomicInt threshold = null ;

    public SymbolTable() {
        this.table = new AtomicObject<AtomicArray<Binding>>( new AtomicArray<Binding>(INITIAL_CAPACITY) ) ;
        this.count = new AtomicInt() ;
        this.threshold = new AtomicInt( (int) (INITIAL_CAPACITY * LOAD_FACTOR) ) ;
        this.parent = null ;
    }

    public SymbolTable(SymbolTable parent) {
        this();
        this.parent = parent;
    }

    /**
     * Gets the value of a variable.
     * 
     * @param interned
     *            the name of the variable, which must be intern'd
     * @return the value
     */
    public Object get(String interned) {
        int hash = interned.hashCode() & 0x7fffffff;
        int i = hash & (table.get().size() - 1);
        Binding b = table.get().get(i);
        while (true) {
            if (b == null) {
                return b;
            }
            if (interned == b.name) {
                return b.value;
            }
            b = b.chain;
        }
    }

    Binding lookup0(String interned) {
        int hash = interned.hashCode() & 0x7fffffff;
        int i = hash & (table.get().size() - 1);
        Binding b = table.get().get(i);
        while (b != null && interned != b.name) {
            b = b.chain;
        }
        return b;
    }

    /**
     * Looks for a name-value binding in the symbol table chain.
     * 
     * @param interned
     *            the name of the variable, which must be intern'd
     * @return a NamedValue
     */
    public NamedValue lookup(String interned) {
        int hash = interned.hashCode() & 0x7fffffff;
        SymbolTable env = this;
        do {
            AtomicArray<Binding> env_tab = env.getTable();
            int i = hash & (env_tab.size() - 1);
            Binding b = env_tab.get(i);
            while (b != null && interned != b.name) {
                b = b.chain;
            }
            if (b != null) {
                return b;
            }
            env = env.parent;
        } while (env != null);
        return null;
    }

    /**
     * Defines a name-value binding in the symbol table.
     * 
     * @param interned  the name of the variable, which must be intern'd
     * @param value the new value
     * @exception IllegalStateException
     *                thrown when the specified symbol has been defined as a
     *                constant.
     */
    public void set(String interned, Object value) {
        int hash = interned.hashCode() & 0x7fffffff;
        int i = hash & (table.get().size() - 1);

        for (Binding b = table.get().get(i); b != null; b = b.chain) {
            if (interned == b.name) {
                b.set(value);
                return;
            }
        }
        addBinding(hash, interned, value, i);
    }

    /**
     * Defines a constant in the symbol table.
     * 
     * @param interned the name of the variable, which must be intern'd
     * @param value the constant value
     * @exception IllegalStateException
     *                thrown when the specified symbol has been defined as a
     *                constant
     */
    public void setConstant(String interned, Object value) {
        int hash = interned.hashCode() & 0x7fffffff;
        int i = hash & (table.get().size() - 1);

        for (Binding b = table.get().get(i); b != null; b = b.chain) {
            if (interned == b.name) {
                if (b instanceof ImmutableBinding) {
                    throw new IllegalStateException();
                }
                removeBinding(interned);
            }
        }
        addConstant(hash, interned, value, i);
    }

    void assign(String interned, Object value) {
        int hash = interned.hashCode() & 0x7fffffff;
        SymbolTable env = this;
        do {
            AtomicArray<Binding> env_tab = env.getTable();
            int i = hash & (env_tab.size() - 1);
            Binding b = env_tab.get(i);
            while (b != null && interned != b.name) {
                b = b.chain;
            }
            if (b != null) {
                b.set(value);
                return;
            }
            env = env.parent;
        } while (env != null);

        addBinding(hash, interned, value, hash & (table.get().size() - 1));
    }

    void addBinding(int hash, String interned, Object value, int index) {
        table.get().set( new Binding(hash, interned, value, table.get().get(index)),index ) ;
        if (count.inc() >= threshold.get()) {
            ensureCapacity(table.get().size() * 2);
        }
    }

    void addConstant(int hash, String interned, Object value, int index) {
        table.get().set( new ImmutableBinding(hash, interned, value, table.get().get(index)),index ) ;
        if (count.inc() >= threshold.get()) {
            ensureCapacity(table.get().size() * 2);
        }
    }

    void ensureCapacity(int newCapacity) {
        AtomicArray<Binding> oldTable = table.get();
        int oldCapacity = oldTable.size();
        if (oldCapacity == 1 << 30) {
            threshold.set(Integer.MAX_VALUE);
            return;
        }

        AtomicArray<Binding> newTable = new AtomicArray<Binding>(newCapacity);
        AtomicArray<Binding> src = table.get();
        for (int j = 0; j < src.size(); j++) {
            Binding b = src.get(j);
            if (b != null) {
                src.set(null,j);
                do {
                    Binding next = b.chain;
                    int i = b.hash & (newCapacity - 1);
                    b.chain = newTable.get(i);
                    newTable.set(b,i);
                    b = next;
                } while (b != null);
            }
        }
        table.set( newTable ) ;
        threshold.set((int) (newCapacity * LOAD_FACTOR));
    }

    Binding removeBinding(String interned) {
        int hash = interned.hashCode() & 0x7fffffff;
        int i = hash & (table.get().size() - 1);

        Binding prev = table.get().get(i);
        Binding b = prev;

        while (b != null) {
            Binding next = b.chain;
            if (interned == b.name) {
                count.dec();
                if (prev == b) {
                    table.get().set(next,i);
                } else {
                    prev.chain = next;
                }
                return b;
            }
            prev = b;
            b = next;
        }

        return b;
    }

    /**
     * Deletes all name-value bindings.
     */
    public void clear() {
        AtomicArray<Binding> tab = table.get();
        for (int i = 0; i < tab.size(); i++) {
            tab.set( null,i );
        }
        count.set(0);
    }

    public int size() {
        return count.get();
    }

    /**
     * Deep copy
     */
    public Object clone() {
        try {
            SymbolTable copy = (SymbolTable) super.clone();
            AtomicArray<Binding> src = table.get();
            int len = src.size() ;
            AtomicArray<Binding> newTable = new AtomicArray<Binding>(len);
            for (int i = 0; i < len; i++) {
                Binding b = src.get(i);
                if (b != null) {
                    newTable.set((Binding) b.clone(),i);
                }
            }
            copy.setTable(newTable);
            if (parent != null) {
                copy.parent = (SymbolTable) parent.clone();
            }
            return copy;
        } catch (CloneNotSupportedException e) {
            throw new InternalError();
        }
    }

    /**
     * Returns an enumeration of the NamedValues in the symbol table.
     * 
     * @return an enumeration of the NamedValues
     * @see pnuts.lang.NamedValue
     */
    public Enumeration bindings() {
        return new Enumerator(0);
    }

    /**
     * Returns an enumeration of the keys in the symbol table.
     * 
     * @return an enumeration of the keys
     */
    public Enumeration keys() {
        return new Enumerator(1);
    }

    /**
     * Returns an enumeration of the values in the symbol table.
     * 
     * @return an enumeration of the values
     */
    public Enumeration values() {
        return new Enumerator(2);
    }

    private class Enumerator implements Enumeration {
        Binding bind = null;
        AtomicArray<Binding> src =null;
        int index = -1;
        int kind;

        Enumerator(int kind) { // 0==Binding, 1==key, 2==value
            this.src = table.get() ;
            this.index = src.size();
            this.kind = kind;
        }

        public boolean hasMoreElements() {
            while (bind == null && index > 0) {
                bind = src.get(--index);
            }
            return bind != null;
        }

        public Object nextElement() {
            while (bind == null && index > 0) {
                bind = src.get(--index);
            }
            if (bind != null) {
                Binding b = bind;
                bind = b.chain;
                if (kind == 0) {
                    return b;
                } else if (kind == 1) {
                    return b.name;
                } else {
                    return b.value;
                }
            }
            throw new NoSuchElementException("SymbolTable Enumerator");
        }
    }

    static final long serialVersionUID = 61380568117862288L;

    private void writeObject(ObjectOutputStream s)
            throws IOException {
        s.defaultWriteObject();
        int cnt = 0;
        for (int i = 0; i < table.get().size(); i++){
        Binding b = table.get().get(i);
        while (b != null){
            if (b.value instanceof Serializable){
            cnt++;
            }
            b = b.chain;
        }
        }
        s.writeInt(cnt);
        for (int i = 0; i < table.get().size(); i++){
        Binding b = table.get().get(i);
        while (b != null){
            if (b.value instanceof Serializable){
            s.writeObject(b);
            }
            b = b.chain;
        }
        }
    }

    private void readObject(ObjectInputStream s)
            throws IOException, ClassNotFoundException {
        s.defaultReadObject();
        int cnt = s.readInt();
        this.table = new AtomicObject<AtomicArray<Binding>>( new AtomicArray<Binding>(INITIAL_CAPACITY) ) ;
        for (int i = 0; i < cnt; i++){
            Binding b = (Binding)s.readObject();
            set(b.name, b.value);
        }
    }
}
